@@ -1,5 +1,26 @@
 /* erlang_APL_nif.c */
 
+/*
+    This file is part of GNU APL, a free implementation of the
+    ISO/IEC Standard 13751, "Programming Language APL, Extended"
+
+    Copyright (C) 2008-2017  Dr. Jürgen Sauermann
+
+    This program is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published
+    by the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+*/
+
+
 #include <semaphore.h>
 #include <stdio.h>
 #include <string.h>
@@ -12,7 +33,7 @@
 # define HAVE_DIRTY   /* we hope so */
 #endif
 
-#if HAVE_DIRTY
+#ifdef HAVE_DIRTY
 # define DIRTY_FLAG , ERL_NIF_DIRTY_JOB_CPU_BOUND
 #else
 # define DIRTY_FLAG
