@@ -18,7 +18,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-// #define USE_POLL   /* use poll() instead of select() */
+#define USE_POLL   /* use poll() instead of select() */
 
 #include <errno.h>
 #include <limits.h>
@@ -268,7 +268,7 @@ char * bslash = strrchr(AP_progname, '/');
       {
         snprintf(AP_progname, APL_PATH_MAX, "AP%u", id.proc);
       }
-   
+
    AP_progname[APL_PATH_MAX] = 0;
 
    debug && *debug << "*** starting " << AP_progname << endl;
@@ -828,7 +828,7 @@ AP3_fd * ap_fd = 0;
                Svar_record * svar = db.find_var(key, LOC);
 
                if (svar)   svar->set_state(request->get__SET_STATE__new_state(),
-                                       request->get__SET_STATE__loc().c_str());
+                                       request->get__SET_STATE__sloc().c_str());
                else cerr << "*** key not in db at " << LOC << endl;
              }
              return;
@@ -978,7 +978,6 @@ const char * listen_name = APSERVER_PATH;
 bool got_path = false;
 bool got_port = false;
 bool auto_start = false;
-size_t janitor = 0;
 
    for (int a = 1; a < argc; )
        {
@@ -1096,8 +1095,6 @@ const int listen_sock = got_path ? open_UNIX_socket(listen_name)
 
    memset(&db, 0, sizeof(db));
 
-int max_fd = listen_sock;
-
    (verbosity > 0) && cerr << prog << ": entering main loop..." << endl;
    for (;;)
       {
@@ -1107,11 +1104,11 @@ int max_fd = listen_sock;
 
 #ifdef USE_POLL // use poll()
 
-        DynArray(struct pollfd, fds, 2*connected_procs.size() + 1);
+        pollfd fds[2*connected_procs.size() + 1];
         fds[0].fd = listen_sock;
         fds[0].events = POLLIN | POLLPRI;
         int fd_idx = 1;
-        for (int j = 0; j < connected_procs.size(); ++j)
+        for (size_t j = 0; j < connected_procs.size(); ++j)
             {
               TCP_socket fd = connected_procs[j].fd;
               if (fd != NO_TCP_SOCKET)
@@ -1150,22 +1147,51 @@ int max_fd = listen_sock;
 
 #else // use select()
 
+        size_t janitor = 0;
+        int max_fd = listen_sock;
+        char dummy;
+
+        if (0 != read(listen_sock, &dummy, 0))
+           cerr << prog << ": listen socket has died unexpectedly" << endl;
+
         FD_SET(listen_sock, &read_fds);
         for (size_t j = 0; j < connected_procs.size(); ++j)
             {
-              TCP_socket fd = connected_procs[j].fd;
+              {
+              const TCP_socket fd = connected_procs[j].fd;
               if (fd != NO_TCP_SOCKET)
                  {
-                   if (max_fd < fd)   max_fd = fd;
-                   FD_SET(fd, &read_fds);
+                   if (0 != read(fd, &dummy, 0))
+                      {
+                        cerr << prog << ": socket fd[" << j
+                             << "] has died unexpectedly" << endl;
+                        close_fd(fd);
+                      }
+                   else
+                      {
+                        if (max_fd < fd)   max_fd = fd;
+                        FD_SET(fd, &read_fds);
+                      }
                  }
+              }
 
-              fd = connected_procs[j].fd2;
-              if (fd != NO_TCP_SOCKET)
-                 {
-                   if (max_fd < fd)   max_fd = fd;
-                   FD_SET(fd, &read_fds);
-                 }
+              {
+                const TCP_socket fd2 = connected_procs[j].fd2;
+                if (fd2 != NO_TCP_SOCKET)
+                   {
+                     if (0 != read(fd2, &dummy, 0))
+                        {
+                          cerr << prog << ": socket fd2[" << j
+                               << "] has died unexpectedly" << endl;
+                          close_fd(fd2);
+                        }
+                     else
+                        {
+                          if (max_fd < fd2)   max_fd = fd2;
+                          FD_SET(fd2, &read_fds);
+                        }
+                   }
+              }
             }
 
         timeval timeout = { 1, 0 };   // one second
