@@ -48,15 +48,15 @@
 #include "Token.hh"
 #include "UCS_string.hh"
 #include "UserFunction.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Workspace.hh"
 
 using namespace std;
 
 /// if there is less than x chars left on the current line then leave
 /// char mode and start a new line indented.
-#define NEED(x)   if (space < (int)(x))   { leave_char_mode();   out << "\n"; \
-                                            space = do_indent(); } out
+#define NEED(x)   if (space < int(x)) \
+   { leave_char_mode();   out << "\n";   space = do_indent(); } out
 
 //-----------------------------------------------------------------------------
 bool
@@ -103,7 +103,7 @@ XML_Saving_Archive::find_vid(const Value * val)
 const void * item = bsearch(val, values, value_count, sizeof(_val_par),
                       _val_par::compare_val_par1);
    if (item == 0)   return INVALID_VID;
-   return (Vid)((const _val_par *)item - values);
+   return Vid(reinterpret_cast<const _val_par *>(item) - values);
 }
 //-----------------------------------------------------------------------------
 void
@@ -203,20 +203,35 @@ char cc[80];
             case CT_INT:   // uses UNI_PAD_U3
                  space -= leave_char_mode();
                  snprintf(cc, sizeof(cc), "%lld",
-                          (long long)cell.get_int_value());
+                          long_long(cell.get_int_value()));
                  NEED(1 + strlen(cc)) << UNI_PAD_U3 << decr(--space, cc);
                  break;
 
-            case CT_FLOAT:   // uses UNI_PAD_U4
+            case CT_FLOAT:   // uses UNI_PAD_U4 or UNI_PAD_U8
                  space -= leave_char_mode();
-                 snprintf(cc, sizeof(cc), "%.17g", cell.get_real_value());
+#ifdef RATIONAL_NUMBERS_WANTED
+                 {
+                 const FloatCell & flt = cell.cFloatCell();
+                 if (const APL_Integer denom = flt.get_denominator())
+                    {
+                      const APL_Integer numer = flt.get_numerator();
+                      snprintf(cc, sizeof(cc), "%lld÷%lld", long_long(numer),
+                               long_long(denom));
+                      NEED(1 + strlen(cc)) << UNI_PAD_U8 << decr(--space, cc);
+                      break;
+                    }
+                 }
+#endif
+                 snprintf(cc, sizeof(cc), "%.17g",
+                          double(cell.get_real_value()));
                  NEED(1 + strlen(cc)) << UNI_PAD_U4 << decr(--space, cc);
                  break;
 
             case CT_COMPLEX:   // uses UNI_PAD_U5
                  space -= leave_char_mode();
                  snprintf(cc, sizeof(cc), "%17gJ%17g",
-                          cell.get_real_value(), cell.get_imag_value());
+                          double(cell.get_real_value()),
+                          double(cell.get_imag_value()));
                  NEED(1 + strlen(cc)) << UNI_PAD_U5 << decr(--space, cc);
                  break;
 
@@ -235,10 +250,9 @@ char cc[80];
                    const Cell * cp = cell.get_lval_value();
                    if (cp)   // valid Cell *
                       {
-                        const Value * cp_owner =
-                              ((const LvalCell *)&cell)->get_cell_owner();
-                        const long long offset = cp_owner->get_offset(cp);
-                        const Vid vid = find_vid(cp_owner);
+                        const Value * owner = cell.cLvalCell().get_cell_owner();
+                        const long long offset = owner->get_offset(cp);
+                        const Vid vid = find_vid(owner);
                         snprintf(cc, sizeof(cc), "%d[%lld]", vid, offset);
                         NEED(1 + strlen(cc)) << UNI_PAD_U7 << decr(--space, cc);
                       }
@@ -339,17 +353,17 @@ XML_Saving_Archive::save_Parser(const Prefix & prefix)
 void
 XML_Saving_Archive::save_symtab(const SymbolTable & symtab)
 {
-Simple_string<const Symbol *, false> symbols = symtab.get_all_symbols();
+std::vector<const Symbol *> symbols = symtab.get_all_symbols();
 
    // remove erased symbols
    //
-   for (int s = 0; s < symbols.size();)
+   for (size_t s = 0; s < symbols.size();)
       {
         const Symbol * sym = symbols[s];
         if (sym->is_erased())
             {
-              symbols[s] = symbols.last();
-              symbols.pop();
+              symbols[s] = symbols.back();
+              symbols.pop_back();
               continue;
             }
 
@@ -366,7 +380,7 @@ Simple_string<const Symbol *, false> symbols = symtab.get_all_symbols();
         // set idx to the alphabetically smallest name
         //
         int idx = 0;
-        for (int i = 1; i < symbols.size(); ++i)
+        for (size_t i = 1; i < symbols.size(); ++i)
             {
               if (symbols[idx]->compare(*symbols[i]) > 0)   idx = i;
             }
@@ -374,8 +388,8 @@ Simple_string<const Symbol *, false> symbols = symtab.get_all_symbols();
         const Symbol * sym = symbols[idx];
         save_Symbol(*sym);
 
-        symbols[idx] = symbols.last();
-        symbols.pop();
+        symbols[idx] = symbols.back();
+        symbols.pop_back();
       }
 
    --indent;
@@ -474,7 +488,7 @@ XML_Saving_Archive::save_Symbol(const Symbol & sym)
 //-----------------------------------------------------------------------------
 void
 XML_Saving_Archive::save_user_commands(
-               const Simple_string<Command::user_command, false> & cmds)
+               const std::vector<Command::user_command> & cmds)
 {
    if (cmds.size() == 0)   return;
 
@@ -631,8 +645,8 @@ XML_Saving_Archive::_val_par::compare_val_par(const _val_par & A,
 int
 XML_Saving_Archive::_val_par::compare_val_par1(const void * key, const void * B)
 {
-const void * Bv = ((const _val_par *)B)->_val;
-   return (const char *)key - (const char *)Bv;
+const void * Bv = (reinterpret_cast<const _val_par *>(B))->_val;
+   return charP(key) - charP(Bv);
 }
 //-----------------------------------------------------------------------------
 XML_Saving_Archive &
@@ -782,12 +796,13 @@ const int offset = Workspace::get_v_Quad_TZ().get_offset();   // timezone offset
    Value::mark_all_dynamic_values();
    Workspace::unmark_all_values();
 
-   for (const DynamicObject * obj = DynamicObject::get_all_values()->get_next();
-        obj != DynamicObject::get_all_values(); obj = obj->get_next())
+   for (const DynamicObject * dob = DynamicObject::get_all_values()->get_next();
+        dob != DynamicObject::get_all_values(); dob = dob->get_next())
        {
-         const Value & val(*(const Value *)obj);
+         // WARNING: do not use pValue() here !
+         const Value * val = static_cast<const Value *>(dob);
 
-         if (val.is_marked())    continue;   // stale
+         if (val->is_marked())    continue;   // stale
 
          ++value_count;
        }
@@ -795,10 +810,11 @@ const int offset = Workspace::get_v_Quad_TZ().get_offset();   // timezone offset
    values = new _val_par[value_count];
 ShapeItem idx = 0;
 
-   for (const DynamicObject * obj = DynamicObject::get_all_values()->get_next();
-        obj != DynamicObject::get_all_values(); obj = obj->get_next())
+   for (const DynamicObject * dob = DynamicObject::get_all_values()->get_next();
+        dob != DynamicObject::get_all_values(); dob = dob->get_next())
        {
-         const Value * val = (const Value *)obj;
+         // WARNING: do not use pValue() here !
+         const Value * val = static_cast<const Value *>(dob);
 
          if (val->is_marked())    continue;   // stale
 
@@ -835,14 +851,14 @@ ShapeItem idx = 0;
                         // happen. print out some more information about this
                         // case.
                         //
-                        CERR << "*** Sub-Value " << (void *)sub
-                             << " has two parents." << endl
+                        CERR << "*** Sub-Value "
+                             << voidP(sub) << " has two parents." << endl
                              << "Child: vid=" << sub_idx << ", _val="
                              << values[sub_idx]._val << "_par="
-                             << (void *)values[sub_idx]._par  << endl
+                             << values[sub_idx]._par << endl
                              << "Parent 2: vid=" << p <<  ", _val="
                              << values[p]._val << "_par="
-                             << (void *)values[p]._par  << endl
+                             << values[p]._par << endl
                              << "Call stack:" << endl;
                              BACKTRACE
                         CERR << endl << " Running )CHECK..." << endl;
@@ -862,7 +878,7 @@ print_history(CERR, values[p]._val, 0);
 " for BIG trouble!" << endl;
                       }
 
-                   values[sub_idx] = _val_par(values[sub_idx]._val, (Vid)p);
+                   values[sub_idx] = _val_par(values[sub_idx]._val, Vid(p));
                  }
               else if (cP.is_lval_cell())
                  {
@@ -875,11 +891,11 @@ print_history(CERR, values[p]._val, 0);
 
    // save all values (without their ravel)
    //
-   loop(vid, value_count)   save_shape((Vid)vid);
+   loop(vid, value_count)   save_shape(Vid(vid));
 
    // save ravels of all values
    //
-   loop(vid, value_count)   save_Ravel((Vid)vid);
+   loop(vid, value_count)   save_Ravel(Vid(vid));
 
    // save user defined symbols
    //
@@ -969,7 +985,7 @@ struct stat st;
 
    map_length = st.st_size;
    map_start = mmap(0, map_length, PROT_READ, MAP_SHARED, fd, 0);
-   if (map_start == MAP_FAILED)
+   if (map_start == reinterpret_cast<const void *>(-1))
       {
         CERR << "mmap() failed: " << strerror(errno) << endl;
         close(fd);
@@ -979,14 +995,14 @@ struct stat st;
 
    // success
    //
-   file_start = (const UTF8 *)map_start;
-   file_end = file_start + map_length;
+   file_start = charP(map_start);
+   file_end = utf8P(file_start + map_length);
 
    reset();
 
-   if (!strncmp((const char *)file_start, "#!", 2) ||   // )DUMP file
-       !strncmp((const char *)file_start, "<!", 2) ||   // )DUMP-HTML file
-       !strncmp((const char *)file_start, "⍝!", 4))     // a library
+   if (!strncmp(file_start, "#!", 2) ||   // )DUMP file
+       !strncmp(file_start, "<!", 2) ||   // )DUMP-HTML file
+       !strncmp(file_start, "⍝!", 4))     // a library
       {
         // the file was either written with )DUMP or is a library.
         // Return the open file descriptor (the destructor will unmap())
@@ -996,7 +1012,7 @@ struct stat st;
         return;
       }
 
-   if (strncmp((const char *)file_start, "<?xml", 5))   // not an xml file
+   if (strncmp(file_start, "<?xml", 5))   // not an xml file
       {
         CERR << "file " << filename << " does not " << endl
              << "have the format of a GNU APL .xml or .apl file" << endl;
@@ -1008,14 +1024,14 @@ struct stat st;
 //-----------------------------------------------------------------------------
 XML_Loading_Archive::~XML_Loading_Archive()
 {
-   if (map_start)   munmap((char *)map_start, map_length);
-   if (fd != -1)   close(fd);
+   if (map_start)   munmap(map_start, map_length);
+   if (fd != -1)    close(fd);
 }
 //-----------------------------------------------------------------------------
 void
 XML_Loading_Archive::reset()
 {
-   line_start = data = file_start;
+   line_start = data = utf8P(file_start);
    line_no = 1;
    next_tag(LOC);
 }
@@ -1082,7 +1098,7 @@ int len = 0;
 bool
 XML_Loading_Archive::is_tag(const char * prefix) const
 {
-   return !strncmp((const char *)tag_name, prefix, strlen(prefix));
+   return !strncmp(charP(tag_name), prefix, strlen(prefix));
 }
 //-----------------------------------------------------------------------------
 void
@@ -1111,7 +1127,7 @@ const int att_len = strlen(att_name);
 
    for (const UTF8 * d = attributes; d < end_attr; ++d)
        {
-         if (strncmp(att_name, (const char *)d, att_len))   continue;
+         if (strncmp(att_name, charP(d), att_len))   continue;
          const UTF8 * dd = d + att_len;
          while (*dd <= ' ')   ++dd;   // skip whitespaces
          if (*dd++ != '=')   continue;
@@ -1139,15 +1155,15 @@ XML_Loading_Archive::find_int_attr(const char * attrib, bool optional, int base)
 const UTF8 * value = find_attr(attrib, optional);
    if (value == 0)   return -1;   // not found
 
-const int64_t val = strtoll((const char *)value, 0, base);
+const int64_t val = strtoll(charP(value), 0, base);
    return val;
 }
 //-----------------------------------------------------------------------------
-double
+APL_Float
 XML_Loading_Archive::find_float_attr(const char * attrib)
 {
 const UTF8 * value = find_attr(attrib, false);
-const double val = strtod((const char *)value, 0);
+const APL_Float val = strtod(charP(value), 0);
    return val;
 }
 //-----------------------------------------------------------------------------
@@ -1304,7 +1320,7 @@ bool prev_month = false;
    //
    for (const UTF8 * c = file_end - 12; (c > data) && (c > file_end - 200); --c)
        {
-         if (!strncmp((const char *)c, "</Workspace>", 12))
+         if (!strncmp(charP(c), "</Workspace>", 12))
             {
               file_is_complete = true;
               break;
@@ -1412,7 +1428,7 @@ const int  rk  = find_int_attr("rk",  false, 10);
 
    if (reading_vids)
       {
-         parents.append(parent);
+         parents.push_back(parent);
          return;
       }
 
@@ -1422,7 +1438,7 @@ Shape sh_value;
         char sh[20];
         snprintf(sh, sizeof(sh), "sh-%d", int(r));
         const UTF8 * sh_r = find_attr(sh, false);
-        sh_value.add_shape_item(atoll((const char *)sh_r));
+        sh_value.add_shape_item(atoll(charP(sh_r)));
       }
 
    // if we do )COPY or )PCOPY and vid is not in vids_COPY list, then we
@@ -1437,7 +1453,7 @@ bool no_copy = false;   // assume the value is needed
         int parent = vid;
         for (;;)
             {
-              Assert(parent < (int)parents.size());
+              Assert(parent < int(parents.size()));
               if (parents[parent] == -1)   break;   // topmost owner found
               parent = parents[parent];
             }
@@ -1455,14 +1471,14 @@ bool no_copy = false;   // assume the value is needed
 
    if (no_copy)
       {
-        values.append(Value_P());
+        values.push_back(Value_P());
       }
    else
       {
-        Assert(vid == (int)values.size());
+        Assert(vid == int(values.size()));
 
         Value_P val(sh_value, LOC);
-        values.append(val);
+        values.push_back(val);
       }
 }
 //-----------------------------------------------------------------------------
@@ -1495,9 +1511,9 @@ const Unicode type = UTF8_string::toUni(first, len, true);
              first += len;
              {
                char * end = 0;
-               const long long int val = strtoll((const char *)first, &end, 10);
+               const APL_Integer val = strtoll(charP(first), &end, 10);
                new (C++) IntCell(val);
-               first = (const UTF8 *)end;
+               first = utf8P(end);
              }
              break;
 
@@ -1505,9 +1521,9 @@ const Unicode type = UTF8_string::toUni(first, len, true);
              first += len;
              {
                char * end = 0;
-               const double val = strtod((const char *)first, &end);
+               const APL_Float val = strtod(charP(first), &end);
                new (C++) FloatCell(val);
-               first = (const UTF8 *)end;
+               first = utf8P(end);
              }
              break;
 
@@ -1515,13 +1531,13 @@ const Unicode type = UTF8_string::toUni(first, len, true);
              first += len;
              {
                char * end = 0;
-               const double real = strtod((const char *)first, &end);
-               first = (const UTF8 *)end;
+               const APL_Float real = strtod(charP(first), &end);
+               first = utf8P(end);
                Assert(*end == 'J');
                ++end;
-               const double imag = strtod(end, &end);
+               const APL_Float imag = strtod(end, &end);
                new (C++) ComplexCell(real, imag);
-               first = (const UTF8 *)end;
+               first = utf8P(end);
              }
              break;
 
@@ -1529,11 +1545,11 @@ const Unicode type = UTF8_string::toUni(first, len, true);
              first += len;
              {
                char * end = 0;
-               const int vid = strtol((const char *)first, &end, 10);
+               const int vid = strtoll(charP(first), &end, 10);
                Assert(vid >= 0);
-               Assert(vid < (int)values.size());
-               new (C++) PointerCell(values[vid], C_owner);
-               first = (const UTF8 *)end;
+               Assert(vid < int(values.size()));
+               C++->init_from_value(values[vid].get(), C_owner, LOC);
+               first = utf8P(end);
              }
              break;
 
@@ -1547,18 +1563,44 @@ const Unicode type = UTF8_string::toUni(first, len, true);
              else
                 {
                   char * end = 0;
-                  const int vid = strtol((const char *)first, &end, 16);
+                  const int vid = strtoll(charP(first), &end, 16);
                   Assert(vid >= 0);
-                  Assert(vid < (int)values.size());
+                  Assert(vid < int(values.size()));
                   Assert(*end == '[');   ++end;
                   const ShapeItem offset = strtoll(end, &end, 16);
                   Assert(*end == ']');   ++end;
                   new (C++) LvalCell(&values[vid]->get_ravel(offset),
                                      values[vid].get());
-                  first = (const UTF8 *)end;
+                  first = utf8P(end);
                 }
              break;
 
+        case UNI_PAD_U8: // rational quotient
+             //
+             // we should understand rational quotients even if we are
+             // not ./configured for them..
+             //
+             first += len;
+             {
+               char * end = 0;
+               const uint64_t numer = strtoll(charP(first), &end, 10);
+               first = utf8P(end);
+
+               // skip ÷ (which is is C3 B7 in UTF8)
+               Assert((*end++ & 0xFF) == 0xC3);
+               Assert((*end++ & 0xFF) == 0xB7);
+               const uint64_t denom = strtoll(end, &end, 10);
+               Assert(denom > 0);
+#ifdef RATIONAL_NUMBERS_WANTED
+               new (C++) FloatCell(numer, denom);
+#else
+               new (C++) FloatCell((1.0*(numer))/denom);
+#endif
+               first = utf8P(end);
+             }
+             break;
+
+
         default: Q1(type) Q1(line_no) DOMAIN_ERROR;
       }
 }
@@ -1594,9 +1636,9 @@ XML_Loading_Archive::read_chars(UCS_string & ucs, const UTF8 * & utf)
               utf += len;   // skip UNI_PAD_U1
               char_mode = false;
               char * end = 0;
-              const int hex = strtol((const char *)utf, &end, 16);
+              const int hex = strtoll(charP(utf), &end, 16);
               ucs.append(Unicode(hex));
-              utf = (const UTF8 *)end;
+              utf = utf8P(end);
               continue;
             }
 
@@ -1635,7 +1677,7 @@ const UTF8 * cells = find_attr("cells", false);
 
    Log(LOG_archive)   CERR << "    read_Ravel() vid=" << vid << endl;
 
-   Assert(vid < (int)values.size());
+   Assert(vid < int(values.size()));
 Value_P val = values[vid];
 
    if (!val)   // )COPY with vids_COPY or static value
@@ -1682,18 +1724,19 @@ void
 XML_Loading_Archive::read_Variable(int d, Symbol & symbol)
 {
 const int vid = find_int_attr("vid", false, 10);
-   Assert(vid < (int)values.size());
+   Assert(vid < int(values.size()));
 
    Log(LOG_archive)   CERR << "      [" << d << "] read_Variable() vid=" << vid
                            << " name=" << symbol.get_name() << endl;
 
-   // some system variables are  saved for troubleshooting purposes, but
+   // some system variables are saved for troubleshooting purposes, but
    // should not be loaded...
    //
    if (symbol.is_readonly())                     return;
-   if (symbol.get_name().starts_iwith("⎕NLT"))   return;
-   if (symbol.get_Id() == ID::Quad_SVE)           return;
-   if (symbol.get_Id() == ID::Quad_SYL)           return;
+   if (symbol.get_name().starts_iwith("⎕NLT"))   return;   // extinct
+   if (symbol.get_Id() == ID_Quad_SVE)           return;
+   if (symbol.get_Id() == ID_Quad_SYL)           return;
+   if (symbol.get_Id() == ID_Quad_PS)            return;
 
    if (vid == -1)   // stale variable
       {
@@ -1735,7 +1778,7 @@ int eprops[4] = { 0, 0, 0, 0 };
 const UTF8 * ep = find_attr("exec-properties", true);
    if (ep)
       {
-        sscanf((const char *)ep, "%d,%d,%d,%d",
+        sscanf(charP(ep), "%d,%d,%d,%d",
                eprops, eprops + 1, eprops + 2, eprops+ 3);
       }
 
@@ -1871,17 +1914,17 @@ const int depth = find_int_attr("stack-size", false, 10);
    // lookup symbol, trying ⎕xx first
    //
 Symbol * symbol;
-   if (name_UCS == ID::get_name(ID::LAMBDA))
+   if (name_UCS == ID::get_name_UCS(ID_LAMBDA))
       symbol = &Workspace::get_v_LAMBDA();
-   else if (name_UCS == ID::get_name(ID::ALPHA))
+   else if (name_UCS == ID::get_name_UCS(ID_ALPHA))
       symbol = &Workspace::get_v_ALPHA();
-   else if (name_UCS == ID::get_name(ID::ALPHA_U))
+   else if (name_UCS == ID::get_name_UCS(ID_ALPHA_U))
       symbol = &Workspace::get_v_ALPHA_U();
-   else if (name_UCS == ID::get_name(ID::CHI))
+   else if (name_UCS == ID::get_name_UCS(ID_CHI))
       symbol = &Workspace::get_v_CHI();
-   else if (name_UCS == ID::get_name(ID::OMEGA))
+   else if (name_UCS == ID::get_name_UCS(ID_OMEGA))
       symbol = &Workspace::get_v_OMEGA();
-   else if (name_UCS == ID::get_name(ID::OMEGA_U))
+   else if (name_UCS == ID::get_name_UCS(ID_OMEGA_U))
       symbol = &Workspace::get_v_OMEGA_U();
    else
       symbol = Workspace::lookup_existing_symbol(name_UCS);
@@ -1913,7 +1956,7 @@ bool no_copy = is_protected || (have_allowed_objects && !is_selected);
         if (is_tag("Variable"))
            {
              const int vid = find_int_attr("vid", false, 10);
-             vids_COPY.append(vid);
+             vids_COPY.push_back(vid);
            }
         skip_to_tag("/Symbol");
         return;
@@ -1924,13 +1967,13 @@ bool no_copy = is_protected || (have_allowed_objects && !is_selected);
    //
    if (!have_allowed_objects       &&   // no dedicated object list
         copying                    &&   // )COPY
-        (name_UCS == ID::get_name(ID::Quad_CT) ||
-         name_UCS == ID::get_name(ID::Quad_FC) ||
-         name_UCS == ID::get_name(ID::Quad_IO) ||
-         name_UCS == ID::get_name(ID::Quad_LX) ||
-         name_UCS == ID::get_name(ID::Quad_PP) ||
-         name_UCS == ID::get_name(ID::Quad_PR) ||
-         name_UCS == ID::get_name(ID::Quad_RL)
+        (name_UCS == ID::get_name_UCS(ID_Quad_CT) ||
+         name_UCS == ID::get_name_UCS(ID_Quad_FC) ||
+         name_UCS == ID::get_name_UCS(ID_Quad_IO) ||
+         name_UCS == ID::get_name_UCS(ID_Quad_LX) ||
+         name_UCS == ID::get_name_UCS(ID_Quad_PP) ||
+         name_UCS == ID::get_name_UCS(ID_Quad_PR) ||
+         name_UCS == ID::get_name_UCS(ID_Quad_RL)
         ))
       {
         Log(LOG_archive)   CERR << name_UCS << " not copied at " << LOC << endl;
@@ -1940,8 +1983,8 @@ bool no_copy = is_protected || (have_allowed_objects && !is_selected);
    // in a )LOAD silently ignore session variables (⎕PW and ⎕TZ)
    //
    if (!copying &&   // )LOAD
-        (name_UCS == ID::get_name(ID::Quad_PW) ||
-         name_UCS == ID::get_name(ID::Quad_TZ)))
+        (name_UCS == ID::get_name_UCS(ID_Quad_PW) ||
+         name_UCS == ID::get_name_UCS(ID_Quad_TZ)))
       {
         skip_to_tag("/Symbol");
         return;
@@ -1963,8 +2006,8 @@ bool no_copy = is_protected || (have_allowed_objects && !is_selected);
       {
         if (allowed_objects[a] == name_UCS)
              {
-               allowed_objects[a] = allowed_objects.last();
-               allowed_objects.pop();
+               allowed_objects[a] = allowed_objects.back();
+               allowed_objects.pop_back();
                break;
            }
       }
@@ -2035,7 +2078,7 @@ UCS_string  fun_UCS(fun_UTF);
 const int mode = find_int_attr("mode", false, 10);
 
 Command::user_command ucmd = { name_UCS, fun_UCS, mode };
-   Workspace::get_user_commands().append(ucmd);
+   Workspace::get_user_commands().push_back(ucmd);
 }
 //-----------------------------------------------------------------------------
 void
@@ -2103,7 +2146,7 @@ Executable * exec = 0;
    Workspace::push_SI(exec, LOC);
 StateIndicator * si = Workspace::SI_top();
    Assert(si);
-   si->set_PC((Function_PC)pc);
+   si->set_PC(Function_PC(pc));
    read_Parser(*si, lev);
 
    for (;;)
@@ -2153,7 +2196,8 @@ Executable *
 XML_Loading_Archive::read_UserFunction()
 {
 const int macro_num = find_int_attr("macro-num", true, 10);
-   if (macro_num != -1)   return Macro::get_macro((Macro::Macro_num)macro_num);
+   if (macro_num != -1)
+      return Macro::get_macro(Macro::Macro_num(macro_num));
 
 const UTF8 * lambda_name = find_attr("lambda-name", true);
    if (lambda_name)   return read_lambda(lambda_name);
@@ -2184,7 +2228,7 @@ XML_Loading_Archive::read_lambda(const UTF8 * lambda_name)
 {
 UCS_string lambda = read_UCS();
 
-Symbol dummy(ID::No_ID);
+Symbol dummy(ID_No_ID);
 UserFunction * ufun = UserFunction::fix_lambda(dummy, lambda);
    Assert(ufun);
    ufun->increment_refcount(LOC);   // since we push it below
@@ -2219,9 +2263,9 @@ const int action     = find_int_attr("action",         false, 10);
 
 Prefix & parser = si.current_stack;
 
-   parser.set_assign_state((Assign_state)ass_state);
-   parser.action = (R_action)action;
-   parser.lookahead_high = (Function_PC)lah_high;
+   parser.set_assign_state(Assign_state(ass_state));
+   parser.action = R_action(action);
+   parser.lookahead_high = Function_PC(lah_high);
 
    for (;;)
        {
@@ -2275,15 +2319,15 @@ const TokenTag tag = TokenTag(find_int_attr("tag", false, 16));
 
         case TV_FLT:   
              {
-               const double fval = find_float_attr("float");
-               new (&tloc.tok) Token(tag, fval);
+               const APL_Float val = find_float_attr("float");
+               new (&tloc.tok) Token(tag, val);
              }
              break;
 
         case TV_CPX:
              {
-               const double real = find_float_attr("real");
-               const double imag = find_float_attr("imag");
+               const APL_Float real = find_float_attr("real");
+               const APL_Float imag = find_float_attr("imag");
                new (&tloc.tok) Token(tag, real, imag);
              }
              break;
@@ -2313,14 +2357,14 @@ const TokenTag tag = TokenTag(find_int_attr("tag", false, 16));
         case TV_VAL:   
              {
                const int vid = find_int_attr("vid", false, 10);
-               Assert(vid < (int)values.size());
+               Assert(vid < int(values.size()));
                new (&tloc.tok) Token(tag, values[vid]);
              }
              break;
 
         case TV_INDEX: 
              {
-               char * vids = (char *)find_attr("index", false);
+               const UTF8 * vids = find_attr("index", false);
                IndexExpr & idx = *new IndexExpr(ASS_none, LOC);
                while (*vids != '"')
                   {
@@ -2331,13 +2375,15 @@ const TokenTag tag = TokenTag(find_int_attr("tag", false, 16));
                        }
                     else                // value
                        {
+                         char * end = 0;
                          Assert1(*vids == 'v');   ++vids;
                          Assert1(*vids == 'i');   ++vids;
                          Assert1(*vids == 'd');   ++vids;
                          Assert1(*vids == '_');   ++vids;
-                         const int vid = strtol(vids, &vids, 10);
-                         Assert(vid < (int)values.size());
+                         const int vid = strtoll(charP(vids), &end, 10);
+                         Assert(vid < int(values.size()));
                          idx.add(values[vid]);
+                         vids = utf8P(end);
                        }
                   }
                new (&tloc.tok) Token(tag, idx);
@@ -2347,7 +2393,8 @@ const TokenTag tag = TokenTag(find_int_attr("tag", false, 16));
         case TV_FUN:
              {
                Function * fun = read_Function_name("ufun-name",
-                                                   "symbol-level", "fun-id");
+                                                   "symbol-level",
+                                                   "fun-id");
                Assert(fun);
                new (&tloc.tok) Token(tag, fun);
              }
@@ -2368,12 +2415,12 @@ const UTF8 * fun_name = find_attr(name_attr, true);
 
    if (fun_name)   // user defined function
       {
-        int level = find_int_attr(level_attr, false, 10);
+        const int level = find_int_attr(level_attr, false, 10);
         const UTF8 * end = fun_name;
         while (*end != '"')   ++end;
         UTF8_string name_UTF(fun_name, end - fun_name);
         UCS_string name_UCS(name_UTF);
-        if (name_UCS == ID::get_name(ID::LAMBDA))
+        if (name_UCS == ID::get_name_UCS(ID_LAMBDA))
            {
              Assert(level == -1);
              return find_lambda(name_UCS);
@@ -2393,7 +2440,7 @@ const UTF8 * fun_name = find_attr(name_attr, true);
 const int fun_id = find_int_attr(id_attr, true, 16);
    if (fun_id != -1)
       {
-        Function * sysfun = ID::get_system_function(ID::Id(fun_id));
+        Function * sysfun = ID::get_system_function(Id(fun_id));
         Assert(sysfun);
         return sysfun;
       }
