@@ -38,7 +38,7 @@ const int loc_len = strlen(file) + 40;
 char * loc = new char[loc_len + 1];
 
    Log(LOG_delete)
-      CERR << "new    " << (const void *)loc << " at " LOC << endl;
+      CERR << "new    " << voidP(loc) << " at " LOC << endl;
 
    snprintf(loc, loc_len, "%s:%d", file, line);
    loc[loc_len] = 0;
@@ -80,8 +80,8 @@ char * loc = new char[loc_len + 1];
    // count errors
    IO_Files::assert_error();
 
-Error * err = Workspace::get_error();
-   if (err)   err->init(E_ASSERTION_FAILED, loc);
+   if (Error * err = Workspace::get_error())
+      new (err) Error(E_ASSERTION_FAILED, loc);
 
    asserting = false;
 
