@@ -26,9 +26,9 @@
 #include "CharCell.hh"
 #include "PrimitiveFunction.hh"
 #include "PrintOperator.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Workspace.hh"
-                                            
+
 Bif_F12_FORMAT   Bif_F12_FORMAT::_fun;       // ⍕
 Bif_F12_FORMAT * Bif_F12_FORMAT::fun = &Bif_F12_FORMAT::_fun;
 
@@ -175,7 +175,7 @@ Bif_F12_FORMAT::monadic_format(Value_P B)
 {
    Assert(B->get_rank() <= 2);
 
-const PrintStyle style((PrintStyle)(PR_APL | PST_NO_FRACT_0));
+const PrintStyle style(PrintStyle(PR_APL | PST_NO_FRACT_0));
 const PrintContext pctx = Workspace::get_PrintContext(style);
 
 const PrintBuffer pb(*B, pctx, 0);
@@ -219,7 +219,7 @@ const ShapeItem rows = B->get_rows();
    // split format into format fields, one per column.
    // If there is only one format field, then repeat it cols times.
    //
-UCS_string_vector col_formats;
+vector<UCS_string> col_formats;
    split_example_into_columns(format, col_formats);
    if (col_formats. size() == 1)
       {
@@ -228,17 +228,17 @@ UCS_string_vector col_formats;
         loop(c, cols - 1)
            {
              format.append(f);
-             col_formats.append(col0);
+             col_formats.push_back(col0);
            }
       }
 
-   if (cols != (ShapeItem)(col_formats.size()))   LENGTH_ERROR;
+   if (cols != ShapeItem(col_formats.size()))   LENGTH_ERROR;
 
    // convert each column format string into a Format_LIFER
    //
-Simple_string<Format_LIFER, false> col_items;
+vector<Format_LIFER> col_items;
    loop(c, col_formats.size())
-       col_items.append(Format_LIFER(col_formats[c]));
+       col_items.push_back(Format_LIFER(col_formats[c]));
 
    Log(LOG_Bif_F12_FORMAT)
       {
@@ -303,7 +303,7 @@ Value_P Z(shape_Z, LOC);
 //-----------------------------------------------------------------------------
 void
 Bif_F12_FORMAT::split_example_into_columns(const UCS_string & format,
-                                   UCS_string_vector & col_formats)
+                                   vector<UCS_string> & col_formats)
 {
 bool fmt_seen = false;
 UCS_string fmt;
@@ -317,8 +317,8 @@ UCS_string fmt;
 
          if ((cc == UNI_ASCII_SPACE) && fmt_seen)   // end of field
             {
-              col_formats.append(fmt);
-              fmt.shrink(0);    // start a new field;
+              col_formats.push_back(fmt);
+              fmt.clear();    // start a new field;
               fmt_seen = false;
               continue;   // next char
             }
@@ -328,8 +328,8 @@ UCS_string fmt;
               ++f;   // next char is right decorator (and end of field)
               if (f < format.size())   fmt.append(format[f]);
 
-              col_formats.append(fmt);
-              fmt.shrink(0);    // start a new field;
+              col_formats.push_back(fmt);
+              fmt.clear();    // start a new field;
               fmt_seen = false;
               continue;   // next char
             }
@@ -337,11 +337,11 @@ UCS_string fmt;
 
    if ((!fmt_seen) && (col_formats.size() > 0))
       {
-        col_formats.last().append(fmt);
+        col_formats.back().append(fmt);
       }
    else
       {
-        col_formats.append(fmt);
+        col_formats.push_back(fmt);
       }
 }
 //-----------------------------------------------------------------------------
@@ -351,7 +351,7 @@ Bif_F12_FORMAT::Format_LIFER::format_example(APL_Float value)
 UCS_string data_int;
 UCS_string data_fract;
 UCS_string data_expo;
-const double val_1 = (value < 0) ? -value : value;
+const APL_Float val_1 = (value < 0.0) ? APL_Float(-value) : value;
 
 bool overflow = false;
    fill_data_fields(val_1, data_int, data_fract, data_expo, overflow);
@@ -364,11 +364,11 @@ bool overflow = false;
       }
    if (overflow)   return UCS_string(out_size(), Workspace::get_FC(3));
 
-const UCS_string left = format_left_side(data_int, value < 0, overflow);
+const UCS_string left = format_left_side(data_int, value < 0.0, overflow);
    if (overflow)   return UCS_string(out_size(), Workspace::get_FC(3));
    Assert(left.size() == (left_deco.out_len + int_part.out_len));
 
-const UCS_string right = format_right_side(data_fract, value<0, data_expo);
+const UCS_string right = format_right_side(data_fract, value < 0.0, data_expo);
    Assert(right.size() == (fract_part.out_len + expo_deco.out_len +
                            exponent.out_len + right_deco.out_len));
 
@@ -657,7 +657,7 @@ Format_sub::print(ostream & out) const
 }
 //-----------------------------------------------------------------------------
 void
-Bif_F12_FORMAT::Format_LIFER::fill_data_fields(double value,
+Bif_F12_FORMAT::Format_LIFER::fill_data_fields(APL_Float value,
                 UCS_string & data_int, UCS_string & data_fract,
                 UCS_string & data_expo, bool & overflow)
 {
@@ -676,8 +676,8 @@ char * fract_end = 0;
         // create a format like %.5E (if fract_part has 5 digits)
         //
         const int flen = snprintf(format, sizeof(format), "%%.%luE",
-                                  (unsigned long)fract_part.size());
-        Assert(flen < (int)sizeof(format));   // format was big enough.
+                                unsigned_long(fract_part.size()));
+        Assert(flen < int(sizeof(format)));   // format was big enough.
 
         const int dlen = snprintf(&data_buf[0], data_buf_len, format, value);
         Assert(dlen < data_buf_len);
@@ -699,8 +699,8 @@ char * fract_end = 0;
    else   // no exponent in format string.
       {
         const int flen = snprintf(format, sizeof(format), "%%.%luf",
-                                  (unsigned long)fract_part.size());
-        Assert(flen < (int)sizeof(format));   // assume no snprintf() overflow
+                                  unsigned_long(fract_part.size()));
+        Assert(flen < int(sizeof(format)));   // assume no snprintf() overflow
 
         const int dlen = snprintf(&data_buf[0], data_buf_len, format, value);
         data_buf[data_buf_len - 1] = 0;
@@ -755,7 +755,7 @@ const int ilen = int_end - &data_buf[0];
    if (  data_int.size() == 1
       && data_int[0] == UNI_ASCII_0
       && int_part.min_len == 0
-      && fract_part.size())   data_int.shrink(0);
+      && fract_part.size())   data_int.clear();
 
    Log(LOG_Bif_F12_FORMAT)
       {
@@ -836,7 +836,7 @@ size_t d = data.size();
          else                             overflow = true;
       }
 
-   if ((flt_mask & BIT_9) && ucs.all_zeroes())   ucs.shrink(0);
+   if ((flt_mask & BIT_9) && ucs.all_zeroes())   ucs.clear();
 
    return ucs.reverse();
 }
@@ -866,7 +866,7 @@ int d = 0;
             }
       }
 
-   if ((flt_mask & BIT_9) && ucs.all_zeroes())   ucs.shrink(0);
+   if ((flt_mask & BIT_9) && ucs.all_zeroes())   ucs.clear();
 
    return ucs;
 }
