@@ -18,6 +18,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#include "Bif_F12_TAKE_DROP.hh"
 #include "Bif_OPER1_EACH.hh"
 #include "Macro.hh"
 #include "PointerCell.hh"
@@ -32,7 +33,7 @@ Bif_OPER1_EACH * Bif_OPER1_EACH::fun = &Bif_OPER1_EACH::_fun;
 Token
 Bif_OPER1_EACH::eval_ALB(Value_P A, Token & _LO, Value_P B)
 {
-   // dyadic EACH
+   // dyadic EACH: call _LO for corresponding items of A and B
 
    if (!(A->is_scalar() || B->is_scalar() || A->same_shape(*B)))
       {
@@ -43,6 +44,11 @@ Bif_OPER1_EACH::eval_ALB(Value_P A, Token & _LO, Value_P B)
 Function * LO = _LO.get_function();
    Assert1(LO);
 
+   // for the ambiguous operators /. ⌿, \, and ⍀ is_operator() returns true,
+   // which is incorrect in this context. We use get_signature() instead.
+   //
+   if ((LO->get_signature() & SIG_DYA) != SIG_DYA)   VALENCE_ERROR;
+
    if (A->is_empty() || B->is_empty())
       {
         if (!LO->has_result())   return Token(TOK_VOID);
@@ -58,9 +64,10 @@ Function * LO = _LO.get_function();
         else if (!B->is_scalar())   DOMAIN_ERROR;
 
         Value_P Z1 = LO->eval_fill_AB(Fill_A, Fill_B).get_apl_val();
+        if (Z1->is_simple_scalar())   return Token(TOK_APL_VALUE1, Z1);
 
         Value_P Z(shape_Z, LOC);
-        new (&Z->get_ravel(0)) PointerCell(Z1, Z.getref());
+        new (&Z->get_ravel(0)) PointerCell(Z1.get(), Z.getref());
         Z->check_value(LOC);
         return Token(TOK_APL_VALUE1, Z);
       }
@@ -80,21 +87,28 @@ Function * LO = _LO.get_function();
         if (scalar_B && B->get_ravel(0).is_pointer_cell())
            B = B->get_ravel(0).get_pointer_value();
 
-        Macro * macro = 0; 
+        Macro * macro = 0;
         if (LO->has_result())
            if (scalar_A)
-              if (scalar_B)   macro = Macro::Z__sA_LO_EACH_sB;
-              else            macro = Macro::Z__sA_LO_EACH_vB;
+              if (scalar_B)   macro = Macro::get_macro(
+                                      Macro::MAC_Z__sA_LO_EACH_sB);
+              else            macro = Macro::get_macro(
+                                      Macro::MAC_Z__sA_LO_EACH_vB);
            else
-              if (scalar_B)   macro = Macro::Z__vA_LO_EACH_sB;
-              else            macro = Macro::Z__vA_LO_EACH_vB;
+              if (scalar_B)   macro = Macro::get_macro(
+                                      Macro::MAC_Z__vA_LO_EACH_sB);
+              else            macro = Macro::get_macro(
+                                      Macro::MAC_Z__vA_LO_EACH_vB);
         else
            if (scalar_A)
               if (scalar_B)   LO->eval_ALB(A, _LO, B);
-              else                  macro = Macro::sA_LO_EACH_vB;
+              else                  macro = Macro::get_macro(
+                                            Macro::MAC_sA_LO_EACH_vB);
            else
-              if (scalar_B)   macro = Macro::vA_LO_EACH_sB;
-              else            macro = Macro::vA_LO_EACH_vB;
+              if (scalar_B)   macro = Macro::get_macro(
+                                      Macro::MAC_vA_LO_EACH_sB);
+              else            macro = Macro::get_macro(
+                                      Macro::MAC_vA_LO_EACH_vB);
 
         return macro->eval_ALB(A, _LO, B);
       }
@@ -143,7 +157,7 @@ ShapeItem len_Z = 0;
              if (vZ->is_simple_scalar())
                 cZ->init(vZ->get_ravel(0), Z.getref(), LOC);
              else
-                new (cZ)   PointerCell(vZ, Z.getref());
+                new (cZ)   PointerCell(vZ.get(), Z.getref());
 
             continue;   // next z
            }
@@ -163,10 +177,12 @@ ShapeItem len_Z = 0;
 Token
 Bif_OPER1_EACH::eval_LB(Token & _LO, Value_P B)
 {
-   // monadic EACH
+   // monadic EACH: call _LO for every item of B
 
 Function * LO = _LO.get_function();
    Assert1(LO);
+   if (LO->is_operator())                SYNTAX_ERROR;
+   if (!(LO->get_signature() & SIG_B))   VALENCE_ERROR;
 
    if (B->is_empty())
       {
@@ -174,16 +190,29 @@ Function * LO = _LO.get_function();
 
         Value_P Fill_B = Bif_F12_TAKE::first(B);
         Token tZ = LO->eval_fill_B(Fill_B);
-        Value_P Z = tZ.get_apl_val();
-        Z->set_shape(B->get_shape());
+        Value_P Z1 = tZ.get_apl_val();
+
+        if (Z1->is_simple_scalar())
+           {
+             Z1->set_shape(B->get_shape());
+             return Token(TOK_APL_VALUE1, Z1);
+           }
+
+        Value_P Z(B->get_shape(), LOC);
+        new (&Z->get_ravel(0)) PointerCell(Z1.get(), Z.getref());
+        Z->check_value(LOC);
         return Token(TOK_APL_VALUE1, Z);
       }
 
    if (LO->may_push_SI())   // user defined LO
       {
-        if (!LO->has_result())         return Macro::LO_EACH_B->eval_LB(_LO, B);
-        if (LO == Bif_F1_EXECUTE::fun)  return Macro::Z__EXEC_EACH_B->eval_B(B);
-        return Macro::Z__LO_EACH_B->eval_LB(_LO, B);
+        if (!LO->has_result())
+           return Macro::get_macro(Macro::MAC_LO_EACH_B)->eval_LB(_LO, B);
+
+        if (LO == Bif_F1_EXECUTE::fun)
+           return Macro::get_macro(Macro::MAC_Z__EXEC_EACH_B)->eval_B(B);
+
+        return Macro::get_macro(Macro::MAC_Z__LO_EACH_B)->eval_LB(_LO, B);
       }
 
 const ShapeItem len_Z = B->element_count();
@@ -207,7 +236,7 @@ Value_P Z;
                   if (vZ->is_simple_scalar())
                      cZ->init(vZ->get_ravel(0), Z.getref(), LOC);
                   else
-                     new (cZ)  PointerCell(vZ, Z.getref());
+                     new (cZ)  PointerCell(vZ.get(), Z.getref());
 
                   continue;   // next z
            }
@@ -230,7 +259,7 @@ Value_P Z;
              else
                 {
                   LO_B = Value_P(LOC);
-   
+
                   LO_B->get_ravel(0).init(*cB, LO_B.getref(), LOC);
                   LO_B->set_complete();
                 }
@@ -238,7 +267,7 @@ Value_P Z;
              Token result = LO->eval_B(LO_B);
              if (result.get_Class() == TC_VALUE)
                 {
-                  Value_P vZ = result.get_apl_val();
+                  Value * vZ = result.get_apl_val().get();
 
                   Cell * cZ = Z->next_ravel();
                   if (0)
