@@ -24,7 +24,7 @@
 #include "FloatCell.hh"
 #include "IntCell.hh"
 #include "PointerCell.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
@@ -34,7 +34,7 @@ CDR::to_CDR(CDR_string & result, const Value & value)
 const CDR_type type = value.get_CDR_type();
 const int len = value.total_size_brutto(type);
    result.reserve(len + 1);
-   result.shrink(0);
+   result.clear();
 
    fill(result, type, len, value);
 }
@@ -60,42 +60,42 @@ CDR::fill(CDR_string & result, int type, int len, const Value & val)
    // ptr
    //
 const uint32_t ptr = 0x00002020;
-   result.append(Unicode(ptr >> 24 & 0xFF));
-   result.append(Unicode(ptr >> 16 & 0xFF));
-   result.append(Unicode(ptr >>  8 & 0xFF));
-   result.append(Unicode(ptr       & 0xFF));
+   result.push_back(Unicode(ptr >> 24 & 0xFF));
+   result.push_back(Unicode(ptr >> 16 & 0xFF));
+   result.push_back(Unicode(ptr >>  8 & 0xFF));
+   result.push_back(Unicode(ptr       & 0xFF));
 
    // nb (header + data + sub-values + padding)
    //
-   result.append(Unicode(len >> 24 & 0xFF));
-   result.append(Unicode(len >> 16 & 0xFF));
-   result.append(Unicode(len >>  8 & 0xFF));
-   result.append(Unicode(len       & 0xFF));
+   result.push_back(Unicode(len >> 24 & 0xFF));
+   result.push_back(Unicode(len >> 16 & 0xFF));
+   result.push_back(Unicode(len >>  8 & 0xFF));
+   result.push_back(Unicode(len       & 0xFF));
 
    // nelm
    //
 const uint32_t nelm = val.element_count();
-   result.append(Unicode(nelm >> 24 & 0xFF));
-   result.append(Unicode(nelm >> 16 & 0xFF));
-   result.append(Unicode(nelm >>  8 & 0xFF));
-   result.append(Unicode(nelm       & 0xFF));
+   result.push_back(Unicode(nelm >> 24 & 0xFF));
+   result.push_back(Unicode(nelm >> 16 & 0xFF));
+   result.push_back(Unicode(nelm >>  8 & 0xFF));
+   result.push_back(Unicode(nelm       & 0xFF));
 
    // type, rank, 0, 0
    //
-   result.append(Unicode(type));
-   result.append(Unicode(val.get_rank()));
-   result.append(Unicode_0);
-   result.append(Unicode_0);
+   result.push_back(Unicode(type));
+   result.push_back(Unicode(val.get_rank()));
+   result.push_back(Unicode_0);
+   result.push_back(Unicode_0);
 
    // shape
    //
    loop(r, val.get_rank())
       {
         const uint32_t sh = val.get_shape_item(r);
-        result.append(Unicode(sh >> 24 & 0xFF));
-        result.append(Unicode(sh >> 16 & 0xFF));
-        result.append(Unicode(sh >>  8 & 0xFF));
-        result.append(Unicode(sh       & 0xFF));
+        result.push_back(Unicode(sh >> 24 & 0xFF));
+        result.push_back(Unicode(sh >> 16 & 0xFF));
+        result.push_back(Unicode(sh >>  8 & 0xFF));
+        result.push_back(Unicode(sh       & 0xFF));
       }
 
    // body
@@ -112,14 +112,14 @@ const uint32_t nelm = val.element_count();
              if (i)   accu |= 0x80 >> bit;
              if (bit == 7)
                 {
-                  result.append(Unicode(accu));
+                  result.push_back(Unicode(accu));
                   accu = 0;
                 }
            }
 
         if (nelm % 8)   // partly filled
            {
-             result.append(Unicode(accu));
+             result.push_back(Unicode(accu));
            }
 
       }
@@ -128,10 +128,10 @@ const uint32_t nelm = val.element_count();
         loop(e, nelm)
            {
              uint64_t i = val.get_ravel(e).get_int_value();
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
              Assert(i == 0 || i == 0xFFFFFFFF);
            }
       }
@@ -139,46 +139,46 @@ const uint32_t nelm = val.element_count();
       {
         loop(e, nelm)
            {
-             const double v = val.get_ravel(e).get_real_value();
-             const double * dv = &v;
-             uint64_t i = *(const uint64_t *)dv;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
+             const APL_Float v = val.get_ravel(e).get_real_value();
+             const APL_Float * dv = &v;
+             uint64_t i = *reinterpret_cast<const uint64_t *>(dv);
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
            }
       }
    else if (type == 3)   // 16 byte complex vector
       {
         loop(e, nelm)
            {
-             double v = val.get_ravel(e).get_real_value();
-             const double * dv = &v;
-             uint64_t i = *(uint64_t *)dv;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
+             APL_Float v = val.get_ravel(e).get_real_value();
+             const APL_Float * dv = &v;
+             uint64_t i = *reinterpret_cast<const uint64_t *>(dv);
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
 
              v = val.get_ravel(e).get_imag_value();
              dv = &v;
-             i = *(uint64_t *)dv;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
+             i = *reinterpret_cast<const uint64_t *>(dv);
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
            }
       }
    else if (type == 4)   // 1 byte char vector
@@ -188,7 +188,7 @@ const uint32_t nelm = val.element_count();
              const Unicode uni = val.get_ravel(e).get_char_value();
              Assert(uni >= 0);
              Assert(uni < 256);
-             result.append(uni);
+             result.push_back(uni);
            }
       }
    else if (type == 5)   // 4 byte UNICODE vector
@@ -196,10 +196,10 @@ const uint32_t nelm = val.element_count();
         loop(e, nelm)
            {
              uint32_t i = val.get_ravel(e).get_char_value();
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
-             result.append(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
+             result.push_back(Unicode(i & 0xFF));   i >>= 8;
            }
       }
    else if (type == 7)   // packed vector with 4 bytes offsets.
@@ -235,10 +235,10 @@ const uint32_t nelm = val.element_count();
         while (offset & 0x0F)   ++offset;
         loop(e, nelm)
            {
-             result.append(Unicode(offset       & 0xFF));
-             result.append(Unicode(offset >>  8 & 0xFF));
-             result.append(Unicode(offset >> 16 & 0xFF));
-             result.append(Unicode(offset >> 24 & 0xFF));
+             result.push_back(Unicode(offset       & 0xFF));
+             result.push_back(Unicode(offset >>  8 & 0xFF));
+             result.push_back(Unicode(offset >> 16 & 0xFF));
+             result.push_back(Unicode(offset >> 24 & 0xFF));
 
              const Cell & cell = val.get_ravel(e);
              if (cell.is_simple_cell())
@@ -261,7 +261,7 @@ const uint32_t nelm = val.element_count();
         // pad top level to 16 bytes. end is aligned to 16 byte so we
         // case use it as reference.
         //
-        while (result.size() & 0x0F)   result.append(Unicode_0);
+        while (result.size() & 0x0F)   result.push_back(Unicode_0);
 
         // append sub-values.
         //
@@ -295,7 +295,7 @@ const uint32_t nelm = val.element_count();
 
    // final padding
    //
-   while (result.size() & 15)   result.append(Unicode_0);
+   while (result.size() & 15)   result.push_back(Unicode_0);
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -319,7 +319,7 @@ struct tf3_header
 const uint8_t * data = cdr.get_items();
 
 const uint32_t nelm = get_4_be(data + 8);
-const CDR_type vtype = (CDR_type)(data[12]);
+const CDR_type vtype = CDR_type(data[12]);
 const Rank rank = data[13];
 Shape shape;
    loop(r, rank)
@@ -345,7 +345,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
       {
         loop(n, nelm)
            {
-             APL_Integer d = *(const uint32_t *)(ravel + 4*n);
+             APL_Integer d = *reinterpret_cast<const uint32_t *>(ravel + 4*n);
              if (d & 0x80000000)   d |= 0xFFFFFFFF00000000ULL;
              new (&ret->get_ravel(n)) IntCell(d);
            }
@@ -354,7 +354,8 @@ const uint8_t * ravel = data + 16 + 4*rank;
       {
         loop(n, nelm)
            {
-             const double v = *(const double *)(ravel + 8*n);
+             const APL_Float v =
+                  *reinterpret_cast<const APL_Float *>(ravel + 8*n);
              new (&ret->get_ravel(n)) FloatCell(v);
            }
       }
@@ -362,8 +363,10 @@ const uint8_t * ravel = data + 16 + 4*rank;
       {
         loop(n, nelm)
            {
-             const double vr = *(const double *)(ravel + 16*n);
-             const double vi = *(const double *)(ravel + 16*n + 8);
+             const APL_Float vr =
+                  *reinterpret_cast<const APL_Float *>(ravel + 16*n);
+             const APL_Float vi =
+                  *reinterpret_cast<const APL_Float *>(ravel + 16*n + 8);
              new (&ret->get_ravel(n)) ComplexCell(vr, vi);
            }
       }
@@ -380,7 +383,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
       {
         loop(n, nelm)
            {
-             const uint32_t d = *(const uint32_t *)(ravel + 4*n);
+             const uint32_t d = *reinterpret_cast<const uint32_t *>(ravel+4*n);
              new (&ret->get_ravel(n)) CharCell(Unicode(d));
            }
       }
@@ -388,7 +391,8 @@ const uint8_t * ravel = data + 16 + 4*rank;
       {
         loop(n, nelm)
             {
-              APL_Integer offset = *(const uint32_t *)(ravel + 4*n);
+              APL_Integer offset =
+                         *reinterpret_cast<const uint32_t *>(ravel + 4*n);
               const uint8_t * sub_data = data + offset;
               const uint32_t sub_vtype = sub_data[12];
               const Rank sub_rank = sub_data[13];
@@ -408,23 +412,26 @@ const uint8_t * ravel = data + 16 + 4*rank;
 
                    if (sub_vtype == 1)        // 4 byte int
                       {
-                        new (&ret->get_ravel(n))
-                            IntCell(*(uint32_t *)(sub_ravel));
+                        new (&ret->get_ravel(n)) IntCell(
+                               *reinterpret_cast<const uint32_t *>(sub_ravel));
                         continue;   // next n
                       }
 
                    if (sub_vtype == 2)        // 8 byte real
                       {
                         new (&ret->get_ravel(n))
-                            FloatCell(*(const double *)sub_ravel);
+                            FloatCell(*reinterpret_cast<const APL_Float *>
+                                                       (sub_ravel));
                         continue;   // next n
                       }
 
                    if (sub_vtype == 3)        // 16 byte complex
                       {
                         new (&ret->get_ravel(n))
-                            ComplexCell(*(const double *)sub_ravel,
-                                        *(const double *)(sub_ravel + 8));
+                            ComplexCell(*reinterpret_cast<const APL_Float *>
+                                                         (sub_ravel),
+                                        *reinterpret_cast<const APL_Float *>
+                                                         (sub_ravel + 8));
                         continue;   // next n
                       }
 
@@ -476,7 +483,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
               CDR_string sub_cdr(sub_data, sub_cdr_len);
               Value_P sub_val = from_CDR(sub_cdr, LOC);
               Assert(!!sub_val);
-              new (&ret->get_ravel(n))   PointerCell(sub_val, ret.getref());
+              new (&ret->get_ravel(n))   PointerCell(sub_val.get(), ret.getref());
             }
       }
    else
