@@ -27,10 +27,12 @@
 #include "Output.hh"
 #include "PointerCell.hh"
 #include "PrintOperator.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "SystemLimits.hh"
 #include "Workspace.hh"
 
+#include "Cell.icc"
+
 //-----------------------------------------------------------------------------
 void *
 Cell::operator new(std::size_t s, void * pos)
@@ -39,11 +41,11 @@ Cell::operator new(std::size_t s, void * pos)
 }
 //-----------------------------------------------------------------------------
 void
-Cell::init_from_value(Value_P value, Value & cell_owner, const char * loc)
+Cell::init_from_value(Value * value, Value & cell_owner, const char * loc)
 {
-   if (value->is_scalar())
+   if (value->is_simple_scalar())
       {
-        init(value->get_ravel(0), cell_owner, loc);
+        value->get_ravel(0).init_other(this, cell_owner, loc);
       }
    else
       {
@@ -62,7 +64,7 @@ Value_P ret;
    else
       {
         ret = Value_P(loc);
-        ret->get_ravel(0).init(*this, ret.getref(), loc);
+        init_other(&ret->get_ravel(0), ret.getref(), loc);
         ret->check_value(LOC);
       }
 
@@ -74,14 +76,16 @@ Cell::init_type(const Cell & other, Value & cell_owner, const char * loc)
 {
    if (other.is_pointer_cell())
       {
-        new (this) PointerCell(other.get_pointer_value()->clone(loc),
-                               cell_owner);
-        get_pointer_value()->to_proto();
+Q(LOC)
+        Value_P sub = other.get_pointer_value()->clone(loc);
+        Assert(!sub->is_simple_scalar());
+        sub->to_proto();
+        new (this) PointerCell(sub.get(), cell_owner);
       }
    else if (other.is_lval_cell())
       {
         new (this) LvalCell(other.get_lval_value(),
-                            ((const LvalCell *)&other)->get_cell_owner());
+                            other.cLvalCell().get_cell_owner());
       }
    else if (other.is_character_cell())
       {
@@ -99,7 +103,7 @@ Cell::copy(Value & val, const Cell * & src, ShapeItem count)
    loop(c, count)
       {
         Assert1(val.more());
-        val.next_ravel()->init(*src++, val, LOC);
+        src++->init_other(val.next_ravel(), val, LOC);
       }
 }
 //-----------------------------------------------------------------------------
@@ -112,7 +116,7 @@ Cell::greater(const Cell & other) const
 }
 //-----------------------------------------------------------------------------
 bool
-Cell::equal(const Cell & other, APL_Float qct) const
+Cell::equal(const Cell & other, double qct) const
 {
    MORE_ERROR() << "Cell::equal() : Objects of class " << get_classname()
                 << " cannot be compared";
@@ -120,66 +124,19 @@ Cell::equal(const Cell & other, APL_Float qct) const
 }
 //-----------------------------------------------------------------------------
 bool
-Cell::tolerantly_equal(APL_Complex A, APL_Complex B, APL_Float C)
+Cell::compare_stable(const Cell * const & A, const Cell * const & B,
+                     const void *)
 {
-   // if A equals B, return true
-   //
-   if (A == B) return true;
-
-   // if A and B are not in the same half-plane, return false.
-   //
-   // Implementation: If A and B are in the same real half-plane then
-   //                 the product of their real parts is ≥ 0,
-   //
-   //                 If A and B are in the same imag half-plane then
-   //                 the product of their imag parts is ≥ 0,
-   //
-   //                 Otherwise: they are not in the same half-plane
-   //                 and we return false;
-   //
-   if (A.real() * B.real() < 0.0 &&
-       A.imag() * B.imag() < 0.0)   return false;
-
-   // If the distance-between A and B is ≤ C times the larger-magnitude
-   // of A and B, return true
-   //
-   // Implementation: Instead of mag(A-B)  ≤ C × max(mag(A),   mag(B))
-   // we compute                 mag²(A-B) ≤ C² × max(mag²(A), mag²(B))
-   //
-   // 1. compute max(mag²A, mag²B)
-   //
-const APL_Float mag2_A   = A.real() * A.real() + A.imag() * A.imag();
-const APL_Float mag2_B   = B.real() * B.real() + B.imag() * B.imag();
-const APL_Float mag2_max = mag2_A > mag2_B ? mag2_A : mag2_B;
-
-   // 2. compute mag²(A-B)
-   //
-const APL_Complex A_B = A - B;
-const APL_Float dist2_A_B = A_B.real() * A_B.real()
-                          + A_B.imag() * A_B.imag();
-   // compare
-   //
-
-   return dist2_A_B <= C*C*mag2_max;
+const Comp_result cr = A->compare(*B);
+   if (cr == COMP_EQ)   return A < B;
+   return cr == COMP_GT;
 }
 //-----------------------------------------------------------------------------
 bool
-Cell::tolerantly_equal(APL_Float A, APL_Float B, APL_Float C)
+Cell::compare_ptr(const Cell * const & A, const Cell * const & B,
+                  const void *)
 {
-   // if the signs of A and B differ then they are unequal (ISO standard
-   // page 19). We treat exact 0.0 as having both signs
-   //
-   if (A == B)               return true;
-   if (A < 0.0 && B > 0.0)   return false;
-   if (A > 0.0 && B < 0.0)   return false;
-
-APL_Float mag_A = A < 0 ? -A : A;
-APL_Float mag_B = B < 0 ? -B : B;
-APL_Float mag_max = (mag_A > mag_B) ? mag_A : mag_B;
-
-const APL_Float dist_A_B = (A > B) ? (A - B) : (B - A);
-
-   return (dist_A_B < C*mag_max);
+   return A > B;
 }
 //-----------------------------------------------------------------------------
 bool
@@ -188,8 +145,22 @@ Cell::is_near_int(APL_Float value)
    if (value > LARGE_INT)   return true;
    if (value < SMALL_INT)   return true;
 
-const double result = nearbyint(value);
-const double diff = value - result;
+const APL_Float result = nearbyint(value);
+const APL_Float diff = value - result;
+   if (diff >= INTEGER_TOLERANCE)    return false;
+   if (diff <= -INTEGER_TOLERANCE)   return false;
+
+   return true;
+}
+//-----------------------------------------------------------------------------
+bool
+Cell::is_near_int64_t(APL_Float value)
+{
+   if (value > LARGE_INT)   return false;
+   if (value < SMALL_INT)   return false;
+
+const APL_Float result = nearbyint(value);
+const APL_Float diff = value - result;
    if (diff >= INTEGER_TOLERANCE)    return false;
    if (diff <= -INTEGER_TOLERANCE)   return false;
 
@@ -202,24 +173,24 @@ Cell::near_int(APL_Float value)
    if (value >= LARGE_INT)   DOMAIN_ERROR;
    if (value <= SMALL_INT)   DOMAIN_ERROR;
 
-const double result = nearbyint(value);
-const double diff = value - result;
-   if (diff > INTEGER_TOLERANCE)    DOMAIN_ERROR;
+const APL_Float result = nearbyint(value);
+const APL_Float diff = value - result;
+   if (diff >  INTEGER_TOLERANCE)   DOMAIN_ERROR;
    if (diff < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
 
-   if (result > 0)   return   APL_Integer(0.3 + result);
-   else              return - APL_Integer(0.3 - result);
+   if (result > 0.0)   return   APL_Integer(0.3 + result);
+   else                return - APL_Integer(0.3 - result);
 }
 //-----------------------------------------------------------------------------
 bool
 Cell::greater_vec(const IntCell & Za, const IntCell & Zb, const void * comp_arg)
 {
 struct _ctx { const Cell * base;   ShapeItem comp_len; };
-const _ctx * ctx = (const _ctx *)comp_arg;
+const _ctx * ctx = reinterpret_cast<const _ctx *>(comp_arg);
 const Cell * ca = ctx->base + ctx->comp_len * Za.get_int_value();
 const Cell * cb = ctx->base + ctx->comp_len * Zb.get_int_value();
 
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
 
    // most frequently comp_len is 1, so we optimize for this case.
    //
@@ -246,11 +217,11 @@ bool
 Cell::smaller_vec(const IntCell & Za, const IntCell & Zb, const void * comp_arg)
 {
 struct _ctx { const Cell * base;   ShapeItem comp_len; };
-const _ctx * ctx = (const _ctx *)comp_arg;
+const _ctx * ctx = reinterpret_cast<const _ctx *>(comp_arg);
 const Cell * ca = ctx->base + ctx->comp_len * Za.get_int_value();
 const Cell * cb = ctx->base + ctx->comp_len * Zb.get_int_value();
 
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
 
    // most frequently comp_len is 1, so we optimize for this case.
    //
