@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2018  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -19,6 +19,7 @@
 */
 
 #include "CharCell.hh"
+#include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
 CellType
@@ -27,40 +28,40 @@ CharCell::get_cell_subtype() const
    if (value.ival < 0)   // negative char (only fits in signed containers)
       {
         if (-value.ival <= 0x80)
-           return (CellType)(CT_CHAR | CTS_S8 | CTS_S16 | CTS_S32 | CTS_S64);
+           return CellType(CT_CHAR | CTS_S8 | CTS_S16 | CTS_S32 | CTS_S64);
 
         if (-value.ival <= 0x8000)
-           return (CellType)(CT_CHAR | CTS_S16 | CTS_S32 | CTS_S64);
+           return CellType(CT_CHAR | CTS_S16 | CTS_S32 | CTS_S64);
 
-        return (CellType)(CT_CHAR | CTS_S32 | CTS_S64);
+        return CellType(CT_CHAR | CTS_S32 | CTS_S64);
       }
 
    // positive char
    //
    if (value.ival <= 0x7F)
-      return (CellType)(CT_CHAR | CTS_X8  | CTS_S8  | CTS_U8  |
-                                  CTS_X16 | CTS_S16 | CTS_U16 |
-                                  CTS_X32 | CTS_S32 | CTS_U32 |
-                                  CTS_X64 | CTS_S64 | CTS_U64);
+      return CellType(CT_CHAR | CTS_X8  | CTS_S8  | CTS_U8  |
+                      CTS_X16 | CTS_S16 | CTS_U16 |
+                      CTS_X32 | CTS_S32 | CTS_U32 |
+                      CTS_X64 | CTS_S64 | CTS_U64);
 
    if (value.ival <= 0xFF)
-      return (CellType)(CT_CHAR |                     CTS_U8  |
-                                  CTS_X16 | CTS_S16 | CTS_U16 |
-                                  CTS_X32 | CTS_S32 | CTS_U32 |
-                                  CTS_X64 | CTS_S64 | CTS_U64);
+      return CellType(CT_CHAR |                     CTS_U8  |
+                      CTS_X16 | CTS_S16 | CTS_U16 |
+                      CTS_X32 | CTS_S32 | CTS_U32 |
+                      CTS_X64 | CTS_S64 | CTS_U64);
 
    if (value.ival <= 0x7FFF)
-      return (CellType)(CT_CHAR | CTS_X16 | CTS_S16 | CTS_U16 |
-                                  CTS_X32 | CTS_S32 | CTS_U32 |
-                                  CTS_X64 | CTS_S64 | CTS_U64);
+      return CellType(CT_CHAR | CTS_X16 | CTS_S16 | CTS_U16 |
+                      CTS_X32 | CTS_S32 | CTS_U32 |
+                      CTS_X64 | CTS_S64 | CTS_U64);
 
    if (value.ival <= 0xFFFF)
-      return (CellType)(CT_CHAR |                     CTS_U16 |
-                                  CTS_X32 | CTS_S32 | CTS_U32 |
-                                  CTS_X64 | CTS_S64 | CTS_U64);
+      return CellType(CT_CHAR |                     CTS_U16 |
+                      CTS_X32 | CTS_S32 | CTS_U32 |
+                      CTS_X64 | CTS_S64 | CTS_U64);
 
-   return (CellType)(CT_CHAR | CTS_X32 | CTS_S32 | CTS_U32 |
-                               CTS_X64 | CTS_S64 | CTS_U64);
+   return CellType(CT_CHAR | CTS_X32 | CTS_S32 | CTS_U32 |
+                   CTS_X64 | CTS_S64 | CTS_U64);
 }
 //-----------------------------------------------------------------------------
 bool
@@ -81,7 +82,7 @@ const Unicode other_val = other.get_char_value();
 }
 //-----------------------------------------------------------------------------
 bool
-CharCell::equal(const Cell & other, APL_Float qct) const
+CharCell::equal(const Cell & other, double qct) const
 {
    if (!other.is_character_cell())   return false;
    return value.aval == other.get_char_value();
@@ -109,9 +110,13 @@ UCS_string ucs;
 ColInfo info;
    info.flags |= CT_CHAR;
 
-   if (pctx.get_style() == PR_APL_FUN)
+PrintStyle style = pctx.get_style();
+Unicode uni = get_char_value();
+   if ((style & PST_PRETTY) && uni < UNI_ASCII_SPACE)
+      uni = Unicode(uni + 0x2400);
+
+   if (style == PR_APL_FUN)
       {
-        Unicode uni = get_char_value();
         if (uni == UNI_SINGLE_QUOTE)
            {
              ucs.append(UNI_SINGLE_QUOTE);
@@ -128,17 +133,16 @@ ColInfo info;
       }
    else
       {
-       if (pctx.get_style() & PST_QUOTE_CHARS)
+       if (style & PST_QUOTE_CHARS)
           {
             ucs.append(UNI_SINGLE_QUOTE);
-            ucs.append(get_char_value());
+            ucs.append(uni);
             ucs.append(UNI_SINGLE_QUOTE);
           }
         else
           {
-            ucs.append(get_char_value());
+            ucs.append(uni);
           }
-
       }
 
    info.real_len = info.int_len = ucs.size();
@@ -147,6 +151,30 @@ ColInfo info;
 }
 //-----------------------------------------------------------------------------
 int
+CharCell::get_byte_value() const
+{
+   if (value.aval < -128)
+      {
+        char cc[20];
+        snprintf(cc, sizeof(cc), "' = U+%4.4X", value.aval & 0x0FFFF);
+        MORE_ERROR() << "Unicode character '" << value.aval << cc
+                     << " is too small (< U+FF80) for a byte value";
+        DOMAIN_ERROR;
+      }
+
+   if (value.aval > 255)
+      {
+        char cc[20];
+        snprintf(cc, sizeof(cc), "' = U+%4.4X", value.aval & 0x0FFFF);
+        MORE_ERROR() << "Unicode character '" << value.aval << cc
+                     << " is too large (> U+00FF) for a byte value";
+        DOMAIN_ERROR;
+      }
+
+   return value.aval;
+}
+//-----------------------------------------------------------------------------
+int
 CharCell::CDR_size() const
 {
    // use 1 byte for small chars and 4 bytes for other UNICODE chars
@@ -157,3 +185,54 @@ const Unicode uni = get_char_value();
    return 1;
 }
 //-----------------------------------------------------------------------------
+ErrorCode
+CharCell::bif_not_bitwise(Cell * Z) const
+{
+   return zv(Z, Unicode(get_char_value() ^ 0xFFFFFFFF));
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+CharCell::bif_and_bitwise(Cell * Z, const Cell * A) const
+{
+   if (A->is_character_cell())
+     return zv(Z, Unicode(value.aval & A->get_char_value()));
+   else if (A->is_numeric())
+      return zv(Z, Unicode(value.aval & A->get_int_value()));
+   else
+      return E_DOMAIN_ERROR;
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+CharCell::bif_or_bitwise(Cell * Z, const Cell * A) const
+{
+   if (A->is_character_cell())
+     return zv(Z, Unicode(value.aval | A->get_char_value()));
+   else if (A->is_numeric())
+      return zv(Z, Unicode(value.aval | A->get_int_value()));
+   else
+      return E_DOMAIN_ERROR;
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+CharCell::bif_equal_bitwise(Cell * Z, const Cell * A) const
+{
+   if (A->is_character_cell())
+     return zv(Z, Unicode(0xFFFFFFFF & (value.aval ^ A->get_char_value())));
+   else if (A->is_numeric())
+      return zv(Z, Unicode(0xFFFFFFFF & (value.aval ^ A->get_int_value())));
+   else
+      return E_DOMAIN_ERROR;
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+CharCell::bif_not_equal_bitwise(Cell * Z, const Cell * A) const
+{
+   if (A->is_character_cell())
+     return zv(Z, Unicode(value.aval ^ A->get_char_value()));
+   else if (A->is_numeric())
+      return zv(Z, Unicode(value.aval ^ A->get_int_value()));
+   else
+      return E_DOMAIN_ERROR;
+}
+//-----------------------------------------------------------------------------
+
