@@ -21,27 +21,28 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <string.h>
-#include <sys/resource.h>
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/types.h>
 
-#include "CharCell.hh"
 #include "Common.hh"
-#include "ComplexCell.hh"
-#include "FloatCell.hh"
 #include "LineInput.hh"
 #include "LibPaths.hh"
 #include "NativeFunction.hh"
 #include "Output.hh"
+#include "Parallel.hh"
 #include "ProcessorID.hh"
+#include "Quad_WA.hh"
 #include "Svar_DB.hh"
-#include "Value.icc"
+#include "Symbol.hh"
+#include "StateIndicator.hh"
+#include "Thread_context.hh"
+#include "Token.hh"
+#include "Value.hh"
+#include "UserFunction.hh"
 #include "UserPreferences.hh"
 #include "ValueHistory.hh"
 
-uint64_t total_memory = 0;
-
 bool got_WINCH = false;
 
 //-----------------------------------------------------------------------------
@@ -80,24 +81,48 @@ bool interrupt_is_raised()
 void
 init_1(const char * argv0, bool log_startup)
 {
-rlimit rl;
-   getrlimit(RLIMIT_AS, &rl);
-   total_memory = rl.rlim_cur;
+   // init the workspace memory limits
+   //
+   Quad_WA::init(log_startup);
 
    if (log_startup)
-      CERR << "sizeof(Svar_record) is    " << sizeof(Svar_record) << endl
-           << "sizeof(Svar_partner) is   " << sizeof(Svar_partner)
-           << endl;
+      CERR << endl
+           << "sizeof(int) is            " << sizeof(int)               << endl
+           << "sizeof(long) is           " << sizeof(long)              << endl
+           << "sizeof(void *) is         " << sizeof(void *)            << endl
+           << endl
+           << "sizeof(Cell) is           " << sizeof(Cell)              << endl
+           << "sizeof(SI stack item) is  " << sizeof(StateIndicator)    << endl
+           << "sizeof(Svar_partner) is   " << sizeof(Svar_partner)      << endl
+           << "sizeof(Svar_record) is    " << sizeof(Svar_record)       << endl
+           << "sizeof(Symbol) is         " << sizeof(Symbol)            << endl
+           << "sizeof(Token) is          " << sizeof(Token)             << endl
+           << "sizeof(Value) is          " << sizeof(Value)
+           << " (including " << SHORT_VALUE_LENGTH_WANTED << " Cells)"  << endl
+           << "sizeof(ValueStackItem) is " << sizeof(ValueStackItem)    << endl
+           << "sizeof(UCS_string) is     " << sizeof(UCS_string)        << endl
+           << "sizeof(UserFunction) is   " << sizeof(UserFunction)      << endl
+           << endl
+           << "⎕WA total memory is       " << Quad_WA::total_memory
+           << " bytes (" << (Quad_WA::total_memory/1000000) << " MB, 0x"
+           << hex << Quad_WA::total_memory << ")" << dec << endl;
 
    // CYGWIN does not have RLIMIT_NPROC
    //
 #ifdef RLIMIT_NPROC
+
+   // unlimit the number of threads and processes...
+   //
+rlimit rl;
    getrlimit(RLIMIT_NPROC, &rl);
    if (log_startup)
       CERR << "increasing rlimit RLIMIT_NPROC from " <<  rl.rlim_cur
            << " to infinity" << endl;
    rl.rlim_cur = RLIM_INFINITY;
    setrlimit(RLIMIT_NPROC, &rl);
+
+   // limit the virtual memory size to avoid new() problem with large values
+   //
 #endif
 
    Avec::init();
@@ -136,6 +161,10 @@ cleanup(bool soft)
         LineInput::close(false);
 
         Output::reset_colors();
+
+        Thread_context::cleanup();
+
+        ID::cleanup();
       }
    else        // minimal clean-up
       {
@@ -180,7 +209,8 @@ void *
 common_new(size_t size)
 {
 void * ret = malloc(size);
-   CERR << "NEW " << HEX(ret) << "-" << HEX((const char *)ret + size)
+const uint64_t iret = uint64_t(ret);
+   CERR << "NEW " << HEX(iret) << "-" << HEX(iret + size)
         << "  (" << HEX(size) << ")" << endl;
    return ret;
 }
@@ -188,7 +218,7 @@ void * ret = malloc(size);
 void
 common_delete(void * p)
 {
-   CERR << "DEL " << HEX(p) << endl;
+   CERR << "DEL " << HEX(uint64_t(p)) << endl;
    free(p);
 }
 //-----------------------------------------------------------------------------
@@ -287,81 +317,3 @@ sixbit(Unicode uni)
    return -1;   // uni is not a hex digit
 }
 //-----------------------------------------------------------------------------
-Value_P
-IntScalar(APL_Integer val, const char * loc)
-{
-Value_P Z(loc);
-   new (Z->next_ravel())   IntCell(val);
-   Z->check_value(LOC);
-   return Z;
-}
-//-----------------------------------------------------------------------------
-Value_P
-FloatScalar(APL_Float val, const char * loc)
-{
-Value_P Z(loc);
-   new (Z->next_ravel())   FloatCell(val);
-   Z->check_value(LOC);
-   return Z;
-}
-//-----------------------------------------------------------------------------
-Value_P
-ComplexScalar(APL_Complex val, const char * loc)
-{
-Value_P Z(loc);
-   new (Z->next_ravel()) ComplexCell(val);
-   Z->check_value(LOC);
-   return Z;
-}
-//-----------------------------------------------------------------------------
-Value_P
-CharScalar(Unicode uni, const char * loc)
-{
-Value_P Z(loc);
-   new (Z->next_ravel()) CharCell(uni);
-   Z->check_value(LOC);
-   return Z;
-}
-//-----------------------------------------------------------------------------
-Value_P
-Idx0(const char * loc)
-{
-Value_P Z((ShapeItem)0, loc);
-   Z->check_value(LOC);
-   return Z;
-}
-//-----------------------------------------------------------------------------
-Value_P
-Str0(const char * loc)
-{
-Value_P Z((ShapeItem)0, loc);
-   Z->set_proto_Spc();
-   Z->check_value(LOC);
-   return Z;
-}
-//-----------------------------------------------------------------------------
-Value_P
-Str0_0(const char * loc)
-{
-Shape sh((ShapeItem)0, (ShapeItem)0);
-Value_P Z(sh, loc);
-   Z->set_proto_Spc();
-   Z->check_value(LOC);
-   return Z;
-}
-//-----------------------------------------------------------------------------
-Value_P
-Idx0_0(const char * loc)
-{
-Shape sh((ShapeItem)0, (ShapeItem)0);
-Value_P Z(sh, loc);
-   Z->check_value(LOC);
-   return Z;
-}
-//-----------------------------------------------------------------------------
-ostream & operator << (ostream & out, const AP_num3 & ap3)
-{
-   return out << ap3.proc << "." << ap3.parent << "." << ap3.grand;
-}
-//-----------------------------------------------------------------------------
-
