@@ -24,6 +24,7 @@
 #include <stdio.h>
 #include <string.h>
 
+#include "Common.hh"
 #include "ComplexCell.hh"
 #include "ErrorCode.hh"
 #include "FloatCell.hh"
@@ -31,73 +32,129 @@
 #include "Output.hh"
 #include "Workspace.hh"
 
+#include "Cell.icc"
+
 //-----------------------------------------------------------------------------
 ComplexCell::ComplexCell(APL_Complex c)
 {
-   value.cval_r  = c.real();
-   value2.cval_i = c.imag();
+   value.cval[0]  = c.real();
+   value.cval[1] = c.imag();
 }
 //-----------------------------------------------------------------------------
 ComplexCell::ComplexCell(APL_Float r, APL_Float i)
 {
-   value.cval_r  = r;
-   value2.cval_i = i;
+   value.cval[0] = r;
+   value.cval[1] = i;
 }
 //-----------------------------------------------------------------------------
 APL_Float
 ComplexCell::get_real_value() const
 {
-   return value.cval_r;
+   return value.cval[0];
 }
 //-----------------------------------------------------------------------------
 APL_Float
 ComplexCell::get_imag_value() const
 {
-   return value2.cval_i;
+   return value.cval[1];
 }
 //-----------------------------------------------------------------------------
 bool
 ComplexCell::is_near_int() const
 {
-   return Cell::is_near_int(value.cval_r) &&
-          Cell::is_near_int(value2.cval_i);
+   return Cell::is_near_int(value.cval[0]) &&
+          Cell::is_near_int(value.cval[1]);
+}
+//-----------------------------------------------------------------------------
+bool
+ComplexCell::is_near_int64_t() const
+{
+   return Cell::is_near_int64_t(value.cval[0]) &&
+          Cell::is_near_int64_t(value.cval[1]);
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+ComplexCell::bif_near_int64_t(Cell * Z) const
+{
+   if (!is_near_int64_t())   return E_DOMAIN_ERROR;
+
+   // return an int Cell if the imaginary part is 0 and the number is not too
+   // large
+   //
+   if (value.cval[1] <  INTEGER_TOLERANCE &&
+       value.cval[1] > -INTEGER_TOLERANCE &&
+       value.cval[0] <  BIG_INT64_F       &&
+       value.cval[0] > -BIG_INT64_F)   return zv(Z, round(value.cval[0]));
+
+   return zv(Z, round(value.cval[0]), round(value.cval[1]));
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+ComplexCell::bif_within_quad_CT(Cell * Z) const
+{
+const double val_r = value.cval[0];
+   if (val_r > LARGE_INT)   return E_DOMAIN_ERROR;
+   if (val_r < SMALL_INT)   return E_DOMAIN_ERROR;
+
+const double val_i = value.cval[1];
+   if (val_i > LARGE_INT)   return E_DOMAIN_ERROR;
+   if (val_i < SMALL_INT)   return E_DOMAIN_ERROR;
+
+const double max_diff_r = Workspace::get_CT() * val_r;   // scale ⎕CT
+const double max_diff_i = Workspace::get_CT() * val_i;   // scale ⎕CT
+
+const APL_Float val_dn_r = floor(val_r);
+const APL_Float val_up_r = ceil(val_r);
+const APL_Float val_dn_i = floor(val_i);
+const APL_Float val_up_i = ceil(val_i);
+
+double z_r;
+   if (val_r < (val_dn_r + max_diff_r))        z_r = val_dn_r;
+   else if (val_r < (val_up_r - max_diff_r))   z_r = val_up_r;
+   else                                        return E_DOMAIN_ERROR;
+
+double z_i;
+   if (val_i < (val_dn_i + max_diff_i))        z_i = val_dn_i;
+   else if (val_i < (val_up_i - max_diff_i))   z_i = val_up_i;
+   else                                        return E_DOMAIN_ERROR;
+
+   return zv(Z, z_r, z_i);
 }
 //-----------------------------------------------------------------------------
 bool
 ComplexCell::is_near_zero() const
 {
-   if (value.cval_r  >=  INTEGER_TOLERANCE)   return false;
-   if (value.cval_r  <= -INTEGER_TOLERANCE)   return false;
-   if (value2.cval_i >=  INTEGER_TOLERANCE)   return false;
-   if (value2.cval_i <= -INTEGER_TOLERANCE)   return false;
+   if (value.cval[0]  >=  INTEGER_TOLERANCE)   return false;
+   if (value.cval[0]  <= -INTEGER_TOLERANCE)   return false;
+   if (value.cval[1] >=  INTEGER_TOLERANCE)   return false;
+   if (value.cval[1] <= -INTEGER_TOLERANCE)   return false;
    return true;
 }
 //-----------------------------------------------------------------------------
 bool
 ComplexCell::is_near_one() const
 {
-   if (value.cval_r >  (1.0 + INTEGER_TOLERANCE))   return false;
-   if (value.cval_r <  (1.0 - INTEGER_TOLERANCE))   return false;
-   if (value2.cval_i >  INTEGER_TOLERANCE)           return false;
-   if (value2.cval_i < -INTEGER_TOLERANCE)           return false;
+   if (value.cval[0] >  (1.0 + INTEGER_TOLERANCE))   return false;
+   if (value.cval[0] <  (1.0 - INTEGER_TOLERANCE))   return false;
+   if (value.cval[1] >  INTEGER_TOLERANCE)           return false;
+   if (value.cval[1] < -INTEGER_TOLERANCE)           return false;
    return true;
 }
 //-----------------------------------------------------------------------------
 bool
 ComplexCell::is_near_real() const
 {
-const APL_Float B = REAL_TOLERANCE*REAL_TOLERANCE;
-const double I = value2.cval_i * value2.cval_i;
+const APL_Float B2 = REAL_TOLERANCE*REAL_TOLERANCE;
+const APL_Float I2 = value.cval[1] * value.cval[1];
 
-   if (I < B)     return true;   // I is absolutely small
+   if (I2 < B2)     return true;   // I is absolutely small
 
-const double R = value.cval_r * value.cval_r;
-   if (I < R*B)   return true;   // I is relatively small
-   return false;
+const APL_Float R2 = value.cval[0] * value.cval[0];
+   return (I2 < R2*B2);   // I is relatively small
 }
 //-----------------------------------------------------------------------------
 bool
-ComplexCell::equal(const Cell & A, APL_Float qct) const
+ComplexCell::equal(const Cell & A, double qct) const
 {
    if (!A.is_numeric())   return false;
    return tolerantly_equal(A.get_complex_value(), get_complex_value(), qct);
@@ -111,17 +168,13 @@ ComplexCell::bif_factorial(Cell * Z) const
    if (is_near_real())
       {
         const FloatCell fc(get_real_value());
-        fc.bif_factorial(Z);
-        return E_NO_ERROR;
+        return fc.bif_factorial(Z);
       }
 
-const APL_Float zr = get_real_value() + 1.0;
-const APL_Float zi = get_imag_value();
-const APL_Complex z(zr, zi);
-
-   new (Z) ComplexCell(gamma(get_real_value(), get_imag_value()));
+ErrorCode ret = ComplexCell::zv(Z, gamma(get_real_value() + 1.0,
+                                         get_imag_value()));
    if (errno)   return E_DOMAIN_ERROR;
-   return E_NO_ERROR;
+   return ret;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -142,16 +195,15 @@ ErrorCode
 ComplexCell::bif_direction(Cell * Z) const
 {
 const APL_Float mag = abs(cval());
-   if (mag == 0.0)   new (Z) IntCell(0);
-   else              new (Z) ComplexCell(cval()/mag);
-   return E_NO_ERROR;
+   if (mag == 0.0)   return IntCell::zv(Z, 0);
+   else              return ComplexCell::zv(Z, get_real_value()/mag,
+                                               get_imag_value()/mag);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_magnitude(Cell * Z) const
 {
-   new (Z) FloatCell(abs(cval()));
-   return E_NO_ERROR;
+   return FloatCell::zv(Z, abs(cval()));
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -161,14 +213,19 @@ ComplexCell::bif_reciprocal(Cell * Z) const
 
    if (is_near_real())
       {
-        const APL_Float z = 1.0/value.cval_r;
+        const APL_Float z = 1.0/value.cval[0];
         if (!isfinite(z))   return E_DOMAIN_ERROR;
         return FloatCell::zv(Z, z);
       }
 
-const APL_Complex z = 1.0 / cval();
-   if (!isfinite(z.real()))   return E_DOMAIN_ERROR;
-   if (!isfinite(z.imag()))   return E_DOMAIN_ERROR;
+const APL_Float denom = mag2();
+const APL_Float r = value.cval[0]/denom;
+   if (!isfinite(r))   return E_DOMAIN_ERROR;
+
+const APL_Float i = value.cval[1]/denom;
+   if (!isfinite(i))   return E_DOMAIN_ERROR;
+
+const APL_Complex z(r, -i);
    return zv(Z, z);
 }
 //-----------------------------------------------------------------------------
@@ -182,31 +239,31 @@ const APL_Integer set_size = get_checked_near_int();
    if (set_size <= 0)   return E_DOMAIN_ERROR;
 
 const uint64_t rnd = Workspace::get_RL(set_size);
-   new (Z) IntCell(qio + (rnd % set_size));
-   return E_NO_ERROR;
+   return IntCell::zv(Z, qio + (rnd % set_size));
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_pi_times(Cell * Z) const
 {
-   new (Z) ComplexCell(M_PI * cval());
+APL_Float pi(M_PI);
+   return zv(Z, value.cval[0] * pi, value.cval[1] * pi);
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_pi_times_inverse(Cell * Z) const
 {
-   new (Z) ComplexCell(cval() / M_PI);
-   return E_NO_ERROR;
+APL_Float pi(M_PI);
+   return zv(Z, value.cval[0] / pi, value.cval[1] / pi);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_ceiling(Cell * Z) const
 {
-const APL_Float cr = ceil(value.cval_r);    // cr ≥ value.cval_r
-const APL_Float Dr = cr - value.cval_r;     // Dr ≥ 0
-const APL_Float ci = ceil(value2.cval_i);   // ci ≥ value2.cval_i
-const APL_Float Di = ci - value2.cval_i;    // Di ≥ 0
+const APL_Float cr = ceil(value.cval[0]);    // cr ≥ value.cval[0]
+const APL_Float Dr = cr - value.cval[0];     // Dr ≥ 0
+const APL_Float ci = ceil(value.cval[1]);   // ci ≥ value.cval[1]
+const APL_Float Di = ci - value.cval[1];    // Di ≥ 0
 const APL_Float D = Dr + Di;                // 0 ≤ D < 2
 
    // ISO: if D is tolerantly less than 1 return fr + 0J1×fi
@@ -218,34 +275,49 @@ const APL_Float D = Dr + Di;                // 0 ≤ D < 2
 
    if (Di > Dr)   return zv(Z, cr, ci - 1.0);
    else           return zv(Z, cr - 1.0, ci);
-
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_floor(Cell * Z) const
 {
-const APL_Float fr = floor(value.cval_r);    // fr ≤ value.cval_r
-const APL_Float Dr = value.cval_r - fr;      // Dr ≥ 0
-const APL_Float fi = floor(value2.cval_i);   // fi ≤ value2.cval_i
-const APL_Float Di = value2.cval_i - fi;     // Di ≥ 0
-const APL_Float D = Dr + Di;                 // 0 ≤ D < 2
+APL_Float fr = floor(value.cval[0]);    // fr ≤ value.cval[0]
+APL_Float Dr = value.cval[0] - fr;      // 0 ≤ Dr < 1
+APL_Float fi = floor(value.cval[1]);   // fi ≤ value.cval[1]
+APL_Float Di = value.cval[1] - fi;     // 0 ≤ Di < 1
+const double qct = Workspace::get_CT();   // ⎕CT
+const double limit = 1.0 - qct;
 
-   // ISO: if D is tolerantly less than 1 return fr + 0J1×fi
-   // IBM: if D is            less than 1 return fr + 0J1×fi
-   // However, ISO examples follow IBM (and so do we)
-   //
-// if (D < 1.0 + Workspace::get_CT())   return zv(Z, fr, fi);   // ISO
-   if (D < 1.0)    return zv(Z, fr, fi);   // IBM and examples in ISO
+   if (Dr > limit)
+      {
+        fr = fr + 1.0;
+        Dr = 0.0;
+      }
 
-   if (Di > Dr)   return zv(Z, fr, fi + 1.0);
-   else           return zv(Z, fr + 1.0, fi);
+   if (Di > limit)
+      {
+        fi = fi + 1.0;
+        Di = 0.0;
+      }
+
+   if ((Dr + Di) < limit)
+      {
+        return zv(Z, fr, fi);
+      }
+
+   if (Dr < (Di - qct))
+      {
+        return zv(Z, fr, fi + 1.0);
+      }
+   else
+      {
+        return zv(Z, fr + 1.0, fi);
+      }
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_exponential(Cell * Z) const
 {
-   new (Z) ComplexCell(exp(cval()));
-   return E_NO_ERROR;
+   return ComplexCell::zv(Z, complex_exponent(cval()));
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -260,15 +332,19 @@ ComplexCell::bif_nat_log(Cell * Z) const
 ErrorCode
 ComplexCell::bif_add(Cell * Z, const Cell * A) const
 {
-   new (Z) ComplexCell(A->get_complex_value() + get_complex_value());
-   return E_NO_ERROR;
+   return ComplexCell::zv(Z, A->get_complex_value() + get_complex_value());
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+ComplexCell::bif_add_inverse(Cell * Z, const Cell * A) const
+{
+   return A->bif_subtract(Z, this);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_subtract(Cell * Z, const Cell * A) const
 {
-   new (Z) ComplexCell(A->get_complex_value() - get_complex_value());
-   return E_NO_ERROR;
+   return ComplexCell::zv(Z, A->get_complex_value() - get_complex_value());
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -283,13 +359,19 @@ const APL_Complex z = A->get_complex_value() * cval();
 }
 //-----------------------------------------------------------------------------
 ErrorCode
+ComplexCell::bif_multiply_inverse(Cell * Z, const Cell * A) const
+{
+   return A->bif_divide(Z, this);
+}
+//-----------------------------------------------------------------------------
+ErrorCode
 ComplexCell::bif_divide(Cell * Z, const Cell * A) const
 {
    if (cval().real() == 0.0 && cval().imag() == 0.0)   // A ÷ 0
       {
         if (A->get_real_value() != 0.0)   return E_DOMAIN_ERROR;
         if (A->get_imag_value() != 0.0)   return E_DOMAIN_ERROR;
-        new (Z) IntCell(1);   // 0÷0 is 1 in APL
+        return IntCell::zv(Z, 1);   // 0÷0 is 1 in APL
       }
 
    new (Z) ComplexCell(A->get_complex_value() / get_complex_value());
@@ -301,43 +383,31 @@ ComplexCell::bif_residue(Cell * Z, const Cell * A) const
 {
    if (!A->is_numeric())   return E_DOMAIN_ERROR;
 
-const APL_Complex a = A->get_complex_value();
+const APL_Complex mod = A->get_complex_value();
+const APL_Complex b = get_complex_value();
 
    // if A is zero , return B
    //
-   if (a.real() == 0.0 && a.imag() == 0.0)
-      return zv(Z, value.cval_r, value2.cval_i);
+   if (mod.real() == 0.0 && mod.imag() == 0.0)
+      return zv(Z, b);
 
-   // IBM: if B is zero , return B
+   // IBM: if B is zero , return 0
    //
-   if (value.cval_r == 0.0 && value2.cval_i == 0.0)   return IntCell::z0(Z);
+   if (b.real() == 0.0 && b.imag() == 0.0)
+      return IntCell::z0(Z);
 
-   // if ⎕CT != 0 and B ÷ A is close to an integer within ⎕CT then return 0.
+   // compliant implementation: B-A×⌊B÷A+A=0
+   // The b=0 case may still exist due to ⎕CT
    //
-   // Note: In that case, the integer to which A ÷ B is close is either
-   // floor(A ÷ B) or ceil(A ÷ B).
-   //
-const APL_Float qct = Workspace::get_CT();
-   if (qct != 0)
-      {
-        const APL_Complex quot = cval() / a;
-        const APL_Float qfr = floor(quot.real());
-        const APL_Float qfi = floor(quot.imag());
-        const APL_Float qcr = ceil(quot.real());
-        const APL_Float qci = ceil(quot.imag());
-
-       if (quot.real() > (qcr - qct) && quot.imag() > (qci - qct)) 
-          return IntCell::z0(Z);   // quot is close to its ceiling
-
-       if (quot.real() < (qfr - qct) && quot.imag() < (qfi - qct)) 
-          return IntCell::z0(Z);   // quot is close to its floor
-      }
-
-   // We divide A by B, round down the result, and subtract.
-   //
-   new (Z) ComplexCell(cval() / a);
-   Z->bif_floor(Z);            // Z = A/B rounded down.
-   return zv(Z, cval() - a * Z->get_complex_value());
+   //                          	// op       Z before     Z after
+   new (Z) IntCell(0);          // Z←0      any          0
+   Z->bif_equal(Z, A);          // Z←A=Z    0            A=0
+   Z->bif_add(Z, A);           	// Z←A+Z    A=0          A+A=0
+   Z->bif_divide(Z, this);      // Z←B÷Z    A+A=0        B÷A+A=0
+   Z->bif_floor(Z);             // Z←⌊Z     B÷A+A=0      ⌊B÷A+A=0
+   Z->bif_multiply(Z, A);       // Z←A×Z    ⌊B÷A+A=0     A×⌊B÷A+A=0
+   Z->bif_subtract(Z, this);    // Z←A-Z    A×⌊B÷A+A=0   B-A×⌊B÷A+A=0
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -346,31 +416,22 @@ ComplexCell::bif_maximum(Cell * Z, const Cell * A) const
    // maximum of complex numbers gives DOMAN ERROR if one of the cells
    // is not near-real.
    //
-   if (!is_near_real())      return E_DOMAIN_ERROR;
+   if (!is_near_real())         return E_DOMAIN_ERROR;
+   if (!A->is_near_real())      return E_DOMAIN_ERROR;
 
 const APL_Float breal = get_real_value();
 
    if (A->is_integer_cell())
       {
-         const APL_Integer a = A->get_int_value();
-         if (APL_Float(a) >= breal)   new (Z) IntCell(a);
-         else                        new (Z) FloatCell(breal);
-      }
-   else if (A->is_float_cell())
-      {
-         const APL_Float a = A->get_real_value();
-         if (a >= breal)   new (Z) FloatCell(a);
-         else             new (Z) FloatCell(breal);
-      }
-   else
-      {
-        if (!A->is_near_real())   return E_DOMAIN_ERROR;
-        const APL_Float a = A->get_real_value();
-        if (a >= breal)   new (Z) FloatCell(a);
-        else             new (Z) FloatCell(breal);
+        const APL_Integer a = A->get_int_value();
+        return a >= breal ? IntCell::zv(Z, a) : FloatCell::zv(Z, breal);
       }
 
-   return E_NO_ERROR;
+   // both A and B are near-real, therefore they must be numeric and have no
+   // non-0 imag part
+   //
+const APL_Float a = A->get_real_value();
+   return FloatCell::zv(Z, a >= breal ? a : breal);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -380,59 +441,49 @@ ComplexCell::bif_minimum(Cell * Z, const Cell * A) const
    // is not near real.
    //
    if (!is_near_real())      return E_DOMAIN_ERROR;
+   if (!A->is_near_real())   return E_DOMAIN_ERROR;
 
 const APL_Float breal = get_real_value();
 
    if (A->is_integer_cell())
       {
-         const APL_Integer a = A->get_int_value();
-         if (APL_Float(a) <= breal)   new (Z) IntCell(a);
-         else                        new (Z) FloatCell(breal);
+        const APL_Integer a = A->get_int_value();
+        return a <= breal ? IntCell::zv(Z, a) : FloatCell::zv(Z, breal);
       }
-   else if (A->is_float_cell())
-      {
-         const APL_Float a = A->get_real_value();
-         if (a <= breal)   new (Z) FloatCell(a);
-         else             new (Z) FloatCell(breal);
-      }
-   else
-      {
-        if (!A->is_near_real())   return E_DOMAIN_ERROR;
 
-        const APL_Float a = A->get_real_value();
-        if (a <= breal)   new (Z) FloatCell(a);
-        else             new (Z) FloatCell(breal);
-      }
-   return E_NO_ERROR;
+   // both A and B are near-real, therefore they must be numeric and have no
+   // non-0 imag part
+   //
+const APL_Float a = A->get_real_value();
+   return FloatCell::zv(Z, a <= breal ? a : breal);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_equal(Cell * Z, const Cell * A) const
 {
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
    if (A->is_complex_cell())
       {
         const APL_Complex diff = A->get_complex_value() - get_complex_value();
-        if      (diff.real()      >  qct)   new (Z) IntCell(0);
-        else if (diff.real()      < -qct)   new (Z) IntCell(0);
-        else if (diff.imag()      >  qct)   new (Z) IntCell(0);
-        else if (diff.imag()      < -qct)   new (Z) IntCell(0);
-        else                                new (Z) IntCell(1);
+        if (diff.real() >  qct)   return IntCell::zv(Z, 0);
+        if (diff.real() < -qct)   return IntCell::zv(Z, 0);
+        if (diff.imag() >  qct)   return IntCell::zv(Z, 0);
+        if (diff.imag() < -qct)   return IntCell::zv(Z, 0);
+        return IntCell::zv(Z, 1);
       }
-   else if (A->is_numeric())
+
+   if (A->is_numeric())
       {
+        if (get_imag_value() >  qct)   return IntCell::zv(Z, 0);
+        if (get_imag_value() < -qct)   return IntCell::zv(Z, 0);
+
         const APL_Float diff = A->get_real_value() - get_real_value();
-        if      (diff             >  qct)   new (Z) IntCell(0);
-        else if (diff             < -qct)   new (Z) IntCell(0);
-        else if (get_imag_value() >  qct)   new (Z) IntCell(0);
-        else if (get_imag_value() < -qct)   new (Z) IntCell(0);
-        else                                new (Z) IntCell(1);
+        if (diff >  qct)   return IntCell::zv(Z, 0);
+        if (diff < -qct)   return IntCell::zv(Z, 0);
+        return IntCell::zv(Z, 1);
       }
-   else
-      {
-        new (Z) IntCell(0);
-      }
-   return E_NO_ERROR;
+
+   return IntCell::zv(Z, 1);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -450,14 +501,14 @@ const APL_Float ai = A->get_imag_value();
    if (ar == 0.0 && ai == 0.0)
        {
          if (cval().real() == 0.0)   return IntCell::z1(Z);   // 0⋆0 is 1
-         if (cval().imag() > 0)      return IntCell::z0(Z);   // 0⋆N is 0
+         if (cval().imag()  > 0.0)   return IntCell::z0(Z);   // 0⋆N is 0
          return E_DOMAIN_ERROR;                               // 0⋆¯N = 1÷0
        }
 
    // 2. complex result
    {
      const APL_Complex a(ar, ai);
-     const APL_Complex z = pow(a, cval());
+     const APL_Complex z = complex_power(a, cval());
      if (!isfinite(z.real()))   return E_DOMAIN_ERROR;
      if (!isfinite(z.imag()))   return E_DOMAIN_ERROR;
 
@@ -477,12 +528,14 @@ ComplexCell::bif_logarithm(Cell * Z, const Cell * A) const
        get_imag_value() == A->get_imag_value())   return IntCell::z1(Z);
 
 
-   if (value.cval_r == 0.0 && value2.cval_i == 0.0)   return E_DOMAIN_ERROR;
+   if (value.cval[0] == 0.0 && value.cval[1] == 0.0)   return E_DOMAIN_ERROR;
    if (A->is_near_one())                              return E_DOMAIN_ERROR;
 
    if (A->is_real_cell())
       {
-        const APL_Complex z = log(cval()) / log(A->get_real_value());
+        const APL_Float lg_a = log(A->get_real_value());
+        const APL_Complex lg_z = log(cval());
+        const APL_Complex z(lg_z.real() / lg_a, lg_z.imag() / lg_a);
         if (!isfinite(z.real()))   return E_DOMAIN_ERROR;
         if (!isfinite(z.imag()))   return E_DOMAIN_ERROR;
         return zv(Z, z);
@@ -549,6 +602,7 @@ ErrorCode ret = E_DOMAIN_ERROR;
 ErrorCode
 ComplexCell::do_bif_circle_fun(Cell * Z, int fun, APL_Complex b)
 {
+const APL_Complex one(1.0, 0.0);
    switch(fun)
       {
         case -12: {
@@ -562,15 +616,22 @@ ComplexCell::do_bif_circle_fun(Cell * Z, int fun, APL_Complex b)
         case -10: return zv(Z, b.real(), -b.imag());
 
         case  -9: return zv(Z,             b   );
+        case  -8: { const APL_Complex par = -(b*b + one);     // (¯1 - R⋆2)
+                    const APL_Complex sq = complex_sqrt(par); // (¯1 + R⋆2)⋆0.5
+                    if ((b.real()  > 0.0 && b.imag() > 0.0) ||
+                        (b.real() == 0.0 && b.imag() > 1.0) ||
+                        (b.real()  < 0.0 && b.imag() >= 0.0)) return zv(Z,  sq);
+                    else                                      return zv(Z, -sq);
+                  }
 
-        case  -7: // arctanh(z) = 0.5 (ln(1 + z) - ln(1 - z))
+        case  -7: // arctanh(z) = 0.5 (ln(1.0 + z) - ln(1.0 - z))
                   {
                     const APL_Complex b1      = ONE() + b;
                     const APL_Complex b_1     = ONE() - b;
                     const APL_Complex log_b1  = log(b1);
                     const APL_Complex log_b_1 = log(b_1);
                     const APL_Complex diff    = log_b1 - log_b_1;
-                    const APL_Complex half    = 0.5 * diff;
+                    const APL_Complex half(diff.real()*0.5, diff.imag()*0.5);
                     return zv(Z, half);
                   }
 
@@ -578,8 +639,8 @@ ComplexCell::do_bif_circle_fun(Cell * Z, int fun, APL_Complex b)
                   {
                     const APL_Complex b1     = b + ONE();
                     const APL_Complex b_1    = b - ONE();
-                    const APL_Complex root1  = sqrt(b1);
-                    const APL_Complex root_1 = sqrt(b_1);
+                    const APL_Complex root1  = complex_sqrt(b1);
+                    const APL_Complex root_1 = complex_sqrt(b_1);
                     const APL_Complex prod   = root1 * root_1;
                     const APL_Complex sum    = b + prod;
                     const APL_Complex loga   = log(sum);
@@ -590,16 +651,16 @@ ComplexCell::do_bif_circle_fun(Cell * Z, int fun, APL_Complex b)
                   {
                     const APL_Complex b2 = b*b;
                     const APL_Complex b2_1 = b2 + ONE();
-                    const APL_Complex root = sqrt(b2_1);
+                    const APL_Complex root = complex_sqrt(b2_1);
                     const APL_Complex sum =  b + root;
                     const APL_Complex loga = log(sum);
                     return zv(Z, loga);
                   }
 
-        case  -4: if (b.real() >= 0 ||
-                      (b.real() > -1 && Cell::is_near_zero(b.imag()))
-                     )   return zv(Z,  sqrt(b*b - 1.0));
-                  else   return zv(Z, -sqrt(b*b - 1.0));
+        case  -4: if (b.real() >= 0.0 ||
+                      (b.real() > -1.0 && Cell::is_near_zero(b.imag()))
+                     )   return zv(Z,  complex_sqrt(b*b - one));
+                  else   return zv(Z, -complex_sqrt(b*b - one));
 
         case  -3: // arctan(z) = i/2 (ln(1 - iz) - ln(1 + iz))
                   {
@@ -617,7 +678,7 @@ ComplexCell::do_bif_circle_fun(Cell * Z, int fun, APL_Complex b)
                   {
                     const APL_Complex b2 = b*b;
                     const APL_Complex diff = b2 - ONE();
-                    const APL_Complex root = sqrt(diff);
+                    const APL_Complex root = complex_sqrt(diff);
                     const APL_Complex sum = b + root;
                     const APL_Complex loga = log(sum);
                     const APL_Complex prod = MINUS_i() * loga;
@@ -628,7 +689,7 @@ ComplexCell::do_bif_circle_fun(Cell * Z, int fun, APL_Complex b)
                   {
                     const APL_Complex b2 = b*b;
                     const APL_Complex diff = ONE() - b2;
-                    const APL_Complex root = sqrt(diff);
+                    const APL_Complex root = complex_sqrt(diff);
                     const APL_Complex sum  = APL_Complex(-b.imag(), b.real())
                                            + root;
                     const APL_Complex loga = log(sum);
@@ -636,7 +697,7 @@ ComplexCell::do_bif_circle_fun(Cell * Z, int fun, APL_Complex b)
                     return zv(Z, prod);
                   }
 
-        case   0: return zv(Z, sqrt(1.0 - b*b));
+        case   0: return zv(Z, complex_sqrt(one - b*b));
 
         case   1: return zv(Z, sin       (b));
 
@@ -644,7 +705,7 @@ ComplexCell::do_bif_circle_fun(Cell * Z, int fun, APL_Complex b)
 
         case   3: return zv(Z, tan       (b));
 
-        case   4: return zv(Z, sqrt(1.0 + b*b));
+        case   4: return zv(Z, complex_sqrt(one + b*b));
 
         case   5: return zv(Z, sinh      (b));
 
@@ -652,29 +713,18 @@ ComplexCell::do_bif_circle_fun(Cell * Z, int fun, APL_Complex b)
 
         case   7: return zv(Z, tanh      (b));
 
-        case  -8: // ¯8○B is - 8○B
-        case   8: {
-                    // Let b = X + iY and sq = (¯1 - R*2)*.5
-                    // Then 8○X+JY:  Z is: + sq  if  X > 0 and Y > 0
-                    //                           or  X = 0 and Y > 1
-                    //                           or  X < 0 and Y ≥ 0, and
-                    //                     - sq  otherwise
-                    bool pos_8 = false;
-                    if      (b.real() >  0) { if (b.imag() > 0)  pos_8 = true; }
-                    else if (b.real() == 0) { if (b.imag() > 1)  pos_8 = true; }
-                    else                    { if (b.imag() >= 0) pos_8 = true; }
-
-                    if (fun == 8)   pos_8 = ! pos_8;
-
-                    const APL_Complex sq = sqrt(-(1.0 + b*b));  // (¯1 - R*2)*.5
-
-                    return zv(Z, pos_8 ? sq : -sq);
+        case   8: { const APL_Complex par = -(b*b + one);      // (¯1 - R⋆2)
+                    const APL_Complex sq = complex_sqrt(par);  // (¯1 + R⋆2)⋆0.5
+                    if ((b.real()  > 0.0 && b.imag() > 0.0) ||
+                        (b.real() == 0.0 && b.imag() > 1.0) ||
+                        (b.real()  < 0.0 && b.imag() >= 0.0)) return zv(Z, -sq);
+                    else                                      return zv(Z,  sq);
                   }
 
         case   9: return FloatCell::zv(Z,      b.real());
 
-        case  10: return FloatCell::zv(Z, sqrt(b.real()*b.real()
-                                             + b.imag()* b.imag()));
+        case  10: return FloatCell::zv(Z, sqrt(mag2(b)));
+
         case  11: return FloatCell::zv(Z, b.imag());
 
         case  12: return FloatCell::zv(Z, atan2(b.imag(), b.real()));
@@ -686,10 +736,11 @@ ComplexCell::do_bif_circle_fun(Cell * Z, int fun, APL_Complex b)
 }
 //-----------------------------------------------------------------------------
 APL_Complex
-ComplexCell::gamma(APL_Float x, const APL_Float y)
+ComplexCell::gamma(APL_Float x, const APL_Float & y)
 {
+const APL_Complex pi(M_PI, 0);
    if (x < 0.5)
-      return M_PI / sin(M_PI * APL_Complex(x, y)) * gamma(1 - x, -y);
+      return pi / (sin(pi * APL_Complex(x, y)) * gamma(1.0 - x, -y));
 
    // coefficients for lanczos approximation of the gamma function.
    //
@@ -704,20 +755,20 @@ ComplexCell::gamma(APL_Float x, const APL_Float y)
    errno = 0;
    feclearexcept(FE_ALL_EXCEPT);
 
-   x += 1.0;
-
 const APL_Complex z(x, y);
+const APL_Complex z1(x + 5.5, y);
+const APL_Complex z2(x + 0.5, y);
 
-const APL_Complex ret( (sqrt(2*M_PI) / z)
+const APL_Complex ret( (complex_sqrt(APL_Complex(2*M_PI)) / z)
                       * (p0                            +
-                          p1 / APL_Complex(x + 1.0, y) +
-                          p2 / APL_Complex(x + 2.0, y) +
-                          p3 / APL_Complex(x + 3.0, y) +
-                          p4 / APL_Complex(x + 4.0, y) +
-                          p5 / APL_Complex(x + 5.0, y) +
-                          p6 / APL_Complex(x + 6.0, y))
-                       * pow(z + 5.5, z + 0.5)
-                       * exp(-(z + 5.5))
+                         p1 / APL_Complex(x + 1.0, y) +
+                         p2 / APL_Complex(x + 2.0, y) +
+                         p3 / APL_Complex(x + 3.0, y) +
+                         p4 / APL_Complex(x + 4.0, y) +
+                         p5 / APL_Complex(x + 5.0, y) +
+                         p6 / APL_Complex(x + 6.0, y))
+                       * complex_power(z1, z2)
+                       * complex_exponent(-z1)
                      );
 
    // errno may be set and is checked by the caller
@@ -743,18 +794,18 @@ const APL_Complex ret( (sqrt(2*M_PI) / z)
 bool
 ComplexCell::get_near_bool()  const   
 {
-   if (value2.cval_i >  INTEGER_TOLERANCE)   DOMAIN_ERROR;
-   if (value2.cval_i < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
+   if (value.cval[1] >  INTEGER_TOLERANCE)   DOMAIN_ERROR;
+   if (value.cval[1] < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
 
-   if (value.cval_r > INTEGER_TOLERANCE)   // 1 or invalid
+   if (value.cval[0] > INTEGER_TOLERANCE)   // 1 or invalid
       {
-        if (value.cval_r < (1.0 - INTEGER_TOLERANCE))   DOMAIN_ERROR;
-        if (value.cval_r > (1.0 + INTEGER_TOLERANCE))   DOMAIN_ERROR;
+        if (value.cval[0] < (1.0 - INTEGER_TOLERANCE))   DOMAIN_ERROR;
+        if (value.cval[0] > (1.0 + INTEGER_TOLERANCE))   DOMAIN_ERROR;
         return true;
       }
    else
       {
-        if (value.cval_r < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
+        if (value.cval[0] < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
         return false;
       }
 }
@@ -762,12 +813,12 @@ ComplexCell::get_near_bool()  const
 APL_Integer
 ComplexCell::get_near_int() const
 {
-// if (value2.cval_i >  qct)   DOMAIN_ERROR;
-// if (value2.cval_i < -qct)   DOMAIN_ERROR;
+// if (value.cval[1] >  qct)   DOMAIN_ERROR;
+// if (value.cval[1] < -qct)   DOMAIN_ERROR;
 
-const double val = value.cval_r;
-const double result = round(val);
-const double diff = val - result;
+const APL_Float val = value.cval[0];
+const APL_Float result = round(val);
+const APL_Float diff = val - result;
    if (diff > INTEGER_TOLERANCE)    DOMAIN_ERROR;
    if (diff < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
 
@@ -799,7 +850,7 @@ ComplexCell::compare(const Cell & other) const
    if (other.is_character_cell())   return COMP_GT;
    if (!other.is_numeric())   DOMAIN_ERROR;
 
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
    if (equal(other, qct))   return COMP_EQ;
 
 APL_Float areal = other.get_real_value();
@@ -826,22 +877,22 @@ ComplexCell::character_representation(const PrintContext & pctx) const
          // 10⋆get_PP()
          //
          APL_Float ten_to_PP = 1.0;
-         loop(p, pctx.get_PP())   ten_to_PP *= 10.0;
-      
+         loop(p, pctx.get_PP())   ten_to_PP = ten_to_PP * 10.0;
+
          // lrm p. 13: In J notation, the real or imaginary part is not
          // displayed if it is less than the other by more than ⎕PP orders
          // of magnitude (unless ⎕PP is at its maximum).
          //
-         const APL_Float pos_real = value.cval_r < 0
-                                  ? -value.cval_r : value.cval_r;
-         const APL_Float pos_imag = value2.cval_i < 0
-                                  ? -value2.cval_i : value2.cval_i;
+         const APL_Float pos_real = value.cval[0] < 0.0
+                                  ? -value.cval[0] : value.cval[0];
+         const APL_Float pos_imag = value.cval[1] < 0.0
+                                  ? -value.cval[1] : value.cval[1];
 
          if (pos_real >= pos_imag)   // pos_real dominates pos_imag
             {
               if (pos_real > pos_imag*ten_to_PP)
                  {
-                   const FloatCell real_cell(value.cval_r);
+                   const FloatCell real_cell(value.cval[0]);
                    return real_cell.character_representation(pctx);
                  }
             }
@@ -849,7 +900,7 @@ ComplexCell::character_representation(const PrintContext & pctx) const
             {
               if (pos_imag > pos_real*ten_to_PP)
                  {
-                   const FloatCell imag_cell(value2.cval_i);
+                   const FloatCell imag_cell(value.cval[1]);
                    PrintBuffer ret = imag_cell.character_representation(pctx);
                    ret.pad_l(UNI_ASCII_J, 1);
                    ret.pad_l(UNI_ASCII_0, 1);
@@ -865,7 +916,7 @@ ComplexCell::character_representation(const PrintContext & pctx) const
       }
 
 bool scaled_real = pctx.get_scaled();   // may be changed by print function
-UCS_string ucs(value.cval_r, scaled_real, pctx);
+UCS_string ucs(value.cval[0], scaled_real, pctx);
 
 ColInfo info;
    info.flags |= CT_COMPLEX;
@@ -895,7 +946,7 @@ int int_fract = ucs.size();
       {
         ucs.append(UNI_ASCII_J);
         bool scaled_imag = pctx.get_scaled();  // may be changed by UCS_string()
-        const UCS_string ucs_i(value2.cval_i, scaled_imag, pctx);
+        const UCS_string ucs_i(value.cval[1], scaled_imag, pctx);
 
         ucs.append(ucs_i);
 
@@ -911,8 +962,8 @@ ComplexCell::need_scaling(const PrintContext &pctx) const
 {
    // a complex number needs scaling if the real part needs it, ot
    // the complex part is significant and needs it.
-   return FloatCell::need_scaling(value.cval_r, pctx.get_PP()) ||
+   return FloatCell::need_scaling(value.cval[0], pctx.get_PP()) ||
           (!is_near_real() && 
-          FloatCell::need_scaling(value2.cval_i, pctx.get_PP()));
+          FloatCell::need_scaling(value.cval[1], pctx.get_PP()));
 }
 //-----------------------------------------------------------------------------
