@@ -29,7 +29,7 @@
 //=============================================================================
 DerivedFunction::DerivedFunction(Token & lfun, Function * dyop, Token & rfun,
                                  const char * loc)
-   : Function(ID::USER_SYMBOL, TOK_FUN2),
+   : Function(ID_USER_SYMBOL, TOK_FUN2),
      left_fun(lfun),
      oper(dyop),
      right_fun(rfun),
@@ -46,7 +46,7 @@ DerivedFunction::DerivedFunction(Token & lfun, Function * dyop, Token & rfun,
 //-----------------------------------------------------------------------------
 DerivedFunction::DerivedFunction(Token & lfun, Function * dyop, Value_P X,
                                  Token & rval, const char * loc)
-   : Function(ID::USER_SYMBOL, TOK_FUN2),
+   : Function(ID_USER_SYMBOL, TOK_FUN2),
      left_fun(lfun),
      oper(dyop),
      right_fun(rval),
@@ -56,7 +56,7 @@ DerivedFunction::DerivedFunction(Token & lfun, Function * dyop, Value_P X,
 //-----------------------------------------------------------------------------
 DerivedFunction::DerivedFunction(Token & lfun, Function * monop,
                                  const char * loc)
-   : Function(ID::USER_SYMBOL, TOK_FUN2),
+   : Function(ID_USER_SYMBOL, TOK_FUN2),
      left_fun(lfun),
      oper(monop),
      right_fun(TOK_VOID),
@@ -73,7 +73,7 @@ DerivedFunction::DerivedFunction(Token & lfun, Function * monop,
 //-----------------------------------------------------------------------------
 DerivedFunction::DerivedFunction(Token & lfun, Function * monop,
                                  Value_P X, const char * loc)
-   : Function(ID::USER_SYMBOL, TOK_FUN2),
+   : Function(ID_USER_SYMBOL, TOK_FUN2),
      left_fun(lfun),
      oper(monop),
      right_fun(TOK_VOID),
@@ -89,7 +89,7 @@ DerivedFunction::DerivedFunction(Token & lfun, Function * monop,
 }
 //-----------------------------------------------------------------------------
 DerivedFunction::DerivedFunction(Function * fun, Value_P X, const char * loc)
-   : Function(ID::USER_SYMBOL, TOK_FUN2),
+   : Function(ID_USER_SYMBOL, TOK_FUN2),
      left_fun(TOK_VOID),
      oper(fun),
      right_fun(TOK_VOID),
@@ -110,7 +110,8 @@ DerivedFunction::eval_B(Value_P B)
    Log(LOG_FunOperX)
       {
         print(CERR << "entering DerivedFunction");
-        CERR << "::eval_B() , this = " << (void *)this << endl;
+        CERR << "::eval_B() , this = "
+             << voidP(this) << endl;
       }
 
    if (left_fun.get_tag() == TOK_VOID)   // function bound to axis
@@ -135,7 +136,8 @@ DerivedFunction::eval_XB(Value_P X, Value_P B)
    Log(LOG_FunOperX)
       {
         print(CERR << "entering DerivedFunction");
-        CERR << "::eval_XB() , this = " << (void *)this << endl;
+        CERR << "::eval_XB() , this = "
+             << voidP(this) << endl;
       }
 
    if (right_fun.get_tag() != TOK_VOID)   // dyadic operator
@@ -241,8 +243,10 @@ DerivedFunctionCache::DerivedFunctionCache()
 {
    Log(LOG_FunOperX)
       {
-         CERR << "DerivedFunctionCache created, cache at " << (const void *)cache
-              << "..." << (const void *)(cache + MAX_FUN_OPER) << endl;
+         CERR << "DerivedFunctionCache created, cache at "
+              << voidP(cache) << "..."
+              << voidP(cache + MAX_FUN_OPER)
+              << endl;
       }
 }
 //-----------------------------------------------------------------------------
@@ -250,8 +254,10 @@ DerivedFunctionCache::~DerivedFunctionCache()
 {
    Log(LOG_FunOperX)
       {
-         CERR << "DerivedFunctionCache deleted, cache at " << (const void *)cache
-              << "..." << (const void *)(cache + MAX_FUN_OPER) << endl;
+         CERR << "DerivedFunctionCache deleted, cache at "
+              << voidP(cache) << "..."
+              << voidP(cache + MAX_FUN_OPER)
+              << endl;
       }
 }
 //-----------------------------------------------------------------------------
@@ -262,8 +268,10 @@ DerivedFunctionCache::reset()
 
    Log(LOG_FunOperX)
       {
-         CERR << "DerivedFunctionCache reset, cache at " << (const void *)cache
-              << "..." << (const void *)(cache + MAX_FUN_OPER) << endl;
+         CERR << "DerivedFunctionCache reset, cache at "
+              << voidP(cache) << "..."
+              << voidP(cache + MAX_FUN_OPER)
+              << endl;
       }
 }
 //-----------------------------------------------------------------------------
@@ -275,8 +283,8 @@ DerivedFunctionCache::get(const char * loc)
    Log(LOG_FunOperX)
       {
          CERR << "DerivedFunctionCache get( " << idx << " ), cache at "
-              << (const void *)cache << "..."
-              << (const void *)(cache + MAX_FUN_OPER)
+              << voidP(cache) << "..."
+              << voidP(cache + MAX_FUN_OPER)
               << " at " << loc << endl;
       }
 
