@@ -21,8 +21,9 @@
 
 #include "Common.hh"
 #include "DynamicObject.hh"
+#include "IndexExpr.hh"
 #include "PrintOperator.hh"
-#include "Value.icc"   // so that casting to Value * works
+#include "Value.hh"
 
 //-----------------------------------------------------------------------------
 ostream &
@@ -39,8 +40,9 @@ int pos = 0;
    for (const DynamicObject * p = this; ;)
        {
          out << "    Chain[" << setw(2) << pos++ << "]  " 
-             << (void *)p->prev << " --> "
-             << (void *)p << " --> "   << (void *)p->next << "    "
+             << voidP(p->prev) << " --> "
+             << voidP(p)       << " --> "
+             << voidP(p->next) << "    "
              << p->where_allocated() << endl;
 
           p = p->next;
@@ -51,18 +53,49 @@ int pos = 0;
 void
 DynamicObject::print_new(ostream & out, const char * loc) const
 {
-   out << "new    " << (const void *)(const Value *)this
-       << " at " << loc << endl;
+   out << "new    " << voidP(this) << " at " << loc << endl;
 }
 //-----------------------------------------------------------------------------
 void
 DynamicObject::print(ostream & out) const
 {
-   out << "DynamicObject: " << (void *)this
-       << " (Value: " << (void *)(Value *)this << ") :" << endl
-       << "    prev:      " << (void *)prev         << endl
-       << "    next:      " << (void *)next         << endl
+   out << "DynamicObject: " << voidP(this)
+       << " (Value: " << voidP(this) << ") :"       << endl
+       << "    prev:      " << voidP(prev)          << endl
+       << "    next:      " << voidP(next)          << endl
        << "    allocated: " << where_allocated()    << endl;
 }
 //-----------------------------------------------------------------------------
+/// cast DynamicObject to derived class Value.
+// This only works properly after #include Value.hh !
+Value *
+DynamicObject::pValue()
+{
+   return static_cast<Value *>(this);
+}
+//-----------------------------------------------------------------------------
+/// cast DynamicObject to derived class Value.
+// This only works properly after #include Value.hh !
+const Value *
+DynamicObject::pValue() const
+{
+  return static_cast<const Value *>(this);
+}
+//-----------------------------------------------------------------------------
+/// cast DynamicObject to derived class IndexExpr.
+// This only works properly after #include IndexExpr.hh !
+const IndexExpr *
+DynamicObject::pIndexExpr() const
+{
+  return static_cast<const IndexExpr *>(this);
+}
+//-----------------------------------------------------------------------------
+/// cast DynamicObject to derived class Value.
+// This only works properly after #include IndexExpr.hh !
+IndexExpr *
+DynamicObject::pIndexExpr()
+{
+  return static_cast<IndexExpr *>(this);
+}
+//-----------------------------------------------------------------------------
 
