@@ -23,13 +23,40 @@
 #include "Output.hh"
 #include "PrintOperator.hh"
 #include "Symbol.hh"
+#include "StateIndicator.hh"
 #include "UserFunction.hh"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
-void
-Error::print(ostream & out) const
+Error::Error(ErrorCode ec, const char * loc)
+   : error_code(ec),
+     throw_loc(loc),
+     parser_loc(0),
+     show_locked(false),
+     left_caret(-1),
+     right_caret(-1),
+     print_loc(0)
 {
+const char more = Workspace::more_error().size() ? '+' : 0;
+   snprintf(error_message_1, sizeof(error_message_1), "%s%c",
+            error_name(error_code), more);
+
+   *symbol_name = 0;
+   *error_message_2 = 0;
+}
+//-----------------------------------------------------------------------------
+void
+Error::print(ostream & out, const char * loc) const
+{
+   if (print_loc)
+      {
+        CERR << endl << "*** Error printed twice; first printed at "
+             << print_loc << endl
+             << "now printed at " << loc << endl;
+
+        return;
+      }
+
    Log(LOG_verbose_error)
       {
         out << endl
@@ -43,7 +70,7 @@ Error::print(ostream & out) const
              return;
            }
 
-        if (error_message_1.size())
+        if (*error_message_1)
            {
              out << error_message_1 << endl;
            }
@@ -56,8 +83,10 @@ Error::print(ostream & out) const
 
         if (parser_loc)   out << "   Parser LOC: " << parser_loc  << endl;
         if (print_loc)    out << "   Print LOC:  " << print_loc   << endl;
+        out                   << "   loc:        " << loc         << endl;
+        loc = print_loc;
 
-        if (symbol_name.size())
+        if (*symbol_name)
            out                << "   Symbol:     " << symbol_name << endl;
 
         out <<                   "   Thrown at:  " << throw_loc   << endl
@@ -66,19 +95,19 @@ Error::print(ostream & out) const
       }
 }
 //-----------------------------------------------------------------------------
-const UCS_string
+const char *
 Error::error_name(ErrorCode err)
 {
    switch(err)
       {
    /// the cases
 #define err_def(c, txt, _maj, _min) \
-   case E_ ## c:   return UCS_string(UTF8_string(txt));
+   case E_ ## c:   return txt;
 
 #include "Error.def"
       }
 
-   return UCS_string("Unknown Error");
+   return "Unknown Error";
 }
 //-----------------------------------------------------------------------------
 bool
@@ -94,6 +123,16 @@ Error::is_known() const
    return false;
 }
 //-----------------------------------------------------------------------------
+void
+Error::set_error_line_2(const UCS_string & ucs, int lcaret, int rcaret)
+{
+UTF8_string utf(ucs);
+   strncpy(error_message_2, utf.c_str(), sizeof(error_message_2));
+   error_message_2[sizeof(error_message_2) - 1] = 0;
+   left_caret = lcaret;
+   right_caret = rcaret;
+}
+//-----------------------------------------------------------------------------
 bool
 Error::is_syntax_or_value_error() const
 {
@@ -102,22 +141,6 @@ Error::is_syntax_or_value_error() const
    return false;
 }
 //-----------------------------------------------------------------------------
-void
-Error::init(ErrorCode ec, const char * loc)
-{
-   error_code = ec;
-   throw_loc = loc;
-   error_message_1 = error_name(error_code);
-   if (Workspace::more_error().size())   error_message_1.append(UNI_ASCII_PLUS);
-   error_message_2.shrink(0);
-   symbol_name.shrink(0);
-   parser_loc = 0;
-   show_locked = false;
-   left_caret = -1;
-   right_caret = -1;
-   print_loc = 0;
-}
-//-----------------------------------------------------------------------------
 UCS_string
 Error::get_error_line_3() const
 {
@@ -140,16 +163,17 @@ const int diff = right_caret - left_caret;
 void
 Error::print_em(ostream & out, const char * loc)
 {
-   if (print_loc) 
+   if (print_loc)
       {
         CERR << endl << "*** Error printed twice; first printed at "
-             << print_loc << endl;
+             << print_loc << endl
+             << "now printed from " << loc << endl;
 
         return;
       }
 
    print_loc = loc;
-   if (get_error_line_1().size())   out << get_error_line_1() << endl;
+   if (*get_error_line_1())   out << get_error_line_1() << endl;
 
    out << get_error_line_2() << endl
        << get_error_line_3() << endl;
@@ -184,27 +208,14 @@ StateIndicator * si = Workspace::SI_top();
       }
 
 Error error(code, loc);
-   if (si)   si->update_error_info(error);
-   throw error;
+   if (si)   error.update_error_info(si);
+
+Error & eref = error;
+   throw eref;
 }
 //-----------------------------------------------------------------------------
 void
-throw_tokenize_error(ErrorCode code, const UCS_string & more_error_info)
-{
-   Log(LOG_error_throw)
-      CERR << endl
-           << "throwing " << Error::error_name(code)
-           << " in  Tokenizer" << endl;
-
-   MORE_ERROR() << more_error_info;
-
-Error error(code, LOC);
-
-   throw error;
-}
-//-----------------------------------------------------------------------------
-void
-throw_parse_error(ErrorCode code, const char * par_loc, const char *loc)
+Error::throw_parse_error(ErrorCode code, const char * par_loc, const char *loc)
 {
    Log(LOG_error_throw)
       CERR << endl
@@ -218,12 +229,14 @@ Error error(code, loc);
    error.parser_loc = par_loc;
 
 // StateIndicator * si = Workspace::SI_top();
-//   if (si)   si->update_error_info(error);
-   throw error;
+// if (si)   error.update_error_info(si);
+
+Error & eref = error;
+   throw eref;
 }
 //-----------------------------------------------------------------------------
 void
-throw_symbol_error(const UCS_string & sym_name, const char * loc)
+Error::throw_symbol_error(const UCS_string & sym_name, const char * loc)
 {
    Log(LOG_error_throw)   
       {   
@@ -235,14 +248,18 @@ throw_symbol_error(const UCS_string & sym_name, const char * loc)
    Log(LOG_verbose_error)     Backtrace::show(__FILE__, __LINE__);
 
 Error error(E_VALUE_ERROR, loc);
-   error.symbol_name = sym_name;
-   if (Workspace::SI_top())   Workspace::SI_top()->update_error_info(error);
-   throw error;
+UTF8_string sym_name_utf(sym_name);
+   snprintf(error.symbol_name, sizeof(error.symbol_name), "%s",
+            sym_name_utf.c_str());
+   if (StateIndicator * si = Workspace::SI_top())   error.update_error_info(si);
+
+Error & eref = error;
+   throw eref;
 }
 //-----------------------------------------------------------------------------
 void
-throw_define_error(const UCS_string & fun_name, const UCS_string & cmd,
-                   const char * loc)
+Error::throw_define_error(const UCS_string & fun_name, const UCS_string & cmd,
+                          const char * loc)
 {
    Log(LOG_error_throw)   
       {   
@@ -253,12 +270,83 @@ throw_define_error(const UCS_string & fun_name, const UCS_string & cmd,
    Log(LOG_verbose_error)     Backtrace::show(__FILE__, __LINE__);
 
 Error error(E_DEFN_ERROR, loc);
-   error.symbol_name = fun_name;
-   error.error_message_2 = UCS_string(6, UNI_ASCII_SPACE);
-   error.error_message_2.append(cmd);   // something like ∇FUN[⎕]∇
-   error.left_caret = error.error_message_2.size() - 1;
-   if (Workspace::SI_top())   *Workspace::get_error() = error;
-   throw error;
+Error & eref = error;
+UTF8_string fun_name_utf(fun_name);
+   snprintf(error.symbol_name, sizeof(error.symbol_name), "%s",
+            fun_name_utf.c_str());
+
+UTF8_string cmd_utf(cmd);   // cmd is something like ∇FUN[⎕]∇
+   snprintf(error.error_message_2, sizeof(error.error_message_2),
+            " aaa  %s", cmd_utf.c_str());
+   eref.left_caret = 5 + cmd.size();
+   if (Workspace::SI_top())   *Workspace::get_error() = eref;
+   throw eref;
+}
+//-----------------------------------------------------------------------------
+void
+Error::update_error_info(StateIndicator * si)
+{
+bool locked = false;
+const UserFunction * ufun = si->get_executable()->get_ufun();
+
+   // prepare second error line (failed statement)
+   //
+   if (ufun)
+      {
+        if (get_show_locked() || ufun->get_exec_properties()[1])
+           {
+             locked = true;
+             set_error_line_2("      ");
+             set_left_caret(6);
+           }
+        else
+           {
+             UCS_string ucs(ufun->get_name_and_line(si->get_PC()));
+             ucs.append(UNI_ASCII_SPACE);
+             ucs.append(UNI_ASCII_SPACE);
+             UTF8_string utf(ucs);
+             set_error_line_2(utf.c_str());
+             set_left_caret(ucs.size());
+           }
+      }
+   else
+      {
+        set_error_line_2("      ");
+        set_left_caret(6);
+      }
+
+   // prepare third line (carets)
+   //
+   set_right_caret(-1);
+
+   if (locked)
+      {
+        si->get_executable()->get_ufun()->set_locked_error_info(*this);
+      }
+   else
+      {
+        const Function_PC from = si->get_prefix().get_range_low();
+        Function_PC to   = si->get_prefix().get_range_high();
+        const Function_PC2 error_range(from, to);
+        si->get_executable()->set_error_info(*this, error_range);
+      }
+
+   // print error, unless we are in safe execution mode.
+   //
+   {
+     bool print_error = true;
+     for (const StateIndicator * si1 = si; si1; si1 = si1->get_parent())
+         {
+           if (si1->get_safe_execution())
+              {
+                print_error = false;
+                break;
+              }
+         }
+     if (print_error)   print_em(UERR, LOC);
+   }
+
+   StateIndicator::get_error(si) = *this;
 }
 //-----------------------------------------------------------------------------
 
