@@ -24,7 +24,7 @@
 #include "PrintOperator.hh"
 #include "UCS_string.hh"
 #include "UserFunction.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
@@ -48,17 +48,17 @@ Executable::Executable(const UCS_string & ucs,  bool multi_line,
              switch (uni)
                 {
                   case UNI_ASCII_CR:                         break;
-                  case UNI_ASCII_LF: text.append(line);
-                                     line.shrink(0);         break;
+                  case UNI_ASCII_LF: text.push_back(line);
+                                     line.clear();           break;
                   default:           line.append(uni);
                }
            }
 
-        if (line.size())   text.append(line);
+        if (line.size())   text.push_back(line);
       }
    else
       {
-        text.append(ucs);
+        text.push_back(ucs);
       }
 }
 //-----------------------------------------------------------------------------
@@ -87,17 +87,17 @@ ShapeItem last_semi = -1;
       {
         for (ShapeItem t = last_semi; t < lambda_text.size(); ++t)
             header.append(lambda_text[t]);
-        text.append(header);
+        text.push_back(header);
 
         UCS_string new_body(lambda_text);
-        new_body.shrink(last_semi);
+        new_body.resize(last_semi);
         new_body.remove_trailing_whitespaces();
-        text.append(new_body);
+        text.push_back(new_body);
       }
    else
       {
-        text.append(header);
-        text.append(lambda_text);
+        text.push_back(header);
+        text.push_back(lambda_text);
      }
 }
 //-----------------------------------------------------------------------------
@@ -105,7 +105,7 @@ Executable::~Executable()
 {
    Log(LOG_UserFunction__fix)
       {
-        CERR << "deleting Executable " << (const void *)this
+        CERR << "deleting Executable " << voidP(this)
              << " (body size=" << body.size() << ")" << endl;
       }
 
@@ -128,7 +128,7 @@ Executable::clear_body()
            }
       }
 
-   body.shrink(0);
+   body.clear();
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -152,46 +152,47 @@ ErrorCode ec = parser.parse(ucs_line, in);
              throw error;
            }
 
-        if (ec != E_NO_ERROR)   throw_parse_error(ec, LOC, LOC);
+        if (ec != E_NO_ERROR)   Error::throw_parse_error(ec, LOC, LOC);
       }
 
    return parse_body_line(line, in, trace, tolerant, loc);
 } 
 //-----------------------------------------------------------------------------
 ErrorCode
-Executable::parse_body_line(Function_Line line, const Token_string & in,
+Executable::parse_body_line(Function_Line line, const Token_string & input,
                             bool trace,  bool tolerant, const char * loc)
 {
-Source<Token, true> src(in);
+ShapeItem idx = 0;
+const ShapeItem end = input.size();
 
    // handle labels (if any)
    //
-   if (get_parse_mode() == PM_FUNCTION &&
-       src.rest() >= 2                 &&
-       src[0].get_tag() == TOK_SYMBOL  &&
-       src[1].get_tag() == TOK_COLON)
+   if (get_parse_mode() == PM_FUNCTION  &&
+       end > 1                          &&
+       input[0].get_tag() == TOK_SYMBOL &&
+       input[1].get_tag() == TOK_COLON)
       {
-        Token tok = src.get();   // get the label symbol
-        ++src;                   // skip the :
+        Token tok_sym = input[0];   // get the label symbol
+        idx = 2;                   // skip the :
 
         UserFunction * ufun = get_ufun();
         Assert(ufun);
-        ufun->add_label(tok.get_sym_ptr(), line);
+        ufun->add_label(tok_sym.get_sym_ptr(), line);
       }
 
 Token_string out;
-   while (src.rest())
+   while ((end - idx))
       {
         // find next diamond. If there is one, copy statement in reverse order,
         // append TOK_ENDL, and continue.
-        
+
         // 1. determine statement length
         //
         ShapeItem stat_len = 0;   // statement length, not counting ◊
         int diamond = 0;
-        loop(t, src.rest())
+        loop(t, (end - idx))
             {
-              if (src[t].get_tag() == TOK_DIAMOND)
+              if (input[idx + t].get_tag() == TOK_DIAMOND)
                  {
                    diamond = 1;
                    break;
@@ -206,7 +207,7 @@ Token_string out;
         //
         loop(s, stat_len)
             {
-              const Token & tok = src[stat_len - s - 1];
+              const Token & tok = input[idx + stat_len - s - 1];
               if (tok.get_Class() >= TC_MAX_PERM &&
                   tok.get_tag() != TOK_L_CURLY &&
                   tok.get_tag() != TOK_R_CURLY)
@@ -217,35 +218,38 @@ Token_string out;
                         << "Offending token: (tag > TC_MAX_PERM) "
                         << tok.get_tag() << " " << tok << endl
                         << "Statement: ";
-                   Token::print_token_list(CERR, in);
+                   loop(t, input.size())   CERR << "`" << input[t] << "  ";
+                   CERR << endl;
                    SYNTAX_ERROR;
                  }
-              out.append(tok);
+              out.push_back(tok);
             }
-        src.skip(stat_len + diamond);   // skip statement and maybe ◊
+        idx += stat_len + diamond;   // skip statement and maybe ◊
 
         // 3. maybe add a TOK_END or TOK_ENDL
         //
         if (stat_len && (diamond || get_parse_mode() != PM_EXECUTE))
            {
              const int64_t tr = trace ? 1 : 0;
-             out.append(Token(TOK_END, tr), LOC);
+             out.push_back(Token(TOK_END, tr));
            }
       }
 
    // replace the trailing TOK_END (if any) into TOK_ENDL
    //
-   if (out.size() && out.last().get_tag() == TOK_END)
-      out.last().ChangeTag(TOK_ENDL);
+   if (out.size() && out.back().get_tag() == TOK_END)
+      out.back().ChangeTag(TOK_ENDL);
 
    Log(LOG_UserFunction__set_line)
       {
         CERR << "[final line " << line << "] ";
-        out.print(CERR);
+        out.print(CERR, false);
       } 
 
-   loop(t, out.size())   body.append(out[t], LOC);
+   loop(t, out.size())   body.push_back(out[t]);
    return E_NO_ERROR;
+
+#undef get
 }
 //-----------------------------------------------------------------------------
 Token
@@ -253,10 +257,10 @@ Executable::execute_body() const
 {
 StateIndicator & si = *Workspace::SI_top();
 
-   try                 { return si.run();                         }
-   catch (Error err)   { return Token(TOK_ERROR, err.error_code); }
-   catch (Token tok)   { return tok;                              }
-   catch (...)         { return Token(TOK_ERROR, E_SYSTEM_ERROR); }
+   try                 { return si.run();                               }
+   catch (Error err)   { return Token(TOK_ERROR, err.get_error_code()); }
+   catch (Token tok)   { return tok;                                    }
+   catch (...)         { return Token(TOK_ERROR, E_SYSTEM_ERROR);       }
 
    // not reached
    FIXME;
@@ -268,7 +272,7 @@ Executable::print_token(ostream & out) const
    out << endl
        <<  "Function body [" << body.size() << " token]:" << endl;
 
-   body.print(out);
+   body.print(out, false);
 }
 //-----------------------------------------------------------------------------
 void
@@ -280,9 +284,9 @@ Executable::print_text(ostream & out) const
 UCS_string
 Executable::statement_text(Function_PC pc) const
 {
-   if (pc >= body.size())
+   if (pc >= Function_PC(body.size()))
       {
-        Assert(pc == body.size());
+        Assert(pc == Function_PC(body.size()));
         pc = Function_PC(body.size() - 1);
       }
 
@@ -321,7 +325,7 @@ UCS_string ret;
    while (tidx < line_txt.size())
       {
         if (Avec::is_diamond(line_txt[tidx]))    break;
-        ret.append(line_txt[tidx++], LOC);
+        ret.append(line_txt[tidx++]);
       }
 
    // skip trailing spaces
@@ -334,8 +338,8 @@ UCS_string ret;
 void
 Executable::set_error_info(Error & error, Function_PC2 pc_from_to) const
 {
-const ErrorCode ec = error.error_code;
-UCS_string & message_2 = error.error_message_2;
+const ErrorCode ec = error.get_error_code();
+UCS_string message_2(UTF8_string(error.get_error_line_2()));
 
    // for value errors we point to the failed symbol itself.
    //
@@ -420,13 +424,18 @@ int len_between = 0;   // distance between the carets
          else if (q > pc_from_to.low)   len_between += len;
        }
 
+   {
+     UTF8_string utf(message_2);
+     error.set_error_line_2(utf.c_str());
+   }
+
    // Line 3: carets
    //
-   error.left_caret += len_left;
+   error.set_left_caret(error.get_left_caret() + len_left);
 
    if (pc_from_to.high != pc_from_to.low)   // two carets
       {
-        error.right_caret = error.left_caret + len_between;
+        error.set_right_caret(error.get_left_caret() + len_between);
       }
 }
 //-----------------------------------------------------------------------------
@@ -436,7 +445,7 @@ Executable::get_statement_start(int pc) const
    // this function is used in error reporting so it should
    // not Assert() and the like to avoid infinite recursion.
 
-   if (pc >= body.size())   pc = body.size() - 1;
+   if (pc >= int(body.size()))   pc = body.size() - 1;
 
    // if we are at the end of the statement, move back.
    //
@@ -517,7 +526,7 @@ bool have_curly = false;
    if (!have_curly)   return;   // no { ... } in this body
 
    // undo the token reversion of the body so that the body token and the
-   // text run in the same direction.
+   // function text run in the same direction.
    //
    reverse_statement_token(body);
 
@@ -526,7 +535,25 @@ int lambda_num = 0;
        {
          if (body[b].get_tag() != TOK_L_CURLY)   continue;   // not {
 
-         b = setup_one_lambda(b, ++lambda_num) - 1;   // -1 due to ++b in loop(b) above
+         ShapeItem end = -1;
+         int level = 1;   // since body[b] is {
+         for (ShapeItem b1 = b + 1; b1 < ShapeItem(body.size()); ++b1)
+             {
+               switch(body[b1].get_tag())
+                  {
+                    default:                            continue;
+                    case TOK_L_CURLY:   ++level;        continue;
+                    case TOK_R_CURLY:   if (--level)    continue;
+                                        end = b1;
+                                        b1 = body.size();
+                                        break;
+                    case TOK_END:       MORE_ERROR() << "Illegal ◊ in {}";
+                                        DEFN_ERROR;
+                  }
+                break;   // for b1...
+             }
+         Assert(end != -1);
+         b = setup_one_lambda(b, end, ++lambda_num) - 1;   // -1 due to ++b in loop(b) above
        }
 
    // redo the token reversion of the body so that the body token and the
@@ -540,10 +567,8 @@ int lambda_num = 0;
 }
 //-----------------------------------------------------------------------------
 ShapeItem
-Executable::setup_one_lambda(ShapeItem b, int lambda_num)
+Executable::setup_one_lambda(ShapeItem b, ShapeItem bend, int lambda_num)
 {
-const ShapeItem bend = b + body[b].get_int_val2();   // the corresponding }
-   Assert(bend < body.size());
    Assert(body[bend].get_tag() == TOK_R_CURLY);
 
    body[b++].clear(LOC);    // invalidate {
@@ -569,7 +594,7 @@ UCS_string ufun_text = UserFunction_header::lambda_header(signature,lambda_num);
    ufun_text.append(lambda_text);
    ufun_text.append(UNI_ASCII_LF);
 
-Q(ufun_text)
+Q1(ufun_text)
 
 int error_line = -1;
 const char * error_loc = 0;
@@ -634,13 +659,13 @@ int level = 0;
               // first token: prepend λ ← tokens
               //
               Symbol * sym_Z = &Workspace::get_v_LAMBDA();
-              lambda_body.append(Token(TOK_LAMBDA, sym_Z), LOC);
-              lambda_body.append(Token(TOK_ASSIGN1), LOC);
+              lambda_body.push_back(Token(TOK_LAMBDA, sym_Z));
+              lambda_body.push_back(Token(TOK_ASSIGN1));
 
               signature |= SIG_Z;
             }
 
-         lambda_body.append(t, LOC);
+         lambda_body.push_back(t);
        }
 
    if ((signature & SIG_B) == 0 &&   // niladic
@@ -657,30 +682,30 @@ int level = 0;
    if (signature & SIG_Z)
       {
         const int64_t trace = 0;
-        lambda_body.append(Token(TOK_ENDL, trace));
+        lambda_body.push_back(Token(TOK_ENDL, trace));
 
         Token ret_lambda(TOK_RETURN_SYMBOL, &Workspace::get_v_LAMBDA());
-        lambda_body.append(ret_lambda, LOC);
+        lambda_body.push_back(ret_lambda);
       }
    else
       {
         Token ret_void(TOK_RETURN_VOID);
-        lambda_body.append(ret_void, LOC);
+        lambda_body.push_back(ret_void);
       }
 
-   return (Fun_signature)signature;
+   return Fun_signature(signature);
 }
 //-----------------------------------------------------------------------------
 UCS_string
 Executable::extract_lambda_text(Fun_signature signature, int skip) const
 {
 UCS_string lambda_text;
-   if (signature & SIG_Z)   lambda_text.append_utf8("λ←");
+   if (signature & SIG_Z)   lambda_text.append_UTF8("λ←");
 
 int level = 0;   // {/} nesting level
 bool copying = false;
-int tidx = 0;    // the current line in text[]
-int tcol = 0;    // the current column in text[tidx];
+size_t tidx = 0;    // the current line in text[]
+size_t tcol = 0;    // the current column in text[tidx];
 bool in_single_quotes = false;
 bool in_double_quotes = false;
 
@@ -695,7 +720,7 @@ bool in_double_quotes = false;
 
          const UCS_string & line = text[tidx];
 
-         if (tcol >= line.size())   // end of line: wrap to next line
+         if (tcol >= size_t(line.size()))   // end of line: wrap to next line
             {
               ++tidx;     // next line
               tcol = 0;   // first column
@@ -753,7 +778,7 @@ bool in_double_quotes = false;
                    --skip;                       // next {...} at top-level
                    if (!copying)  continue;
 
-                   lambda_text.pop();            // the last }
+                   lambda_text.pop_back();       // the last }
                    return lambda_text;
 
               default: continue;
@@ -770,14 +795,14 @@ Executable::reverse_statement_token(Token_string & tos)
 {
 ShapeItem from = 0;
 
-   for (ShapeItem to = from + 1; to < tos.size(); ++to)
+   for (ShapeItem to = from + 1; to < ShapeItem(tos.size()); ++to)
        {
          if (tos[to].get_Class() == TC_END)
             {
               tos.reverse_from_to(from, to - 1);   // except TC_END
               from = to + 1;
             }
-         else if (to == (tos.size() - 1))
+         else if (to == ShapeItem(tos.size() - 1))
             {
               tos.reverse_from_to(from, to);
               from = to + 1;
@@ -788,16 +813,8 @@ ShapeItem from = 0;
 void
 Executable::reverse_all_token(Token_string & tos)
 {
-Token * t1 = &tos[0];
-Token * t2 = &tos[tos.size() - 1];
-
-   for (;t1 < t2; ++t1, --t2)
-       {
-         char tt[sizeof(Token)];
-         memcpy(tt, t1, sizeof(Token));
-         memcpy(t1, t2,   sizeof(Token));
-         memcpy(t2, tt, sizeof(Token));
-       }
+   for (Token * t1 = &tos[0], * t2 = &tos[tos.size() - 1]; t1 < t2;)
+       t1++->Hswap(*t2--);
 }
 //-----------------------------------------------------------------------------
 void
@@ -845,31 +862,28 @@ ExecuteList *
 ExecuteList::fix(const UCS_string & data, const char * loc)
 {
    // clear errors that may have occured before
-   {
-     Error * err = Workspace::get_error();
-     if (err) err->error_code = E_NO_ERROR;
-   }
+   if (Error * err = Workspace::get_error())   err->clear_error_code();
 
 ExecuteList * fun = new ExecuteList(data, loc);
 
    Log(LOG_UserFunction__fix)
       {
         CERR << "fix pmode=execute list:" << endl << data
-             << " addr " << (const void *)fun << endl
+             << " addr " << voidP(fun) << endl
              << "------------------- ExecuteList::fix() --" << endl;
       }
 
    {
      Error * err = Workspace::get_error();
-     if (err && err->error_code)
+     if (err && err->get_error_code())
         {
           Log(LOG_UserFunction__fix)
              {
                 CERR << "fix pmode=execute list failed with error "
-                     << Error::error_name(err->error_code) << endl;
+                     << Error::error_name(err->get_error_code()) << endl;
              }
 
-          err->parser_loc = 0;
+          err->set_parser_loc(0);
           delete fun;
           return 0;
         }
@@ -881,8 +895,8 @@ ExecuteList * fun = new ExecuteList(data, loc);
       }
    catch (Error err)
       {
-        Error * werr = Workspace::get_error();
-        if (werr)   *werr = err;
+        if (Error * werr = Workspace::get_error())   *werr = err;
+
         delete fun;
         return 0;
       }
@@ -894,8 +908,8 @@ ExecuteList * fun = new ExecuteList(data, loc);
         CERR << "fun->body.size() is " << fun->body.size() << endl;
       }
 
-   // for ⍎ we don't append TOK_END, but only TOK_RETURN_EXEC.
-   fun->body.append(Token(TOK_RETURN_EXEC), LOC);
+   // for ⍎ we do not append TOK_END, but only TOK_RETURN_EXEC.
+   fun->body.push_back(Token(TOK_RETURN_EXEC));
 
    Log(LOG_UserFunction__fix)   fun->print(CERR);
    return fun;
@@ -909,24 +923,31 @@ StatementList * fun = new StatementList(data, loc);
    Log(LOG_UserFunction__fix)
       {
         CERR << "fix pmode=statement list:" << endl << data << endl
-             << " addr " << (const void *)fun << endl
+             << " addr " << voidP(fun) << endl
              << "------------------- StatementList::fix() --" << endl;
       }
 
-   {
-     Error * err = Workspace::get_error();
-     if (err)   err->parser_loc = 0;
-   }
+   if (Error * err = Workspace::get_error())   err->set_parser_loc(0);
 
-   fun->parse_body_line(Function_Line_0, data, false, false, loc, false);
-   fun->setup_lambdas();
+   try
+      {
+        fun->parse_body_line(Function_Line_0, data, false, false, loc, false);
+        fun->setup_lambdas();
+      }
+   catch (Error & e)
+      {
+        Log(LOG_UserFunction__fix)
+           CERR << "parse_body_line(line 0) failed" << endl;
+        delete fun;
+        throw e;
+      }
 
    Log(LOG_UserFunction__fix)
       {
         CERR << "fun->body.size() is " << fun->body.size() << endl;
       }
 
-   fun->body.append(Token(TOK_RETURN_STATS), LOC);
+   fun->body.push_back(Token(TOK_RETURN_STATS));
 
    Log(LOG_UserFunction__fix)   fun->print(CERR);
    return fun;
