@@ -31,7 +31,7 @@
 #include "Parser.hh"
 #include "PrintOperator.hh"
 #include "Symbol.hh"
-#include "Value.icc"
+#include "Value.hh"
 
 //-----------------------------------------------------------------------------
 void
@@ -143,7 +143,7 @@ int sig = SIG_FUN;
    if (get_fun_valence() == 2)    sig |= SIG_A;
    if (get_fun_valence() >= 1)    sig |= SIG_B;
 
-   return (Fun_signature)sig;
+   return Fun_signature(sig);
 }
 //-----------------------------------------------------------------------------
 ostream &
