@@ -27,6 +27,7 @@
 #include "IO_Files.hh"
 #include "Output.hh"
 #include "UTF8_string.hh"
+#include "UserPreferences.hh"
 #include "Workspace.hh"
 
 int IO_Files::testcase_count = 0;
@@ -140,7 +141,7 @@ InputFile * input = InputFile::current_file();
                     const bool allowed = input->check_filter(file_line);
                     if (!allowed)
                        {
-                         file_line.shrink(0);
+                         file_line.clear();
                          continue;
                        }
                  }
@@ -149,7 +150,7 @@ InputFile * input = InputFile::current_file();
 
          if (cc == '\r')   continue;   // ignore carriage returns
 
-          file_line.append((UTF8)cc);
+          file_line += cc;
        }
 
    Log(LOG_test_execution)
@@ -230,7 +231,7 @@ IO_Files::end_of_current_file()
                 }
            }
       }
-             
+
    InputFile::close_current_file();
    ++testcases_done;
 
@@ -269,13 +270,15 @@ IO_Files::end_of_current_file()
                   << "Failed testcase is " << InputFile::current_filename()
                   << endl 
                   << endl;
-                  
-             InputFile::files_todo.shrink(0);
+
+             InputFile::files_todo.clear();
              return false;
            }
       }
 
-   InputFile::files_todo.erase(0);
+   InputFile::files_todo.erase(InputFile::files_todo.begin());
+
+   if (uprefs.auto_OFF && !InputFile::files_todo.size())   Command::cmd_OFF(0);
 
    Output::reset_dout();
    reset_errors();
@@ -348,12 +351,12 @@ IO_Files::open_next_file()
               {
                 CERR << "could not open "
                      << InputFile::current_filename() << endl;
-                InputFile::files_todo.erase(0);
+                InputFile::files_todo.erase(InputFile::files_todo.begin());
                 continue;
               }
 
            Log(LOG_test_execution)
-              CERR << "openened testcase file "
+              CERR << "opened testcase file "
                    << InputFile::current_filename() << endl;
 
            Output::reset_dout();
