@@ -21,7 +21,7 @@
 #include "IndexExpr.hh"
 #include "PrintBuffer.hh"
 #include "PrintOperator.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
@@ -40,7 +40,7 @@ IndexExpr::~IndexExpr()
 }
 //-----------------------------------------------------------------------------
 Value_P
-IndexExpr::extract_value(Rank rk)
+IndexExpr::extract_value(uAxis rk)
 {
    Assert1(rk < rank);
 
@@ -101,7 +101,7 @@ Rank axis = I->get_ravel(0).get_near_int() - qio;
 }
 //-----------------------------------------------------------------------------
 void
-IndexExpr::set_value(Axis axis, Value_P val)
+IndexExpr::set_value(uAxis axis, Value_P val)
 {
    // Note: we expect that all axes have been allocated with add(0)
    // Note also that IndexExpr is in parsing order, ie. 0 is the lowest axis.
@@ -137,7 +137,7 @@ int count = 0;
    for (const DynamicObject * dob = all_index_exprs.get_next();
         dob != &all_index_exprs; dob = dob->get_next())
        {
-         const IndexExpr * idx = (const IndexExpr *)dob;
+         const IndexExpr * idx = dob->pIndexExpr();
 
          out << dob->where_allocated();
 
@@ -158,20 +158,20 @@ IndexExpr::erase_stale(const char * loc)
    Log(LOG_Value__erase_stale)
       CERR << endl << endl << "erase_stale() called from " << loc << endl;
 
-   for (DynamicObject * vb = all_index_exprs.get_next();
-        vb != &all_index_exprs; vb = vb->get_next())
+   for (DynamicObject * dob = all_index_exprs.get_next();
+        dob != &all_index_exprs; dob = dob->get_next())
        {
-         IndexExpr * v = (IndexExpr *)vb;
+         IndexExpr * idx = dob->pIndexExpr();
 
          Log(LOG_Value__erase_stale)
             {
               CERR << "Erasing stale IndexExpr:" << endl
-                   << "  Allocated by " << v->alloc_loc << endl;
+                   << "  Allocated by " << idx->alloc_loc << endl;
             }
 
-         vb = vb->get_prev();
-         ((IndexExpr *)(vb->get_next()))->unlink();
-         delete (IndexExpr *)(vb->get_next());
+         dob = dob->get_prev();
+         idx->unlink();
+         delete idx;
        }
 }
 //-----------------------------------------------------------------------------
