@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2017  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -26,7 +26,7 @@
 
 //-----------------------------------------------------------------------------
 void
-IndexIterator::increment()
+IndexIterator::operator ++()
 {
    ++pos;
    if (pos >= get_index_count())
@@ -34,11 +34,11 @@ IndexIterator::increment()
          if (upper)
             {
               pos = 0;
-              upper->increment();
+              ++*upper;
             }
          else
             {
-              pos = get_index_count();   // so that done() works
+              pos = get_index_count();   // so that more() works
             }
       }
 
@@ -48,10 +48,10 @@ IndexIterator::increment()
 ostream &
 IndexIterator::print(ostream & out) const
 {
-   out << "Iterator "        << HEX(this) << ":"  << endl
-       << "   upper:       " << HEX(upper)        << endl
-       << "   value:       " << get_value()       << endl
-       << "   indices[" << get_index_count() << "] :";
+   out << "Iterator "     << voidP(this) << ":"  << endl
+       << "   upper:       " << voidP(upper)     << endl
+       << "   value:       " << get_value()      << endl
+       << "   indices[" << get_index_count()     << "] :";
 
    loop(i, get_index_count())   out << " " << get_pos(i);
 
@@ -80,7 +80,7 @@ TrueIndexIterator::TrueIndexIterator(ShapeItem w, Value_P value,
 //=============================================================================
 MultiIndexIterator::MultiIndexIterator(const Shape & shape,
                                        const IndexExpr & IDX)
-   : last_it(0),
+   : highest_it(0),
      lowest_it(0),
      empty(false)
 {
@@ -97,7 +97,7 @@ MultiIndexIterator::MultiIndexIterator(const Shape & shape,
    // ---------------
    //
 
-   if (shape.get_rank() != IDX.value_count())
+   if (shape.get_rank() != uRank(IDX.value_count()))
       {
         Log(LOG_error_throw)
            {
@@ -118,24 +118,23 @@ ShapeItem weight = 1;
          IndexIterator * new_it;
          if (!I)   // elided index
             {
-              new_it = (IndexIterator *)(new ElidedIndexIterator(weight, sh_r));
+              new_it = new ElidedIndexIterator(weight, sh_r);
               if (sh_r == 0)   empty = true;
             }
          else
             {
-              new_it = (IndexIterator *)
-                       (new TrueIndexIterator(weight, I, IDX.quad_io, sh_r));
+              new_it = new TrueIndexIterator(weight, I, IDX.quad_io, sh_r);
               if (I->element_count() == 0)   empty = true;
             }
 
          Log(LOG_delete)
-            CERR << "new    " << (const void *)new_it << " at " LOC << endl;
+            CERR << "new    " << voidP(new_it) << " at " LOC << endl;
 
-         if (last_it)   last_it->set_upper(new_it);
-         else           lowest_it = new_it;
+         if (highest_it)   highest_it->set_upper(new_it);
+         else              lowest_it = new_it;
 
          weight *= sh_r;
-         last_it = new_it;
+         highest_it = new_it;
        }
 }
 //-----------------------------------------------------------------------------
@@ -145,30 +144,24 @@ MultiIndexIterator::~MultiIndexIterator()
        {
          IndexIterator * del = it;
          it = it->get_upper();
-         Log(LOG_delete)   CERR << "delete " HEX(del) << " at " LOC << endl;
+         Log(LOG_delete)
+            CERR << "delete " HEX(uint64_t(del))
+                 << " at " LOC << endl;
          delete del;
        }
 }
 //-----------------------------------------------------------------------------
 ShapeItem
-MultiIndexIterator::next()
+MultiIndexIterator::operator ++(int)
 {
-   Assert(!lowest_it || !done());
+   Assert(!lowest_it || more());
 
 ShapeItem ret = 0;
    for (IndexIterator * it = lowest_it; it; it = it->get_upper())
             ret += it->get_value();
 
-   if (lowest_it)   lowest_it->increment();
+   if (lowest_it)   ++*lowest_it;
    return ret;
 }
-//-----------------------------------------------------------------------------
-bool
-MultiIndexIterator::done() const
-{
-   if (last_it == 0)   return true;
-   if (empty)          return true;
-   return last_it->done();
-}
 //=============================================================================
 
