@@ -24,6 +24,8 @@
 #include <string.h>
 #include <unistd.h>
 
+#include <string>
+
 #include "Error.hh"
 #include "LibPaths.hh"
 #include "PrintOperator.hh"
@@ -50,8 +52,8 @@ LibPaths::init(const char * argv0, bool logit)
 
    loop(d, LIB_MAX)
       {
-        if (root_from_env)        lib_dirs[d].cfg_src = LibDir::CSRC_ENV;
-        else if (root_from_pwd)   lib_dirs[d].cfg_src = LibDir::CSRC_ARGV0;
+        if      (root_from_env)   lib_dirs[d].cfg_src = LibDir::CSRC_ENV;
+        else if (root_from_pwd)   lib_dirs[d].cfg_src = LibDir::CSRC_PWD;
         else                      lib_dirs[d].cfg_src = LibDir::CSRC_NONE;
       }
 }
@@ -82,23 +84,25 @@ LibPaths::compute_bin_path(const char * argv0, bool logit)
               // string.
               //
               const size_t plen = strlen(path);
-              Simple_string<char, false>   path1;
+              std::string   path1;
               path1.reserve(plen + 1);
-              loop(p, (plen + 1))   path1.append(path[p]);
+              loop(p, (plen + 1))   path1 += path[p];
               char * next = &path1[0];
               for (;;)
                   {
                     char * semi = strchr(next, ':');
                     if (semi)   *semi = 0;
                     UTF8_string filename;
-                    for (const char * n = next; *n; ++n)   filename.append(*n);
-                    filename.append('/');
-                    for (const char * a = argv0; *a; ++a)   filename.append(*a);
+                    for (const char * n = next; *n; ++n)   filename += *n;
+                    filename += '/';
+                    for (const char * a = argv0; *a; ++a)   filename += *a;
 
                     if (access(filename.c_str(), X_OK) == 0)
                        {
                          strncpy(APL_bin_path, filename.c_str(),
-                                 sizeof(APL_bin_path));
+                                 sizeof(APL_bin_path) - 1);
+                         APL_bin_path[sizeof(APL_bin_path) - 1] = 0;
+
                          char * slash =   strrchr(APL_bin_path, '/');
                          Assert(slash);   // due to %s/%s above
                          *slash = 0;
@@ -166,6 +170,8 @@ char filename[APL_PATH_MAX + 1];
 void
 LibPaths::search_APL_lib_root()
 {
+   APL_lib_root[0] = 0;
+
 const char * path = getenv("APL_LIB_ROOT");
    if (path)
       {
@@ -174,36 +180,25 @@ const char * path = getenv("APL_LIB_ROOT");
         return;
       }
 
+   root_from_pwd = true;
+
    // search from "." to "/" for  a valid lib-root
    //
-int last_len = 2*APL_PATH_MAX;
-   APL_lib_root[0] = '.';
-   APL_lib_root[1] = 0;
-   for (;;)
-       {
-         unused = realpath(APL_lib_root, APL_lib_root);
-         int len = strlen(APL_lib_root);
-
-         if (is_lib_root(APL_lib_root))   return;   // lib-rot found
-
-         // if length has not decreased then we stop in order to
-         // avoid an endless loop.
-         //
-         if (len >= last_len)   break;
-         last_len = len;
-
-         // move up by appending /..
-         //
-         APL_lib_root[len++] = '/';
-         APL_lib_root[len++] = '.';
-         APL_lib_root[len++] = '.';
-         APL_lib_root[len++] = 0;
-       }
-
-   // no lib-root found. use ".";
-   //
    unused = realpath(".", APL_lib_root);
-   root_from_pwd = true;
+   while (strlen(APL_lib_root))
+      {
+        if (is_lib_root(APL_lib_root))   return;   // lib-root found
+        if (char * s = strrchr(APL_lib_root, '/'))         *s = 0;
+        else if (char * s = strrchr(APL_lib_root, '\\'))   *s = 0;
+        else
+           {
+             CERR << "*** Cannot locate APL_lib_root: no / or \\ in "
+                  << APL_lib_root << endl;
+             break;
+           }
+      }
+
+   unused = realpath(".", APL_lib_root);
 }
 //-----------------------------------------------------------------------------
 void
@@ -227,24 +222,24 @@ LibPaths::get_lib_dir(LibRef libref)
         case LibDir::CSRC_NONE:      return UTF8_string();
 
         case LibDir::CSRC_ENV:
-        case LibDir::CSRC_ARGV0:     break;   // continue below
+        case LibDir::CSRC_PWD:       break;   // continue below
 
         case LibDir::CSRC_PREF_SYS:
         case LibDir::CSRC_PREF_HOME:
-        case LibDir::CSRC_CMD:     return lib_dirs[libref].dir_path;
+        case LibDir::CSRC_CMD:       return lib_dirs[libref].dir_path;
       }
 
 UTF8_string ret(APL_lib_root);
    if (libref == LIB0)   // workspaces
       {
         const UTF8_string subdir("/workspaces");
-        ret.append(subdir);
+        ret.append_UTF8(subdir);
       }
    else                  // wslibN
       {
         const UTF8_string subdir("/wslib");
-        ret.append(subdir);
-        ret.append(Unicode(libref + '0'));
+        ret.append_UTF8(subdir);
+        ret += libref + '0';
       }
 
    return ret;
@@ -258,7 +253,7 @@ LibPaths::maybe_warn_ambiguous(int name_has_extension, const UTF8_string name,
    if (ext2 == 0)            return;   // no second extension
 
 UTF8_string filename_ext2 = name;
-   filename_ext2.append_str(ext2);
+   filename_ext2.append_ASCII(ext2);
    if (access(filename_ext2.c_str(), F_OK))   return;   // not existing
 
    CERR << endl 
@@ -283,7 +278,7 @@ int name_has_extension = 0;   // assume name has neither extension ext1 nor ext2
        name.starts_with("../"))
       {
         // paths from / or ./ are fallbacks for the case where the library
-        // path setup id wrong. So that the user can survive by using an
+        // path setup is wrong. So that the user can survive by using an
         // explicit path
         //
         if (name_has_extension)   return name;
@@ -294,7 +289,7 @@ int name_has_extension = 0;   // assume name has neither extension ext1 nor ext2
         if (ext1)
            {
              UTF8_string filename_ext1 = name;
-             filename_ext1.append_str(ext1);
+             filename_ext1.append_ASCII(ext1);
              if (!access(filename_ext1.c_str(), F_OK))
                 {
                    // filename_ext1 exists, but filename_ext2 may exist as well.
@@ -308,7 +303,7 @@ int name_has_extension = 0;   // assume name has neither extension ext1 nor ext2
         if (ext2)
            {
              UTF8_string filename_ext2 = name;
-             filename_ext2.append_str(ext2);
+             filename_ext2.append_ASCII(ext2);
              if (!access(filename_ext2.c_str(), F_OK))   return filename_ext2;
            }
 
@@ -319,8 +314,8 @@ int name_has_extension = 0;   // assume name has neither extension ext1 nor ext2
       }
 
 UTF8_string filename = get_lib_dir(lib);
-   filename.append(UNI_ASCII_SLASH);
-   filename.append(name);
+   filename += '/';
+   filename.append_UTF8(name);
 
    if (name_has_extension)   return filename;
 
@@ -334,7 +329,7 @@ UTF8_string filename = get_lib_dir(lib);
         if (ext1)
            {
              UTF8_string filename_ext1 = filename;
-             filename_ext1.append_str(ext1);
+             filename_ext1.append_ASCII(ext1);
              if (!access(filename_ext1.c_str(), F_OK))
                 {
                   maybe_warn_ambiguous(name_has_extension,
@@ -346,7 +341,7 @@ UTF8_string filename = get_lib_dir(lib);
         if (ext2)
            {
              UTF8_string filename_ext2 = filename;
-             filename_ext2.append_str(ext2);
+             filename_ext2.append_ASCII(ext2);
              if (!access(filename_ext2.c_str(), F_OK))   return filename_ext2;
            }
 
@@ -360,8 +355,8 @@ UTF8_string filename = get_lib_dir(lib);
         //
         if (name_has_extension)   return filename;
 
-        if      (ext1) filename.append_str(ext1);
-        else if (ext2) filename.append_str(ext2);
+        if      (ext1) filename.append_ASCII(ext1);
+        else if (ext2) filename.append_ASCII(ext2);
         return filename;
       }
 }
