@@ -23,6 +23,7 @@
 
 #include "Assert.hh"
 #include "Command.hh"
+#include "Common.hh"
 #include "Error.hh"
 #include "InputFile.hh"
 #include "IO_Files.hh"
@@ -30,6 +31,7 @@
 #include "Nabla.hh"
 #include "Parallel.hh"
 #include "SystemVariable.hh"
+#include "TabExpansion.hh"
 #include "UserPreferences.hh"
 #include "Workspace.hh"
 
@@ -170,8 +172,8 @@ int count = 0;
          buffer[sizeof(buffer) - 1] = 0;
 
          int slen = strlen(buffer);
-         if (buffer[slen - 1] == '\n')   buffer[--slen] = 0;
-         if (buffer[slen - 1] == '\r')   buffer[--slen] = 0;
+         if (slen && (buffer[slen - 1] == '\n'))   buffer[--slen] = 0;
+         if (slen && (buffer[slen - 1] == '\r'))   buffer[--slen] = 0;
          ++count;
          UTF8_string utf(buffer);
          UCS_string ucs(utf);
@@ -195,7 +197,7 @@ ofstream outf(filename);
       }
 
 int count = 0;
-   for (int p = put + 1; p < hist_lines.size(); ++p)
+   for (size_t p = put + 1; p < hist_lines.size(); ++p)
       {
         outf << hist_lines[p] << endl;
         ++count;
@@ -215,7 +217,7 @@ LineHistory::clear_history(ostream & out)
 {
    current_line = 0;
    put = 0;
-   hist_lines.shrink(0);
+   hist_lines.clear();
 UCS_string u("xxx");
    add_line(u);
 }
@@ -223,7 +225,7 @@ UCS_string u("xxx");
 void
 LineHistory::print_history(ostream & out)
 {
-   for (int p = put + 1; p < hist_lines.size(); ++p)
+   for (size_t p = put + 1; p < hist_lines.size(); ++p)
       {
         out << "      " << hist_lines[p] << endl;
       }
@@ -236,16 +238,17 @@ LineHistory::print_history(ostream & out)
 void
 LineHistory::add_line(const UCS_string & line)
 {
+   if (max_lines == 0)      return;   // no history
    if (!line.has_black())   return;
 
-   if ((int)hist_lines.size() < max_lines)   // append
+   if (int(hist_lines.size()) < max_lines)   // append
       {
-        hist_lines.append(line);
+        hist_lines.push_back(line);
         put = 0;
       }
    else                            // override
       {
-        if (put >= (int)hist_lines.size())   put = 0;   // wrap
+        if (put >= int(hist_lines.size()))   put = 0;   // wrap
         hist_lines[put++] = line;
       }
 
@@ -256,7 +259,7 @@ void
 LineHistory::replace_line(const UCS_string & line)
 {
    if (put > 0)   hist_lines[put - 1] = line;
-   else           hist_lines.last() = line;
+   else           hist_lines.back() = line;
 }
 //-----------------------------------------------------------------------------
 const UCS_string *
@@ -278,7 +281,8 @@ LineHistory::down()
    if (current_line == put)   return 0;
 
 int new_current_line = current_line + 1;
-   if (new_current_line >= (int)hist_lines.size())   new_current_line = 0;   // wrap
+   if (new_current_line >= int(hist_lines.size()))
+      new_current_line = 0;   // wrap
    if (new_current_line == put)   return 0;
 
    return &hist_lines[current_line = new_current_line];
@@ -316,7 +320,7 @@ LineEditContext::~LineEditContext()
    //
    if (!Output::use_curses)
       {
-        if (!ins_mode)   CIN << (char)0x1B << "[1 q" << flush;
+        if (!ins_mode)   CIN << "\x1B[1 q" << flush;
       }
 }
 //-----------------------------------------------------------------------------
@@ -407,7 +411,7 @@ LineEditContext::delete_char()
    if (uidx == (user_line.size() - 1))   // cursor on last char
       {
         CIN << ' ' << UNI_ASCII_BS;
-        user_line.pop();
+        user_line.pop_back();
       }
    else
       {
@@ -430,7 +434,7 @@ LineEditContext::insert_char(Unicode uni)
       }
    else if (ins_mode)              // insert char
       {
-        user_line.insert_before(uidx, uni);
+        user_line.insert(uidx, uni);
         adjust_allocated_height();
         refresh_wrapped_cursor();
         refresh_from_cursor();
@@ -452,7 +456,7 @@ LineEditContext::cut_to_EOL()
    if (uidx >= user_line.size())   return;   // nothing to cut
 
    cut_buffer = UCS_string(user_line, uidx, user_line.size() - uidx);
-   user_line.shrink(uidx);
+   user_line.resize(uidx);
    refresh_from_cursor();
 }
 //-----------------------------------------------------------------------------
@@ -468,7 +472,7 @@ LineEditContext::paste()
    else                            // insert cut buffer
       {
         const UCS_string rest(user_line, uidx, user_line.size() - uidx);
-        user_line.shrink(uidx);
+        user_line.resize(uidx);
         user_line.append(cut_buffer);
         user_line.append(rest);
       }
@@ -492,8 +496,8 @@ LineEditContext::toggle_ins_mode()
 
    if (!Output::use_curses)
       {
-        if (ins_mode)   CIN << (char)0x1B << "[0 q" << flush;
-        else            CIN << (char)0x1B << "[3 q" << flush;
+        if (ins_mode)   CIN << "\x1B[0 q" << flush;
+        else            CIN << "\x1B[3 q" << flush;
       }
 }
 //-----------------------------------------------------------------------------
@@ -502,9 +506,9 @@ LineEditContext::tab_expansion(LineInputMode mode)
 {
    if (mode != LIM_ImmediateExecution)   return;
 
-int replace_count = 0;
 UCS_string line = user_line;
-const ExpandResult expand_result = Command::expand_tab(line, replace_count);
+TabExpansion tab_exp(line);
+const ExpandResult expand_result = tab_exp.expand_tab(line);
 
    switch(expand_result)
       {
@@ -519,7 +523,7 @@ const ExpandResult expand_result = Command::expand_tab(line, replace_count);
              return;
 
         case ER_REPLACE:
-             user_line.shrink(user_line.size() - replace_count);
+             user_line.clear();
              user_line.append(line);
              uidx = 0;
              refresh_from_cursor();
@@ -611,7 +615,9 @@ LineInput::LineInput(bool do_read_history)
    current_termios.c_lflag |= ISIG;
 
 #ifndef WANT_LIBAPL
+# ifndef WANT_PYTHON
    tcsetattr(STDIN_FILENO, TCSANOW, &current_termios);
+# endif // WANT_PYTHON
 #endif // WANT_LIBAPL
 }
 //-----------------------------------------------------------------------------
@@ -622,11 +628,14 @@ LineInput::~LineInput()
    if (write_history)   history.save_history(uprefs.line_history_path.c_str());
 
 #ifndef WANT_LIBAPL
+# ifndef WANT_PYTHON
    tcsetattr(STDIN_FILENO, TCSANOW, &initial_termios);
+# endif // WANT_PYTHON
 #endif // WANT_LIBAPL
 }
 //-----------------------------------------------------------------------------
-void LineInput::init(bool do_read_history)
+void
+LineInput::init(bool do_read_history)
 {
    the_line_input = new LineInput(do_read_history);
 }
@@ -679,9 +688,9 @@ InputMux::get_line(LineInputMode mode, const UCS_string & prompt,
                               file_line[0] == UNI_ASCII_BS)
                              {
                                file_line.erase(0);
-                               line.pop();
+                               line.pop_back();
                              }
-                       line.append_utf8(file_line.c_str());
+                       line.append_UTF8(file_line.c_str());
                        break;
 
                   case LIM_Nabla:
@@ -713,8 +722,8 @@ InputMux::get_line(LineInputMode mode, const UCS_string & prompt,
         buffer[sizeof(buffer) - 1] = 0;
 
         int slen = strlen(buffer);
-        if (buffer[slen - 1] == '\n')   buffer[--slen] = 0;
-        if (buffer[slen - 1] == '\r')   buffer[--slen] = 0;
+        if (slen && buffer[slen - 1] == '\n')   buffer[--slen] = 0;
+        if (slen && buffer[slen - 1] == '\r')   buffer[--slen] = 0;
 
         UTF8_string line_utf(buffer);
         line = UCS_string(line_utf);
@@ -802,17 +811,13 @@ LineInput::get_terminal_line(LineInputMode mode, const UCS_string & prompt,
         case LIM_Quote_Quad:
         case LIM_Quad_Quad:
         case LIM_Quad_INP:
-             {
-               Output::set_color_mode(Output::COLM_INPUT);
-               edit_line(mode, prompt, line, eof, hist);
-               return;
-             }
+             Output::set_color_mode(Output::COLM_INPUT);
+
+             /* fall through */
 
         case LIM_Nabla:
-             {
-               edit_line(LIM_Nabla, prompt, line, eof, hist);
-               return;
-             }
+             edit_line(mode, prompt, line, eof, hist);
+             return;
 
         default: FIXME;
       }
@@ -822,15 +827,14 @@ LineInput::get_terminal_line(LineInputMode mode, const UCS_string & prompt,
 //-----------------------------------------------------------------------------
 void
 LineInput::edit_line(LineInputMode mode, const UCS_string & prompt,
-                       UCS_string & user_line, bool & eof,
-                       LineHistory & hist)
+                     UCS_string & user_line, bool & eof, LineHistory & hist)
 {
    the_line_input->current_termios.c_lflag &= ~ISIG;   // disable ^C
 #ifndef WANT_LIBAPL
    tcsetattr(STDIN_FILENO, TCSANOW, &the_line_input->current_termios);
 #endif // WANT_LIBAPL
 
-   user_line.shrink(0);
+   user_line.clear();
 
 LineEditContext lec(mode, 24, Workspace::get_PW(), hist, prompt);
 
@@ -909,7 +913,7 @@ LineEditContext lec(mode, 24, Workspace::get_PW(), hist, prompt);
 
               case Invalid_Unicode:
                    continue;
- 
+
               default:  // regular APL character
                    lec.insert_char(uni);
                    continue;
@@ -956,22 +960,32 @@ bool add_hist = false;
    CIN << endl;
 }
 //-----------------------------------------------------------------------------
+int
+LineInput::safe_fgetc()
+{
+   for (;;)
+       {
+          const int ret = fgetc(stdin);
+          if (ret != EOF)       return ret;
+          if (errno == EINTR)   continue;
+
+          if (got_WINCH)
+             {
+               got_WINCH = false;
+               continue;
+             }
+
+         return EOF;
+       }
+}
+//-----------------------------------------------------------------------------
 Unicode
 LineInput::get_uni()
 {
 again:
 
-const int b0 = fgetc(stdin);
-   if (b0 == EOF)
-      {
-       if (got_WINCH)
-          {
-            got_WINCH = false;
-            goto again;
-          }
-
-         return UNI_EOF;
-      }
+const int b0 = safe_fgetc();
+   if (b0 == EOF)   return UNI_EOF;
 
    if (b0 & 0x80)   // non-ASCII unicode
       {
@@ -991,7 +1005,7 @@ const int b0 = fgetc(stdin);
         uint32_t uni = 0;
         loop(l, len - 1)
             {
-              const UTF8 subc = fgetc(stdin);
+              const UTF8 subc = safe_fgetc();
               if ((subc & 0xC0) != 0x80)
                  {
                    CERR << "Bad UTF8 sequence: " << HEX(b0)
@@ -1004,7 +1018,7 @@ const int b0 = fgetc(stdin);
               uni |= subc & 0x3F;
             }
 
-        return (Unicode)(bx | uni);
+        return Unicode(bx | uni);
       }
 
    if (b0 == UNI_ASCII_ESC)
@@ -1012,7 +1026,7 @@ const int b0 = fgetc(stdin);
         char seq[Output::MAX_ESC_LEN];   seq[0] = UNI_ASCII_ESC;
         for (int s = 1; s < Output::MAX_ESC_LEN; ++s)
             {
-              const int bs = fgetc(stdin);
+              const int bs = safe_fgetc();
               if (bs == EOF)   return UNI_EOF;
               seq[s] = bs;
 
@@ -1057,6 +1071,6 @@ const int b0 = fgetc(stdin);
       }
    else if (b0 == UNI_ASCII_DELETE)   return UNI_ASCII_BS;
 
-   return (Unicode)b0;
+   return Unicode(b0);
 }
 //-----------------------------------------------------------------------------
