@@ -21,13 +21,13 @@
 #include "LvalCell.hh"
 #include "PrintOperator.hh"
 #include "UTF8_string.hh"
-#include "Value.icc"
+#include "Value.hh"
 
 //-----------------------------------------------------------------------------
 LvalCell::LvalCell(Cell * cell, Value * cell_owner)
 {
    value.lval = cell;
-   value2.owner = cell_owner;
+   value.pval.owner = cell_owner;
 }
 //-----------------------------------------------------------------------------
 Cell *
