@@ -32,8 +32,6 @@
 #include "UserPreferences.hh"
 #include "Workspace.hh"
 
-#define NABLA_ERROR throw_edit_error(LOC)
-
 //-----------------------------------------------------------------------------
 void
 Nabla::edit_function(const UCS_string & cmd)
@@ -55,7 +53,7 @@ Nabla::Nabla(const UCS_string & cmd)
      current_line(1),
      first_command(cmd)
 {
-   Workspace::more_error().shrink(0);
+   Workspace::more_error().clear();
 }
 //-----------------------------------------------------------------------------
 void
@@ -71,7 +69,7 @@ Nabla::throw_edit_error(const char * why)
         MORE_ERROR() << why;
       }
 
-   throw_define_error(fun_header, first_command, why);
+   Error::throw_define_error(fun_header, first_command, why);
 }
 //-----------------------------------------------------------------------------
 void
@@ -134,7 +132,10 @@ int control_D_count = 0;
        }
 
    Log(LOG_nabla)
-      UERR << "done '" << fun_header << "'" << endl;
+      {
+        UERR << "done: '" << fun_header << "'" << endl;
+        loop(l, lines.size())   UERR << lines[l].text << endl;
+      }
 
 UCS_string fun_text;
    loop(l, lines.size())
@@ -160,11 +161,11 @@ UCS_string fun_text;
           LineInput::replace_history_line(line_0);
         }
 
-        for (int l = 1; l < lines.size(); ++l)
+        for (size_t l = 1; l < lines.size(); ++l)
             {
               UCS_string line_l("[");
               line_l.append_number(l);
-              line_l.append_utf8("]  ");
+              line_l.append_UTF8("]  ");
               while (line_l.size() < 6)   line_l.append(UNI_ASCII_SPACE);
               line_l.append(lines[l].text);
               line_l.remove_trailing_whitespaces();
@@ -185,7 +186,8 @@ UTF8_string creator_utf8(creator);
 
 UserFunction * ufun = UserFunction::fix(fun_text, error_line, false,
                                         LOC, creator_utf8, false);
-   if (ufun == 0)   NABLA_ERROR;
+
+   if (ufun == 0)   throw_edit_error(LOC);
 
    if (locked)
       {
@@ -195,12 +197,12 @@ UserFunction * ufun = UserFunction::fix(fun_text, error_line, false,
 
    // set stop and trace vectors
    //
-Simple_string<Function_Line, false> stop_vec;
-Simple_string<Function_Line, false> trace_vec;
+std::vector<Function_Line> stop_vec;
+std::vector<Function_Line> trace_vec;
    loop(l, lines.size())
        {
-         if (lines[l].stop_flag)    stop_vec.append(Function_Line(l));
-         if (lines[l].trace_flag)   trace_vec.append(Function_Line(l));
+         if (lines[l].stop_flag)    stop_vec.push_back(Function_Line(l));
+         if (lines[l].trace_flag)   trace_vec.push_back(Function_Line(l));
        }
 
    ufun->set_trace_stop(stop_vec,  true);
@@ -269,7 +271,7 @@ UserFunction_header hdr(fun_header, false);
       {
         UCS_string oper;
         Unicode cc;
-        do 
+        do
           {
             if (!c.more())   return "no ] in ∇-command";
             oper.append(cc = c.next());
@@ -353,29 +355,30 @@ Nabla::parse_oper(UCS_string & oper, bool initial)
 
    // skip trailing spaces
    //
-   while (oper.size() > 0 && oper.last() <= ' ')   oper.pop();
-   if (oper.size() > 0 && oper.last() == UNI_NABLA)
+   while (oper.size() > 0 && oper.back() <= ' ')   oper.pop_back();
+   if (oper.size() > 0 && oper.back() == UNI_NABLA)
       {
         do_close = true;
-        oper.pop();
-        while (oper.size() > 0 && oper.last() <= ' ')   oper.pop();
+        oper.pop_back();
+        while (oper.size() > 0 && oper.back() <= ' ')   oper.pop_back();
       }
-   else if (oper.size() > 0 && oper.last() == UNI_DEL_TILDE)
+   else if (oper.size() > 0 && oper.back() == UNI_DEL_TILDE)
       {
         do_close = true;
         locked = true;
-        oper.pop();
-        while (oper.size() > 0 && oper.last() <= ' ')   oper.pop();
+        oper.pop_back();
+        while (oper.size() > 0 && oper.back() <= ' ')   oper.pop_back();
       }
 
-   current_text.shrink(0);
+   current_text.clear();
    ecmd = ECMD_NOP;
 
    if (oper.size() == 0 && do_close)   return 0;
 
 UCS_string::iterator c(oper.begin());
 Unicode cc = c.next();
-   while (cc == ' ')   cc = c.next();
+UCS_string text = oper;
+   while (cc == ' ')   cc = c.next();   // skip leading whitespace
 
    // we expect one of the following:
    //
@@ -385,11 +388,12 @@ Unicode cc = c.next();
    // [n]                                           (goto)
    // text                                          (override text)
 
-   if (cc != UNI_ASCII_L_BRACK)
+   if (cc != UNI_ASCII_L_BRACK)   // override text
       {
         ecmd = ECMD_EDIT;
         edit_from = current_line;
-        for (; cc != Invalid_Unicode; cc = c.next())   current_text.append(cc);
+        current_text = text;
+//      for (; cc != Invalid_Unicode; cc = c.next())   current_text.append(cc);
         return 0;
       }
 
@@ -569,7 +573,7 @@ Nabla::open_new_function()
       UERR << "creating new function '" << fun_symbol->get_name() 
            << "' with header '" << fun_header << "'" << endl;
 
-   lines.append(FunLine(0, fun_header));
+   lines.push_back(FunLine(0, fun_header));
    return 0;
 }
 //-----------------------------------------------------------------------------
@@ -636,7 +640,7 @@ UCS_string_vector tlines;
                   }
              }
 
-         lines.append(fl);
+         lines.push_back(fl);
        }
 
    current_line = LineLabel(tlines.size());
@@ -696,7 +700,7 @@ const LineLabel user_edit_to = edit_to;
 
    if (idx_from == 0)   COUT << "    ∇" << endl;               // if header line
    for (int e = idx_from; e <= idx_to; ++e)   lines[e].print(COUT);
-   if (idx_to == lines.size() - 1)   COUT << "    ∇" << endl;  // if last line
+   if (idx_to == int(lines.size() - 1))   COUT << "    ∇" << endl;  // if last line
 
    if (user_edit_to.valid())   // eg. [⎕42] or [2⎕42]
       {
@@ -713,7 +717,7 @@ const LineLabel user_edit_to = edit_to;
       }
    else                      // eg. [42⎕] or [⎕]
       {
-        current_line = lines.last().label;
+        current_line = lines.back().label;
         current_line.next();
       }
 
@@ -734,20 +738,20 @@ const int idx_to = find_line(LineLabel(edit_to));
 
    modified = true;
 
-   if (edit_from == -1)   // delete single line
+   if (edit_from == -1)   // [∆N] : delete single line
       {
-        lines.erase(idx_to);
+        lines.erase(lines.begin() + idx_to);
         return 0;
       }
 
-   // delete multiple lines
+   // [N∆M] : delete multiple lines
    //
 const int idx_from = find_line(LineLabel(edit_from));
    if (idx_from == -1)       return "Bad line number M in [M∆N] ";
    if (idx_from >= idx_to)   return "M ≥ N in [M∆N] ";
 
-   loop(j, idx_to - idx_from)   lines.erase(idx_from);
-   current_line = lines.last().label;
+   loop(j, idx_to - idx_from)   lines.erase(lines.begin() + idx_from);
+   current_line = lines.back().label;
    return 0;
 }
 //-----------------------------------------------------------------------------
@@ -820,9 +824,19 @@ Nabla::edit_body_line()
 const Parser parser(PM_FUNCTION, LOC, false);
 Token_string in;
 ErrorCode ec = parser.parse(current_text, in);
+   if (ec == E_NO_STRING_END && uprefs.multi_line_strings)
+      {
+        ec = E_NO_ERROR;
+        Workspace::more_error().clear();
+      }
+
    if (ec)
       {
         CERR << "SYNTAX ERROR";
+        if (Workspace::more_error().size())
+           {
+             CERR << "+" << endl << Workspace::more_error();
+           }
         COUT << endl;
         return 0;
       }
@@ -837,14 +851,14 @@ const int idx_from = find_line(edit_from);
         // find the largest label before edit_from (if any)
         //
         int before_idx = -1;
-        for (int i = 0; i < lines.size(); ++i)
+        for (size_t i = 0; i < lines.size(); ++i)
             {
               if (lines[i].label < edit_from)   before_idx = i;
-              else                                 break;
+              else                              break;
             }
 
         FunLine fl(edit_from, current_text);
-        lines.insert_before(before_idx + 1, fl);
+        lines.insert(lines.begin() + before_idx + 1, fl);
       }
    else
       {
@@ -861,8 +875,8 @@ const int idx_from = find_line(edit_from);
 const char *
 Nabla::execute_escape()
 {
-   lines.shrink(0);
-   lines.append(FunLine(0, fun_header));
+   lines.clear();
+   lines.push_back(FunLine(0, fun_header));
    return 0;
 }
 //-----------------------------------------------------------------------------
@@ -894,12 +908,12 @@ UCS_string ucs("[");
    ucs.append_number(ln_major);
    if (ln_minor.size())
       {
-        ucs.append_utf8(".");
+        ucs.append_UTF8(".");
         ucs.append(ln_minor);
       }
-   ucs.append_utf8("]");
+   ucs.append_UTF8("]");
 
-   while (ucs.size() < 5)   ucs.append_utf8(" ");
+   while (ucs.size() < 5)   ucs.append_UTF8(" ");
    out << ucs;
 }
 //-----------------------------------------------------------------------------
@@ -915,7 +929,7 @@ UCS_string ret("[");
         loop(s, ln_minor.size())   ret.append(Unicode(char(ln_minor[s])));
       }
 
-   ret.append_utf8("] ");
+   ret.append_UTF8("] ");
    while (ret.size() < min_size)   ret.append(UNI_ASCII_SPACE);
    return ret;
 }
