@@ -30,12 +30,12 @@
 #include "Symbol.hh"
 #include "Workspace.hh"
 
-Simple_string<NativeFunction *, false> NativeFunction::valid_functions;
+std::vector<NativeFunction *> NativeFunction::valid_functions;
 
 //-----------------------------------------------------------------------------
 NativeFunction::NativeFunction(const UCS_string & so_name,
                                const UCS_string & apl_name)
-   : Function(ID::USER_SYMBOL, TOK_FUN2),
+   : Function(ID_USER_SYMBOL, TOK_FUN2),
      handle(0),
      name(apl_name),
      original_so_path(so_name),
@@ -66,7 +66,8 @@ void * fmux = dlsym(handle, "get_function_mux");
           return;
         }
 
-void * (*get_function_mux)(const char *) = (void * (*)(const char *))fmux;
+void * (*get_function_mux)(const char *) =
+                 reinterpret_cast<void * (*)(const char *)>(fmux);
 
    // get the mandatory function get_signature() which returns
    //  the function signature
@@ -83,14 +84,15 @@ void * (*get_function_mux)(const char *) = (void * (*)(const char *))fmux;
         }
 
 
-     signature = ((Fun_signature (*)())get_sig)();
+     signature = reinterpret_cast<Fun_signature (*)()>(get_sig)();
    }
 
    // get the optional function close_fun(), which is called before
    // this function disappears
    {
      void * cfun = get_function_mux("close_fun");
-     if (cfun)   close_fun = (bool (*)(Cause, const NativeFunction *))cfun;
+     if (cfun)   close_fun = reinterpret_cast
+                             <bool (*)(Cause, const NativeFunction *)>(cfun);
      else        close_fun = 0;
    }
 
@@ -110,7 +112,8 @@ const char * why = sym->cant_be_defined();
    //
 #define Th const NativeFunction * th
 
-#define ev(fun, args) f_ ## fun = (Token (*) args ) get_function_mux(#fun)
+#define ev(fun, args) \
+   f_ ## fun = reinterpret_cast<Token (*) args>(get_function_mux(#fun))
 
    ev(eval_        , (                                Th));
 
@@ -144,9 +147,9 @@ const char * why = sym->cant_be_defined();
    if (is_operator())   sym->set_nc(NC_OPERATOR, this);
    else                 sym->set_nc(NC_FUNCTION, this);
 
-   Workspace::more_error().shrink(0);
+   Workspace::more_error().clear();
    valid = true;
-   valid_functions.append(this);
+   valid_functions.push_back(this);
 }
 //-----------------------------------------------------------------------------
 NativeFunction::~NativeFunction()
@@ -158,7 +161,7 @@ NativeFunction::~NativeFunction()
       {
         if (valid_functions[v] == this)
            {
-             valid_functions.erase(v);
+             valid_functions.erase(valid_functions.begin() + v);
            }
       }
 }
@@ -167,15 +170,15 @@ void *
 NativeFunction::try_one_file(const char * filename, UCS_string & t4)
 {
 const int t4_len = t4.size();
-   t4.append_utf8("    file ");
-   t4.append_utf8(filename);
+   t4.append_UTF8("    file ");
+   t4.append_UTF8(filename);
 
    if (access(filename, R_OK) != 0)
       {
-        while (t4.size() < t4_len + 44)     t4.append_utf8(" ");
-        t4.append_utf8(" (");
-        t4.append_utf8(strerror(errno));
-        t4.append_utf8(")\n");
+        while (t4.size() < t4_len + 44)     t4.append_UTF8(" ");
+        t4.append_UTF8(" (");
+        t4.append_UTF8(strerror(errno));
+        t4.append_UTF8(")\n");
         return 0;
       }
 
@@ -185,10 +188,10 @@ void * handle = dlopen(filename, RTLD_NOW);
         const char * err = dlerror();
         if (strrchr(err, ':'))   err = 1 + strrchr(err, ':');
 
-        while (t4.size() < t4_len + 44)     t4.append_utf8(" ");
-        t4.append_utf8(" (");
-        t4.append_utf8(err);
-        t4.append_utf8(" )\n");
+        while (t4.size() < t4_len + 44)     t4.append_UTF8(" ");
+        t4.append_UTF8(" (");
+        t4.append_UTF8(err);
+        t4.append_UTF8(" )\n");
       }
 
    return handle;
@@ -199,10 +202,10 @@ NativeFunction::open_so_file(UCS_string & t4, UCS_string & so_path)
 {
    // prepare a )MORE error message containing the file names tried.
    //
-   t4.shrink(0);
-   t4.append_utf8("Could not find shared library '");
+   t4.clear();
+   t4.append_UTF8("Could not find shared library '");
    t4.append(so_path);
-   t4.append_utf8("'\n"
+   t4.append_UTF8("'\n"
                   "The following directories and file names were tried:\n");
 
    // if the name starts with / (or \ on Windows) or .
@@ -217,7 +220,7 @@ NativeFunction::open_so_file(UCS_string & t4, UCS_string & so_path)
 
         if (handle == 0)
            {
-             t4.append_utf8(
+             t4.append_UTF8(
             "NOTE: Filename extensions are NOT automatically added "
             "when a full path\n"
             "      (i.e. a path starting with / or .) is used.");
@@ -250,16 +253,16 @@ const char * dirs[] =
          if (dirs[d] == 0)   continue;
 
          UTF8_string dir_so_path(dirs[d]);
-         dir_so_path.append(UTF8_string("/"));
-         dir_so_path.append(utf_so_path);
+         dir_so_path += '/';
+         dir_so_path.append_UTF8(utf_so_path);
 
          UTF8_string dir_only(dir_so_path);
          dir_only[strrchr(dir_only.c_str(), '/') - dir_only.c_str()] = 0;
          if (access(dir_only.c_str(), R_OK | X_OK))
             {
-              t4.append_utf8("    directory ");
-              t4.append_utf8(dir_only.c_str());
-              t4.append_utf8("\n");
+              t4.append_UTF8("    directory ");
+              t4.append_UTF8(dir_only.c_str());
+              t4.append_UTF8("\n");
               continue;   // new directory
             }
 
@@ -289,7 +292,7 @@ const char * dirs[] =
                if (has_extension && *exts[e])   continue;
 
                UTF8_string filename(dir_so_path);
-               if (exts[e])   filename.append(UTF8_string(exts[e]));
+               if (exts[e])   filename.append_UTF8(UTF8_string(exts[e]));
 
                void * handle = try_one_file(filename.c_str(), t4);
                if (handle)   // found library
@@ -310,9 +313,9 @@ NativeFunction::cleanup()
    //
    while(valid_functions.size())
       {
-        NativeFunction * fun = valid_functions.last();
-        valid_functions.pop();
-        
+        NativeFunction * fun = valid_functions.back();
+        valid_functions.pop_back();
+
         if (fun->close_fun && fun->handle)
            {
              const bool do_dlclose = (*fun->close_fun)(CAUSE_SHUTDOWN, fun);
@@ -331,7 +334,6 @@ NativeFunction::cleanup()
 void
 NativeFunction::destroy()
 {
-   if (close_fun)   (*close_fun)(CAUSE_ERASED, this);
    if (close_fun && handle)
       {
         const bool do_dlclose = (*close_fun)(CAUSE_ERASED, this);
@@ -339,8 +341,15 @@ NativeFunction::destroy()
            {
              dlclose(handle);
              handle = 0;
+             close_fun = 0;
            }
       }
+   else if (handle)
+      {
+        dlclose(handle);
+        handle = 0;
+      }
+
    delete this;
 }
 //-----------------------------------------------------------------------------
@@ -376,13 +385,13 @@ NativeFunction::fix(const UCS_string & so_name,
 
 NativeFunction * new_function = new NativeFunction(so_name, function_name);
    Log(LOG_delete)
-      CERR << "new    " << (const void *)new_function << " at " LOC << endl;
+      CERR << "new    " << voidP(new_function) << " at " LOC << endl;
 
 
    if (!new_function->valid)   // something went wrong
       {
         Log(LOG_delete)
-          CERR << "delete " << (const void *)new_function << " at " LOC << endl;
+          CERR << "delete " << voidP(new_function) << " at " LOC << endl;
         delete new_function;
         return 0;
       }
@@ -449,23 +458,24 @@ void * handle = open_so_file(t4, so_path);
 void * emacs_start = dlsym(handle, "emacs_start");
    if (emacs_start == 0)
       {
-        t4.append_utf8(", but it\n   it is lacking the mandatory "
+        t4.append_UTF8(", but it\n   it is lacking the mandatory "
                        "function emacs_start()\n");
         return t4;
       }
 
 UTF8_string so_path_utf(so_path);
-const int error = ((int (*)(const char *, const char *))emacs_start)
-                  (emacs_arg, (const char *)so_path_utf.c_str());
+const int error =
+    reinterpret_cast<int (*)(const char *, const char *)>(emacs_start)
+            (emacs_arg, charP(so_path_utf.c_str()));
    if (error)
       {
-        t4.append_utf8(", but emacs_start()  returned error ");
+        t4.append_UTF8(", but emacs_start()  returned error ");
         t4.append_number(error);
-        t4.append_utf8("\n");
+        t4.append_UTF8("\n");
         return t4;
       }
 
-   t4.shrink(0);   // success
+   t4.clear();   // success
    return t4;
 }
 //-----------------------------------------------------------------------------
