@@ -94,7 +94,7 @@ bool ErrOut::used = false;
 DiffOut DOUT_filebuf(false);
 DiffOut UERR_filebuf(true);
 
-// Android defines its own CIN, COUT, CERR, and UERR ostreams
+// Android is supposed to define its own CIN, COUT, CERR, and UERR ostreams
 #ifndef WANT_ANDROID
 
 CinOut CIN_filebuf;
@@ -121,7 +121,7 @@ Output::ColorMode Output::color_mode = COLM_UNDEF;
 char Output::color_CIN[MAX_ESC_LEN] = CSI "0;30;47m";
 
 /// VT100 escape sequence to change to cout color
-char Output::color_COUT[MAX_ESC_LEN] = CSI "0;30;49m";
+char Output::color_COUT[MAX_ESC_LEN] = CSI "0;39;49m";
 
 /// VT100 escape sequence to change to cerr color
 char Output::color_CERR[MAX_ESC_LEN] = CSI "0;35;49m";
@@ -170,7 +170,7 @@ PERFORMANCE_START(cerr_perf)
    if (!InputFile::echo_current_file())   return 0;
 
    Output::set_color_mode(Output::COLM_INPUT);
-   cerr << (char)c;
+   cerr << char(c);
 PERFORMANCE_END(fs_CERR_B, cerr_perf, 1)
 
    return 0;
@@ -182,7 +182,7 @@ ErrOut::overflow(int c)
 PERFORMANCE_START(cerr_perf)
 
    Output::set_color_mode(Output::COLM_ERROR);
-   cerr << (char)c;
+   cerr << char(c);
 PERFORMANCE_END(fs_CERR_B, cerr_perf, 1)
 
    return 0;
@@ -289,7 +289,7 @@ Output::read_ESC_sequence(char * dest, int destlen, int append,
         return 1;
       }
 
-   if (str == (char *)-1)
+   if (str == reinterpret_cast<const char *>(-1))
       {
         const char * term = getenv("TERM");
         CERR << "capability '" << capname 
