@@ -19,6 +19,7 @@
 */
 
 #include <string.h>
+#include <vector>
 
 #include "CharCell.hh"
 #include "ComplexCell.hh"
@@ -32,7 +33,7 @@
 #include "Symbol.hh"
 #include "SystemLimits.hh"
 #include "SystemVariable.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Tokenizer.hh"
 #include "Workspace.hh"
 
@@ -55,7 +56,7 @@ Token_string tos1;
    if (tos1.size() == 1)
       {
         const ErrorCode err = parse_statement(tos1);
-        if (err == E_NO_ERROR)   tos.append(tos1[0]);
+        if (err == E_NO_ERROR)   tos.push_back(tos1[0]);
         return err;
       }
 
@@ -68,39 +69,37 @@ Parser::parse(const Token_string & input, Token_string & tos) const
    Log(LOG_parse)
       {
         CERR << "parse 1 [" << input.size() << "]: ";
-        input.print(CERR);
+        input.print(CERR, true);
         CERR << endl;
       }
 
-   // split input into statements.
+   // split input line into statements (separated by ◊)
    //
-Simple_string<Token_string *, false> statements;
+std::vector<Token_string *> statements;
    {
-     Source<Token, true> src(input);
-     Token_string  * stat = new Token_string();
-     while (src.rest())
+     Token_string * stat = new Token_string();
+     loop(idx, input.size())
         {
-          const Token tok = src.get();
+          const Token & tok = input[idx];
           if (tok.get_tag() == TOK_DIAMOND)
              {
-               statements.append(stat);
+               statements.push_back(stat);
                stat = new Token_string();
              }
           else
              {
-               stat->append(tok, LOC);
+               stat->push_back(tok);
              }
         }
-     statements.append(stat);
+     statements.push_back(stat);
    }
 
    loop(s, statements.size())
       {
         Token_string * stat = statements[s];
-        ErrorCode err = parse_statement(*stat);
-        if (err)
+        if (const ErrorCode err = parse_statement(*stat))
            {
-             while (s < statements.size())
+             while (size_t(s) < statements.size())
                {
                  stat = statements[s++];
                  delete stat;
@@ -108,11 +107,11 @@ Simple_string<Token_string *, false> statements;
              return err;
            }
 
-        if (s)   tos.append(Token(TOK_DIAMOND));
+        if (s)   tos.push_back(Token(TOK_DIAMOND));
 
         loop(t, stat->size())
            {
-             tos.append(Token());
+             tos.push_back(Token());
              tos[tos.size() - 1].move_1((*stat)[t], LOC);
            }
         delete stat;
@@ -132,7 +131,7 @@ Parser::parse_statement(Token_string & tos)
    Log(LOG_parse)
       {
         CERR << "parse 2 [" << tos.size() << "]: ";
-        tos.print(CERR);
+        tos.print(CERR, true);
       }
 
    // 2. convert groups like '(' val val...')' into single APL values
@@ -148,7 +147,7 @@ Parser::parse_statement(Token_string & tos)
    Log(LOG_parse)
       {
         CERR << "parse 3 [" << tos.size() << "]: ";
-        tos.print(CERR);
+        tos.print(CERR, true);
       }
 
    // 3. convert vectors like 1 2 3 or '1' 2 '3' into single APL values
@@ -167,7 +166,7 @@ Parser::parse_statement(Token_string & tos)
    Log(LOG_parse)
       {
         CERR << "parse 4 [" << tos.size() << "]: ";
-        tos.print(CERR);
+        tos.print(CERR, true);
       }
 
    // 4. mark symbol left of ← as LSYMB
@@ -176,10 +175,20 @@ Parser::parse_statement(Token_string & tos)
    Log(LOG_parse)
       {
         CERR << "parse 5 [" << tos.size() << "]: ";
-        tos.print(CERR);
+        tos.print(CERR, true);
       }
 
-   // 5. update distances between (), [], and {}
+   // 5. replace bitwise functons ∧∧, ∨∨, ⍲⍲, and ⍱⍱ by their bitwise variant
+   //
+   replace_bitwise_functions(tos);
+   remove_void_token(tos);
+   Log(LOG_parse)
+      {
+        CERR << "parse 6 [" << tos.size() << "]: ";
+        tos.print(CERR, true);
+      }
+
+   // 6. update distances between (), [], and {}
    //
    {
      const ErrorCode ec = match_par_bra(tos, false);
@@ -199,22 +208,22 @@ Parser::collect_constants(Token_string & tos)
    Log(LOG_collect_constants)
       {
         CERR << "collect_constants [" << tos.size() << " token] in: ";
-        tos.print(CERR);
+        tos.print(CERR, true);
       }
 
    // convert several items in vector notation into a single APL value
    //
    loop (t, tos.size())
       {
-        int to;
+        size_t to;
         switch(tos[t].get_tag())
            {
-             case TOK_CHARACTER:
-             case TOK_INTEGER:
-             case TOK_REAL:
-             case TOK_COMPLEX:
              case TOK_APL_VALUE1:
              case TOK_APL_VALUE3:
+             case TOK_CHARACTER:
+             case TOK_COMPLEX:
+             case TOK_INTEGER:
+             case TOK_REAL:
                   break;          // continue below
 
              default: continue;   // nex token
@@ -236,8 +245,8 @@ Parser::collect_constants(Token_string & tos)
               // if tos[to + 1] is [ then [ binds stronger than
               // vector notation and we stop collecting.
               //
-              if ((to + 1) < tos.size() && tos[to + 1].get_tag() == TOK_L_BRACK)
-                 break;
+              if (size_t(to + 1) < tos.size() &&
+                  tos[to + 1].get_tag() == TOK_L_BRACK)   break;
 #endif
               const TokenTag tag = tos[to].get_tag();
 
@@ -259,7 +268,7 @@ Parser::collect_constants(Token_string & tos)
    Log(LOG_collect_constants)
       {
         CERR << "collect_constants [" << tos.size() << " token] out: ";
-        tos.print(CERR);
+        tos.print(CERR, true);
       }
 }
 //-----------------------------------------------------------------------------
@@ -269,7 +278,7 @@ Parser::collect_groups(Token_string & tos)
    Log(LOG_collect_constants)
       {
         CERR << "collect_groups [" << tos.size() << " token] in: ";
-        tos.print(CERR);
+        tos.print(CERR, true);
       }
 
 int opening = -1;
@@ -309,7 +318,7 @@ int opening = -1;
                       {
                         CERR << "collect_groups [" << tos.size()
                              << " token] out: ";
-                        tos.print(CERR);
+                        tos.print(CERR, true);
                       }
                    return true;
 
@@ -329,7 +338,7 @@ Parser::find_closing_bracket(const Token_string & tos, int pos)
 
 int others = 0;
 
-   for (int p = pos + 1; p < tos.size(); ++p)
+   for (size_t p = pos + 1; p < tos.size(); ++p)
        {
          Log(LOG_find_closing)
             CERR << "find_closing_bracket() sees " << tos[p] << endl;
@@ -375,7 +384,7 @@ Parser::find_closing_parent(const Token_string & tos, int pos)
 
 int others = 0;
 
-   for (int p = pos + 1; p < tos.size(); ++p)
+   for (size_t p = pos + 1; p < tos.size(); ++p)
        {
          Log(LOG_find_closing)
             CERR << "find_closing_bracket() sees " << tos[p] << endl;
@@ -475,6 +484,50 @@ Parser::remove_nongrouping_parantheses(Token_string & tos)
        }
 }
 //-----------------------------------------------------------------------------
+void
+Parser::replace_bitwise_functions(Token_string & tos)
+{
+   loop(t, int(tos.size()) - 1)
+       {
+         if (tos[t].get_tag() != TOK_F12_ENCODE)   continue;
+
+         switch(tos[t+1].get_tag())
+            {
+              case TOK_F2_AND:
+                   tos[t] = Token(TOK_F2_AND_B, &Bif_F2_AND_B::_fun);
+                   tos[++t].clear(LOC);
+                   continue;
+
+              case TOK_F2_OR:
+                   tos[t] = Token(TOK_F2_OR_B, &Bif_F2_OR_B::_fun);
+                   tos[++t].clear(LOC);
+                   continue;
+
+              case TOK_F2_NAND:
+                   tos[t] = Token(TOK_F2_NAND_B,&Bif_F2_NAND_B::_fun);
+                   tos[++t].clear(LOC);
+                   continue;
+
+              case TOK_F2_NOR:
+                   tos[t] = Token(TOK_F2_NOR_B, &Bif_F2_NOR_B::_fun);
+                   tos[++t].clear(LOC);
+                   continue;
+
+              case TOK_F2_EQUAL:
+                   tos[t] = Token(TOK_F2_EQUAL_B, &Bif_F2_EQUAL_B::_fun);
+                   tos[++t].clear(LOC);
+                   continue;
+
+              case TOK_F2_UNEQ:
+                   tos[t] = Token(TOK_F2_UNEQ_B, &Bif_F2_UNEQ_B::_fun);
+                   tos[++t].clear(LOC);
+                   continue;
+
+              default: break;
+            }
+       }
+}
+//-----------------------------------------------------------------------------
 bool
 Parser::check_if_value(const Token_string & tos, int pos)
 {
@@ -524,22 +577,22 @@ Parser::check_if_value(const Token_string & tos, int pos)
 void
 Parser::remove_void_token(Token_string & tos)
 {
-int dst = 0;
+size_t dst = 0;
 
    loop(src, tos.size())
        {
          if (tos[src].get_tag() == TOK_VOID)   continue;
-         if (src != dst)   tos[dst].move_1(tos[src], LOC);
+         if (src != ShapeItem(dst))   tos[dst].move_1(tos[src], LOC);
          ++dst;
        }
 
-   tos.shrink(dst);
+   tos.resize(dst);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 Parser::match_par_bra(Token_string & tos, bool backwards)
 {
-Simple_string<ShapeItem, false> stack;
+std::vector<ShapeItem> stack;
    loop(s, tos.size())
        {
          const ShapeItem t = backwards ? (tos.size() - 1) - s : s;
@@ -550,7 +603,7 @@ Simple_string<ShapeItem, false> stack;
              // for [ ( or { push the position onto stack
              case TC_L_BRACK:  if (tos[t].get_tag() != TOK_L_BRACK)   continue;
              case TC_L_PARENT:
-             case TC_L_CURLY:  stack.append(t);
+             case TC_L_CURLY:  stack.push_back(t);
                                continue;
 
              case TC_R_BRACK:  ec = E_UNBALANCED_BRACKET;
@@ -572,8 +625,8 @@ Simple_string<ShapeItem, false> stack;
           //
           if (stack.size() == 0)   return ec;
 
-           const ShapeItem t1 = stack.last();
-           stack.pop();
+           const ShapeItem t1 = stack.back();
+           stack.pop_back();
 
           if (tos[t1].get_Class() != tc_peer)   return ec;
 
@@ -602,7 +655,7 @@ Parser::create_value(Token_string & tos, int pos, int count)
       {
         CERR << "create_value(" << __LINE__ << ") tos[" << tos.size()
              <<  "]  pos " << pos << " count " << count << " in:";
-        tos.print(CERR);
+        tos.print(CERR, true);
       }
 
    if (count == 1)   create_scalar_value(tos[pos]);
@@ -612,7 +665,7 @@ Parser::create_value(Token_string & tos, int pos, int count)
       {
         CERR << "create_value(" << __LINE__ << ") tos[" << tos.size()
              <<  "]  pos " << pos << " count " << count << " out:";
-        tos.print(CERR);
+        tos.print(CERR, true);
       }
 }
 //-----------------------------------------------------------------------------
@@ -625,7 +678,7 @@ Parser::create_scalar_value(Token & output)
              {
                Value_P scalar(LOC);
 
-               new (&scalar->get_ravel(0))  CharCell(output.get_char_val());
+               new (scalar->next_ravel())  CharCell(output.get_char_val());
                scalar->check_value(LOC);
                Token tok(TOK_APL_VALUE3, scalar);
                output.move_1(tok, LOC);
@@ -636,7 +689,7 @@ Parser::create_scalar_value(Token & output)
              {
                Value_P scalar(LOC);
 
-               new (&scalar->get_ravel(0))   IntCell(output.get_int_val());
+               new (scalar->next_ravel())   IntCell(output.get_int_val());
                scalar->check_value(LOC);
                Token tok(TOK_APL_VALUE3, scalar);
                output.move_1(tok, LOC);
@@ -647,7 +700,7 @@ Parser::create_scalar_value(Token & output)
              {
                Value_P scalar(LOC);
 
-               new (&scalar->get_ravel(0))  FloatCell(output.get_flt_val());
+               new (scalar->next_ravel())  FloatCell(output.get_flt_val());
                scalar->check_value(LOC);
                Token tok(TOK_APL_VALUE3, scalar);
                output.move_1(tok, LOC);
@@ -658,8 +711,8 @@ Parser::create_scalar_value(Token & output)
              {
                Value_P scalar(LOC);
 
-               new (&scalar->get_ravel(0))   ComplexCell(output.get_cpx_real(),
-                                                         output.get_cpx_imag());
+               new (scalar->next_ravel())   ComplexCell(output.get_cpx_real(),
+                                                        output.get_cpx_imag());
                scalar->check_value(LOC);
                Token tok(TOK_APL_VALUE3, scalar);
                output.move_1(tok, LOC);
@@ -713,7 +766,8 @@ Value_P vector(count, LOC);
 
             case TOK_APL_VALUE1:
             case TOK_APL_VALUE3:
-                 new (addr) PointerCell(tok.get_apl_val(), vector.getref());
+                 new (addr) PointerCell(tok.get_apl_val().get(),
+                                        vector.getref());
                  tok.clear(LOC);   // invalidate token
                  break;
 
@@ -729,7 +783,7 @@ Token tok(TOK_APL_VALUE3, vector);
    Log(LOG_create_value)
       {
         CERR << "create_value [" << tos.size() << " token] out: ";
-        tos.print(CERR);
+        tos.print(CERR, true);
       }
 }
 //-----------------------------------------------------------------------------
