@@ -32,15 +32,19 @@
 #include "ValueHistory.hh"
 #include "Workspace.hh"
 
+uint64_t Prefix::instance_counter = 0;
+
 //-----------------------------------------------------------------------------
 Prefix::Prefix(StateIndicator & _si, const Token_string & _body)
-   : si(_si),
+   : instance(++instance_counter),
+     si(_si),
      put(0),
      saved_lookahead(Token(TOK_VOID), Function_PC_invalid),
      body(_body),
      PC(Function_PC_0),
      assign_state(ASS_none),
-     lookahead_high(Function_PC_invalid)
+     lookahead_high(Function_PC_invalid),
+     action(RA_FIXME)
 {
 }
 //-----------------------------------------------------------------------------
@@ -119,7 +123,7 @@ Prefix::uses_function(const UserFunction * ufun) const
 }
 //-----------------------------------------------------------------------------
 bool
-Prefix::is_value_parent(int pc) const
+Prefix::is_value_parenthesis(int pc) const
 {
    // we have ) XXX with XXX on the stack and need to know if the evaluation
    // of (... ) will be a value as in e.g. (1 + 1) or a function as in (+/).
@@ -127,7 +131,7 @@ Prefix::is_value_parent(int pc) const
    Assert1(body[pc].get_Class() == TC_R_PARENT);
 
    ++pc;
-   if (pc >= body.size())   return true;   // syntax error
+   if (pc >= int(body.size()))   return true;   // syntax error
 
 TokenClass next = body[pc].get_Class();
 
@@ -136,7 +140,7 @@ TokenClass next = body[pc].get_Class();
         const int offset = body[pc].get_int_val2();
         pc += offset;
         Assert1(body[pc].get_Class() == TC_L_BRACK);   // opening [
-        if (pc >= body.size())   return true;   // syntax error
+        if (pc >= Function_PC(body.size()))   return true;   // syntax error
         next = body[pc].get_Class();
       }
 
@@ -144,7 +148,7 @@ TokenClass next = body[pc].get_Class();
       {
         const Symbol * sym = body[pc].get_sym_ptr();
         const NameClass nc = sym->get_nc();
-                      
+
         if (nc == NC_FUNCTION)   return false;
         if (nc == NC_OPERATOR)   return false;
         return true;
@@ -157,14 +161,14 @@ TokenClass next = body[pc].get_Class();
    if (next == TC_L_PARENT)   // )) XXX
       {
         ++pc;
-        if (!is_value_parent(pc))   return false;   // (fun)) XXX
+        if (!is_value_parenthesis(pc))   return false;   // (fun)) XXX
         const int offset = body[pc].get_int_val2();
         pc += offset;
-        if (pc >= body.size())   return true;   // syntax error
+        if (pc >= Function_PC(body.size()))   return true;   // syntax error
         next = body[pc].get_Class();
         Assert1(next == TC_L_PARENT);   // opening (
         ++pc;
-        if (pc >= body.size())   return true;   // syntax error
+        if (pc >= Function_PC(body.size()))   return true;   // syntax error
 
         //   (val)) XXX
         //  ^
@@ -187,8 +191,8 @@ TokenClass next = body[pc].get_Class();
 
    // dyadic operator with numeric function argument, for example:  ⍤ 0 
    //
-   if (next == TC_VALUE       &&
-       pc < (body.size() - 1) &&
+   if (next == TC_VALUE                  &&
+       pc < Function_PC(body.size() - 1) &&
        body[pc+1].get_Class() == TC_OPER2)   return false;
 
    return true;
@@ -211,7 +215,7 @@ TokenTag tag_LO = body[pc].get_tag();
         tag_LO = body[pc].get_tag();
       }
 
-   if (tag_LO == TOK_R_PARENT)   return !is_value_parent(pc);
+   if (tag_LO == TOK_R_PARENT)   return !is_value_parenthesis(pc);
    if (body[pc].get_Class() == TC_OPER2)   return false;   // make / a function
 
    if ((tag_LO & TV_MASK) == TV_FUN)   return true;
@@ -247,7 +251,7 @@ Prefix::vector_ass_count() const
 {
 int count = 0;
 
-   for (int pc = PC; pc < body.size(); ++pc)
+   for (Function_PC pc = PC; pc < Function_PC(body.size()); ++pc)
        {
          if (body[pc].get_tag() != TOK_LSYMB2)   break;
          ++count;
@@ -343,7 +347,7 @@ Prefix::value_expected()
    //
    if (saved_lookahead.tok.get_ValueType() == TV_INDEX)   return true;
 
-   for (int pc = PC; pc < body.size();)
+   for (int pc = PC; pc < int(body.size());)
       {
         const Token & tok = body[pc++];
         switch(tok.get_Class())
@@ -362,12 +366,12 @@ Prefix::value_expected()
                     {
                       const Symbol * sym = tok.get_sym_ptr();
                       const NameClass nc = sym->get_nc();
-                      
+
                       if (nc == NC_FUNCTION)   return false;
                       if (nc == NC_OPERATOR)   return false;
                       return true;   // value
                     }
-             
+
                case TC_RETURN:  return false;   // syntax error
                case TC_VALUE:   return true;
 
@@ -430,6 +434,28 @@ grow:
      if (size() > 0 && at0().get_Class() == TC_END)
         {
           Log(LOG_prefix_parser)   print_stack(CERR, LOC);
+
+          // provide help on some common cases...
+          //
+          for (int j = 1; j < (size() - 1); ++j)
+              {
+                if ( (at(j).tok.get_Class() == TC_ASSIGN)    &&
+                     (at(j + 1).tok.get_Class() == TC_VALUE))
+                   {
+                     if (at(j - 1).tok.get_Class() == TC_FUN0 ||
+                         at(j - 1).tok.get_Class() == TC_FUN12)
+                        {
+                           MORE_ERROR() <<
+                           "Cannot assign a value to a function";
+                        }
+                     else if (at(j - 1).tok.get_Class() == TC_OPER1 ||
+                             at(j - 1).tok.get_Class() == TC_OPER1)
+                        {
+                           MORE_ERROR() <<
+                           "Cannot assign a value to an operator";
+                        }
+                   }
+              }
           syntax_error(LOC);   // no more token
         }
 
@@ -527,40 +553,40 @@ again:
    // search prefixes in phrase table...
    //
    {
-     const int idx_0 = at0().get_Class();
+     const int hash_0 = at0().get_Class();
 
      if (size() >= 3)
         {
-          const int idx_01  = idx_0  | at1().get_Class() <<  5;
-          const int idx_012 = idx_01 | at2().get_Class() << 10;
+          const int hash_01  = hash_0  | at1().get_Class() <<  5;
+          const int hash_012 = hash_01 | at2().get_Class() << 10;
 
           if (size() >= 4)
              {
-               const int idx_0123 = idx_012 | at3().get_Class() << 15;
-               best = hash_table + idx_0123 % PHRASE_MODU;
-               if (best->idx == idx_0123)   goto found_prefix;
+               const int hash_0123 = hash_012 | at3().get_Class() << 15;
+               best = hash_table + hash_0123 % PHRASE_MODU;
+               if (best->phrase_hash == hash_0123)   goto found_prefix;
              }
 
-          best = hash_table + idx_012 % PHRASE_MODU;
-          if (best->idx == idx_012)   goto found_prefix;
+          best = hash_table + hash_012 % PHRASE_MODU;
+          if (best->phrase_hash == hash_012)   goto found_prefix;
 
-          best = hash_table + idx_01 % PHRASE_MODU;
-          if (best->idx == idx_01)   goto found_prefix;
+          best = hash_table + hash_01 % PHRASE_MODU;
+          if (best->phrase_hash == hash_01)   goto found_prefix;
 
-          best = hash_table + idx_0 % PHRASE_MODU;
-          if (best->idx != idx_0)   goto grow;   // no matching phrase
+          best = hash_table + hash_0 % PHRASE_MODU;
+          if (best->phrase_hash != hash_0)   goto grow;   // no matching phrase
         }
      else   // 0 < size() < 3
         {
           if (size() >= 2)
              {
-               const int idx_01 = idx_0 | at1().get_Class() << 5;
-               best = hash_table + idx_01 % PHRASE_MODU;
-               if (best->idx == idx_01)   goto found_prefix;
+               const int hash_01 = hash_0 | at1().get_Class() << 5;
+               best = hash_table + hash_01 % PHRASE_MODU;
+               if (best->phrase_hash == hash_01)   goto found_prefix;
              }
 
-          best = hash_table + idx_0 % PHRASE_MODU;
-          if (best->idx != idx_0)   goto grow;
+          best = hash_table + hash_0 % PHRASE_MODU;
+          if (best->phrase_hash != hash_0)   goto grow;
         }
    }
 
@@ -571,7 +597,7 @@ found_prefix:
    //
    {
      TokenClass next = TC_INVALID;
-     if (PC < body.size())
+     if (PC < Function_PC(body.size()))
         {
           const Token & tok = body[PC];
 
@@ -645,8 +671,18 @@ found_prefix:
         --prefix_len;
       }
 
+const uint64_t inst = instance;
    (this->*best->reduce_fun)();
 
+   if (inst != Workspace::SI_top()->get_prefix().instance)
+      {
+        // the reduce_fun() above has changed the )SI stack. As a consequence
+        // the 'this' pointer is no longer valid and we must not access members
+        // of this Prefix instance.
+        //
+        return Token(TOK_SI_PUSHED);
+      }
+
    Log(LOG_prefix_parser)
       CERR << "   reduce_" << best->reduce_name << "() returned: ";
 
@@ -700,7 +736,7 @@ Prefix::dont_reduce(TokenClass next) const
            }
         else if (next == TC_R_PARENT)   // ) B
            {
-             if (is_value_parent(PC))     // e.g. (X+Y) B
+             if (is_value_parenthesis(PC))     // e.g. (X+Y) B
                 {
                   return best->prio < BS_VAL_VAL;
                 }
@@ -840,12 +876,6 @@ Token result = at1();   // B or F
 }
 //-----------------------------------------------------------------------------
 void
-Prefix::reduce_LPAR_F_RPAR_()
-{
-   reduce_LPAR_B_RPAR_();   // same as ( B )
-}
-//-----------------------------------------------------------------------------
-void
 Prefix::reduce_LPAR_F_C_RPAR()
 {
    Assert1(prefix_len == 4);
@@ -855,10 +885,14 @@ Prefix::reduce_LPAR_F_C_RPAR()
    if (at2().get_ValueType() != TV_VAL)   SYNTAX_ERROR;
    if (!at2().get_apl_val())              SYNTAX_ERROR;
 
+   //     at: 0 1 2 3
+   // before: ( F C )
+   // after:  F C
+   //
    at3().move_1(at2(), LOC);
    at2().move_1(at1(), LOC);
-   pop();    // pop C
-   pop();    // pop RPAR
+   pop_and_discard();    // pop old RPAR
+   pop_and_discard();    // pop old C
    action = RA_CONTINUE;
 }
 //-----------------------------------------------------------------------------
@@ -901,54 +935,119 @@ Prefix::reduce_MISC_F_B_()
 Token result = at0().get_function()->eval_B(at1().get_apl_val());
    if (result.get_Class() == TC_SI_LEAVE)
       {
-        if (result.get_tag() == TOK_EA_EXEC)
-           {
-             // the macro Quad_EA was called and has returned an action
-             // to be performed in the caller. We pop the )SI entry for
-             // Quad_EA continue in the calling prefix parser
-             //
-             action = RA_SI_PUSHED;
-             const Cell * info = &result.get_apl_val()->get_ravel(0);
-             Workspace::pop_SI(LOC);
-             UCS_string stat(*info[1].get_pointer_value());
-             if (stat.size())
-                {
-                  // the B side of A ⎕EA B has failed or returned → or →N
-                  // so that stat (which is A or → or →N) shall be executed
-                  //
-//                CERR << "⎕ES: ⍎" << stat << endl;
-                  Token exec = Bif_F1_EXECUTE::execute_statement(stat);
-                  if (exec.get_Class() == TC_VALUE)   // ⍎ literal
-                     {
-                       Workspace::SI_top()->get_prefix().at0().move_1(exec,LOC);
-                       return;
-                     }
+        if (result.get_tag() == TOK_SI_PUSHED)   goto done;
 
-                  const APL_Integer major = info[2].get_int_value();
-                  const APL_Integer minor = info[3].get_int_value();
-                  const ErrorCode ec = (ErrorCode)(major << 16 | minor);
-                  Workspace::SI_top()->get_parent()->get_error().init(ec, LOC);
-                }
-              else              // committed value
+        /* NOTE: the tags TOK_QUAD_ES_COM, TOK_QUAD_ES_ESC, TOK_QUAD_ES_BRA,
+                 and TOK_QUAD_ES_ERR below can only occur if:
+
+            1. ⎕EA resp. ⎕EB is called; each is implemented as macro
+               Z__A_Quad_EA_B resp. Z__A_Quad_EB_B.
+
+            2. The macro calls ⎕ES 100 ¯1...¯4, which brings us here.
+
+            Token result is the return token of Quad_ES::eval_AB() or
+            Quad_ES::eval_B() and contains the right argument B (as set in
+            the macro).
+
+            We must check that ⎕ES 100 was not called directly, but only via
+            ⎕EA or ⎕EB.
+         */
+
+        if (result.get_tag() == TOK_QUAD_ES_COM)
+           {
+             // make sure that ⎕ES was called from a macro (implies parent)
+             //
+             if (Workspace::SI_top()->function_name()[0] != UNI_MUE)
+                DOMAIN_ERROR;
+
+             Workspace::pop_SI(LOC);   // discard ⎕EA/⎕EB context
+
+             const Cell & QES_arg2 = result.get_apl_val()->get_ravel(2);
+             Token & si_pushed = Workspace::SI_top()->get_prefix().at0();
+             Assert(si_pushed.get_tag() == TOK_SI_PUSHED);
+             if (QES_arg2.is_pointer_cell())
                 {
-                  Token & si_pushed = Workspace::SI_top()->get_prefix().at0();
-                  Assert(si_pushed.get_tag() == TOK_SI_PUSHED);
-                  if (info[3].is_pointer_cell())
-                     {
-                       Value_P val = info[3].get_pointer_value();
-                       new (&si_pushed)  Token(TOK_APL_VALUE2, val);
-                     }
-                  else
-                     {
-                       Value_P scalar(LOC);
-                       scalar->next_ravel()->init(info[3], scalar.getref(),LOC);
-                       scalar->check_value(LOC);
-                       new (&si_pushed)  Token(TOK_APL_VALUE2, scalar);
-                     }
+                  Value_P val = QES_arg2.get_pointer_value();
+                  new (&si_pushed)  Token(TOK_APL_VALUE2, val);
+                }
+             else
+                {
+                  Value_P scalar(LOC);
+                  scalar->next_ravel()->init(QES_arg2, scalar.getref(),LOC);
+                  scalar->check_value(LOC);
+                  new (&si_pushed)  Token(TOK_APL_VALUE2, scalar);
                 }
              return;
            }
 
+        if (result.get_tag() == TOK_QUAD_ES_ESC)
+           {
+             // make sure that ⎕ES was called from a macro (implies parent)
+             //
+             if (Workspace::SI_top()->function_name()[0] != UNI_MUE)
+                DOMAIN_ERROR;
+
+             Workspace::pop_SI(LOC);   // discard the ⎕EA/⎕EB context
+
+             Token & si_pushed = Workspace::SI_top()->get_prefix().at0();
+             Assert(si_pushed.get_tag() == TOK_SI_PUSHED);
+             new (&si_pushed)  Token(TOK_ESCAPE);
+             return;
+           }
+
+        if (result.get_tag() == TOK_QUAD_ES_BRA)
+           {
+             // make sure that ⎕ES was called from a macro (implies parent)
+             //
+             if (Workspace::SI_top()->function_name()[0] != UNI_MUE)
+                DOMAIN_ERROR;
+
+             Workspace::pop_SI(LOC);   // discard the ⎕EA/⎕EB context
+
+             const Cell & QES_arg2 = result.get_apl_val()->get_ravel(2);
+             const APL_Integer line = QES_arg2.get_int_value();
+
+             Token & si_pushed = Workspace::SI_top()->get_prefix().at0();
+             Assert(si_pushed.get_tag() == TOK_SI_PUSHED);
+
+             Workspace::SI_top()->jump(IntScalar(line, LOC));
+             return;
+           }
+
+        if (result.get_tag() == TOK_QUAD_ES_ERR)
+           {
+             // this case can only occur with ⎕EA, but not with ⎕EB.
+
+             // make sure that ⎕ES was called from a macro (implies parent)
+             //
+             if (Workspace::SI_top()->function_name()[0] != UNI_MUE)
+                DOMAIN_ERROR;
+
+             Workspace::pop_SI(LOC);   // discard the ⎕EA/⎕EB context
+             StateIndicator * top = Workspace::SI_top();
+
+             Token & si_pushed = top->get_prefix().at0();
+             Assert(si_pushed.get_tag() == TOK_SI_PUSHED);
+
+             const Cell * QES_arg = &result.get_apl_val()->get_ravel(0);
+             UCS_string statement_A(   QES_arg[2].get_pointer_value().getref());
+             const APL_Integer major = QES_arg[3].get_int_value();
+             const APL_Integer minor = QES_arg[4].get_int_value();
+             const ErrorCode ec = ErrorCode(major << 16 | minor);
+
+             Token result_A = Bif_F1_EXECUTE::execute_statement(statement_A);
+             if (result_A.get_Class() == TC_VALUE)   // ⍎ literal
+                {
+                  Workspace::SI_top()->get_prefix().at0().move_1(result_A, LOC);
+                  return;
+                }
+             new (&StateIndicator::get_error(top)) Error(ec, LOC);
+             return;
+           }
+
+        // at this point a normal monadic function (i.e. other than ⎕EA/⎕EB)
+        // has returned an error
+        //
         if (result.get_tag() == TOK_ERROR)
            {
              Token_loc tl(result, get_range_low());
@@ -956,8 +1055,13 @@ Token result = at0().get_function()->eval_B(at1().get_apl_val());
              action = RA_RETURN;
              return;
            }
+
+        // not reached
+        Q1(result.get_tag())
+        FIXME;
       }
 
+done:
    pop_args_push_result(result);
    set_action(result);
 }
@@ -983,7 +1087,7 @@ Prefix::reduce_MISC_F_C_B()
    if (at1().get_ValueType() != TV_VAL)   // [i1;i2...] instead of [axis]
       {
         IndexExpr * idx = &at1().get_index_val();
-        Log(LOG_delete)   CERR << "delete " HEX(idx) << " at " LOC << endl;
+        Log(LOG_delete)   CERR << "delete " << voidP(idx) << " at " LOC << endl;
         delete idx;
          at1().clear(LOC);
          SYNTAX_ERROR;
@@ -1118,9 +1222,17 @@ Prefix::reduce_F_M_C_()
 {
    Assert1(prefix_len == 3);
 
+   if (at2().get_tag() != TOK_AXES)   // e.g. F[;2] instead of F[2]
+      {
+        // the user has provided a TOK_INDEX where TOK_AXES was expected
+        MORE_ERROR() << "illegal ; in axis";
+        AXIS_ERROR;
+      }
+
 DerivedFunction * derived =
    Workspace::SI_top()->fun_oper_cache.get(LOC);
-   new (derived) DerivedFunction(at0(), at1().get_function(),
+   new (derived) DerivedFunction(at0(),
+                                 at1().get_function(),
                                  at2().get_axes(), LOC);
 
    pop_args_push_result(Token(TOK_FUN2, derived));
@@ -1149,9 +1261,17 @@ Prefix::reduce_F_C_M_()
 {
    Assert1(prefix_len == 3);
 
+   if (at1().get_tag() != TOK_AXES)   // e.g. F[;2] instead of F[2]
+      {
+        // the user has provided a TOK_INDEX where TOK_AXES was expected
+        MORE_ERROR() << "illegal ; in axis";
+        AXIS_ERROR;
+      }
+
 DerivedFunction * F_C =
    Workspace::SI_top()->fun_oper_cache.get(LOC);
-   new (F_C) DerivedFunction(at0().get_function(), at1().get_axes(), LOC);
+   new (F_C) DerivedFunction(at0().get_function(),
+                             at1().get_axes(), LOC);
 
 Token tok_F_C(TOK_FUN2, F_C);
 DerivedFunction * derived =
@@ -1184,9 +1304,24 @@ Prefix::reduce_F_C_M_C()
 {
    Assert1(prefix_len == 4);
 
+   if (at1().get_tag() != TOK_AXES)   // e.g. F[;2] instead of F[2]
+      {
+        // the user has provided a TOK_INDEX where TOK_AXES was expected
+        MORE_ERROR() << "illegal ; in axis";
+        AXIS_ERROR;
+      }
+
+   if (at3().get_tag() != TOK_AXES)   // e.g. M[;2] instead of M[2]
+      {
+        // the user has provided a TOK_INDEX where TOK_AXES was expected
+        MORE_ERROR() << "illegal ; in axis";
+        AXIS_ERROR;
+      }
+
 DerivedFunction * F_C =
    Workspace::SI_top()->fun_oper_cache.get(LOC);
-   new (F_C) DerivedFunction(at0().get_function(), at1().get_axes(), LOC);
+   new (F_C) DerivedFunction(at0().get_function(),
+                             at1().get_axes(), LOC);
 
 Token tok_F_C(TOK_FUN2, F_C);
 DerivedFunction * derived =
@@ -1216,23 +1351,11 @@ const TokenTag tag = at0().get_tag();
 }
 //-----------------------------------------------------------------------------
 void
-Prefix::reduce_B_D_B_()
-{
-   reduce_F_D_G_();   // same as F2
-}
-//-----------------------------------------------------------------------------
-void
-Prefix::reduce_B_D_G_()
-{
-   reduce_F_D_G_();   // same as F2
-}
-//-----------------------------------------------------------------------------
-void
 Prefix::reduce_F_D_B_()
 {
    // same as G2, except for ⍤
    //
-   if (at1().get_function()->get_Id() != ID::OPER2_RANK)
+   if (at1().get_function()->get_Id() != ID_OPER2_RANK)
       {
          reduce_F_D_G_();
          return;
@@ -1259,7 +1382,7 @@ Token result = Token(TOK_FUN2, derived);
         // save locations of ⍤ and B
         //
         Function_PC pc_D = at(1).pc;
-        Function_PC pc_B = at(3).pc;
+        Function_PC pc_B = at(2).pc;
 
         pop_and_discard();   // pop F
         pop_and_discard();   // pop C
@@ -1273,7 +1396,6 @@ Token result = Token(TOK_FUN2, derived);
       }
 
    action = RA_CONTINUE;
-
 }
 //-----------------------------------------------------------------------------
 void
@@ -1329,7 +1451,7 @@ Prefix::reduce_F_D_C_B()
    // reduce, except if another dyadic operator is coming. In that case
    // F belongs to the other operator and we simply continue.
    //
-   if (PC < body.size())
+   if (PC < Function_PC(body.size()))
         {
           const Token & tok = body[PC];
           TokenClass next =  tok.get_Class();
@@ -1353,6 +1475,13 @@ Value_P B;
    Bif_OPER2_RANK::split_y123_B(at3().get_apl_val(), y123, B);
 Token new_y123(TOK_APL_VALUE1, y123);
 
+   if (at2().get_tag() != TOK_AXES)   // e.g. D[;2] instead of D[;2]
+      {
+        // the user has provided a TOK_INDEX where TOK_AXES was expected
+        MORE_ERROR() << "illegal ; in axis";
+        AXIS_ERROR;
+      }
+
 Value_P v_idx = at2().get_axes();
 DerivedFunction * derived = Workspace::SI_top()->fun_oper_cache.get(LOC);
    new (derived) DerivedFunction(at0(), at1().get_function(),
@@ -1402,13 +1531,15 @@ Value_P Z;
         try
            {
              Z = A->index(*idx);
-             Log(LOG_delete)   CERR << "delete " HEX(idx) << " at " LOC << endl;
+             Log(LOG_delete)
+                CERR << "delete " << voidP(idx) << " at " LOC << endl;
              delete idx;
            }
         catch (Error err)
            {
-             Token result = Token(TOK_ERROR, err.error_code);
-             Log(LOG_delete)   CERR << "delete " HEX(idx) << " at " LOC << endl;
+             Token result = Token(TOK_ERROR, err.get_error_code());
+             Log(LOG_delete)   CERR << "delete " << voidP(idx)
+                                    << " at " LOC << endl;
              delete idx;
              pop_args_push_result(result);
              set_action(result);
@@ -1448,7 +1579,7 @@ Value_P B = at3().get_apl_val();
            }
         catch (Error err)
            {
-             Token result = Token(TOK_ERROR, err.error_code);
+             Token result = Token(TOK_ERROR, err.get_error_code());
              at1().clear(LOC);
              at3().clear(LOC);
              pop_args_push_result(result);
@@ -1463,15 +1594,17 @@ Value_P B = at3().get_apl_val();
         try
            {
              V->assign_indexed(*idx, B);
-             Log(LOG_delete)   CERR << "delete " HEX(idx) << " at " LOC << endl;
+             Log(LOG_delete)   CERR << "delete " << voidP(idx)
+                                    << " at " LOC << endl;
              delete idx;
            }
         catch (Error err)
            {
-             Token result = Token(TOK_ERROR, err.error_code);
+             Token result = Token(TOK_ERROR, err.get_error_code());
              at1().clear(LOC);
              at3().clear(LOC);
-             Log(LOG_delete)   CERR << "delete " HEX(idx) << " at " LOC << endl;
+             Log(LOG_delete)   CERR << "delete " << voidP(idx)
+                                    << " at " LOC << endl;
              delete idx;
              pop_args_push_result(result);
              set_assign_state(ASS_none);
@@ -1541,7 +1674,7 @@ Function * F = at2().get_function();
    if (!F->is_lambda())   SYNTAX_ERROR;
 
 Symbol * V = at0().get_sym_ptr();
-   V->assign_named_lambda(F, LOC);
+   if (V->assign_named_lambda(F, LOC))   DEFN_ERROR;
 
 Value_P Z(V->get_name(), LOC);
 Token result = Token(TOK_APL_VALUE2, Z);
@@ -1552,30 +1685,6 @@ Token result = Token(TOK_APL_VALUE2, Z);
 }
 //-----------------------------------------------------------------------------
 void
-Prefix::reduce_V_ASS_N_()
-{
-   // named lambda: V ← { ... } niladic (same as reduce_V_ASS_F_)
-   //
-   reduce_V_ASS_F_();
-}
-//-----------------------------------------------------------------------------
-void
-Prefix::reduce_V_ASS_M_()
-{
-   // named lambda: V ← { ... } monadic operator (same as reduce_V_ASS_F_)
-   //
-   reduce_V_ASS_F_();
-}
-//-----------------------------------------------------------------------------
-void
-Prefix::reduce_V_ASS_D_()
-{
-   // named lambda: V ← { ... } dyadic operator (same as reduce_V_ASS_F_)
-   //
-   reduce_V_ASS_F_();
-}
-//-----------------------------------------------------------------------------
-void
 Prefix::reduce_RBRA___()
 {
    Assert1(prefix_len == 1);
@@ -1586,7 +1695,7 @@ Prefix::reduce_RBRA___()
    //
 IndexExpr * idx = new IndexExpr(get_assign_state(), LOC);
    Log(LOG_delete)
-      CERR << "new    " << (const void *)idx << " at " LOC << endl;
+      CERR << "new    " << voidP(idx) << " at " LOC << endl;
 
    new (&at0()) Token(TOK_PINDEX, *idx);
    set_assign_state(ASS_none);
@@ -1609,7 +1718,8 @@ const bool last_index = (at0().get_tag() == TOK_L_BRACK);
         Token result = Token(TOK_INDEX, idx);
         pop_args_push_result(result);
         action = RA_CONTINUE;
-        Log(LOG_delete)   CERR << "delete " HEX(&idx) << " at " LOC << endl;
+        Log(LOG_delete)   CERR << "delete " << voidP(&idx)
+                               << " at " LOC << endl;
         delete &idx;
         return;
       }
@@ -1658,7 +1768,7 @@ const bool last_index = (at0().get_tag() == TOK_L_BRACK);   // ; vs. [
              Assert1(!!X);
              I.move_2(Token(TOK_AXES, X), LOC);
              Log(LOG_delete)
-                CERR << "delete " HEX(&idx) << " at " LOC << endl;
+                CERR << "delete " << voidP(&idx) << " at " LOC << endl;
              delete &idx;
            }
         else
@@ -1715,15 +1825,15 @@ const int count = vector_ass_count();
 
    // vector assignment.
    //
-Simple_string<Symbol *, false> symbols;
-   symbols.append(at0().get_sym_ptr());
+std::vector<Symbol *> symbols;
+   symbols.push_back(at0().get_sym_ptr());
    loop(c, count)
       {
         Token_loc tl = lookahead();
         Assert1(tl.tok.get_tag() == TOK_LSYMB2);   // by vector_ass_count()
         Symbol * V = tl.tok.get_sym_ptr();
         Assert(V);
-        symbols.append(V);
+        symbols.push_back(V);
       }
 
 Value_P B = at3().get_apl_val();
@@ -1838,11 +1948,8 @@ const Token result = si.jump(line);
 
    if (result.get_tag() == TOK_NOBRANCH)   // branch not taken, e.g. →⍬
       {
-        if (trace)
-           {
-             Token bra(TOK_NOBRANCH);
-             si.statement_result(bra, true);
-           }
+        Token bra(TOK_NOBRANCH);
+        si.statement_result(bra, trace);
 
         action = RA_PUSH_NEXT;
         if (attention_is_raised() && end_of_line)
@@ -1906,8 +2013,8 @@ const bool trace = at0().get_Class() == TC_END &&
                   (at0().get_int_val() & 1) != 0;
    if (trace)
       {
-        Token bra(TOK_ESCAPE);
-        si.statement_result(bra, true);
+        Token esc(TOK_ESCAPE);
+        si.statement_result(esc, true);
       }
 
    // the statement is → which could mean TOK_ESCAPE (normal →) or
@@ -1920,8 +2027,8 @@ const bool trace = at0().get_Class() == TC_END &&
            {
               // the function ignores attention (aka. weak interrupt)
               //
-              pop_and_discard();   // pop() END
-              pop_and_discard();   // pop() GOTO
+              pop_and_discard();   // pop END
+              pop_and_discard();   // pop GOTO
               action = RA_CONTINUE;
               return;
            }
