@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2016  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2019  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -24,6 +24,7 @@
 #include "ArrayIterator.hh"
 #include "Avec.hh"
 #include "Bif_F12_SORT.hh"
+#include "Bif_F12_TAKE_DROP.hh"
 #include "Bif_OPER1_EACH.hh"
 #include "Bif_OPER1_REDUCE.hh"
 #include "CharCell.hh"
@@ -42,9 +43,9 @@
 #include "PrintOperator.hh"
 #include "StateIndicator.hh"
 #include "UserFunction.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Workspace.hh"
-                                            
+
 // primitive function instances
 //
 Bif_F0_ZILDE      Bif_F0_ZILDE     ::_fun;    // ⍬
@@ -54,8 +55,6 @@ Bif_F12_PARTITION Bif_F12_PARTITION::_fun;    // ⊂
 Bif_F12_PICK      Bif_F12_PICK     ::_fun;    // ⊃
 Bif_F12_COMMA     Bif_F12_COMMA    ::_fun;    // ,
 Bif_F12_COMMA1    Bif_F12_COMMA1   ::_fun;    // ⍪
-Bif_F12_TAKE      Bif_F12_TAKE     ::_fun;    // ↑
-Bif_F12_DROP      Bif_F12_DROP     ::_fun;    // ↓
 Bif_F12_ELEMENT   Bif_F12_ELEMENT  ::_fun;    // ∈
 Bif_F12_EQUIV     Bif_F12_EQUIV    ::_fun;    // ≡
 Bif_F12_NEQUIV    Bif_F12_NEQUIV   ::_fun;    // ≢
@@ -81,8 +80,6 @@ Bif_F12_PARTITION * Bif_F12_PARTITION::fun = &Bif_F12_PARTITION::_fun;
 Bif_F12_PICK      * Bif_F12_PICK     ::fun = &Bif_F12_PICK     ::_fun;
 Bif_F12_COMMA     * Bif_F12_COMMA    ::fun = &Bif_F12_COMMA    ::_fun;
 Bif_F12_COMMA1    * Bif_F12_COMMA1   ::fun = &Bif_F12_COMMA1   ::_fun;
-Bif_F12_TAKE      * Bif_F12_TAKE     ::fun = &Bif_F12_TAKE     ::_fun;
-Bif_F12_DROP      * Bif_F12_DROP     ::fun = &Bif_F12_DROP     ::_fun;
 Bif_F12_ELEMENT   * Bif_F12_ELEMENT  ::fun = &Bif_F12_ELEMENT  ::_fun;
 Bif_F12_EQUIV     * Bif_F12_EQUIV    ::fun = &Bif_F12_EQUIV    ::_fun;
 Bif_F12_NEQUIV    * Bif_F12_NEQUIV   ::fun = &Bif_F12_NEQUIV   ::_fun;
@@ -153,12 +150,10 @@ Token
 Bif_F12_RHO::eval_AB(Value_P A, Value_P B)
 {
 #ifdef PERFORMANCE_COUNTERS_WANTED
-#ifdef HAVE_RDTSC
 const uint64_t start_1 = cycle_counter();
 #endif
-#endif
 
-const Shape shape_Z(A, 0);
+const Shape shape_Z(A.get(), 0);
 
    // check that shape_Z is positive
    //
@@ -204,11 +199,9 @@ const ShapeItem len_Z = shape_Z.get_volume();
         B->set_shape(shape_Z);
 
 #ifdef PERFORMANCE_COUNTERS_WANTED
-#ifdef HAVE_RDTSC
 const uint64_t end_1 = cycle_counter();
    Performance::fs_F12_RHO_AB.add_sample(end_1 - start_1,
                                          B->nz_element_count());
-#endif
 #endif
 
         return Token(TOK_APL_VALUE1, B);
@@ -217,11 +210,9 @@ const uint64_t end_1 = cycle_counter();
 Token ret = do_reshape(shape_Z, *B);
 
 #ifdef PERFORMANCE_COUNTERS_WANTED
-#ifdef HAVE_RDTSC
 const uint64_t end_1 = cycle_counter();
    Performance::fs_F12_RHO_AB.add_sample(end_1 - start_1,
                                          shape_Z.get_volume());
-#endif
 #endif
 
    return ret;
@@ -243,7 +234,7 @@ Value_P Z(shape_Z, LOC);
    else
       {
         loop(z, len_Z)
-            Z->next_ravel()->init(B.get_ravel(z % len_B), Z.getref(), LOC);
+            B.get_ravel(z % len_B).init_other(Z->next_ravel(), Z.getref(), LOC);
       }
 
    Z->set_default(B, LOC);
@@ -279,12 +270,12 @@ const ShapeItem ec = B->element_count();
    if (ec == 0)
       {
         Value_P Z(LOC);
-        new (Z->next_ravel()) PointerCell(Idx0(LOC), Z.getref());
+        new (Z->next_ravel()) PointerCell(Idx0(LOC).get(), Z.getref());
         Z->check_value(LOC);
         return Token(TOK_APL_VALUE1, Z);
       }
 
-Shape sh(B, 0);
+Shape sh(B.get(), 0);
    loop(b, ec)
       {
         if (sh.get_shape_item(b) < 0)   DOMAIN_ERROR;
@@ -304,7 +295,7 @@ Value_P Z(sh, LOC);
               N /= q;
             }
         ZZ->check_value(LOC);
-        new (Z->next_ravel())   PointerCell(ZZ, Z.getref());
+        new (Z->next_ravel())   PointerCell(ZZ.get(), Z.getref());
       }
 
    if (Z->element_count() == 0)   // empty result
@@ -312,7 +303,7 @@ Value_P Z(sh, LOC);
         Value_P ZZ(ec, LOC);
         loop(r, ec)   new (ZZ->next_ravel())   IntCell(0);
         ZZ->check_value(LOC);
-        new (&Z->get_ravel(0))   PointerCell(ZZ, Z.getref());   // prototype
+        new (&Z->get_ravel(0))   PointerCell(ZZ.get(), Z.getref());   // prototype
       }
 
    Z->check_value(LOC);
@@ -322,7 +313,7 @@ Value_P Z(sh, LOC);
 Token
 Bif_F12_INDEX_OF::eval_AB(Value_P A, Value_P B)
 {
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
 const APL_Integer qio = Workspace::get_IO();
 
    // Index of
@@ -369,7 +360,7 @@ const ShapeItem count = B->element_count();
 }
 //-----------------------------------------------------------------------------
 Value_P
-Bif_COMMA::prepend_scalar(const Cell & cell_A, Axis axis, Value_P B)
+Bif_COMMA::prepend_scalar(const Cell & cell_A, uAxis axis, Value_P B)
 {
    if (B->is_empty())
       {
@@ -422,7 +413,7 @@ const Cell * cB = &B->get_ravel(0);
 }
 //-----------------------------------------------------------------------------
 Value_P
-Bif_COMMA::append_scalar(Value_P A, Axis axis, const Cell & cell_B)
+Bif_COMMA::append_scalar(Value_P A, uAxis axis, const Cell & cell_B)
 {
    if (A->is_empty())
       {
@@ -469,6 +460,9 @@ const Cell * cA = &A->get_ravel(0);
 Token
 Bif_COMMA::catenate(Value_P A, Axis axis, Value_P B)
 {
+   // NOTE: the case A->is_scalar() && B->is_scalar() was supposedly ruled out
+   //       before calling catenate()
+
    if (A->is_scalar())
       {
         const Cell & cell_A = A->get_ravel(0);
@@ -689,7 +683,7 @@ const Cell * cB = &B->get_ravel(0);
 }
 //-----------------------------------------------------------------------------
 Token
-Bif_COMMA::ravel_axis(Value_P X, Value_P B, Axis axis)
+Bif_COMMA::ravel_axis(Value_P X, Value_P B, uAxis axis)
 {
 const APL_Integer qio = Workspace::get_IO();
 
@@ -723,7 +717,7 @@ const APL_Integer qio = Workspace::get_IO();
         if (B->get_rank() == MAX_RANK)   INDEX_ERROR;
 
         const APL_Float new_axis = X->get_ravel(0).get_real_value() - qio;
-        const Axis axis((new_axis < 0) ? -1 : new_axis);
+        Axis axis = new_axis;   if (new_axis < 0.0)   axis = -1;
         const Shape shape_Z = B->get_shape().insert_axis(axis + 1, 1);
         return ravel(shape_Z, B);
       }
@@ -738,7 +732,7 @@ const APL_Integer qio = Workspace::get_IO();
 
    // case 3b: ,[n1 ... nk]B : combine axes.
    //
-const Shape axes(X, qio);
+const Shape axes(X.get(), qio);
 
 const ShapeItem from = axes.get_shape_item(0);
    if (from < 0)   AXIS_ERROR;
@@ -759,7 +753,7 @@ ShapeItem count = 1;
 Shape shape_Z;
    loop (r, from)   shape_Z.add_shape_item(B->get_shape_item(r));
    shape_Z.add_shape_item(count);
-   for (Rank r = to + 1; r < B->get_rank(); ++r)
+   for (uRank r = to + 1; r < B->get_rank(); ++r)
        shape_Z.add_shape_item(B->get_shape_item(r));
 
    return ravel(shape_Z, B);
@@ -796,7 +790,16 @@ const Shape shape_Z(B->element_count());
 Token
 Bif_F12_COMMA::eval_AB(Value_P A, Value_P B)
 {
-Rank max_rank = A->get_rank();
+  if (A->is_scalar() && B->is_scalar())
+     {
+       Value_P Z(2, LOC);
+       Z->next_ravel()->init(A->get_ravel(0), Z.getref(), LOC);
+       Z->next_ravel()->init(B->get_ravel(0), Z.getref(), LOC);
+       Z->check_value(LOC);
+       return Token(TOK_APL_VALUE1, Z);
+     }
+
+uRank max_rank = A->get_rank();
    if (max_rank < B->get_rank())  max_rank = B->get_rank(); 
    return catenate(A, max_rank - 1, B);
 }
@@ -804,6 +807,8 @@ Rank max_rank = A->get_rank();
 Token
 Bif_F12_COMMA::eval_AXB(Value_P A, Value_P X, Value_P B)
 {
+  if (A->is_scalar() && B->is_scalar())   RANK_ERROR;
+
    // catenate or laminate
    //
    if (!X->is_scalar_or_len1_vector())   AXIS_ERROR;
@@ -815,14 +820,16 @@ const APL_Integer qio = Workspace::get_IO();
       {
         const Axis axis = cX.get_checked_near_int() - qio;
         if (axis < 0)                                         AXIS_ERROR;
-        if (axis >= A->get_rank() && axis >= B->get_rank())   AXIS_ERROR;
+        if (uAxis(axis) >= A->get_rank() && uAxis(axis) >= B->get_rank())
+           AXIS_ERROR;
         return catenate(A, axis, B);
       }
 
 const APL_Float axis = cX.get_real_value() - qio;
-   if (axis <= -1.0)                                                 AXIS_ERROR;
-   if (axis >= (A->get_rank() + 1) && axis >= (B->get_rank() + 1))   AXIS_ERROR;
-   return laminate(A, Axis(axis + 1), B);
+   if (axis <= -1.0)   AXIS_ERROR;
+   if (axis >= (A->get_rank() + 1.0) &&
+       axis >= (B->get_rank() + 1.0))   AXIS_ERROR;
+   return laminate(A, Axis(axis + 1.0), B);
 }
 //-----------------------------------------------------------------------------
 Token
@@ -836,7 +843,7 @@ ShapeItem c2 = 1;   // assume B is scalar;
    if (B->get_rank() >= 1)
       {
         c1 = B->get_shape_item(0);
-        for (Rank r = 1; r < B->get_rank(); ++r)   c2 *= B->get_shape_item(r);
+        for (uRank r = 1; r < B->get_rank(); ++r)   c2 *= B->get_shape_item(r);
       }
 
 Shape shape_Z(c1, c2);
@@ -854,6 +861,15 @@ Shape shape_Z(c1, c2);
 Token
 Bif_F12_COMMA1::eval_AB(Value_P A, Value_P B)
 {
+  if (A->is_scalar() && B->is_scalar())
+     {
+       Value_P Z(2, LOC);
+       Z->next_ravel()->init(A->get_ravel(0), Z.getref(), LOC);
+       Z->next_ravel()->init(B->get_ravel(0), Z.getref(), LOC);
+       Z->check_value(LOC);
+       return Token(TOK_APL_VALUE1, Z);
+     }
+
    return catenate(A, 0, B);
 }
 //-----------------------------------------------------------------------------
@@ -877,7 +893,7 @@ Bif_F12_DOMINO::eval_B(Value_P B)
 
    if (B->get_rank() == 1)   // inversion at the unit sphere
       {
-        const APL_Float qct = Workspace::get_CT();
+        const double qct = Workspace::get_CT();
         const ShapeItem len = B->get_shape_item(0);
         APL_Complex r2(0.0);
         loop(l, len)
@@ -1146,7 +1162,7 @@ Bif_F12_TRANSPOSE::eval_AB(Value_P A, Value_P B)
         return Token(TOK_APL_VALUE1, Z);
       }
 
-const Shape shape_A(A, Workspace::get_IO());
+const Shape shape_A(A.get(), Workspace::get_IO());
    if (shape_A.get_rank() != B->get_rank())   LENGTH_ERROR;
 
    // the elements in A shall be valid axes of B->
@@ -1171,7 +1187,7 @@ Bif_F12_TRANSPOSE::transpose(const Shape & A, Value_P B)
 const Shape shape_Z = permute(B->get_shape(), inverse_permutation(A));
 Value_P Z(shape_Z, LOC);
 
-   if (Z->is_empty())
+   if (shape_Z.is_empty())
       {
          Z->set_default(*B.get(), LOC);
          return Z;
@@ -1179,9 +1195,9 @@ Value_P Z(shape_Z, LOC);
 
 const Cell * cB = &B->get_ravel(0);
 
-   for (ArrayIterator it_Z(shape_Z); !it_Z.done(); ++it_Z)
+   for (ArrayIterator it_Z(shape_Z); it_Z.more(); ++it_Z)
        {
-         const Shape idx_B = permute(it_Z.get_values(), A);
+         const Shape idx_B = permute(it_Z.get_offsets(), A);
          const ShapeItem b = B->get_shape().ravel_pos(idx_B);
          Z->next_ravel()->init(cB[b], Z.getref(), LOC);
        }
@@ -1239,9 +1255,9 @@ Value_P Z(shape_Z, LOC);
 
 const Cell * cB = &B->get_ravel(0);
 
-   for (ArrayIterator it_Z(shape_Z); !it_Z.done(); ++it_Z)
+   for (ArrayIterator it_Z(shape_Z); it_Z.more(); ++it_Z)
        {
-         const Shape idx_B = permute(it_Z.get_values(), A);
+         const Shape idx_B = permute(it_Z.get_offsets(), A);
          const ShapeItem b = B->get_shape().ravel_pos(idx_B);
          Z->next_ravel()->init(cB[b], Z.getref(), LOC);
        }
@@ -1338,7 +1354,7 @@ const ShapeItem l_len_B = shape_B1.get_volume();
 
    if ((l_len_A != 1) && (h_len_B != 1) && (l_len_A != h_len_B))   LENGTH_ERROR;
 
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
 
 const Shape shape_Z = shape_A1 + shape_B1;
 
@@ -1426,13 +1442,13 @@ const ShapeItem len = (len_A == 1) ? len_B : len_A;
         if (weight_f < SMALL_INT)   return true;
 
         const APL_Integer vB = cB[0].get_near_int();
-        value   += weight*vB;
-        value_f += weight_f*vB;
+        value   = value   + weight   * vB;
+        value_f = value_f + weight_f * vB;
         if (value_f > LARGE_INT)   return true;
         if (value_f < SMALL_INT)   return true;
 
-        weight   *= cA[0].get_near_int();
-        weight_f *= cA[0].get_near_int();
+        weight   = weight   * cA[0].get_near_int();
+        weight_f = weight_f * cA[0].get_near_int();
         if (len_A != 1)   --cA;
         if (len_B != 1)   cB -= dB;
       }
@@ -1461,8 +1477,10 @@ const ShapeItem len = (len_A == 1) ? len_B : len_A;
         if (len_B != 1)   cB -= dB;
       }
 
-   if (value < LARGE_INT && value > SMALL_INT && Cell::is_near_int(value))
-      new (cZ)   IntCell(value);
+   if (value < LARGE_INT &&
+       value > SMALL_INT &&
+       Cell::is_near_int(value))
+      new (cZ)   IntCell(Cell::near_int(value));
    else
       new (cZ)   FloatCell(value);
 }
@@ -1490,8 +1508,9 @@ const ShapeItem len = (len_A == 1) ? len_B : len_A;
       new (cZ)   ComplexCell(value);
    else if (value.imag() < -qct)
       new (cZ)   ComplexCell(value);
-   else if (value.real() < LARGE_INT && value.real() > SMALL_INT
-                              && Cell::is_near_int(value.real()))
+   else if (value.real() < LARGE_INT
+         && value.real() > SMALL_INT
+         && Cell::is_near_int(value.real()))
       new (cZ)   IntCell(value.real());
    else
       new (cZ)   FloatCell(value.real());
@@ -1515,7 +1534,7 @@ const ShapeItem ec_B = B->element_count();
 const ShapeItem ah = A->get_shape_item(0);    // first dimension
 const ShapeItem al = A->element_count()/ah;   // remaining dimensions
 
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
 
 const Shape shape_Z = A->get_shape() + B->get_shape();
 const ShapeItem dZ = shape_Z.get_volume()/shape_Z.get_shape_item(0);
@@ -1656,6 +1675,70 @@ Token
 Bif_F12_ELEMENT::eval_B(Value_P B)
 {
    // enlist
+   //
+   if (B->element_count() == 0)   // empty argument
+      {
+        ShapeItem N = 0;
+        Value_P Z(N, LOC);   // empty vector with prototype ' ' or '0'
+        const Cell * C = &B->get_ravel(0);
+        bool left = false;
+        for (;;)
+            {
+              if (C->is_pointer_cell())
+                 {
+                   C = &C->get_pointer_value()->get_ravel(0);
+                   continue;
+                 }
+
+              if (left && C->is_lval_cell())
+                 {
+                   C = C->get_lval_value();
+                   if (C == 0)
+                      {
+                        CERR << "0-pointer at " LOC << endl;
+                        FIXME;
+                      }
+                   else if (C->is_pointer_cell())
+                      {
+                        C = &C->get_pointer_value()->get_ravel(0);
+                      }
+                   else
+                      {
+                        Value * owner = C->cLvalCell().get_cell_owner();
+                        new (&Z->get_ravel(0))
+                            LvalCell(C->get_lval_value(), owner);
+                        break;
+                      }
+                 }
+
+              if (C->is_numeric())
+                 {
+                   new (&Z->get_ravel(0)) CharCell(UNI_ASCII_0);
+                   break;
+                 }
+
+              if (C->is_character_cell())
+                 {
+                   new (&Z->get_ravel(0)) CharCell(UNI_ASCII_SPACE);
+                   break;
+                 }
+
+              if (C->is_lval_cell())
+                 {
+                   left = C->cLvalCell().get_cell_owner();
+                   C = C->get_lval_value();
+                   continue;
+                 }
+
+               // not reached
+               //
+               FIXME;
+            }
+
+        Z->check_value(LOC);
+        return Token(TOK_APL_VALUE1, Z);
+      }
+
 const ShapeItem len_Z = B->get_enlist_count();
 Value_P Z(len_Z, LOC);
 
@@ -1672,7 +1755,7 @@ Bif_F12_ELEMENT::eval_AB(Value_P A, Value_P B)
 {
    // member
    //
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
 const ShapeItem len_Z = A->element_count();
 const ShapeItem len_B = B->element_count();
 Value_P Z(A->get_shape(), LOC);
@@ -1701,24 +1784,23 @@ const Rank axis = Value::get_single_axis(X.get(), B->get_rank());
    return partition(A, B, axis);
 }
 //-----------------------------------------------------------------------------
-Token
-Bif_F12_PARTITION::eval_B(Value_P B)
+Value_P
+Bif_F12_PARTITION::do_eval_B(Value_P B)
 {
-   if (B->is_simple_scalar())   return Token(TOK_APL_VALUE1, B);
+   if (B->is_simple_scalar())   return B;
 
 Value_P Z(LOC);
-   new (Z->next_ravel()) PointerCell(B->clone(LOC), Z.getref());
+   new (Z->next_ravel()) PointerCell(B->clone(LOC).get(), Z.getref());
    Z->check_value(LOC);
-   return Token(TOK_APL_VALUE1, Z);
+   return Z;
 }
 //-----------------------------------------------------------------------------
-Token
-Bif_F12_PARTITION::eval_XB(Value_P X, Value_P B)
+Value_P
+Bif_F12_PARTITION::do_eval_XB(Value_P X, Value_P B)
 {
 const Shape shape_X = Value::to_shape(X.get());
 
-Value_P Z = enclose_with_axes(shape_X, B);
-   return Token(TOK_APL_VALUE1, Z);
+   return enclose_with_axes(shape_X, B);
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -1779,12 +1861,12 @@ Value_P Z(shape_Z, LOC);
          return Z;
       }
 
-   for (ArrayIterator it_Z(shape_Z); !it_Z.done(); ++it_Z)
+   for (ArrayIterator it_Z(shape_Z); it_Z.more(); ++it_Z)
       {
         const ShapeItem off_Z = it_Z.multiply(weight_Z);   // offset in Z
 
         Value_P vZ(item_shape, LOC);
-        new (Z->next_ravel()) PointerCell(vZ, Z.getref());
+        new (Z->next_ravel()) PointerCell(vZ.get(), Z.getref());
 
         if (item_shape.is_empty())
            {
@@ -1792,7 +1874,7 @@ Value_P Z(shape_Z, LOC);
            }
         else
            {
-             for (ArrayIterator it_it(item_shape); !it_it.done(); ++it_it)
+             for (ArrayIterator it_it(item_shape); it_it.more(); ++it_it)
                  {
                    const ShapeItem off_B =  // offset in B
                          it_it.multiply(it_weight);
@@ -1928,7 +2010,7 @@ Cell * vv = &V->get_ravel(0);
        }
 
    V->check_value(LOC);
-   new (dest) PointerCell(V, dest_owner);
+   new (dest) PointerCell(V.get(), dest_owner);
 }
 //-----------------------------------------------------------------------------
 Token
@@ -1936,10 +2018,19 @@ Bif_F12_PICK::disclose(Value_P B, bool rank_tolerant)
 {
    if (B->is_simple_scalar())   return Token(TOK_APL_VALUE1, B);
 
-const ShapeItem len_B = B->element_count();
 const Shape item_shape = compute_item_shape(B, rank_tolerant);
 const Shape shape_Z = B->get_shape() + item_shape;
 
+const ShapeItem len_B = B->element_count();
+   if (len_B == 0)
+      {
+         Value_P first = Bif_F12_TAKE::first(B);
+         Token result = disclose(first, rank_tolerant);
+         if (result.get_Class() == TC_VALUE)   // success
+            result.get_apl_val()->set_shape(shape_Z);
+         return result;
+      }
+
 Value_P Z(shape_Z, LOC);
 
 const ShapeItem llen = item_shape.get_volume();
@@ -2077,19 +2168,19 @@ Value_P Z(shape_Z, LOC);
    //
 PermutedArrayIterator it_Z(shape_Z, perm);
 
-   for (ArrayIterator it_B(B->get_shape()); !it_B.done(); ++it_B)
+   for (ArrayIterator it_B(B->get_shape()); it_B.more(); ++it_B)
       {
-        const Cell & B_item = B->get_ravel(it_B.get_total());
+        const Cell & B_item = B->get_ravel(it_B());
         const Cell * src = 0;
         if (B_item.is_pointer_cell())
            {
              Value_P vB = B_item.get_pointer_value();
              ArrayIterator vB_it(vB->get_shape());
-             for (ArrayIterator it_it(item_shape); !it_it.done(); ++it_it)
+             for (ArrayIterator it_it(item_shape); it_it.more(); ++it_it)
                  {
-                   if (vB->get_shape().contains(it_it.get_values()))
+                   if (vB->get_shape().contains(it_it.get_offsets()))
                       {
-                        src = &vB->get_ravel(vB_it.get_total());
+                        src = &vB->get_ravel(vB_it());
                         ++vB_it;
                       }
                    else if (vB->get_ravel(0).is_character_cell())  // char
@@ -2097,15 +2188,15 @@ PermutedArrayIterator it_Z(shape_Z, perm);
                    else                                   // simple numeric
                         src = &n_filler;
 
-                   Z->get_ravel(it_Z.get_total()).init(*src, Z.getref(), LOC);
+                   Z->get_ravel(it_Z()).init(*src, Z.getref(), LOC);
                    ++it_Z;
                  }
            }
         else
            {
-             for (ArrayIterator it_it(item_shape); !it_it.done(); ++it_it)
+             for (ArrayIterator it_it(item_shape); it_it.more(); ++it_it)
                  {
-                   if (it_it.get_total() == 0)   // first element: use B_item
+                   if (it_it() == 0)   // first element: use B_item
                       {
                         src = &B_item;
                       }
@@ -2118,7 +2209,7 @@ PermutedArrayIterator it_Z(shape_Z, perm);
                         src = &n_filler;
                       }
 
-                   Z->get_ravel(it_Z.get_total()).init(*src, Z.getref(), LOC);
+                   Z->get_ravel(it_Z()).init(*src, Z.getref(), LOC);
                    ++it_Z;
                  }
             }
@@ -2228,7 +2319,7 @@ ShapeItem c = 0;
         if (B->get_rank() != len_A)   RANK_ERROR;
 
         const Shape weight = B->get_shape().reverse_scan();
-        const Shape A_as_shape(A, qio);
+        const Shape A_as_shape(A.get(), qio);
 
         loop(r, A->element_count())
             {
@@ -2412,265 +2503,6 @@ IndexExpr index_expr(ASS_none, LOC);
         return Token(TOK_APL_VALUE1, Z);
       }
 }
-//-----------------------------------------------------------------------------
-Value_P
-Bif_F12_TAKE::first(Value_P B)
-{
-const Cell & first_B = B->get_ravel(0);
-   if (B->element_count() == 0)   // empty value: return prototype
-      {
-        if (first_B.is_lval_cell())   // (
-            {
-              Value_P Z(LOC);
-              Z->next_ravel()->init(first_B, B.getref(), LOC);
-              return Z;
-            }
-        Value_P Z = B->prototype(LOC);
-        Z->check_value(LOC);
-        return Z;
-      }
-
-   if (!first_B.is_pointer_cell())   // simple cell
-      {
-        Value_P Z(LOC);
-        Z->get_ravel(0).init(first_B, Z.getref(), LOC);
-        Z->check_value(LOC);
-        return Z;
-      }
-
-Value_P v1 = first_B.get_pointer_value();
-Value * v1_owner = v1->get_lval_cellowner();
-   if (v1_owner)   // B is a left value
-      {
-        Value_P B1(LOC);
-        new (&B1->get_ravel(0))   PointerCell(v1, B1.getref());
-
-        Value_P Z(LOC);
-        new (&Z->get_ravel(0))   LvalCell(&B1->get_ravel(0), v1_owner);
-
-        Z->check_value(LOC);
-        return Z;
-      }
-   else
-      {
-        const ShapeItem ec = v1->element_count();
-        Value_P Z(v1->get_shape(), LOC);
-        if (ec == 0)   Z->get_ravel(0).init(v1->get_ravel(0), Z.getref(), LOC);
-
-        loop(e, ec)   Z->next_ravel()->init(v1->get_ravel(e), Z.getref(), LOC);
-
-        Z->check_value(LOC);
-        return Z;
-      }
-}
-//-----------------------------------------------------------------------------
-Token
-Bif_F12_TAKE::eval_AXB(Value_P A, Value_P X, Value_P B)
-{
-   if (X->element_count() == 0)   // no axes
-      {
-        Token result(TOK_APL_VALUE1, B->clone(LOC));
-        return result;
-      }
-
-   // A↑[X]B ←→ ⊃[X](⊂A)↑¨⊂[X]B
-   //
-Value_P cA = Bif_F12_PARTITION::fun->eval_B(A).get_apl_val();           // ⊂A
-Value_P cB = Bif_F12_PARTITION::fun->eval_XB(X, B).get_apl_val();       // ⊂[X]B
-Token take(TOK_FUN2, Bif_F12_TAKE::fun);
-Value_P cT = Bif_OPER1_EACH::fun->eval_ALB(cA, take, cB).get_apl_val(); // cA↑¨cB
-
-Token result = Bif_F12_PICK::fun->eval_XB(X, cT);
-   return result;
-}
-//-----------------------------------------------------------------------------
-Token
-Bif_F12_TAKE::eval_AB(Value_P A, Value_P B)
-{
-Shape ravel_A1(A, /* ⎕IO */ 0);   // checks that 1 ≤ ⍴⍴A and ⍴A ≤ MAX_RANK
-
-   if (B->is_scalar())
-      {
-        Shape shape_B1;
-        loop(a, ravel_A1.get_rank())   shape_B1.add_shape_item(1);
-        Value_P B1 = B->clone(LOC);
-        B1->set_shape(shape_B1);
-        return do_take(ravel_A1, B1);
-      }
-   else
-      {
-        if (ravel_A1.get_rank() != B->get_rank())   LENGTH_ERROR;
-        return do_take(ravel_A1, B);
-      }
-}
-//-----------------------------------------------------------------------------
-Token
-Bif_F12_TAKE::do_take(const Shape shape_Zi, Value_P B)
-{
-   // shape_Zi can have negative items (for take from the end). Create
-   // shape_Z which is a true shape with the absolute values of shape_Zi
-   //
-Shape shape_Z(shape_Zi);
-
-   loop(r, shape_Zi.get_rank())
-       {
-         ShapeItem a = shape_Zi.get_shape_item(r);
-
-         if (a < 0)    shape_Z.set_shape_item(r, -a);
-       }
-
-Value_P Z(shape_Z, LOC);
-
-   fill(shape_Zi, &Z->get_ravel(0), Z.getref(), B);
-
-   Z->set_default(*B.get(), LOC);
-   Z->check_value(LOC);
-   return Token(TOK_APL_VALUE1, Z);
-}
-//-----------------------------------------------------------------------------
-void
-Bif_F12_TAKE::fill(const Shape & shape_Zi, Cell * cZ, Value & Z_owner,
-                   Value_P B)
-{
-   if (shape_Zi.is_empty())   return;
-
-   Assert(shape_Zi.get_rank() == B->get_rank());
-
-   // compute shape_Z = abs(shape_Zi), and Q with Z[i] = B[Q + i].
-   //
-Shape shape_Z;
-Shape Q;
-
-   loop(r, shape_Zi.get_rank())
-       {
-         const ShapeItem Zi = shape_Zi.get_shape_item(r);
-
-         if (Zi >= 0)
-            {
-              shape_Z.add_shape_item(Zi);
-              Q.add_shape_item(0);
-            }
-         else
-            {
-              shape_Z.add_shape_item(-Zi);
-              Q.add_shape_item(B->get_shape_item(r) + Zi);
-            }
-       }
-
-   if (shape_Z.is_empty())   return;
-
-const Shape weight_B = B->get_shape().reverse_scan();
-
-   for (ArrayIterator it(shape_Z); !it.done(); ++it)
-      {
-        ShapeItem bpos = 0;
-        bool fill = false;
-        loop(r, shape_Z.get_rank())
-            {
-              const ShapeItem idx_B = it.get_value(r) + Q.get_shape_item(r);
-              if ((idx_B <  0) || (idx_B >= B->get_shape_item(r)))
-                 {
-                   fill = true;
-                   break;
-                 }
-
-              bpos += weight_B.get_shape_item(r) * idx_B;
-            }
-
-        if (fill)   cZ->init_type(B->get_ravel(0), Z_owner, LOC);
-        else        cZ->init(B->get_ravel(bpos),   Z_owner, LOC);
-        ++cZ;
-      }
-}
-//-----------------------------------------------------------------------------
-Token
-Bif_F12_DROP::eval_AB(Value_P A, Value_P B)
-{
-Shape ravel_A(A, /* ⎕IO */ 0);
-   if (A->get_rank() > 1)   RANK_ERROR;
-
-   if (B->is_scalar())
-      {
-        // if B is a scalar then the result rank shall be the length of A->
-        // the result may be empty (shape 0 0 ... 0) if we drop something
-        // or non-empty (shape 1 1 ... 1) if we drop nothing.
-        //
-        const ShapeItem len_Z = ravel_A.get_volume() ? 0 : 1;
-
-        Shape shape_Z;
-        loop(r, ravel_A.get_rank())   shape_Z.add_shape_item(len_Z);
-
-        Value_P Z(shape_Z, LOC);
-
-        Z->get_ravel(0).init(B->get_ravel(0), Z.getref(), LOC);
-        Z->check_value(LOC);
-        return Token(TOK_APL_VALUE1, Z);
-      }
-
-   if (ravel_A.get_rank() == 0)   ravel_A.add_shape_item(1);   // A = ,A
-
-   if (ravel_A.get_rank() != B->get_rank())   LENGTH_ERROR;
-
-   loop(r, ravel_A.get_rank())
-       {
-         const APL_Integer a = ravel_A.get_shape_item(r);
-         const ShapeItem amax = B->get_shape_item(r);
-
-         if      (a >= amax)  ravel_A.set_shape_item(r, 0);
-         else if (a >= 0)     ravel_A.set_shape_item(r, a - amax);
-         else if (a > -amax)  ravel_A.set_shape_item(r, amax + a);
-         else                 ravel_A.set_shape_item(r, 0);
-       }
-
-   return Bif_F12_TAKE::do_take(ravel_A, B);
-}
-//-----------------------------------------------------------------------------
-Token
-Bif_F12_DROP::eval_AXB(Value_P A, Value_P X, Value_P B)
-{
-   if (X->element_count() == 0)   // no axes
-      {
-        Token result(TOK_APL_VALUE1, B->clone(LOC));
-        return result;
-      }
-
-   if (X->get_rank() > 1)    INDEX_ERROR;
-
-const uint64_t len_X = X->element_count();
-   if (len_X > MAX_RANK)     INDEX_ERROR;
-   if (len_X == 0)           INDEX_ERROR;
-
-   if (A->get_rank() > 1)    RANK_ERROR;
-
-uint64_t len_A = A->element_count();
-   if (len_A != len_X)   LENGTH_ERROR;
-
-const APL_Integer qio = Workspace::get_IO();
-
-   // init ravel_A = shape_B and seen.
-   //
-Shape ravel_A(B->get_shape());
-bool seen[MAX_RANK];
-   loop(r, B->get_rank())   seen[r] = false;
-
-   loop(r, len_X)
-       {
-         const APL_Integer a = A->get_ravel(r).get_near_int();
-         const APL_Integer x = X->get_ravel(r).get_near_int() - qio;
-
-         if (x >= B->get_rank())   INDEX_ERROR;
-         if (seen[x])              INDEX_ERROR;
-         seen[x] = true;
-
-         const ShapeItem amax = B->get_shape_item(x);
-         if      (a >= amax)   ravel_A.set_shape_item(x, 0);
-         else if (a >= 0)      ravel_A.set_shape_item(x, a - amax);
-         else if (a > -amax)   ravel_A.set_shape_item(x, amax + a);
-         else                  ravel_A.set_shape_item(x, 0);
-       }
-
-   return Bif_F12_TAKE::do_take(ravel_A, B);
-}
 //=============================================================================
 Token
 Bif_F12_EQUIV::eval_B(Value_P B)
@@ -2689,7 +2521,7 @@ Bif_F12_EQUIV::eval_AB(Value_P A, Value_P B)
    // match
    //
 
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
 const ShapeItem count = A->nz_element_count();  // compare at least prototype
 
    if (!A->same_shape(*B))   return Token(TOK_APL_VALUE1, IntScalar(0, LOC));   // no match
@@ -2719,7 +2551,7 @@ Bif_F12_NEQUIV::eval_AB(Value_P A, Value_P B)
    // match
    //
 
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
 const ShapeItem count = A->nz_element_count();  // compare at least prototype
 
    if (!A->same_shape(*B))   return Token(TOK_APL_VALUE1, IntScalar(1, LOC));   // no match
@@ -2817,21 +2649,21 @@ const bool user_cmd = Command::do_APL_command(out, command);
 
 UTF8_string result_utf8 = out.get_data();
    if (result_utf8.size() == 0 ||
-       result_utf8.last() != UNI_ASCII_LF)
-      result_utf8.append(UNI_ASCII_LF);
+       result_utf8.back() != UNI_ASCII_LF)
+      result_utf8 += '\n';
 
-Simple_string<ShapeItem, false> line_starts;
-   line_starts.append(0);
+std::vector<ShapeItem> line_starts;
+   line_starts.push_back(0);
    loop(r, result_utf8.size())
       {
-        if (result_utf8[r] == UNI_ASCII_LF)   line_starts.append(r + 1);
+        if (result_utf8[r] == UNI_ASCII_LF)   line_starts.push_back(r + 1);
       }
 
-Value_P Z((ShapeItem)line_starts.size() - 1, LOC);
+Value_P Z(ShapeItem(line_starts.size() - 1), LOC);
    loop(l, line_starts.size() - 1)
       {
         ShapeItem len;
-        if (l < line_starts.size() - 1)
+        if (l < ShapeItem(line_starts.size() - 1))
            len = line_starts[l + 1] - line_starts[l] - 1;
         else
            len = result_utf8.size() - line_starts[l];
@@ -2839,7 +2671,7 @@ Value_P Z((ShapeItem)line_starts.size() - 1, LOC);
         UTF8_string line_utf8(&result_utf8[line_starts[l]], len);
         UCS_string line_ucs(line_utf8);
         Value_P ZZ(line_ucs, LOC);
-        new (Z->next_ravel())   PointerCell(ZZ, Z.getref());
+        new (Z->next_ravel())   PointerCell(ZZ.get(), Z.getref());
       }
 
    Z->check_value(LOC);
@@ -2851,21 +2683,48 @@ Bif_F12_UNION::eval_B(Value_P B)
 {
    if (B->get_rank() > 1)   RANK_ERROR;
 
-const APL_Float qct = Workspace::get_CT();
+const double qct = Workspace::get_CT();
 const ShapeItem len_B = B->element_count();
    if (len_B <= 1)   return Token(TOK_APL_VALUE1, B->clone(LOC));
-   if (len_B >= SHORT_VALUE_LENGTH_WANTED && !B->is_complex(false))
-      return Macro::Z__UNIQUE_B->eval_B(B);
+   if (len_B >= 20)
+      {
+        const Cell ** cells_B = new const Cell *[2*len_B];
+        const Cell ** cells_Z = cells_B + len_B;
 
-   // B is small, so an itertive search of unique elements is faster
+        loop(b, len_B)   cells_B[b] = &B->get_ravel(b);
+
+        Heapsort<const Cell *>::sort(cells_B, len_B, 0, Cell::compare_stable);
+
+        ShapeItem idx_B = 0;
+        ShapeItem len_Z = 0;
+        while(idx_B < len_B)
+            {
+              const Cell * ref = cells_B[idx_B++];
+              cells_Z[len_Z++] = ref;   // always take first
+              while (idx_B < len_B && ref->equal(*cells_B[idx_B], qct)) ++idx_B;
+            }
+
+        // sort cells_Z by position so that the original order in B is
+        // reconstructed
+        //
+        Heapsort<const Cell *>::sort(cells_Z, len_Z, 0, Cell::compare_ptr);
+
+        Value_P Z(len_Z, LOC);
+        loop(z, len_Z)   Z->next_ravel()->init(*cells_Z[z], Z.getref(), LOC);
+        delete[] cells_B;
+        Z->check_value(LOC);
+        return Token(TOK_APL_VALUE1, Z);
+      }
+
+   // B is small, so an iterative search of unique elements is faster
    //
-Simple_string<const Cell *, false> items_Z;
+std::vector<const Cell *> items_Z;
    items_Z.reserve(len_B);
 
    loop(b, len_B)
       {
         const Cell & cell = B->get_ravel(b);
-        if (is_unique(cell, items_Z, qct))   items_Z.append(&cell);
+        if (is_unique(cell, items_Z, qct))   items_Z.push_back(&cell);
       }
 
    // build result value Z
@@ -2883,10 +2742,82 @@ Bif_F2_INTER::eval_AB(Value_P A, Value_P B)
    if (A->get_rank() > 1)   RANK_ERROR;
    if (B->get_rank() > 1)   RANK_ERROR;
 
-   // A ∩ B ←→ (A∈B)/A
-   //
-Token AinB = Bif_F12_ELEMENT::fun->eval_AB(A, B);
-   return Bif_OPER1_REDUCE::fun->eval_AB(AinB.get_apl_val(), A);
+const ShapeItem len_A = A->element_count();
+const ShapeItem len_B = B->element_count();
+
+const double qct = Workspace::get_CT();
+
+   if (len_A*len_B > 60*60)
+      {
+        // large A or B: sort A and B to speed up searches
+        //
+        const Cell ** cells_A = new const Cell *[2*len_A + len_B];
+        const Cell ** cells_Z = cells_A + len_A;
+        const Cell ** cells_B = cells_A + 2*len_A;
+
+        loop(a, len_A)   cells_A[a] = &A->get_ravel(a);
+        loop(b, len_B)   cells_B[b] = &B->get_ravel(b);
+
+        Heapsort<const Cell *>::sort(cells_A, len_A, 0, Cell::compare_stable);
+        Heapsort<const Cell *>::sort(cells_B, len_B, 0, Cell::compare_stable);
+
+        ShapeItem len_Z = 0;
+        ShapeItem idx_B = 0;
+        loop(idx_A, len_A)
+            {
+              const Cell * ref = cells_A[idx_A];
+              while (idx_B < len_B)
+                  {
+                    if (ref->equal(*cells_B[idx_B], qct))
+                       {
+                         cells_Z[len_Z++] = ref;   // A is in B
+                         break;   // for idx_B → next idx_A
+                       }
+
+                    // B is much (by ⎕CT) smaller or greater than A
+                    //
+                    if (ref->greater(*cells_B[idx_B]))    ++idx_B;
+                    else                                  break;
+                 }
+            }
+
+        // sort cells_Z by position so that the original order in A is
+        //  reconstructed
+        //
+        Heapsort<const Cell *>::sort(cells_Z, len_Z, 0, Cell::compare_ptr);
+        Value_P Z(len_Z, LOC);
+        loop(z, len_Z)   Z->next_ravel()->init(*cells_Z[z], Z.getref(), LOC);
+
+        Z->set_default(*B, LOC);
+        Z->check_value(LOC);
+        delete[] cells_A;
+        return Token(TOK_APL_VALUE1, Z);
+      }
+    else
+      {
+        // small A and B: use quadratic algorithm.
+        //
+        const Cell ** cells_Z = new const Cell *[len_A];
+        ShapeItem len_Z = 0;
+
+        loop(a, len_A)
+        loop(b, len_B)
+            {
+              if (A->get_ravel(a).equal(B->get_ravel(b), qct))
+                 {
+                   cells_Z[len_Z++] = &A->get_ravel(a);
+                   break;   // loop(b)
+                 }
+            }
+        Value_P Z(len_Z, LOC);
+        loop(z, len_Z)
+            Z->next_ravel()->init(*cells_Z[z], Z.getref(), LOC);
+
+        Z->set_default(*B, LOC);
+        Z->check_value(LOC);
+        delete[] cells_Z;
+        return Token(TOK_APL_VALUE1, Z);
+      }
 }
 //-----------------------------------------------------------------------------
 Token
@@ -2898,7 +2829,62 @@ Bif_F12_UNION::eval_AB(Value_P A, Value_P B)
    // A ∪ B ←→ A,B∼A
    //
 Token BwoA = Bif_F12_WITHOUT::fun->eval_AB(B, A);
+
+const ShapeItem len_A = A->element_count();
+const ShapeItem len_B = BwoA.get_apl_val()->element_count();
+Value_P Z(len_A + len_B, LOC);
+
+   loop(a, len_A)   Z->next_ravel()->init(A->get_ravel(a), Z.getref(), LOC);
+   loop(b, len_B)   Z->next_ravel()->init(B->get_ravel(b), Z.getref(), LOC);
+   Z->set_default(*B, LOC);
+   Z->check_value(LOC);
    return Bif_F12_COMMA::fun->eval_AB(A, BwoA.get_apl_val());
 }
 //=============================================================================
+Token
+Bif_F2_RIGHT::eval_AXB(Value_P A, Value_P X, Value_P B)
+{
+   // select corresponding items of A or B according to X. A, B, and X
+   // must have matching shapes
+   //
+const int inc_A = A->is_scalar_extensible() ? 0 : 1;
+const int inc_B = B->is_scalar_extensible() ? 0 : 1;
+const int inc_X = X->is_scalar_extensible() ? 0 : 1;
+
+   if (inc_X == 0)   // single item X: pick entire A or B according to X
+      {
+        const APL_Integer x0 = X->get_ravel(0).get_int_value();
+        if (x0 == 0)   return Token(TOK_APL_VALUE1, A->clone(LOC));
+        if (x0 == 1)   return Token(TOK_APL_VALUE1, B->clone(LOC));
+        DOMAIN_ERROR;
+      }
+
+   // X is non-scalar, so it must match any non-scalar A and B
+   //
+   if (inc_A && ! X->same_shape(*A))
+      {
+        if (A->get_rank() != X->get_rank())   RANK_ERROR;
+        else                                  LENGTH_ERROR;
+      }
+
+   if (inc_B && ! X->same_shape(*B))
+      {
+        if (B->get_rank() != X->get_rank())   RANK_ERROR;
+        else                                  LENGTH_ERROR;
+      }
+
+Value_P Z(X->get_shape(), LOC);
+   loop(z, Z->nz_element_count())
+       {
+        const APL_Integer xz = X->get_ravel(z).get_int_value();
+        Cell * cZ = Z->next_ravel();
+        if      (xz == 0)   cZ->init(A->get_ravel(z*inc_A), Z.getref(), LOC);
+        else if (xz == 1)   cZ->init(B->get_ravel(z*inc_B), Z.getref(), LOC);
+        else                DOMAIN_ERROR;
+       }
+
+   Z->check_value(LOC);
+   return Token(TOK_APL_VALUE1, Z);
+}
+//=============================================================================
 
