@@ -27,7 +27,7 @@
 #include "PointerCell.hh"
 #include "PrintBuffer.hh"
 #include "PrintOperator.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Workspace.hh"
 
 /// max sizes for arrays on the stack. Larger values are allocated with new()
@@ -49,7 +49,7 @@ PrintBuffer::PrintBuffer(const UCS_string & ucs, const ColInfo & ci)
    : col_info(ci),
      complete(true)
 {
-   buffer.append(ucs);
+   buffer.push_back(ucs);
 }
 //-----------------------------------------------------------------------------
 PrintBuffer::PrintBuffer(const Value & value, const PrintContext & _pctx,
@@ -59,7 +59,7 @@ PrintBuffer::PrintBuffer(const Value & value, const PrintContext & _pctx,
 PERFORMANCE_START(start_0)
 
    // Note: if ostream is non-0 then this value may be incomplete
-   // (as indicated by member complete if it is huge. This is to speed
+   // (as indicated by member complete if it is huge). This is to speed
    // up printing if the value is discarded after having been printed
 
 const PrintStyle outer_style = _pctx.get_style();
@@ -149,13 +149,13 @@ const ShapeItem ec = value.element_count();
    //
 const ShapeItem cols = value.get_last_shape_item();
 
-Simple_string<bool, false> scaling;
+vector<bool> scaling;
    scaling.reserve(cols);
-   loop(c, cols)   scaling.append(false);
+   loop(c, cols)   scaling.push_back(false);
 
-Simple_string<PrintBuffer, true> pcols;
+vector<PrintBuffer> pcols;
    pcols.reserve(cols);
-   loop(c, cols)   pcols.append(PrintBuffer());
+   loop(c, cols)   pcols.push_back(PrintBuffer());
 
 PrintBuffer * item_matrix = 0;
    try { item_matrix = new PrintBuffer[ec]; }
@@ -166,9 +166,8 @@ PrintBuffer * item_matrix = 0;
         WS_FULL;
       }
 
-   do_PrintBuffer(value, pctx, out, outer_style,
-                            &scaling[0], &pcols[0],
-                            item_matrix);
+   do_PrintBuffer(value, pctx, out, outer_style, scaling, pcols, item_matrix);
+
    delete [] item_matrix;
 
    PERFORMANCE_END(fs_PrintBuffer_B, start_0, ec)
@@ -177,7 +176,7 @@ PrintBuffer * item_matrix = 0;
 void
 PrintBuffer::do_PrintBuffer(const Value & value, const PrintContext & pctx,
                          ostream * out, PrintStyle outer_style,
-                         bool * scaling, PrintBuffer * pcols,
+                         vector<bool> & scaling, vector<PrintBuffer> & pcols,
                          PrintBuffer * item_matrix)
 {
 const bool framed = outer_style & (PST_CS_MASK | PST_CS_OUTER);
@@ -213,12 +212,12 @@ const bool nested = !value.is_simple();
      //    therefore we have (⍴,value) items. Items are rectangular.
      //
      PERFORMANCE_START(start_2)
-     Simple_string<Rank, false> max_row_ranks;
+     vector<Rank> max_row_ranks;
      max_row_ranks.reserve(rows);
      loop(y, rows)
         {
           ShapeItem max_row_height = 0;
-          max_row_ranks.append(0);
+          max_row_ranks.push_back(0);
 
           loop(x, cols)
               {
@@ -234,8 +233,8 @@ const bool nested = !value.is_simple();
                    {
                      Value_P sub_val = cell.get_pointer_value();
                      const Rank sub_rank = sub_val->get_rank();
-                     if (max_row_ranks.last() < sub_rank)
-                        max_row_ranks.last() = sub_rank;
+                     if (max_row_ranks.back() < sub_rank)
+                        max_row_ranks.back() = sub_rank;
                    }
 
                 if (max_row_height < item.get_height())
@@ -269,6 +268,8 @@ const bool nested = !value.is_simple();
                 PrintBuffer * item_row = item_matrix + y*cols;
                 col_info_x.consider(item_row[x].get_info());
               }
+           if (col_info_x.real_len<(col_info_x.int_len + col_info_x.denom_len))
+              col_info_x.real_len = col_info_x.int_len + col_info_x.denom_len;
 
           loop(y, rows)
               {
@@ -292,7 +293,7 @@ const bool nested = !value.is_simple();
           // PrintBuffer pcol. Insert separator rows as needed.
           //
           PrintBuffer & dest = pcols[x];
-          new (pcols + x) PrintBuffer;
+          pcols[x] = PrintBuffer();
 
           // compute the final height of dest and reserve enough rows as to
           // avoid unnecessary copies
@@ -603,16 +604,16 @@ ShapeItem ret = 0;
 Unicode
 PrintBuffer::get_char(int x, int y) const
 { 
-   Assert(y < (int)buffer.size());
-   Assert(x < (int)buffer[y].size());
+   Assert(y < int(buffer.size()));
+   Assert(x < int(buffer[y].size()));
    return buffer[y][x];
 }
 //-----------------------------------------------------------------------------
 void
 PrintBuffer::set_char(int x, int y, Unicode uc)
 {
-   Assert(y < (int)buffer.size());
-   Assert(x < (int)buffer[y].size());
+   Assert(y < int(buffer.size()));
+   Assert(x < int(buffer[y].size()));
    buffer[y][x] = uc;
 }
 //-----------------------------------------------------------------------------
@@ -643,7 +644,7 @@ PrintBuffer::pad_height(Unicode pad, ShapeItem height)
    if (height > get_height())
       {
         UCS_string ucs(get_width(0), pad);
-        while (height > get_height())   buffer.append(ucs);
+        while (height > get_height())   buffer.push_back(ucs);
       }
 }
 //-----------------------------------------------------------------------------
@@ -661,7 +662,7 @@ PrintBuffer::pad_height_above(Unicode pad, ShapeItem height)
    if (height > get_height())
       {
         UCS_string ucs(get_width(0), pad);
-        while (height > get_height())   buffer.insert_before(0, ucs);
+        while (height > get_height())   buffer.insert(buffer.begin(), ucs);
       }
 }
 //-----------------------------------------------------------------------------
@@ -733,12 +734,12 @@ Unicode HORI, VERT, NW, NE, SE, SW;
         UCS_string upper;
         upper.append(NE);
         upper.append(NW);
-        buffer.append(upper);
+        buffer.push_back(upper);
 
         UCS_string lower;
         lower.append(SE);
         lower.append(SW);
-        buffer.append(lower);
+        buffer.push_back(lower);
 
         Assert(is_rectangular());
         return;
@@ -762,8 +763,8 @@ Unicode HORI, VERT, NW, NE, SE, SW;
    //
 UCS_string hori(get_width(0), HORI);
 
-   buffer.insert_before(0, hori);
-   buffer.append(hori);
+   buffer.insert(buffer.begin(), hori);
+   buffer.push_back(hori);
 
    // draw the corners
    //
@@ -840,12 +841,12 @@ Unicode HORI, VERT, NW, NE, SE, SW;
         UCS_string upper;
         upper.append(NE);
         upper.append(NW);
-        buffer.append(upper);
+        buffer.push_back(upper);
 
         UCS_string lower;
         lower.append(SE);
         lower.append(SW);
-        buffer.append(lower);
+        buffer.push_back(lower);
 
         Assert(is_rectangular());
         return;
@@ -868,8 +869,8 @@ Unicode HORI, VERT, NW, NE, SE, SW;
    //
 UCS_string hori(get_width(0), HORI);
 
-   buffer.insert_before(0, hori);
-   buffer.append(hori);
+   buffer.insert(buffer.begin(), hori);
+   buffer.push_back(hori);
 
    // draw the corners
    //
@@ -907,10 +908,11 @@ PrintBuffer::debug(ostream & out, const char * title) const
    out << UNI_LINE_UP_RIGHT;
    loop(w, get_width(0))   out << UNI_LINE_HORI;
    out << UNI_LINE_UP_LEFT
-       << "  flags=" << HEX(col_info.flags)
-       << "  ilen="  << col_info.int_len
-       << "  flen="   << col_info.fract_len
-       << "  rlen="   << col_info.real_len
+       << " flg=" << HEX(col_info.flags)
+       << " il="  << col_info.int_len
+       << " fl="   << col_info.fract_len
+       << " rl="   << col_info.real_len
+       << " ÷l="   << col_info.denom_len
        << endl << endl;
 
    return out;
@@ -929,24 +931,24 @@ PrintBuffer::append_ucs(const UCS_string & ucs)
 {
    if (buffer.size() == 0)   // empty
       {
-        buffer.append(ucs);
+        buffer.push_back(ucs);
       }
    else if (ucs.size() < get_width(0))
       {
         UCS_string ucs1(ucs);
         UCS_string pad(get_width(0) - ucs.size(), UNI_iPAD_L1);
         ucs1.append(pad);
-        buffer.append(ucs1);
+        buffer.push_back(ucs1);
       }
    else if (ucs.size() > get_width(0))
       {
         UCS_string pad(ucs.size() - get_width(0), UNI_iPAD_L2);
         loop(h, get_height())   buffer[h].append(pad);
-        buffer.append(ucs);
+        buffer.push_back(ucs);
       }
    else
       {
-        buffer.append(ucs);
+        buffer.push_back(ucs);
       }
 }
 //-----------------------------------------------------------------------------
@@ -1048,7 +1050,7 @@ UCS_string ucs1;
    if (this_l > 0)   pad_l(UNI_ASCII_SPACE, this_l);
    if (this_r > 0)   pad_r(UNI_ASCII_SPACE, this_r);
 
-   buffer.append(ucs1);
+   buffer.push_back(ucs1);
 
    Assert(is_rectangular());
 }
@@ -1076,13 +1078,13 @@ PrintBuffer::add_column(Unicode pad, int32_t pad_count, const PrintBuffer & pb)
 void PrintBuffer::add_row(const PrintBuffer & pb)
 {
    buffer.reserve(buffer.size() + pb.get_height());
-   loop(h, pb.get_height())   buffer.append(pb.buffer[h]);
+   loop(h, pb.get_height())   buffer.push_back(pb.buffer[h]);
 }
 //-----------------------------------------------------------------------------
 void
 PrintBuffer::align(ColInfo & cols)
 {
-   // this PrintBuffer is a (possibly nested) APL value.
+   // this PrintBuffer is one (possibly nested) APL value.
    // Align the buffer:
    //
    // to the J (in a column containing complex numbers), or
@@ -1103,59 +1105,89 @@ PrintBuffer::align(ColInfo & cols)
 void
 PrintBuffer::align_dot(ColInfo & COL_INFO)
 {
+   // align this PrintBuffer (one value) at the decimal dot of COL_INFO.
+   //
+   // COL_INFO is the desired ColInfo of the entire column while
+   // col_info (a member of this PrintBuffer) is the (smaller) item
+   // being aligned
+   //
    Log(LOG_printbuf_align)
       {
-        CERR << "before align_dot(), COL_INFO = " << COL_INFO.int_len
-             << ":" << COL_INFO.fract_len
-             << ":" << COL_INFO.real_len
-             << ", this col = " << col_info.int_len
-             << ":" << col_info.fract_len
-             << ":" << col_info.real_len << endl;
+        CERR << "before align_dot():" << endl
+             << "desired COL_INFO = "
+                "i-"  << COL_INFO.int_len
+             << " f-" << COL_INFO.fract_len
+             << " r-" << COL_INFO.real_len
+             << " ÷"  << COL_INFO.denom_len << endl
+             << "this row         = "
+                "i-"  << col_info.int_len
+             << " f-" << col_info.fract_len
+             << " r-" << col_info.real_len
+             << " ÷"  << col_info.denom_len << endl;
         debug(CERR, 0);
       }
 
    Assert(buffer.size() > 0);
 
-   Assert(COL_INFO.int_len   >= col_info.int_len);
-   Assert(COL_INFO.fract_len >= col_info.fract_len);
-   Assert(COL_INFO.real_len  >= col_info.real_len);
+   // make sure that consider() has worked.
+   //
+   Assert(COL_INFO.int_len    >= col_info.int_len);
+   Assert(COL_INFO.fract_len  >= col_info.fract_len);
+   Assert(COL_INFO.real_len   >= col_info.real_len);
+   Assert(COL_INFO.denom_len  >= col_info.denom_len);
 
    if (col_info.flags & CT_NUMERIC)
       {
-        // dot-align numeric items
+        // numeric items are aligned at the decimal dot. First pad the
+        // integer part with spaces to the left
+        //
         if (COL_INFO.int_len > col_info.int_len)
            {
              const size_t diff = COL_INFO.int_len - col_info.int_len;
              pad_l(UNI_iPAD_L3, diff);
              col_info.real_len += diff;
-             col_info.int_len = COL_INFO.int_len;
+             col_info.int_len  += diff;
            }
 
-        if (COL_INFO.fract_len > col_info.fract_len)
-           {
-             pad_fraction(COL_INFO.fract_len, COL_INFO.have_expo());
-           }
-
-        if (COL_INFO.real_len > col_info.real_len)
+        if (col_info.denom_len)   // quotient: maybe pad right with spaces
            {
              const size_t diff = COL_INFO.real_len - col_info.real_len;
-             if (col_info.have_expo())   // alrady have an expo field
+             if (diff)
                 {
                   pad_r(UNI_iPAD_L4, diff);
+                  col_info.real_len += diff;
                 }
-             else                        // no expo yet: create one
+           }
+        else
+           {
+             if (COL_INFO.fract_len > col_info.fract_len)
                 {
-                  Assert1(diff >= 2);
-                  pad_r(UNI_ASCII_E, 1);
-                  pad_r(UNI_ASCII_0, 1);
-                  pad_r(UNI_iPAD_L4, diff - 2);
+                  pad_fraction(COL_INFO.fract_len, COL_INFO.have_expo());
+                }
+
+             if (COL_INFO.real_len > col_info.real_len)
+                {
+                  const size_t diff = COL_INFO.real_len - col_info.real_len;
+                  if (!(COL_INFO.flags & (real_has_E | imag_has_E))
+                   || col_info.have_expo())       // or has one already 
+                     {
+                       pad_r(UNI_iPAD_L4, diff);
+                     }
+                  else                        // no expo yet: create one
+                     {
+                       Assert1(diff >= 2);
+                       pad_r(UNI_ASCII_E, 1);
+                       pad_r(UNI_ASCII_0, 1);
+                       pad_r(UNI_iPAD_L4, diff - 2);
+                     }
+                  col_info.real_len = COL_INFO.real_len;
                 }
-             col_info.real_len = COL_INFO.real_len;
            }
       }
    else
       {
-        // right-align char items
+        // char items are right aligned
+        //
         size_t LEN = COL_INFO.total_len();
         size_t len = col_info.total_len();
         if (LEN > len)
@@ -1174,6 +1206,9 @@ PrintBuffer::align_dot(ColInfo & COL_INFO)
 void
 PrintBuffer::align_j(ColInfo & COL_INFO)
 {
+   // align all items in this PrintBuffer (= one APL output column)
+   // at the complex J.
+   //
    Log(LOG_printbuf_align)
       {
         CERR << "before align_j(), COL_INFO = " << COL_INFO.int_len
@@ -1240,11 +1275,11 @@ UCS_string new_buf(buffer[0], 0, col_info.int_len);
       //
       loop(f, col_info.fract_len)
           new_buf.append(buffer[0][col_info.int_len + f]);
-      if (!want_expo)
+      if (!want_expo)                     // no exponent, e.g. 1,0
          {
            loop(d, diff)   new_buf.append(UNI_iPAD_L4);
          }
-      else if (col_info.fract_len == 0)   // no fractional part yet
+      else if (col_info.fract_len == 0)   // no fractional part (yet), e.g. 1E2
          {
            new_buf.append(UNI_ASCII_FULLSTOP);
            loop(d, diff - 1)   new_buf.append(UNI_ASCII_0);
@@ -1299,7 +1334,7 @@ PrintBuffer::align_left(ColInfo & COL_INFO)
 const size_t diff = COL_INFO.int_len - col_info.int_len;
 
    if (buffer.size())   pad_r(UNI_iPAD_L3, diff);
-   else                 buffer.append(UCS_string(diff, UNI_iPAD_L3));
+   else                 buffer.push_back(UCS_string(diff, UNI_iPAD_L3));
 
    col_info.int_len = COL_INFO.int_len;
 
@@ -1322,30 +1357,41 @@ PrintBuffer::is_rectangular() const
 }
 //-----------------------------------------------------------------------------
 void
-ColInfo::consider(ColInfo & other)
+ColInfo::consider(const ColInfo & item)
 {
-   flags |= other.flags;
-   if (other.imag_len)   flags |= has_j;
+   // this is the collective ColInfo of an entire column, and item
+   // is a new, not yet considered item in the columns.
+   //
+   flags |= item.flags;
+   if (item.imag_len)   flags |= has_j;
 
-const int expo_len = real_len - fract_len - int_len;
-const int expo_len1 = other.real_len - other.fract_len - other.int_len;
-
-   if (int_len   < other.int_len)
+   if (int_len < item.int_len)
       {
-        real_len += other.int_len - int_len;
-           int_len = other.int_len;
+        real_len += item.int_len - int_len;
+        int_len = item.int_len;
       }
 
-   if (fract_len < other.fract_len)
+   if (item.denom_len)
       {
-        real_len += other.fract_len - fract_len;
-        fract_len = other.fract_len;
+        if (denom_len < item.denom_len)   denom_len = item.denom_len;
+        if (real_len < int_len + denom_len)   real_len = int_len + denom_len;
+      }
+   else
+      {
+        const int EXPO_LEN = real_len - fract_len - int_len;
+        const int expo_len = item.real_len - item.fract_len - item.int_len;
+
+        if (fract_len < item.fract_len)
+           {
+             fract_len = item.fract_len;
+             if (fract_len + expo_len > denom_len)
+             real_len = int_len + fract_len + EXPO_LEN;
+           }
+
+        if (EXPO_LEN < expo_len)    real_len += expo_len - EXPO_LEN;
       }
 
-   real_len = int_len + fract_len + expo_len;
-   if (expo_len < expo_len1)    real_len += expo_len1 - expo_len;
-
-   if (imag_len  < other.imag_len)    imag_len  = other.imag_len;
+   if (imag_len  < item.imag_len)    imag_len  = item.imag_len;
 }
 //-----------------------------------------------------------------------------
 ostream &
