@@ -19,8 +19,8 @@
 */
 
 #include <unistd.h>
+#include <vector>
 
-#include "ArrayIterator.hh"
 #include "Avec.hh"
 #include "Bif_F12_FORMAT.hh"
 #include "CDR.hh"
@@ -36,11 +36,15 @@
 #include "PrintOperator.hh"
 #include "QuadFunction.hh"
 #include "Quad_FX.hh"
+#include "Quad_FFT.hh"
+#include "Quad_GTK.hh"
+#include "Quad_PLOT.hh"
+#include "Quad_RE.hh"
 #include "Quad_SQL.hh"
 #include "Quad_TF.hh"
 #include "Tokenizer.hh"
 #include "UserFunction.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Workspace.hh"
 
 extern char **environ;
@@ -100,15 +104,15 @@ Value_P Z(B->get_shape(), LOC);
             {
               const Unicode uni = cell_B.get_char_value();
               int32_t pos = Avec::find_av_pos(uni);
-              if (pos < 0)   new (Z->next_ravel()) IntCell(MAX_AV);
-              else           new (Z->next_ravel()) IntCell(pos);
+              if (pos < 0)   Z->next_ravel_Int(MAX_AV);
+              else           Z->next_ravel_Int(pos);
               continue;
             }
 
          if (cell_B.is_integer_cell())
             {
               const APL_Integer idx = cell_B.get_near_int();
-              new (Z->next_ravel()) CharCell(Quad_AV::indexed_at(idx));
+              Z->next_ravel_Char(Quad_AV::indexed_at(idx));
               continue;
             }
 
@@ -154,7 +158,7 @@ Value_P Z(shape_Z, LOC);
            }
 
         NamedObject * obj = Workspace::lookup_existing_name(symbol_name);
-        if (obj == 0)   throw_symbol_error(symbol_name, LOC);
+        if (obj == 0)   Error::throw_symbol_error(symbol_name, LOC);
 
         const Function * function = obj->get_function();
         if (function)             // user defined or system function.
@@ -231,7 +235,7 @@ const APL_time_us end = start + 1000000 * B->get_ravel(0).get_real_value();
    // return time elapsed.
    //
 Value_P Z(LOC);
-   new (Z->next_ravel()) FloatCell(0.000001*(now() - start));
+   Z->next_ravel_Float(0.000001*(now() - start));
 
    Z->check_value(LOC);
    return Token(TOK_APL_VALUE1, Z);
@@ -252,7 +256,7 @@ Quad_EA::eval_AB(Value_P A, Value_P B)
         else                     DOMAIN_ERROR;
       }
 
-   return Macro::Z__A_Quad_EA_B->eval_AB(A, B);
+   return Macro::get_macro(Macro::MAC_Z__A_Quad_EA_B)->eval_AB(A, B);
 }
 //=============================================================================
 Token
@@ -270,7 +274,7 @@ Quad_EB::eval_AB(Value_P A, Value_P B)
         else                     DOMAIN_ERROR;
       }
 
-   return Macro::Z__A_Quad_EB_B->eval_AB(A, B);
+   return Macro::get_macro(Macro::MAC_Z__A_Quad_EB_B)->eval_AB(A, B);
 }
 //=============================================================================
 Token
@@ -279,22 +283,27 @@ Quad_EC::eval_B(Value_P B)
 const UCS_string statement_B(*B.get());
 
 ExecuteList * fun = 0;
-   try { fun = ExecuteList::fix(statement_B, LOC); }    catch (...) {}
+   try {
+         fun = ExecuteList::fix(statement_B, LOC);
+       }
+     catch (...) {}
 
    if (fun == 0)
       {
         // syntax error in B
         //
         Value_P Z2(2, LOC);
-            new (Z2->next_ravel())  IntCell(Error::error_major(E_SYNTAX_ERROR));
-            new (Z2->next_ravel())  IntCell(Error::error_minor(E_SYNTAX_ERROR));
-         Z2->check_value(LOC);
+            Z2->next_ravel_Int(Error::error_major(E_SYNTAX_ERROR));
+            Z2->next_ravel_Int(Error::error_minor(E_SYNTAX_ERROR));
+            Z2->check_value(LOC);
 
-        Value_P Z3(Error::error_name(E_SYNTAX_ERROR), LOC);
+        UTF8_string Z3_utf(Error::error_name(E_SYNTAX_ERROR));
+        UCS_string Z3_ucs(Z3_utf);
+        Value_P Z3(Z3_ucs, LOC);
         Value_P Z(3, LOC);
-        new (Z->next_ravel()) IntCell(0);        // return code = error
-        new (Z->next_ravel()) PointerCell(Z2, Z.getref());   // ⎕ET value
-        new (Z->next_ravel()) PointerCell(Z3, Z.getref());   // ⎕EM
+        Z->next_ravel_Int(0);              // return code = error
+        Z->next_ravel_Pointer(Z2.get());   // ⎕ET value
+        Z->next_ravel_Pointer(Z3.get());   // ⎕EM
 
         Z->check_value(LOC);
         return Token(TOK_APL_VALUE1, Z);
@@ -314,18 +323,18 @@ Token
 Quad_EC::eval_fill_B(Value_P B)
 {
 Value_P Z2(2, LOC);
-   new(Z2->next_ravel())   IntCell(0);
-   new(Z2->next_ravel())   IntCell(0);
+   Z2->next_ravel_Int(0);
+   Z2->next_ravel_Int(0);
    Z2->check_value(LOC);
 
 Value_P Zsub(3, LOC);
-   new (Zsub->next_ravel())   IntCell(3);   // statement without result
-   new (Zsub->next_ravel())   PointerCell(Z2, Zsub.getref());
-   new (Zsub->next_ravel())   PointerCell(Idx0(LOC), Zsub.getref());
+   Zsub->next_ravel_Int(3);   // statement without result
+   Zsub->next_ravel_Pointer(Z2.get());
+   Zsub->next_ravel_Pointer(Idx0(LOC).get());
    Zsub->check_value(LOC);
 
-Value_P Z((ShapeItem)0, LOC);
-  new (&Z->get_ravel(0))   PointerCell(Zsub, Z.getref());
+Value_P Z(ShapeItem(0), LOC);
+  new (&Z->get_ravel(0))   PointerCell(Zsub.get(), Z.getref());
    Z->check_value(LOC);
    return Token(TOK_APL_VALUE1, Z);
 }
@@ -333,36 +342,51 @@ Value_P Z((ShapeItem)0, LOC);
 void
 Quad_EC::eoc(Token & result)
 {
-   // set result to an APL value Z = (Z1 Z2 Z3) where:
+   // set Token result to an APL value Z = (Z1 Z2 Z3) where:
+   //
+   // Z1 is an integer scalar 0-5 (return code):
+   //
+   //    0: Error
+   //    1: Value
+   //    2: Committed Value
+   //    3: Void
+   //    4: Branch
+   //    5: Escape
    //
-   // Z1 is an integer scalar 0-5 (return code),
    // Z2 is a two-element integer vector with the value that ⎕ET would have,
-   // Z3 is some value
+   // ⎕ET is not set while in ⎕EC.
    //
+   // Z3 is a value that depends on Z1:
+   //
+   // 0:      ⎕EM
+   // 1 or 2: the value
+   // 3 or 5: 0⍴0
+   // 4:      the branch line
+
 Value_P Z(3, LOC);
    if (result.get_tag() == TOK_ERROR)
       {
         StateIndicator * si = Workspace::SI_top();
         si->clear_safe_execution();
 
-        const Error & err = si->get_error();
+        const Error & err = StateIndicator::get_error(si);
         const ErrorCode ec = ErrorCode(result.get_int_val());
 
         PrintBuffer pb;
         pb.append_ucs(Error::error_name(ec));
-        pb.append_ucs(err.get_error_line_2());
+        pb.append_ucs(UCS_string(UTF8_string(err.get_error_line_2())));
         pb.append_ucs(err.get_error_line_3());
 
         Value_P Z2(2, LOC);
-            new (Z2->next_ravel()) IntCell(Error::error_major(ec));
-            new (Z2->next_ravel()) IntCell(Error::error_minor(ec));
+            Z2->next_ravel_Int(Error::error_major(ec));
+            Z2->next_ravel_Int(Error::error_minor(ec));
         Z2->check_value(LOC);
 
         Value_P Z3(pb, LOC);   // 3 line message like ⎕EM
 
-        new (Z->next_ravel()) IntCell(0);
-        new (Z->next_ravel()) PointerCell(Z2, Z.getref());
-        new (Z->next_ravel()) PointerCell(Z3, Z.getref());
+        Z->next_ravel_Int(0);
+        Z->next_ravel_Pointer(Z2.get());
+        Z->next_ravel_Pointer(Z3.get());
 
         Z->check_value(LOC);
         result.move_2(Token(TOK_APL_VALUE1, Z), LOC);
@@ -372,42 +396,44 @@ Value_P Z(3, LOC);
    // all other cases have Z2 = 0 0
    //
 Value_P Z2(2, LOC);
-   new (Z2->next_ravel()) IntCell(0);
-   new (Z2->next_ravel()) IntCell(0);
-        Z2->check_value(LOC);
+   Z2->next_ravel_Int(0);
+   Z2->next_ravel_Int(0);
+   Z2->check_value(LOC);
 
    switch(result.get_tag())
       {
         case TOK_APL_VALUE1:
         case TOK_APL_VALUE3:
-             new (Z->next_ravel()) IntCell(1);
-             new (Z->next_ravel()) PointerCell(Z2, Z.getref());
-             new (Z->next_ravel()) PointerCell(result.get_apl_val(),Z.getref());
+             Z->next_ravel_Int(1);
+             Z->next_ravel_Pointer(Z2.get());
+             Z->next_ravel()->init_from_value(result.get_apl_val().get(),
+                                              Z.getref(), LOC);
              break;
 
         case TOK_APL_VALUE2:
-             new (Z->next_ravel()) IntCell(2);
-             new (Z->next_ravel()) PointerCell(Z2, Z.getref());
-             new (Z->next_ravel()) PointerCell(result.get_apl_val(),Z.getref());
+             Z->next_ravel_Int(2);
+             Z->next_ravel_Pointer(Z2.get());
+             Z->next_ravel()->init_from_value(result.get_apl_val().get(),
+                                              Z.getref(), LOC);
              break;
 
         case TOK_NO_VALUE:
         case TOK_VOID:
-             new (Z->next_ravel()) IntCell(3);
-             new (Z->next_ravel()) PointerCell(Z2, Z.getref());
-             new (Z->next_ravel()) PointerCell(Idx0_0(LOC), Z.getref());
+             Z->next_ravel_Int(3);
+             Z->next_ravel_Pointer(Z2.get());
+             Z->next_ravel_Pointer(Idx0_0(LOC).get());
              break;
 
         case TOK_BRANCH:
-             new (Z->next_ravel()) IntCell(4);
-             new (Z->next_ravel()) PointerCell(Z2, Z.getref());
-             new (Z->next_ravel()) IntCell(result.get_int_val());
+             Z->next_ravel_Int(4);
+             Z->next_ravel_Pointer(Z2.get());
+             Z->next_ravel_Int(result.get_int_val());
              break;
 
         case TOK_ESCAPE:
-             new (Z->next_ravel()) IntCell(5);
-             new (Z->next_ravel()) PointerCell(Z2, Z.getref());
-             new (Z->next_ravel()) PointerCell(Idx0_0(LOC), Z.getref());
+             Z->next_ravel_Int(5);
+             Z->next_ravel_Pointer(Z2.get());
+             Z->next_ravel_Pointer(Idx0_0(LOC).get());
              break;
 
         default: CERR << "unexpected result tag " << result.get_tag()
@@ -426,7 +452,7 @@ Quad_ENV::eval_B(Value_P B)
 
 const ShapeItem ec_B = B->element_count();
 
-Simple_string<const char *, false> evars;
+std::vector<const char *> evars;
 
    for (char **e = environ; *e; ++e)
        {
@@ -444,7 +470,7 @@ Simple_string<const char *, false> evars;
                  }
             }
 
-         if (match)   evars.append(env);
+         if (match)   evars.push_back(env);
        }
 
 const Shape sh_Z(evars.size(), 2);
@@ -463,13 +489,13 @@ Value_P Z(sh_Z, LOC);
 
         Value_P varname(ucs, LOC);
 
-        ucs.shrink(0);
+        ucs.clear();
         while (*env)   ucs.append(Unicode(*env++));
 
         Value_P varval(ucs, LOC);
 
-        new (Z->next_ravel()) PointerCell(varname, Z.getref());
-        new (Z->next_ravel()) PointerCell(varval, Z.getref());
+        Z->next_ravel_Pointer(varname.get());
+        Z->next_ravel_Pointer(varval.get());
       }
 
    Z->set_default_Spc();
@@ -483,7 +509,7 @@ Quad_ES::eval_AB(Value_P A, Value_P B)
 const UCS_string ucs(*A.get());
 Error error(E_NO_ERROR, LOC);
 const Token ret = event_simulate(&ucs, B, error);
-   if (error.error_code == E_NO_ERROR)   return ret;
+   if (error.get_error_code() == E_NO_ERROR)   return ret;
 
    throw error;
 }
@@ -493,7 +519,7 @@ Quad_ES::eval_B(Value_P B)
 {
 Error error(E_NO_ERROR, LOC);
 const Token ret = event_simulate(0, B, error);
-   if (error.error_code == E_NO_ERROR)   return ret;
+   if (error.get_error_code() == E_NO_ERROR)              return ret;
    if (Workspace::SI_top()->get_safe_execution())   return ret;
 
    throw error;
@@ -507,17 +533,20 @@ Quad_ES::event_simulate(const UCS_string * A, Value_P B, Error & error)
    if (B->element_count() == 0)   return Token();
 
 const ErrorCode ec = get_error_code(B);
-   if (ec == E_QUAD_EA_EXEC)   return Token(TOK_EA_EXEC, B->clone(LOC));
+   if (ec == E_QUAD_ES_BRA)   return Token(TOK_QUAD_ES_BRA, B->clone(LOC));
+   if (ec == E_QUAD_ES_COM)   return Token(TOK_QUAD_ES_COM, B->clone(LOC));
+   if (ec == E_QUAD_ES_ERR)   return Token(TOK_QUAD_ES_ERR, B->clone(LOC));
+   if (ec == E_QUAD_ES_ESC)   return Token(TOK_QUAD_ES_ESC, B->clone(LOC));
 
-   error.init(ec, error.throw_loc);
+   new (&error)   Error(ec, error.get_throw_loc());
 
-   if (error.error_code == E_NO_ERROR)   // B = 0 0: reset ⎕ET and ⎕EM.
+   if (error.get_error_code() == E_NO_ERROR)   // B = 0 0: reset ⎕ET and ⎕EM.
       {
         Workspace::clear_error(LOC);
         return Token();
       }
 
-   if (error.error_code == E_ASSERTION_FAILED)   // B = 0 ASSERTION_FAILED
+   if (error.get_error_code() == E_ASSERTION_FAILED)   // B = 0 ASSERTION_FAILED
       {
         Assert(0 && "simulated ASSERTION_FAILED in ⎕ES");
       }
@@ -528,11 +557,15 @@ const ErrorCode ec = get_error_code(B);
    //
    if (A)                                 // A ⎕ES B
       {
-        error.error_message_1 = *A;
+        UCS_string msg1_ucs(*A);
+        UTF8_string msg1_utf(msg1_ucs);
+        error.set_error_line_1(msg1_utf.c_str());
       }
-   else if (error.error_code == E_USER_DEFINED_ERROR)   // ⎕ES with character B
+   else if (error.get_error_code() == E_USER_DEFINED_ERROR)   // ⎕ES with character B
       {
-        error.error_message_1 = UCS_string(*B.get());
+        UCS_string msg1_ucs(*B.get());
+        UTF8_string msg1_utf(msg1_ucs);
+        error.set_error_line_1(msg1_utf.c_str());
       }
    else if (error.is_known())             //  ⎕ES B with known major/minor B
       {
@@ -540,10 +573,13 @@ const ErrorCode ec = get_error_code(B);
       }
    else                                   //  ⎕ES B with unknown major/minor B
       {
-        error.error_message_1.shrink(0);
+        char cc[58];
+        snprintf(cc, sizeof(cc), "Unkown error (major %d, minor %d) in ⎕ES B",
+                 error.get_error_code() >> 16, error.get_error_code() & 0xFFFF);
+        error.set_error_line_1(cc);
       }
 
-   error.show_locked = true;
+   error.set_show_locked(true);
 
    Assert(Workspace::SI_top());
    if (StateIndicator * si = Workspace::SI_top()->get_parent())
@@ -555,18 +591,17 @@ const ErrorCode ec = get_error_code(B);
              // and B is not empty, the event action is generated as though
              // the function were primitive.
              //
-             error.error_message_2 = UCS_string(6, UNI_ASCII_SPACE);
-             error.error_message_2.append(ufun->get_name());
-             error.left_caret = 6;
-             error.right_caret = -1;
+             UCS_string ufun_name("      ");
+             ufun_name.append(ufun->get_name());
+             error.set_error_line_2(ufun_name, 6, -1);
              Workspace::pop_SI(LOC);
-             Workspace::SI_top()->get_error() = error;
+             StateIndicator::get_error(Workspace::SI_top()) = error;
              error.print_em(UERR, LOC);
              return Token();
            }
       }
 
-   Workspace::SI_top()->update_error_info(error);
+   error.update_error_info(Workspace::SI_top());
    return Token();
 }
 //-----------------------------------------------------------------------------
@@ -581,19 +616,31 @@ Quad_ES::get_error_code(Value_P B)
    //
    // Otherwise, throw an error
 
+   // We extend the standard error codes by a markers E_QUAD_ES_* (10-13) which
+   // are only used by the Macros Z__A_Quad_EA_B and Z__A_Quad_EB_B in order to
+   // return branches and the like as a result.
+
    if (B->get_rank() > 1)   RANK_ERROR;
 
    if (B->element_count() == 0)   return E_NO_ERROR;
    if (B->is_char_string())       return E_USER_DEFINED_ERROR;
 
-const APL_Integer major = B->get_ravel(0).get_near_int();
-   if (major == (E_QUAD_EA_EXEC >> 16) && B->element_count() == 4)
-      return E_QUAD_EA_EXEC;
+const APL_Integer err = (B->get_ravel(0).get_near_int() << 16)
+                      | (B->get_ravel(1).get_near_int());
+
+   if ((err >> 16) == (E_QUAD_ES_BRA >> 16))   // one of the ⎕EA or ⎕EB events
+      {
+        const ShapeItem len_B = B->element_count();
+        if (err == E_QUAD_ES_COM && len_B == 3)   return E_QUAD_ES_COM;
+        if (err == E_QUAD_ES_ERR && len_B == 5)   return E_QUAD_ES_ERR;
+        if (err == E_QUAD_ES_BRA && len_B == 3)   return E_QUAD_ES_BRA;
+        if (err == E_QUAD_ES_ESC && len_B == 2)   return E_QUAD_ES_ESC;
+        DOMAIN_ERROR;
+      }
 
    if (B->element_count() != 2)   LENGTH_ERROR;
-const APL_Integer minor = B->get_ravel(1).get_near_int();
 
-   return ErrorCode(major << 16 | (minor & 0xFFFF));
+   return ErrorCode(err);
 }
 //=============================================================================
 Token
@@ -608,7 +655,7 @@ Shape sh_Z;
    if (var_count > 1)   sh_Z.add_shape_item(var_count);
 Value_P Z(sh_Z, LOC);
 
-   loop(z, var_count)   new (Z->next_ravel()) IntCell(expunge(vars[z]));
+   loop(z, var_count)   Z->next_ravel_Int(expunge(vars[z]));
 
    Z->check_value(LOC);
    return Token(TOK_APL_VALUE1, Z);
@@ -650,9 +697,9 @@ UCS_string e2;
    esc1 = e1;
    esc2 = e2;
 
-   prefixes.shrink(0);
-   escapes.shrink(0);
-   suffixes.shrink(0);
+   prefixes.clear();
+   escapes.clear();
+   suffixes.clear();
 
    read_strings();    // read lines from file or stdin
    split_strings();   // split lines into prefixes, escapes, and suffixes
@@ -661,24 +708,23 @@ const ShapeItem line_count = raw_lines.size();
 Value_P BB(line_count, LOC);
    loop(l, line_count)
        {
-         Cell * cBB = BB->next_ravel();
          if (escapes[l].size() == 0)   // prefix only
             {
               Value_P c1(prefixes[l], LOC);
               Value_P row(1, LOC);
-              new (row->next_ravel())   PointerCell(c1, row.getref());
+              row->next_ravel_Pointer(c1.get());
               row->check_value(LOC);
-              new (cBB)   PointerCell(row, BB.getref());
+              BB->next_ravel_Pointer(row.get());
             }
          else if (suffixes[l].size() == 0)   // prefix and escape
             {
               Value_P c1(prefixes[l], LOC);
               Value_P c2(escapes [l], LOC);
               Value_P row(2, LOC);
-              new (row->next_ravel())   PointerCell(c1, row.getref());
-              new (row->next_ravel())   PointerCell(c2, row.getref());
+              row->next_ravel_Pointer(c1.get());
+              row->next_ravel_Pointer(c2.get());
               row->check_value(LOC);
-              new (cBB)   PointerCell(row, BB.getref());
+              BB->next_ravel_Pointer(row.get());
             }
          else                                // prefix, escape, and suffix
             {
@@ -686,16 +732,16 @@ Value_P BB(line_count, LOC);
               Value_P c2(escapes [l], LOC);
               Value_P c3(suffixes[l], LOC);
               Value_P row(3, LOC);
-              new (row->next_ravel())   PointerCell(c1, row.getref());
-              new (row->next_ravel())   PointerCell(c2, row.getref());
-              new (row->next_ravel())   PointerCell(c3, row.getref());
+              row->next_ravel_Pointer(c1.get());
+              row->next_ravel_Pointer(c2.get());
+              row->next_ravel_Pointer(c3.get());
               row->check_value(LOC);
-              new (cBB)   PointerCell(row, BB.getref());
+              BB->next_ravel_Pointer(row.get());
             }
        }
    BB->check_value(LOC);
 
-Token ret = Macro::Z__Quad_INP_B->eval_B(BB);
+Token ret = Macro::get_macro(Macro::MAC_Z__Quad_INP_B)->eval_B(BB);
    Assert1(ret.get_tag() == TOK_SI_PUSHED);
 
    loop(l, line_count)   BB->get_ravel(l).release(LOC);
@@ -727,7 +773,7 @@ Value_P Z(raw_lines.size(), LOC);
    loop(l, raw_lines.size())
        {
          Value_P ZZ(raw_lines[l], LOC);
-         new (Z->next_ravel())   PointerCell(ZZ, Z.getref());
+         Z->next_ravel_Pointer(ZZ.get());
        }
 
    Quad_INP_running = false;
@@ -797,7 +843,7 @@ Parser parser(PM_EXECUTE, LOC, false);
                  }
 
             }
-         lines.append(line);
+         lines.push_back(line);
       }
 
 Value_P Z(lines.size(), LOC);
@@ -809,18 +855,18 @@ Value_P Z(lines.size(), LOC);
          Value_P ZZ(val_count, LOC);
          loop(v, val_count)
             {
-              new (ZZ->next_ravel()) PointerCell(tos[2*v].get_apl_val(),
-                                                 ZZ.getref());
+              Value_P val = tos[2*v].get_apl_val();
+              ZZ->next_ravel()->init_from_value(val.get(), ZZ.getref(), LOC);
             }
 
          ZZ->check_value(LOC);
-         new (Z->next_ravel()) PointerCell(ZZ, Z.getref());
+         Z->next_ravel_Pointer(ZZ.get());
       }
 
    if (lines.size() == 0)   // empty result
       {
         Value_P ZZ(UCS_string(), LOC);
-        new(&Z->get_ravel(0)) PointerCell(ZZ, Z.getref());
+        new(&Z->get_ravel(0)) PointerCell(ZZ.get(), Z.getref());
       }
 
    Z->check_value(LOC);
@@ -869,7 +915,7 @@ Quad_INP::read_strings()
 {
    // read lines until an end-maker is detected
    //
-   raw_lines.shrink(0);
+   raw_lines.clear();
    for (;;)
       {
         bool eof = false;
@@ -881,13 +927,13 @@ Quad_INP::read_strings()
         const int end = line.substr_pos(end_marker);
         if (end != -1)   // end marker found
            {
-             line.shrink(end);
-             if (line.size())   raw_lines.append(line);
+             line.resize(end);
+             if (line.size())   raw_lines.push_back(line);
              break;
            }
 
         if (eof && !line.size())   break;
-        raw_lines.append(line);
+        raw_lines.push_back(line);
         if (eof)   break;
       }
 }
@@ -902,17 +948,17 @@ UCS_string empty;
         const ShapeItem epos = line.substr_pos(esc1);
         if (esc1.size() == 0 || epos == -1)   // no escape in this line
            {
-             prefixes.append(line);
-             escapes.append(empty);
-             suffixes.append(empty);
+             prefixes.push_back(line);
+             escapes.push_back(empty);
+             suffixes.push_back(empty);
              continue;
            }
 
         // at this point line did contain an exec string
         //
         UCS_string pref = line;
-        pref.shrink(epos);
-        prefixes.append(pref);
+        pref.resize(epos);
+        prefixes.push_back(pref);
 
         line = line.drop(epos + esc1.size());   // skip prefix and esc1
         if (esc2.size())   // end defined
@@ -920,24 +966,24 @@ UCS_string empty;
              const ShapeItem eend = line.substr_pos(esc2);
              if (eend == -1)   // no exec end in this line
                 {
-                  escapes.append(line);
-                  suffixes.append(empty);
+                  escapes.push_back(line);
+                  suffixes.push_back(empty);
                   continue;
                 }
              else              // found an exec end in this line
                 {
                   UCS_string exec = line;
-                  exec.shrink(eend);
-                  escapes.append(exec);
+                  exec.resize(eend);
+                  escapes.push_back(exec);
                   line = line.drop(eend + esc2.size());   // skip exec and esc2
-                  suffixes.append(line);
+                  suffixes.push_back(line);
                   continue;
                 }
            }
         else               // no end defined
            {
-             escapes.append(line);
-             suffixes.append(empty);
+             escapes.push_back(line);
+             suffixes.push_back(empty);
            }
       }
 
@@ -958,7 +1004,7 @@ Shape sh_Z;
    if (var_count > 1)   sh_Z.add_shape_item(var_count);
 Value_P Z(sh_Z, LOC);
 
-   loop(v, var_count)   new (Z->next_ravel())   IntCell(get_NC(vars[v]));
+   loop(v, var_count)   Z->next_ravel_Int(get_NC(vars[v]));
 
    Z->check_value(LOC);
    return Token(TOK_APL_VALUE1, Z);
@@ -1049,7 +1095,7 @@ int requested_NCs = 0;
    //
 UCS_string_vector names;
    {
-     Simple_string<const Symbol *, false> symbols = Workspace::get_all_symbols();
+     std::vector<const Symbol *> symbols = Workspace::get_all_symbols();
 
      loop(s, symbols.size())
         {
@@ -1067,7 +1113,7 @@ UCS_string_vector names;
                if (!first_chars.contains(first_char))   continue;
              }
 
-          names.append(symbol->get_name());
+          names.push_back(symbol->get_name());
         }
    }
 
@@ -1078,15 +1124,15 @@ UCS_string_vector names;
 #define ro_sv_def(x, _str, _txt)                                   \
    { Symbol * symbol = &Workspace::get_v_ ## x();           \
      if ((requested_NCs & 1 << 5) && symbol->get_nc() != 0) \
-        names.append(symbol->get_name()); }
+        names.push_back(symbol->get_name()); }
 
 #define rw_sv_def(x, _str, _txt)                                   \
    { Symbol * symbol = &Workspace::get_v_ ## x();           \
      if ((requested_NCs & 1 << 5) && symbol->get_nc() != 0) \
-        names.append(symbol->get_name()); }
+        names.push_back(symbol->get_name()); }
 
 #define sf_def(x, _str, _txt)                                      \
-   { if (requested_NCs & 1 << 6)   names.append((x::fun->get_name())); }
+   { if (requested_NCs & 1 << 6)   names.push_back((x::fun->get_name())); }
 #include "SystemVariable.def"
       }
 
@@ -1121,8 +1167,7 @@ Value_P Z(shZ, LOC);
         loop(l, longest)
            {
              const UCS_string & ucs = names[smallest];
-             new (Z->next_ravel())
-                 CharCell(l < ucs.size() ? ucs[l] : UNI_ASCII_SPACE);
+             Z->next_ravel_Char(l < ucs.size() ? ucs[l] : UNI_ASCII_SPACE);
            }
 
         // remove smalles from table
@@ -1178,7 +1223,7 @@ const APL_Integer b = B->get_ravel(0).get_near_int();
                  break;
 
         case 2:  Z = Value_P(LOC);
-                new (Z->next_ravel()) IntCell(fun_line);
+                Z->next_ravel_Int(fun_line);
                 break;
 
         case 3:  {
@@ -1190,24 +1235,26 @@ const APL_Integer b = B->get_ravel(0).get_near_int();
                  }
                  break;
 
-        case 4:  if (si->get_error().error_code)
-                    {
-                      const UCS_string & text =
-                                        si->get_error().get_error_line_2();
-                      Z = Value_P(text, LOC);
-                    }
-                 else
+        case 4: if (StateIndicator::get_error(si).get_error_code())
+                   {
+                     const UCS_string text(UTF8_string(
+                           StateIndicator::get_error(si).get_error_line_2()));
+                     Z = Value_P(text, LOC);
+                   }
+                else
                     {
                       const UCS_string text = exec->statement_text(PC);
                       Z = Value_P(text, LOC);
                     }
-                 break;
+                break;
 
         case 5: Z = Value_P(LOC);
-                new (Z->next_ravel()) IntCell(PC);                      break;
+                Z->next_ravel_Int(PC);
+                break;
 
         case 6: Z = Value_P(LOC);
-                new (Z->next_ravel()) IntCell(pm);                      break;
+                Z->next_ravel_Int(pm);
+                break;
 
         default: DOMAIN_ERROR;
       }
@@ -1250,7 +1297,7 @@ ShapeItem z = 0;
          switch (b)
            {
              case 1:  new (cZ) PointerCell(
-                                Value_P(fun_name, LOC), Z.getref());
+                                Value_P(fun_name, LOC).get(), Z.getref());
                       break;
 
              case 2:  new (cZ) IntCell(fun_line);
@@ -1262,20 +1309,21 @@ ShapeItem z = 0;
                         fun_and_line.append_number(fun_line);
                         fun_and_line.append(UNI_ASCII_R_BRACK);
                         new (cZ) PointerCell(Value_P(
-                                    fun_and_line, LOC), Z.getref()); 
+                                    fun_and_line, LOC).get(), Z.getref()); 
                       }
                       break;
 
-             case 4:  if (si->get_error().error_code)
+             case 4:  if (StateIndicator::get_error(si).get_error_code())
                          {
-                           const UCS_string & text =
-                                        si->get_error().get_error_line_2();
-                           new (cZ) PointerCell(Value_P(text, LOC), Z.getref()); 
+                           const UCS_string text(UTF8_string(
+                             StateIndicator::get_error(si).get_error_line_2()));
+                           new (cZ) PointerCell(Value_P(text, LOC).get(),
+                                                Z.getref());
                          }
                       else
                          {
                            const UCS_string text = exec->statement_text(PC);
-                           new (cZ) PointerCell(Value_P( text, LOC),
+                           new (cZ) PointerCell(Value_P( text, LOC).get(),
                                                 Z.getref()); 
                          }
                       break;
@@ -1298,7 +1346,9 @@ const ShapeItem ec = B->element_count();
    if (ec == 0)
       {
         if (B->get_ravel(0).is_character_cell())   // char to Unicode
-        new (&Z->get_ravel(0))   CharCell(UNI_ASCII_SPACE);
+           new (&Z->get_ravel(0))   IntCell(0);
+        else
+           new (&Z->get_ravel(0))   CharCell(UNI_ASCII_SPACE);
       }
 
    loop(v, ec)
@@ -1308,17 +1358,20 @@ const ShapeItem ec = B->element_count();
          if (cell_B.is_character_cell())   // char to Unicode
             {
               const Unicode uni = cell_B.get_char_value();
-              new (Z->next_ravel()) IntCell(uni);
+              Z->next_ravel_Int(uni);
               continue;
             }
 
          if (cell_B.is_integer_cell())
             {
               const APL_Integer bint = cell_B.get_near_int();
-              new (Z->next_ravel())   CharCell(Unicode(bint));
+              Z->next_ravel_Char(Unicode(bint));
               continue;
             }
 
+         MORE_ERROR() << "⎕UCS got unexpected Cell type "
+                      << cell_B.get_classname()
+                      << ". Expecting int or character";
          DOMAIN_ERROR;
        }
 
@@ -1360,12 +1413,11 @@ UserFunction * ufun = fun->get_ufun1();
 }
 //-----------------------------------------------------------------------------
 Token
-Stop_Trace::reference(const Simple_string<Function_Line, false> & lines,
-                      bool assigned)
+Stop_Trace::reference(const std::vector<Function_Line> & lines, bool assigned)
 {
 Value_P Z(lines.size(), LOC);
 
-   loop(z, lines.size())   new (Z->next_ravel()) IntCell(lines[z]);
+   loop(z, lines.size())   Z->next_ravel_Int(lines[z]);
 
    if (assigned)   return Token(TOK_APL_VALUE2, Z);
    else            return Token(TOK_APL_VALUE1, Z);
@@ -1374,14 +1426,14 @@ Value_P Z(lines.size(), LOC);
 void
 Stop_Trace::assign(UserFunction * ufun, const Value & new_value, bool stop)
 {
-Simple_string<Function_Line, false> lines;
+std::vector<Function_Line> lines;
    lines.reserve(new_value.element_count());
 
    loop(l, new_value.element_count())
       {
          APL_Integer line = new_value.get_ravel(l).get_near_int();
          if (line < 1)   continue;
-         lines.append((Function_Line)line);
+         lines.push_back(Function_Line(line));
       }
 
    ufun->set_trace_stop(lines, stop);
