@@ -18,6 +18,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#include "Bif_F12_TAKE_DROP.hh"
 #include "CDR.hh"
 #include "Macro.hh"
 #include "PointerCell.hh"
@@ -29,11 +30,74 @@ Quad_CR    Quad_CR   ::_fun;
 
 Quad_CR    * Quad_CR   ::fun = &Quad_CR   ::_fun;
 
+//-----------------------------------------------------------------------------
+Token
+Quad_CR::list_functions(ostream & out)
+{
+   out <<
+"   Functions provided by A ⎕CR B...\n"
+"\n"
+"   Legend: b - byte vector (vector of integers between -128 and 255)\n"
+"           h - hex string (characters 0-9 or A-F resp. a-f)\n"
+"           l - string of (\\n-terminated) lines\n"
+"           m - character matrix\n"
+"           n - nested vector of strings\n"
+"           i - integer vector\n"
+"           r - base64 string according to RFC 4648\n"
+"           v - T,V (integer tag T and byte vector V)\n"
+"           s - string\n"
+"\n"
+"   Zm ←  0 ⎕CR B     Zm is B in APL output format\n"
+"   Zs ←  1 ⎕CR B     Zs is B in APL input format\n"
+"   Zm ←  2 ⎕CR B     Zm is B boxed using ASCII characters\n"
+"   Zm ←  3 ⎕CR B     Zm is B boxed using line-drawing characters\n"
+"   Zm ←  4 ⎕CR B     3 ⎕CR B + extra frame\n"
+"   Zs ←  5 ⎕CR B     Zs is B in (uppercase) HEX\n"
+"   Zs ←  6 ⎕CR B     Zs is B in (lowercase) hex\n"
+"   Zm ←  7 ⎕CR B     like 3 ⎕CR B with thin lines\n"
+"   Zm ←  8 ⎕CR B     like 4 ⎕CR B with thin lines\n"
+"   Zm ←  9 ⎕CR B     like 4 ⎕CR B with double-line outer frame\n"
+"   Zs ← 10 ⎕CR Bs    Zs is an APL expression producing variable(-name) Bs\n"
+"   Zs ← 11 ⎕CR B     value B → CDR (Common Data Representation) string Zs\n"
+"   Z  ← 12 ⎕CR Bs    CDR string Bs → value Z\n"
+"   Zb ← 13 ⎕CR Bh    hex string Bh → byte vector Z\n"
+"   Zh ← 14 ⎕CR B     Zs is 6 ⎕CR 11 ⎕CR B (Value → CDR → HEX)\n"
+"   Z  ← 15 ⎕CR Bh    Z is 11 ⎕CR 6 ⎕CR B (HEX → CDR → Value)\n"
+"   Zr ← 16 ⎕CR Bs    string Bs → base64 string Z (RFC 4648)\n"
+"   Zs ← 17 ⎕CR Br    base64 string Br → string Zs (RFC 4648)\n"
+"   Zb ← 18 ⎕CR Bs    UCS string Bs → UTF8 encoded byte vector Zb (⍴Z ≥ ⍴B)\n"
+"   Zs ← 19 ⎕CR Bb    UTF8 encoded byte vector Bb → UCS string Zs (⍴Z ≤ ⍴B)\n"
+"   Zm ← 20 ⎕CR B     4 ⎕CR B, but in NARS style (length instead of ↓ or →)\n"
+"   Zm ← 21 ⎕CR B     20 ⎕CR B but using thin lines\n"
+"   Zm ← 22 ⎕CR B     20 ⎕CR B with double-line outer frame\n"
+"   Zm ← 23 ⎕CR B     20 ⎕CR B with thick lines + extra frame\n"
+"   Zm ← 24 ⎕CR B     20 ⎕CR B using thin lines + extra frame\n"
+"   Zm ← 25 ⎕CR B     20 ⎕CR B with double-line extra frame\n"
+"   Zi ← 26 ⎕CR B     Zi is the cell types of corresponding items in B\n"
+"   Zi ← 27 ⎕CR B     Zi is the primary values of Z items as integer\n"
+"   Zi ← 28 ⎕CR B     Zi is the secondary values of Z items as intege \n"
+"   Zm ← 29 ⎕CR B     4 ⎕CR B with strings being quoted + extra frame\n"
+"   Z  ← 30 ⎕CR B     Z is B with all items expanded to the same shape\n"
+"   Zn ← 31 ⎕CR Bn    internal helper function for the ⎕INP macro\n"
+"   Zn ← 32 ⎕CR Bn    internal helper function for the ⎕INP macro\n"
+"   Zb ← 33 ⎕CR Bv    Zb is a TLV with Tag ↑Bv and Value 1↓Bv\n"
+"   Zv ← 34 ⎕CR Bb    TLV Bb to Tag ↑Z and Value 1↓Z\n"
+"   Zn ← 35 ⎕CR Bl    string of lines Bl → nested vector of lines Zn\n"
+"   Zl ← 36 ⎕CR Bn    nested vector of lines Bn → string of lines Bl\n"
+"\n"
+"   if N ⎕CR has an inverse M ⎕CR then -N can be used instead of M\n";
+
+  return Token(TOK_APL_VALUE1, Str0(LOC));
+}
 //-----------------------------------------------------------------------------
 Token
 Quad_CR::eval_B(Value_P B)
 {
 UCS_string symbol_name(*B.get());
+   if (symbol_name.size() == 0)   // ⎕CR '' : print help
+      {
+        return list_functions(COUT);
+      }
 
    // remove trailing whitespaces in B
    //
@@ -53,7 +117,8 @@ const Function * function = 0;
       {
         loop(m, Macro::MAC_COUNT)
             {
-              Macro * macro = Macro::get_macro((Macro::Macro_num)m);
+              const Macro * macro =
+                    Macro::get_macro(Macro::Macro_num(m));
               if (symbol_name == macro->get_name())
                  {
                    function = macro;
@@ -87,7 +152,7 @@ int max_len = 0;
 Shape shape_Z;
    shape_Z.add_shape_item(tlines.size());
    shape_Z.add_shape_item(max_len);
-   
+
 Value_P Z(shape_Z, LOC);
    loop(row, tlines.size())
       {
@@ -136,6 +201,10 @@ Quad_CR::do_CR(APL_Integer a, const Value * B, PrintContext pctx)
         case -17: a = 16;   break;
         case -18: a = 19;   break;
         case -19: a = 18;   break;
+        case -33: a = 34;   break;
+        case -34: a = 33;   break;
+        case -35: a = 36;   break;
+        case -36: a = 35;   break;
         default: MORE_ERROR() << "A ⎕CR B with invalid A < 0";
                  DOMAIN_ERROR;
       }
@@ -187,6 +256,10 @@ bool extra_frame = false;
         case 30: return do_CR30(*B);             // conform B (for ⍤ macro)
         case 31: return do_CR31_32(true, *B);    // ⎕INP helper
         case 32: return do_CR31_32(false, *B);   // ⎕INP helper
+        case 33: return do_CR33(*B);             // TV to TLV byte vector
+        case 34: return do_CR34(*B);             // TLV byte vector to TV
+        case 35: return do_CR35(*B);             // lines to nested strings
+        case 36: return do_CR36(*B);             // nested strings to lines
 
         default: MORE_ERROR() << "A ⎕CR B with invalid A";
                  DOMAIN_ERROR;
@@ -194,15 +267,13 @@ bool extra_frame = false;
 
    // common code for ⎕CR variants that only differ by print style...
    //
-   if (extra_frame)
+   if (extra_frame && !B->is_simple_scalar())
       {
-         // enclose B so that the entire B is boxed
-         //
-         Value_P B1(LOC);
-         new (&B1->get_ravel(0)) PointerCell(B->clone(LOC), B1.getref());
-         B1->check_value(LOC);
-         PrintBuffer pb(*B1, pctx, 0);
-         return Value_P(pb, LOC);
+        Value_P B1(LOC);
+        new (&B1->get_ravel(0)) PointerCell(B->clone(LOC).get(), B1.getref());
+        B1->check_value(LOC);
+        PrintBuffer pb(*B1, pctx, 0);
+        return Value_P(pb, LOC);
       }
    else   // no frame
       {
@@ -229,11 +300,11 @@ Value_P Z(shape_Z, LOC);
 const Cell * cB = &B.get_ravel(0);
    loop(b, B.element_count())
        {
-         const int val = cB++->get_char_value() & 0x00FF;
+         const int val = cB++->get_byte_value() & 0x00FF;
          const int h = alpha[val >> 4];
          const int l = alpha[val & 0x0F];
-         new (Z->next_ravel())   CharCell((Unicode)h);
-         new (Z->next_ravel())   CharCell((Unicode)l);
+         new (Z->next_ravel())   CharCell(Unicode(h));
+         new (Z->next_ravel())   CharCell(Unicode(l));
        }
 
    Z->set_default_Spc();
@@ -254,7 +325,7 @@ Value_P Z(ucs_vec.size(), LOC);
    loop(line, ucs_vec.size())
       {
          Value_P Z_line(ucs_vec[line], LOC);
-         new (Z->next_ravel())   PointerCell(Z_line, Z.getref());
+         new (Z->next_ravel())   PointerCell(Z_line.get(), Z.getref());
       }
 
    Z->set_default_Spc();
@@ -304,18 +375,18 @@ const Symbol * symbol = Workspace::lookup_existing_symbol(symbol_name);
                         }
                     res.append(UNI_ASCII_R_CURLY);
 
-                    result.append(res);
+                    result.push_back(res);
                   }
                else
                   {
                     UCS_string res("∇");
-               
+
                     loop(u, text.size())
                        {
                          if (text[u] == '\n')
                              {
-                               result.append(res);
-                               res.shrink(0);
+                               result.push_back(res);
+                               res.clear();
                                UCS_string next(text, u+1, text.size()-(u+1));
                                if (!next.is_comment_or_label() &&
                                    u < (text.size() - 1))
@@ -328,7 +399,7 @@ const Symbol * symbol = Workspace::lookup_existing_symbol(symbol_name);
                        }
                     res.append(ufun.get_exec_properties()[0]
                                ? UNI_DEL_TILDE : UNI_NABLA);
-                    result.append(res);
+                    result.push_back(res);
                   }
                return;
              }
@@ -351,14 +422,14 @@ Quad_CR::do_CR10_rec(UCS_string_vector & result, const Value & value,
 {
    /*
       recursively encode var_name with value.
-     
+
       we emit one of 2 formats:
-     
+
       short format: dest ← (⍴ value) ⍴ value
-              
+
       long format:  dest ← (⍴ value) ⍴ 0                 " prolog "
                     dest[] ← partial value ...
-              
+
       short format (the default) requires a reasonably short value
 
       If value is nested then the short or long format is followed
@@ -378,13 +449,13 @@ UCS_string shape_rho;
    {
      if (value.get_rank() == 1)   // true vector
         {
-          if      (value.element_count() == 0)   shape_rho.append_utf8("0⍴");
-          else if (value.element_count() == 1)   shape_rho.append_utf8(",");
+          if      (value.element_count() == 0)   shape_rho.append_UTF8("0⍴");
+          else if (value.element_count() == 1)   shape_rho.append_UTF8(",");
         }
      else if (value.get_rank() > 1)   // matrix or higher
         {
            shape_rho.append_shape(value.get_shape());
-           shape_rho.append_utf8("⍴");
+           shape_rho.append_UTF8("⍴");
          }
    }
 
@@ -398,16 +469,16 @@ UCS_string shape_rho;
         picker.get(proto_name);
         picker.pop();
         do_CR10_rec(result, *proto, picker, pidx);
-        result.last().append_utf8(" ⍝ proto 1");
+        result.back().append_UTF8(" ⍝ proto 1");
 
         // and then another line to reshape the prototype
         //
         reshape.append(proto_name);
-        reshape.append_utf8("←");
+        reshape.append_UTF8("←");
         reshape.append(shape_rho);
         reshape.append(proto_name);
-        result.append(reshape);
-        result.last().append_utf8(" ⍝ proto 2");
+        result.push_back(reshape);
+        result.back().append_UTF8(" ⍝ proto 2");
         return;
       }
 
@@ -418,7 +489,7 @@ bool nested = false;
    //
    if (short_ravel(result, nested, value, picker, left, shape_rho))
       {
-        result.append(compute_prolog(picker.get_level(), left, value));
+        result.push_back(compute_prolog(picker.get_level(), left, value));
 
         ShapeItem pos = 0;
         V_mode mode = Vm_NONE;
@@ -432,7 +503,7 @@ bool nested = false;
              //
              UCS_string line(2*(picker.get_level() - 1), UNI_ASCII_SPACE);
              picker.get_indexed(line, pos, count);
-             line.append_utf8("←");
+             line.append_UTF8("←");
 
              // compute next item
              //
@@ -453,8 +524,8 @@ bool nested = false;
                    pos += count;
                    count = 0;
 
-                   result.append(line);
-                   rhs.shrink(0);
+                   result.push_back(line);
+                   rhs.clear();
                    mode = Vm_NONE;
                  }
 
@@ -471,10 +542,10 @@ bool nested = false;
 
      UCS_string line(2*(picker.get_level() - 1), UNI_ASCII_SPACE);
      picker.get_indexed(line, pos, count);
-     line.append_utf8("←");
+     line.append_UTF8("←");
      line.append(rhs);
 
-     result.append(line);
+     result.push_back(line);
    }
 
    // step 2: nested items
@@ -524,7 +595,7 @@ int len = 0;
 UCS_string line(2*(picker.get_level() - 1), UNI_ASCII_SPACE);  // indent
 
    line.append(left);
-   line.append_utf8("←");
+   line.append_UTF8("←");
    line.append(shape_rho);
 
 V_mode mode = Vm_NONE;
@@ -549,7 +620,7 @@ V_mode mode = Vm_NONE;
    //
    close_mode(line, mode);
 
-   result.append(line);
+   result.push_back(line);
    return false;
 }
 //-----------------------------------------------------------------------------
@@ -565,11 +636,11 @@ APL_Integer default_int  = 0;
 const bool default_is_int = figure_default(value, default_char, default_int);
 
    prolog.append(left);
-   prolog.append_utf8("←");
-   if (pick_level > 1)   prolog.append_utf8("⊂");
+   prolog.append_UTF8("←");
+   if (pick_level > 1)   prolog.append_UTF8("⊂");
    prolog.append_shape(value.get_shape());
-   if (default_is_int)   prolog.append_utf8("⍴0 ⍝ prolog ≡");
-   else                  prolog.append_utf8("⍴' ' ⍝ prolog ≡");
+   if (default_is_int)   prolog.append_UTF8("⍴0 ⍝ prolog ≡");
+   else                  prolog.append_UTF8("⍴' ' ⍝ prolog ≡");
    prolog.append_number(pick_level);
 
    return prolog;
@@ -608,29 +679,29 @@ const int level = shapes.size();
         return;
       }
 
-   result.append_utf8("((⎕IO+");
+   result.append_UTF8("((⎕IO+");
    if (level == 2 && shapes[0].get_rank() > 1)   // need ⊂ for left ⊃ arg
       {
          Shape sh = shapes[0].offset_to_index(indices[1]);
-         result.append_utf8("(⊂");
+         result.append_UTF8("(⊂");
          result.append_shape(sh);
-         result.append_utf8(")");
+         result.append_UTF8(")");
       }
    else
       {
         loop(s, shapes.size() - 1)
            {
-             if (s)   result.append_utf8(" ");
+             if (s)   result.append_UTF8(" ");
              Shape sh = shapes[s].offset_to_index(indices[s + 1]);
-             if (sh.get_rank() > 1)   result.append_utf8("(");
+             if (sh.get_rank() > 1)   result.append_UTF8("(");
              result.append_shape(sh);
-		     if (sh.get_rank() > 1)   result.append_utf8(")");
+             if (sh.get_rank() > 1)   result.append_UTF8(")");
            }
       }
 
-   result.append_utf8(")⊃");
+   result.append_UTF8(")⊃");
    result.append(var_name);
-   result.append_utf8(")");
+   result.append_UTF8(")");
 }
 //-----------------------------------------------------------------------------
 void
@@ -642,22 +713,22 @@ const bool need_disclose = get_level() > 1;
 
    // , is only needed if the picked item is not a true vector
    //
-const bool need_comma = shapes.last().get_rank() != 1;
+const bool need_comma = shapes.back().get_rank() != 1;
 
-   result.append_utf8("(");
-   if (need_comma)      result.append_utf8(",");
-   if (need_disclose)   result.append_utf8("⊃");
+   result.append_UTF8("(");
+   if (need_comma)      result.append_UTF8(",");
+   if (need_disclose)   result.append_UTF8("⊃");
    get(result);
-   result.append_utf8(")");
-   result.append_utf8("[");
+   result.append_UTF8(")");
+   result.append_UTF8("[");
    if (pos)
       {
         result.append_number(pos);
-        result.append_utf8("+");
+        result.append_UTF8("+");
       }
-   result.append_utf8("⍳");
+   result.append_UTF8("⍳");
    result.append_number(len);
-   result.append_utf8("]");
+   result.append_UTF8("]");
 }
 //-----------------------------------------------------------------------------
 Quad_CR::V_mode
@@ -705,7 +776,7 @@ Quad_CR::do_CR10_item(UCS_string & item, const Cell & cell, V_mode mode,
         bool scaled = false;
         PrintContext pctx(PR_APL_MIN, MAX_Quad_PP, MAX_Quad_PW);
         item = UCS_string(cell.get_real_value(), scaled, pctx);
-        item.append_utf8("J");
+        item.append_UTF8("J");
         scaled = false;
         item.append(UCS_string(cell.get_imag_value(), scaled, pctx));
         return Vm_NUM;
@@ -716,9 +787,9 @@ Quad_CR::do_CR10_item(UCS_string & item, const Cell & cell, V_mode mode,
         // cell is a pointer cell which will be replaced later.
         // For now we add an item in the current mode as placeholder
         //
-        if (mode == Vm_QUOT)   { item.append_utf8("∘");     return Vm_QUOT; }
-        if (mode == Vm_UCS)    { item.append_utf8("33");    return Vm_UCS;  }
-        item.append_utf8("00");
+        if (mode == Vm_QUOT)   { item.append_UTF8("∘");     return Vm_QUOT; }
+        if (mode == Vm_UCS)    { item.append_UTF8("33");    return Vm_UCS;  }
+        item.append_UTF8("00");
         return Vm_NUM;
       }
 
@@ -734,7 +805,8 @@ CDR_string cdr;
 
 const ShapeItem len = cdr.size();
 Value_P Z(len, LOC);
-   loop(l, len)   new (Z->next_ravel()) CharCell((Unicode)(0xFF & cdr[l]));
+   loop(l, len)
+       new (Z->next_ravel()) CharCell(Unicode(0xFF & cdr[l]));
 
    Z->set_default_Spc();
    Z->check_value(LOC);
@@ -747,7 +819,7 @@ Quad_CR::do_CR12(const Value & B)
    if (B.get_rank() > 1)   RANK_ERROR;
 
 CDR_string cdr;
-   loop(b, B.element_count())   cdr.append(B.get_ravel(b).get_char_value());
+   loop(b, B.element_count())   cdr.push_back(B.get_ravel(b).get_byte_value());
 Value_P Z = CDR::from_CDR(cdr, LOC);
    return Z;
 }
@@ -811,7 +883,7 @@ const Cell * cB = &B.get_ravel(0);
          const int n1 = nibble(cB++->get_char_value());
          const int n2 = nibble(cB++->get_char_value());
          if (n1 < 0 || n2 < 0)   DOMAIN_ERROR;
-         cdr.append(16*n1 + n2);
+         cdr.push_back(16*n1 + n2);
        }
 
    Value_P Z = CDR::from_CDR(cdr, LOC);
@@ -846,10 +918,10 @@ const Cell * cB = &B.get_ravel(0);
         const int z3 = (b2 & 0x0F) << 2 | (b3 & 0xC0) >> 6;
         const int z4 =  b3 & 0x3F;
 
-        new (Z->next_ravel()) CharCell((Unicode)alpha[z1]);
-        new (Z->next_ravel()) CharCell((Unicode)alpha[z2]);
-        new (Z->next_ravel()) CharCell((Unicode)alpha[z3]);
-        new (Z->next_ravel()) CharCell((Unicode)alpha[z4]);
+        new (Z->next_ravel()) CharCell(Unicode(alpha[z1]));
+        new (Z->next_ravel()) CharCell(Unicode(alpha[z2]));
+        new (Z->next_ravel()) CharCell(Unicode(alpha[z3]));
+        new (Z->next_ravel()) CharCell(Unicode(alpha[z4]));
       }
 
    // process final bytes
@@ -866,10 +938,10 @@ const Cell * cB = &B.get_ravel(0);
                   const int z1 = b1 >> 2;
                   const int z2 = (b1 & 0x03) << 4 | (b2 & 0xF0) >> 4;
 
-                  new (Z->next_ravel()) CharCell((Unicode)alpha[z1]);
-                  new (Z->next_ravel()) CharCell((Unicode)alpha[z2]);
-                  new (Z->next_ravel()) CharCell((Unicode)'=');
-                  new (Z->next_ravel()) CharCell((Unicode)'=');
+                  new(Z->next_ravel())CharCell(Unicode(alpha[z1]));
+                  new(Z->next_ravel())CharCell(Unicode(alpha[z2]));
+                  new (Z->next_ravel()) CharCell(Unicode('='));
+                  new (Z->next_ravel()) CharCell(Unicode('='));
                 }
                 break;
 
@@ -883,10 +955,10 @@ const Cell * cB = &B.get_ravel(0);
                   const int z2 = (b1 & 0x03) << 4 | (b2 & 0xF0) >> 4;
                   const int z3 = (b2 & 0x0F) << 2 | (b3 & 0xC0) >> 6;
 
-                  new (Z->next_ravel()) CharCell((Unicode)alpha[z1]);
-                  new (Z->next_ravel()) CharCell((Unicode)alpha[z2]);
-                  new (Z->next_ravel()) CharCell((Unicode)alpha[z3]);
-                  new (Z->next_ravel()) CharCell((Unicode)'=');
+                  new(Z->next_ravel())CharCell(Unicode(alpha[z1]));
+                  new(Z->next_ravel())CharCell(Unicode(alpha[z2]));
+                  new(Z->next_ravel())CharCell(Unicode(alpha[z3]));
+                  new(Z->next_ravel())CharCell(Unicode('='));
                 }
                 break;
       }
@@ -937,18 +1009,18 @@ const Cell * cB = &B.get_ravel(0);
 
          if (q < (quantums - 1) || missing == 0)
             {
-              new (Z->next_ravel())   CharCell((Unicode)z1);
-              new (Z->next_ravel())   CharCell((Unicode)z2);
-              new (Z->next_ravel())   CharCell((Unicode)z3);
+              new (Z->next_ravel())   CharCell(Unicode(z1));
+              new (Z->next_ravel())   CharCell(Unicode(z2));
+              new (Z->next_ravel())   CharCell(Unicode(z3));
             }
          else if (missing == 1)   // k6 k2l4 l20000 =
             {
-              new (Z->next_ravel())   CharCell((Unicode)z1);
-              new (Z->next_ravel())   CharCell((Unicode)z2);
+              new (Z->next_ravel())   CharCell(Unicode(z1));
+              new (Z->next_ravel())   CharCell(Unicode(z2));
             }
          else                     // k6 k20000 = =
             {
-              new (Z->next_ravel())   CharCell((Unicode)z1);
+              new (Z->next_ravel())   CharCell(Unicode(z1));
             }
        }
 
@@ -970,7 +1042,7 @@ Value_P
 Quad_CR::do_CR19(const Value & B)
 {
 UTF8_string utf(B);
-   for (ShapeItem l = 0; l < utf.size();)
+   for (size_t l = 0; l < utf.size();)
        {
          int len = 0;
          const Unicode uni = UTF8_string::toUni(&utf[l], len, false);
@@ -996,7 +1068,7 @@ Value_P Z(len, LOC);
            {
              Value_P B_sub = cB.get_pointer_value();
              Value_P Z_sub = do_CR26(B_sub.getref());
-             new (Z->next_ravel()) PointerCell(Z_sub, Z.getref());
+             new (Z->next_ravel()) PointerCell(Z_sub.get(), Z.getref());
            }
         else
            {
@@ -1021,7 +1093,7 @@ Value_P Z(len, LOC);
             {
               Value_P B_sub = cB.get_pointer_value();
               Value_P Z_sub = do_CR27_28(primary, B_sub.getref());
-              new (cZ) PointerCell(Z_sub, Z.getref());
+              new (cZ) PointerCell(Z_sub.get(), Z.getref());
             }
          else
             {
@@ -1031,7 +1103,7 @@ Value_P Z(len, LOC);
                    if (cB.get_cell_type() == CT_CHAR)
                       data = cB.get_char_value();
                    else if (cB.get_cell_type() == CT_CELLREF)
-                      data = (APL_Integer)cB.get_lval_value();
+                      data = APL_Integer(cB.get_lval_value());
                    else
                       memcpy(&data, cB.get_u0(), sizeof(data));
                  }
@@ -1040,7 +1112,13 @@ Value_P Z(len, LOC);
                    if (cB.get_cell_type() == CT_COMPLEX)
                       memcpy(&data, cB.get_u1(), sizeof(data));
                    else if (cB.get_cell_type() == CT_CELLREF)
-                      data = (APL_Integer)(*(LvalCell *)&cB).get_cell_owner();
+                      data = APL_Integer(cB.cLvalCell().get_cell_owner());
+#ifdef RATIONAL_NUMBERS_WANTED
+                   else if (cB.get_cell_type() == CT_FLOAT)
+                      memcpy(&data, cB.get_u1(), sizeof(data));
+                   else if (cB.get_cell_type() == CT_INT)
+                      data = 1;
+#endif
                  }
 
               new (cZ)   IntCell(data);
@@ -1100,9 +1178,8 @@ Value_P Z(shape_Z, LOC);
              Shape sh_sub = B_sub->get_shape();
              sh_sub.expand_rank(conformed.get_rank());
              B_sub->set_shape(sh_sub);
-                
-             Token T = Bif_F12_TAKE::do_take(conformed, B_sub);
-             Value_P ZZ = T.get_apl_val();
+
+             Value_P ZZ = Bif_F12_TAKE::do_take(conformed, B_sub);
              loop(zz, conformed_len)
                  Z->next_ravel()->init(ZZ->get_ravel(zz), Z.getref(), LOC);
            }
@@ -1135,7 +1212,7 @@ PrintContext pctx = Workspace::get_PrintContext(PR_APL);
         if (row->element_count() == 1)   // single item
            {
              Value_P Zrow = row->get_ravel(0).get_pointer_value()->clone(LOC);
-             new (Z->next_ravel())   PointerCell(Zrow, Z.getref());
+             new (Z->next_ravel())   PointerCell(Zrow.get(), Z.getref());
              continue;
            }
 
@@ -1154,12 +1231,12 @@ PrintContext pctx = Workspace::get_PrintContext(PR_APL);
         if (pb.get_height() == 1)
            {
              Value_P Zrow(pb.l1(), LOC);
-             new (Z->next_ravel())   PointerCell(Zrow, Z.getref());
+             new (Z->next_ravel())   PointerCell(Zrow.get(), Z.getref());
            }
         else
            {
              Value_P Zrow(pb, LOC);
-             new (Z->next_ravel())   PointerCell(Zrow, Z.getref());
+             new (Z->next_ravel())   PointerCell(Zrow.get(), Z.getref());
            }
       }
 
@@ -1200,8 +1277,8 @@ int ascii_len = 0;
 void
 Quad_CR::close_mode(UCS_string & rhs, V_mode mode)
 {
-   if      (mode == Vm_QUOT)   rhs.append_utf8("'");
-   else if (mode == Vm_UCS)    rhs.append_utf8(")");
+   if      (mode == Vm_QUOT)   rhs.append_UTF8("'");
+   else if (mode == Vm_UCS)    rhs.append_UTF8(")");
 }
 //-----------------------------------------------------------------------------
 void
@@ -1209,17 +1286,161 @@ Quad_CR::item_separator(UCS_string & line, V_mode from_mode, V_mode to_mode)
 {
    if (from_mode == to_mode)   // separator in same mode (if any)
       {
-        if      (to_mode == Vm_UCS)    line.append_utf8(" ");
-        else if (to_mode == Vm_NUM)    line.append_utf8(" ");
+        if      (to_mode == Vm_UCS)    line.append_UTF8(" ");
+        else if (to_mode == Vm_NUM)    line.append_UTF8(" ");
       }
    else                // close old mode and open new one
       {
         close_mode(line, from_mode);
-        if (from_mode != Vm_NONE)   line.append_utf8(",");
+        if (from_mode != Vm_NONE)   line.append_UTF8(",");
 
-        if      (to_mode == Vm_UCS)    line.append_utf8("(,⎕UCS ");
-        else if (to_mode == Vm_QUOT)   line.append_utf8("'");
+        if      (to_mode == Vm_UCS)    line.append_UTF8("(,⎕UCS ");
+        else if (to_mode == Vm_QUOT)   line.append_UTF8("'");
       }
 }
 //-----------------------------------------------------------------------------
+Value_P
+Quad_CR::do_CR33(const Value & B)
+{
+   // convert B = Integer Tag, len bytes Data
+   // to      Z = 4-byte Tag, 4-byte Len, len bytes Data
+   //
+   if (B.get_rank() > 1)   RANK_ERROR;
+const ShapeItem len_B = B.element_count();
+   if (len_B < 1)   LENGTH_ERROR;
+const ShapeItem len_B1 = len_B - 1;
+const Cell * cB = &B.get_ravel(0);
+   if (!cB++->is_integer_cell())   DOMAIN_ERROR;
+   loop (b,  len_B1)   cB++->get_byte_value();   // DOMAIN ERROR if not byte
+
+Value_P Z(len_B + 7, LOC);
+const APL_Integer tag = B.get_ravel(0).get_int_value();
+    new (Z->next_ravel()) CharCell(Unicode(tag >> 24 & 0xFF));
+    new (Z->next_ravel()) CharCell(Unicode(tag >> 16 & 0xFF));
+    new (Z->next_ravel()) CharCell(Unicode(tag >>  8 & 0xFF));
+    new (Z->next_ravel()) CharCell(Unicode(tag       & 0xFF));
+    new (Z->next_ravel()) CharCell(Unicode(len_B1 >> 24 & 0xFF));
+    new (Z->next_ravel()) CharCell(Unicode(len_B1 >> 16 & 0xFF));
+    new (Z->next_ravel()) CharCell(Unicode(len_B1 >>  8 & 0xFF));
+    new (Z->next_ravel()) CharCell(Unicode(len_B1       & 0xFF));
+    loop(z, len_B1)
+       new (Z->next_ravel())
+           CharCell(Unicode(B.get_ravel(z + 1).get_byte_value()));
+
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+Quad_CR::do_CR34(const Value & B)
+{
+   // convert B = 4-byte Tag, 4-byte Len, len bytes Data
+   // to      Z = Integer Tag, len bytes Data
+   //
+   //
+   if (B.get_rank() != 1)   RANK_ERROR;
+const ShapeItem len_B = B.element_count();
+   if (len_B < 8)   LENGTH_ERROR;
+const Cell * cB = &B.get_ravel(0);
+
+   // throwe DOMAIN ERROR if one of the vector items is not a byte
+   loop(b, len_B)
+       {
+         cB++->get_byte_value();
+       }
+
+   cB = &B.get_ravel(0);
+
+int32_t tag = 0;
+   loop(bb, 4)   tag = tag << 8 | cB++->get_byte_value();
+
+uint32_t len = 0;
+   loop(bb, 4)   len = len << 8 | cB++->get_byte_value();
+   if (len != (len_B - 8))   LENGTH_ERROR;
+
+Value_P Z(len_B - 7, LOC);
+   new (Z->next_ravel())   IntCell(tag);
+   loop(z, len_B - 8)
+      new (Z->next_ravel())  CharCell(Unicode(cB++->get_byte_value()));
+
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+Quad_CR::do_CR35(const Value & B)
+{
+   if (B.get_rank() != 1)   RANK_ERROR;
+
+const ShapeItem len_B = B.element_count();
+   if (len_B == 0)
+      {
+        Value_P Z1 = Str0(LOC);
+        Value_P Z(1, LOC);
+        new (&Z->get_ravel(0)) PointerCell(Z1.get(), Z.getref());
+        Z->check_value(LOC);
+        return Z;
+      }
+
+ShapeItem lf_count = 0;
+   loop(b, len_B)
+       {
+         if (B.get_ravel(b).get_char_value() == UNI_ASCII_LF)   ++lf_count;
+       }
+
+   if (B.get_ravel(len_B - 1).get_char_value() != UNI_ASCII_LF)   ++lf_count;
+
+Value_P Z(lf_count, LOC);
+UCS_string line;
+   loop(b, len_B)
+       {
+         const Unicode uni = B.get_ravel(b).get_char_value();
+         if (uni == UNI_ASCII_LF)
+            {
+              Value_P Zb(line, LOC);
+              new (Z->next_ravel()) PointerCell(Zb.get(), Z.getref());
+              line.clear();
+            }
+         else
+            {
+              line.append(uni);
+            }
+       }
+
+   if (line.size())   // incomplete last line
+      {
+        Value_P Zb(line, LOC);
+        new (Z->next_ravel()) PointerCell(Zb.get(), Z.getref());
+        line.clear();
+      }
+
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+Quad_CR::do_CR36(const Value & B)
+{
+   if (B.get_rank() != 1)   RANK_ERROR;
+
+const ShapeItem len_B = B.element_count();
+ShapeItem len_Z = B.element_count();
+   loop(b, len_B)
+       {
+         const Value & Bb = *B.get_ravel(b).get_pointer_value().get();
+         if (Bb.get_rank() > 1)   RANK_ERROR;
+         len_Z += 1 + Bb.element_count();
+       }
+
+UCS_string UZ;
+   UZ.reserve(len_Z);
+   loop(b, len_B)
+       {
+         const Value & Bb = *B.get_ravel(b).get_pointer_value().get();
+         UCS_string Ub(Bb);
+         UZ.append(Ub);
+         UZ.append(UNI_ASCII_LF);
+       }
+
+Value_P Z(UZ, LOC);
+   return Z;
+}
+//-----------------------------------------------------------------------------
 
