@@ -18,6 +18,8 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#include <vector>
+
 #include "Logging.hh"
 #include "PointerCell.hh"
 #include "Quad_DLX.hh"
@@ -54,6 +56,10 @@ public:
       right(r)
    { insert_lr();   insert_ud(); }
 
+   /// assignment
+   void operator = (const DLX_Node & other)
+      { new (this) DLX_Node(other); }
+
    /// (re-)insert \b this node horizontally (between left and right)
    void insert_lr()   { left->right = this;  right->left = this; }
 
@@ -261,7 +267,7 @@ public:
    ShapeItem get_cover_count() const      { return cover_count; }
 
    /// all solutions as len rows... len rows ...
-   Simple_string<ShapeItem, false> all_solutions;
+   std::vector<ShapeItem> all_solutions;
 
 protected:
    /// the max. number of solutions to produce, 0 = all
@@ -274,10 +280,10 @@ protected:
    const ShapeItem cols;
 
    /// the column headers
-   Simple_string<DLX_Header_Node, false> headers;
+   std::vector<DLX_Header_Node> headers;
 
    /// the '1's and '2's in the (sparse) matrix
-   Simple_string<DLX_Node, false> nodes;
+   std::vector<DLX_Node> nodes;
 
    /// the number of primary columns
    ShapeItem primary_count;
@@ -315,7 +321,7 @@ DLX_Root_Node::DLX_Root_Node(ShapeItem rs, ShapeItem cs, ShapeItem max_sol,
      cover_count(0),
      first_0(0)
 {
-   // count the number of non-zero elemnts in the matrix
+   // count the number of non-zero elements in the matrix
    //
 const ShapeItem ec_B = B.element_count();
 const int qio = Workspace::get_IO();
@@ -346,23 +352,23 @@ const Cell * b = &B.get_ravel(0);
 
    Log(LOG_Quad_DLX)   CERR << "Matrix has " << ones << " ones" << endl;
 
-   // set up column headers. Simple_string::append may move the headers so
+   // set up column headers. std::vector.push_back() may move the headers so
    // we first append all headers and initialize then
    //
    headers.reserve(cols);
-   loop (c, cols)   headers.append(DLX_Header_Node());
+   loop (c, cols)   headers.push_back(DLX_Header_Node());
    loop (c, cols)
       {
         if (c) new (&headers[c]) DLX_Header_Node(c, &headers[c - 1], this);
         else   new (&headers[c]) DLX_Header_Node(c, this,            this);
       }
 
-   // set up non-header nodes. Simple_string::append may move the headers so
+   // set up non-header nodes. std::vector.push_back() may move the headers so
    // we first append all nodes and initialize then.
    //
    b = &B.get_ravel(0);
    nodes.reserve(ones);
-   loop(o, ones)   nodes.append(DLX_Node());
+   loop(o, ones)   nodes.push_back(DLX_Node());
 
    DLX_Node * n = &nodes[0];
    loop (r, rows)
@@ -426,7 +432,8 @@ ShapeItem pcnt = 0;
    out << endl << "  N:";
    for (const DLX_Node * x = right; x != this; x = x->right)
        {
-         const DLX_Header_Node & hdr = *(const DLX_Header_Node *)x;
+         const DLX_Header_Node & hdr =
+               *reinterpret_cast<const DLX_Header_Node *>(x);
          Assert(hdr.row == -1);
          if (hdr.col_type == Col_PRIMARY)   ++pcnt;
          out << " " << setw(w) << hdr.count;
@@ -434,7 +441,8 @@ ShapeItem pcnt = 0;
    out << endl << "Col:";
    for (const DLX_Node * x = right; x != this; x = x->right)
        {
-         const DLX_Header_Node & hdr = *(const DLX_Header_Node *)x;
+         const DLX_Header_Node & hdr =
+               *reinterpret_cast<const DLX_Header_Node *>(x);
          Assert(hdr.row == -1);
          out << " " << setw(w) << hdr.col;
        }
@@ -488,7 +496,8 @@ DLX_Root_Node::deep_check() const
 {
    for (const DLX_Node * x = right; x != this; x = x->right)
        {
-         const DLX_Header_Node & hdr = *(const DLX_Header_Node *)x;
+         const DLX_Header_Node & hdr =
+               *reinterpret_cast<const DLX_Header_Node *>(x);
          Assert(hdr.row == -1);
          hdr.check();
          ShapeItem ones = 0;
@@ -505,7 +514,7 @@ void
 DLX_Root_Node::solve()
 {
 new_level:
-   Assert(level < cols);
+   Assert(level <= cols);
 
    deep_check();
 
@@ -532,9 +541,9 @@ DLX_Node * item = right;
 
    if (primary_count == 0)
       {
-        all_solutions.append(level);
+        all_solutions.push_back(level);
         loop(s, level)
-            all_solutions.append(headers[s].item_r->row);
+            all_solutions.push_back(headers[s].item_r->row);
 
         Log(LOG_Quad_DLX)
            {
@@ -792,7 +801,7 @@ Value_P Z(root.get_solution_count(), LOC);
    if (root.get_solution_count() == 0)   // empty result
       {
         Value_P Z1 = Idx0(LOC);   // Z1 is ⍬
-        new (&Z->get_ravel(0))   PointerCell(Z1, Z.getref());
+        new (&Z->get_ravel(0))   PointerCell(Z1.get(), Z.getref());
       }
    else
       {
@@ -806,7 +815,7 @@ Value_P Z(root.get_solution_count(), LOC);
                   new (Zs->next_ravel()) IntCell(qio + root.all_solutions[a++]);
                 }
              Zs->check_value(LOC);
-             new (Z->next_ravel())   PointerCell(Zs, Z.getref());
+             new (Z->next_ravel())   PointerCell(Zs.get(), Z.getref());
            }
       }
 
