@@ -21,14 +21,14 @@
 #include <stdlib.h>
 
 #include "Quad_RL.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "IntCell.hh"
 
 uint64_t Quad_RL::state = 0;
 
 //=============================================================================
 Quad_RL::Quad_RL()
-   : SystemVariable(ID::Quad_RL)
+   : SystemVariable(ID_Quad_RL)
 {
 Value_P value(LOC);
 
@@ -63,9 +63,9 @@ Quad_RL::get_random()
    state += Knuth_c;
 
    Assert(value_stack.size());
-   if (value_stack.last().name_class != NC_VARIABLE)   VALUE_ERROR;
+   if (value_stack.back().name_class != NC_VARIABLE)   VALUE_ERROR;
 
-   new (&value_stack.last().apl_val->get_ravel(0))   IntCell(state);
+   new (&value_stack.back().apl_val->get_ravel(0))   IntCell(state);
    return state;
 }
 //-----------------------------------------------------------------------------
@@ -81,6 +81,6 @@ void
 Quad_RL::pop()
 {
    Symbol::pop();
-   state = value_stack.last().apl_val->get_ravel(0).get_near_int();
+   state = value_stack.back().apl_val->get_ravel(0).get_near_int();
 }
 //=============================================================================
