@@ -121,10 +121,11 @@ char filename[APL_PATH_MAX + 1];
 bool found_executable = false;
    for (int d = 0; d < dircount; ++d)
        {
-         snprintf(filename, APL_PATH_MAX,
+         const int slen = snprintf(filename, APL_PATH_MAX,
                   "%s%s/AP%u --id %u --par %u --gra %u --auto%s",
                   LibPaths::get_APL_bin_path(), dirs[d], ap, ap,
                   own_ID, par_ID, verbose);
+         if (slen >= APL_PATH_MAX)   filename[APL_PATH_MAX] = 0;
 
          found_executable = is_executable(filename);
          if (found_executable)   break;
@@ -145,7 +146,7 @@ FILE * fp = popen(filename, "r");
         return;
       }
 
-   for (int cc; (cc = getc(fp)) != EOF;)   CERR << (char)cc;
+   for (int cc; (cc = getc(fp)) != EOF;)   CERR << char(cc);
    CERR << endl;
 
    pclose(fp);
@@ -199,7 +200,7 @@ Value_P Z(sh_Z, LOC);
         // set control.
         //
         Symbol * sym = Workspace::lookup_existing_symbol(vars[z]);
-        if (sym == 0)   throw_symbol_error(vars[z], LOC);
+        if (sym == 0)   Error::throw_symbol_error(vars[z], LOC);
 
         const SV_key key = sym->get_SV_key();
         Svar_DB::set_control(key, Svar_Control(ctl));
@@ -231,7 +232,7 @@ Value_P Z(sh_Z, LOC);
    loop(z, var_count)
       {
         Symbol * sym = Workspace::lookup_existing_symbol(vars[z]);
-        if (sym == 0)   throw_symbol_error(vars[z], LOC);
+        if (sym == 0)   Error::throw_symbol_error(vars[z], LOC);
 
         const SV_key key = sym->get_SV_key();
         const Svar_Control control = Svar_DB::get_control(key);
@@ -247,7 +248,7 @@ Value_P Z(sh_Z, LOC);
 }
 //=============================================================================
 Quad_SVE::Quad_SVE()
-   : NL_SystemVariable(ID::Quad_SVE)
+   : NL_SystemVariable(ID_Quad_SVE)
 {
    Symbol::assign(IntScalar(0, LOC), false, LOC);
 }
@@ -304,7 +305,7 @@ Quad_SVE::get_apl_value() const
    //
 const APL_Float remaining = timer_end - now();
 
-   if (remaining < 0)   return IntScalar(0, LOC);
+   if (remaining < 0.0)   return IntScalar(0, LOC);
 
 Value_P Z(LOC);
    new (Z->next_ravel())   FloatCell(0.000001 * remaining);
@@ -526,7 +527,7 @@ Quad_SVQ::get_processors()
    //
    // 2. running processors that have offered variables in Svar_DB.
    //
-Simple_string<AP_num, false> processors;
+std::vector<AP_num> processors;
 
    // case 1...
    //
@@ -536,8 +537,10 @@ const char * dirs[] = { "", "/APs" };
    for (int d = 0; d < dircount; ++d)
        {
          char dirname[APL_PATH_MAX + 1];
-         snprintf(dirname, APL_PATH_MAX, "%s%s", LibPaths::get_APL_bin_path(),
+         int slen = snprintf(dirname, APL_PATH_MAX, "%s%s", LibPaths::get_APL_bin_path(),
                   dirs[d]);
+         if (slen >= APL_PATH_MAX)   dirname[APL_PATH_MAX] = 0;
+
          dirname[APL_PATH_MAX] = 0;
          DIR * dir = opendir(dirname);
 
@@ -565,9 +568,9 @@ const char * dirs[] = { "", "/APs" };
 #endif
 
                 char filename[APL_PATH_MAX + 1];
-                snprintf(filename, APL_PATH_MAX, "%s/%s", dirname,
-                         entry->d_name);
-                filename[APL_PATH_MAX] = 0;
+                int slen = snprintf(filename, APL_PATH_MAX, "%s/%s",
+                                    dirname, entry->d_name);
+                if (slen >= APL_PATH_MAX)   filename[APL_PATH_MAX] = 0;
 
                 if (!is_executable(filename))   continue;
 
@@ -575,10 +578,11 @@ const char * dirs[] = { "", "/APs" };
                 if (sscanf(entry->d_name, "AP%u", &apnum) != 1)   continue;
 
                 char expected[APL_PATH_MAX + 1];
-                snprintf(expected, sizeof(expected), "AP%u", apnum);
+                slen = snprintf(expected, sizeof(expected), "AP%u", apnum);
+                if (slen >= APL_PATH_MAX)   expected[APL_PATH_MAX] = 0;
                 if (strcmp(entry->d_name, expected))   continue;
 
-                processors.append((AP_num)apnum);
+                processors.push_back(AP_num(apnum));
              }
 
          closedir(dir);
@@ -590,22 +594,22 @@ const char * dirs[] = { "", "/APs" };
 
    // sort and remove duplicates
    //
-Simple_string<int32_t, false> sorted;
+std::vector<int32_t> sorted;
    while (processors.size())
       {
         // find smallest
         //
         int smallest = processors[0];
-        for (int s = 1; s < processors.size(); ++s)
+        for (size_t s = 1; s < processors.size(); ++s)
             if (smallest > processors[s])   smallest = processors[s];
 
        // add smallest to sorted
        //
-       sorted.append(smallest);
+       sorted.push_back(smallest);
 
        // remove smallest from processors
        //
-        for (int s = 0; s < processors.size();)
+        for (size_t s = 0; s < processors.size();)
             {
               if (processors[s] != smallest)
                  {
@@ -613,8 +617,8 @@ Simple_string<int32_t, false> sorted;
                  }
               else
                  {
-                   processors[s] = processors.last();
-                   processors.pop();
+                   processors[s] = processors.back();
+                   processors.pop_back();
                  }
             }
       }
@@ -627,21 +631,20 @@ Value_P Z(sorted.size(), LOC);
 Value_P
 Quad_SVQ::get_variables(AP_num proc)
 {
-Simple_string<uint32_t, false> varnames;
+std::vector<uint32_t> varnames;
    Svar_DB::get_offered_variables(ProcessorID::get_own_ID(), proc, varnames);
 
    // varnames is a sequence of 0-terminated Unicodes
    //
-Simple_string<int, false> var_lengths;   // including terminating 0
+std::vector<int> var_lengths;   // including terminating 0
 int last_zero = -1;
    loop(v, varnames.size())
       {
         if (varnames[v] == 0)
            {
-             var_lengths.append(v + 1 - last_zero);
+             var_lengths.push_back(v + 1 - last_zero);
              last_zero = v;
            }
-        
       }
 
 const int count = var_lengths.size();
@@ -717,7 +720,7 @@ Value_P Z(sh_Z, LOC);
    loop(z, var_count)
       {
         Symbol * sym = Workspace::lookup_existing_symbol(vars[z]);
-        if (sym == 0)   throw_symbol_error(vars[z], LOC);
+        if (sym == 0)   Error::throw_symbol_error(vars[z], LOC);
 
         const SV_key key = sym->get_SV_key();
         const Svar_state state = Svar_DB::get_state(key);
