@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2017  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
 
 #include <math.h>
 
-#include "Value.icc"
+#include "Value.hh"
 #include "ComplexCell.hh"
 #include "FloatCell.hh"
 #include "IntCell.hh"
@@ -41,7 +41,9 @@ RealCell::bif_logarithm(Cell * Z, const Cell * A) const
    if (get_real_value() == 0.0)   return E_DOMAIN_ERROR;
    if (A->is_near_one())          return E_DOMAIN_ERROR;
 
-   if (A->is_real_cell() && A->get_real_value() >= 0 && get_real_value() >= 0)
+   if (A->is_real_cell()          &&
+       A->get_real_value() >= 0.0 &&
+          get_real_value() >= 0.0)
       {
         const APL_Float z = log(get_real_value()) / log(A->get_real_value());
         if (!isfinite(z))   return E_DOMAIN_ERROR;
@@ -62,7 +64,7 @@ RealCell::bif_circle_fun(Cell * Z, const Cell * A) const
    if (!A->is_near_int())   return E_DOMAIN_ERROR;
 const APL_Integer fun = A->get_checked_near_int();
 
-   new (Z) FloatCell(0);   // prepare for DOMAIN ERROR
+   FloatCell::zv(Z, 0);   // prepare for DOMAIN ERROR
 
 const ErrorCode ret = do_bif_circle_fun(Z, fun);
    if (!Z->is_finite())   return E_DOMAIN_ERROR;
@@ -75,7 +77,7 @@ RealCell::bif_circle_fun_inverse(Cell * Z, const Cell * A) const
    if (!A->is_near_int())   return E_DOMAIN_ERROR;
 const APL_Integer fun = A->get_checked_near_int();
 
-   new (Z) FloatCell(0);   // prepare for DOMAIN ERROR
+   FloatCell::zv(Z, 0);   // prepare for DOMAIN ERROR
 
 ErrorCode ret = E_DOMAIN_ERROR;
    switch(fun)
@@ -107,6 +109,7 @@ ErrorCode
 RealCell::do_bif_circle_fun(Cell * Z, int fun) const
 {
 const APL_Float b = get_real_value();
+const APL_Complex cb(b);
 
    switch(fun)
       {
@@ -114,25 +117,38 @@ const APL_Float b = get_real_value();
         case -11: return ComplexCell::zv(Z, 0.0, b );
         case -10: return FloatCell::zv(Z,       b );
         case  -9: return FloatCell::zv(Z,       b );
-        case  -8: return ComplexCell::do_bif_circle_fun(Z, -8, b);
+        case  -8: { const APL_Float par = -(b*b + 1.0);       // (¯1 + R⋆2)
+                    if (par < 0.0)   // complex square root
+                       {
+                         const APL_Float sq = sqrt(-par);     // (¯1 + R⋆2)⋆0.5
+                         if (b < 0.0)   return ComplexCell::zv(Z, 0.0, -sq);
+                         else           return ComplexCell::zv(Z, 0.0, sq);
+                       }
+                    else           // real square root
+                       {
+                         const APL_Float sq = sqrt(par);      // (¯1 + R⋆2)⋆0.5
+                         if (b < 0.0)   return FloatCell::zv(Z, -sq);
+                         else           return FloatCell::zv(Z, sq);
+                       }
+                  }
 
         case  -7: if (b > -1.0 && b < 1.0)   return FloatCell::zv(Z, atanh(b));
                   if (b == -1.0 || b == 1.0)   return E_DOMAIN_ERROR;
-                  return ComplexCell::do_bif_circle_fun(Z, -7, b);
+                  return ComplexCell::do_bif_circle_fun(Z, -7, cb);
 
         case  -6: if (b > 1.0)   return FloatCell::zv(Z, acosh(b));
-                  return ComplexCell::do_bif_circle_fun(Z, -6, b);
+                  return ComplexCell::do_bif_circle_fun(Z, -6, cb);
 
         case  -5: return FloatCell::zv(Z, asinh(b));
-        case  -4: if (b >= -1)   return FloatCell::zv(Z, sqrt(b*b - 1.0));
-                  return ComplexCell::do_bif_circle_fun(Z, -4, b);
+        case  -4: if (b >= -1.0)   return FloatCell::zv(Z, sqrt(b*b - 1.0));
+                  return ComplexCell::do_bif_circle_fun(Z, -4, cb);
         case  -3: return FloatCell::zv(Z, atan (b));
         case  -2: if (b >= -1.0 && b <= 1.0)  return FloatCell::zv(Z, acos (b));
-                  return ComplexCell::do_bif_circle_fun(Z, -2, b);
+                  return ComplexCell::do_bif_circle_fun(Z, -2, cb);
         case  -1: if (b >= -1.0 && b <= 1.0)  return FloatCell::zv(Z, asin (b));
-                  return ComplexCell::do_bif_circle_fun(Z, -1, b);
-        case   0: if (b*b < 1)   return FloatCell::zv(Z, sqrt (1 - b*b));
-                  return ComplexCell::do_bif_circle_fun(Z, 0, b);
+                  return ComplexCell::do_bif_circle_fun(Z, -1, cb);
+        case   0: if (b*b < 1.0)   return FloatCell::zv(Z, sqrt (1 - b*b));
+                  return ComplexCell::do_bif_circle_fun(Z, 0, cb);
         case   1: return FloatCell::zv(Z,  sin (b));
         case   2: return FloatCell::zv(Z,  cos (b));
         case   3: return FloatCell::zv(Z,  tan (b));
@@ -140,11 +156,25 @@ const APL_Float b = get_real_value();
         case   5: return FloatCell::zv(Z,  sinh(b));
         case   6: return FloatCell::zv(Z,  cosh(b));
         case   7: return FloatCell::zv(Z,  tanh(b));
-        case   8: return ComplexCell::do_bif_circle_fun(Z, 8, b);
+        case   8: { const APL_Float par = -(b*b + 1.0);       // (¯1 - R⋆2)
+                    if (par < 0.0)   // complex square root
+                       {
+                         const APL_Float sq = sqrt(-par);     // (¯1 + R⋆2)⋆0.5
+                         if (b < 0.0)   return ComplexCell::zv(Z, 0.0, sq);
+                         else           return ComplexCell::zv(Z, 0.0, -sq);
+                       }
+                    else           // real square root
+                       {
+                         const APL_Float sq = sqrt(par);      // (¯1 + R⋆2)⋆0.5
+                         if (b < 0.0)   return FloatCell::zv(Z, sq);
+                         else           return FloatCell::zv(Z, -sq);
+                       }
+                  }
         case   9: return FloatCell::zv(Z, b );
-        case  10: return FloatCell::zv(Z, (b < 0) ? -b : b);
+        case  10: if (b < 0.0)   return FloatCell::zv(Z, -b);
+                  else           return FloatCell::zv(Z,  b);
         case  11: return FloatCell::zv(Z, 0.0 );
-        case  12: return FloatCell::zv(Z, (b < 0) ? M_PI : 0.0);
+        case  12: return FloatCell::zv(Z, (b < 0.0) ? M_PI : 0.0);
       }
 
    // invalid fun
