@@ -19,10 +19,10 @@
 */
 
 #include "Shape.hh"
-#include "Value.icc"
+#include "Value.hh"
 
 //-----------------------------------------------------------------------------
-Shape::Shape(Value_P A, int qio_A)
+Shape::Shape(const Value * A, int qio_A)
    : rho_rho(0),
      volume(1)
 {
@@ -89,7 +89,7 @@ Shape::insert_axis(Axis axis, ShapeItem len) const
         return ret + *this;
       }
 
-   if (axis >= get_rank())   // insert after last axis
+   if (uAxis(axis) >= get_rank())   // insert after last axis
       {
         const Shape ret(len);
         return *this + ret;
