@@ -20,12 +20,12 @@
 
 #include <iomanip>
 
+#include "Bif_F12_TAKE_DROP.hh"
 #include "Command.hh"
 #include "Executable.hh"
 #include "IndexExpr.hh"
 #include "Output.hh"
 #include "Prefix.hh"
-#include "SharedValuePointer.hh"
 #include "StateIndicator.hh"
 #include "SystemLimits.hh"
 #include "UserFunction.hh"
@@ -177,9 +177,9 @@ StateIndicator::print(ostream & out) const
    out << "Stat:       " << executable->statement_text(get_PC());
    out << endl;
 
-   out << "err_code:   " << HEX(error.error_code) << endl;
-   if (error.error_code)
-      out << "thrown at:  " << error.throw_loc << endl
+   out << "err_code:   " << HEX(error.get_error_code()) << endl;
+   if (error.get_error_code())
+      out << "thrown at:  " << error.get_throw_loc() << endl
        << "e_msg_1:    '" << error.get_error_line_1() << "'" << endl
        << "e_msg_2:    '" << error.get_error_line_2() << "'" << endl
        << "e_msg_3:    '" << error.get_error_line_3() << "'" << endl;
@@ -210,7 +210,7 @@ StateIndicator::list(ostream & out, SI_mode mode) const
 
              if (mode & SIM_statements)   // )SIS
                 {
-                  if (error.error_code)
+                  if (error.get_error_code())
                      {
                        out << error.get_error_line_2() << endl
                            << error.get_error_line_3();
@@ -258,7 +258,7 @@ StateIndicator::list(ostream & out, SI_mode mode) const
 
                   // )SIS and we have a statement
                   //
-                  if (error.error_code)
+                  if (error.get_error_code())
                      out << error.get_error_line_2() << endl
                          << error.get_error_line_3();
                   else
@@ -271,68 +271,6 @@ StateIndicator::list(ostream & out, SI_mode mode) const
    out << endl;
 }
 //-----------------------------------------------------------------------------
-void
-StateIndicator::update_error_info(Error & err)
-{
-bool locked = false;
-const UserFunction * ufun = executable->get_ufun();
-
-   // prepare second error line (failed statement)
-   //
-   if (ufun)
-      {
-        if (err.show_locked || ufun->get_exec_properties()[1])
-           {
-             locked = true;
-             err.error_message_2 = UCS_string(6, UNI_ASCII_SPACE);
-           }
-        else
-           {
-             err.error_message_2 = ufun->get_name_and_line(get_PC());
-             err.error_message_2.append(UNI_ASCII_SPACE);
-             err.error_message_2.append(UNI_ASCII_SPACE);
-           }
-      }
-   else
-      {
-        err.error_message_2 = UCS_string(6, UNI_ASCII_SPACE);
-      }
-
-   // prepare third line (carets)
-   //
-   err.left_caret = err.error_message_2.size();
-   err.right_caret = -1;
-
-   if (locked)
-      {
-        executable->get_ufun()->set_locked_error_info(err);
-      }
-   else
-      {
-        const Function_PC from = current_stack.get_range_low();
-        Function_PC to   = current_stack.get_range_high();
-        const Function_PC2 error_range(from, to);
-        executable->set_error_info(err, error_range);
-      }
-
-   // print error, unless we are in safe execution mode.
-   //
-   {
-     bool print_error = true;
-     for (const StateIndicator * si = this; si; si = si->get_parent())
-         {
-           if (si->get_safe_execution())
-              {
-                print_error = false;
-                break;
-              }
-         }
-     if (print_error)   err.print_em(UERR, LOC);
-   }
-
-   error = err;
-}
-//-----------------------------------------------------------------------------
 ostream &
 StateIndicator::indent(ostream & out) const
 {
@@ -360,7 +298,7 @@ StateIndicator::jump(Value_P value)
    // function (and then return TOK_BRANCH.). The jump itself (if any)
    // is executed in Prefix.cc.
    //
-   if (value->get_rank() > 1)   DOMAIN_ERROR;
+   if (value->get_rank() > 1)   RANK_ERROR;
 
    if (value->element_count() == 0)     // →''
       {
@@ -389,7 +327,7 @@ const UserFunction * ufun = get_executable()->get_ufun();
 }
 //-----------------------------------------------------------------------------
 void
-StateIndicator::escape()  
+StateIndicator::escape()
 {
 }
 //-----------------------------------------------------------------------------
@@ -489,6 +427,33 @@ Value_P * X = current_stack.locate_X();
    if (X)   *X = new_value;
 }
 //-----------------------------------------------------------------------------
+int
+StateIndicator::nth_push(const Symbol * sym, int from_tos) const
+{
+   if (from_tos == 0)   return 0;
+
+  // collect SI entries in reverse order...
+   //
+std::vector<const StateIndicator *> stack;
+
+   for (const StateIndicator * si = Workspace::SI_top();
+        si; si = si->get_parent())
+      {
+        stack.push_back(si);
+      }
+
+   loop(d, stack.size())
+       {
+         const StateIndicator * si = stack[stack.size() - d - 1];
+         const Executable * exec = si->get_executable();
+         Assert(exec);
+         if (!exec->pushes_sym(sym))   continue;
+         if (0 == --from_tos)   return si->get_level();
+       }
+
+   FIXME;
+}
+//-----------------------------------------------------------------------------
 Function_Line
 StateIndicator::get_line() const
 {
@@ -502,6 +467,11 @@ typedef int (*result_callback)(const Value * result, int committed);
 extern "C" result_callback res_callback;
 result_callback res_callback = 0;
 #endif
+
+#ifdef WANT_PYTHON
+extern bool python_result_callback(Token & result);
+#endif
+
 void
 StateIndicator::statement_result(Token & result, bool trace)
 {
@@ -515,7 +485,7 @@ StateIndicator::statement_result(Token & result, bool trace)
         if (ufun && (ufun->get_exec_properties()[0] == 0))
            {
              const Function_Line line =
-                            executable->get_line((Function_PC)(get_PC() - 1));
+                   executable->get_line(Function_PC(get_PC() - 1));
              result.show_trace(COUT, ufun->get_name(), line);
            }
       }
@@ -527,7 +497,14 @@ StateIndicator::statement_result(Token & result, bool trace)
    // if result is a value then print it, unless it is a committed value
    // (i.e. TOK_APL_VALUE2)
    //
-   if (result.get_ValueType() != TV_VAL)   return;
+   if (result.get_ValueType() != TV_VAL)
+      {
+#ifdef WANT_PYTHON
+         python_result_callback(result);
+#endif
+
+         return;
+      }
 
 const TokenTag tag = result.get_tag();
 Value_P B(result.get_apl_val());
@@ -547,6 +524,10 @@ bool print_value = tag == TOK_APL_VALUE1 || tag == TOK_APL_VALUE3;
       }
 #endif
 
+#ifdef WANT_PYTHON
+   print_value = python_result_callback(result);
+#endif
+
    if (!print_value)   return;
 
    Quad_QUOTE::done(false, LOC);
@@ -574,7 +555,7 @@ const int boxing_format = Command::get_boxing_format();
                   sh.set_shape_item(longest, sh.get_shape_item(longest) / 2);
                 }
 
-             Value_P B1 = Bif_F12_TAKE::do_take(sh, B).get_apl_val();
+             Value_P B1 = Bif_F12_TAKE::do_take(sh, B);
              B1->print(COUT);
 
              CERR << "      *** display of value was truncated (limit "
