@@ -63,7 +63,9 @@ Svar_DB::start_APserver(const char * server_sockname,
    // set APserver_path to the case that applies.
    //
 char APserver_path[APL_PATH_MAX + 1];
-   snprintf(APserver_path, APL_PATH_MAX, "%s/APserver", bin_dir);
+const int slen = snprintf(APserver_path, APL_PATH_MAX, "%s/APserver", bin_dir);
+   if (slen >= APL_PATH_MAX)   APserver_path[APL_PATH_MAX] = 0;
+
    APserver_path[APL_PATH_MAX] = 0;
    if (access(APserver_path, X_OK) != 0)   // no APserver
       {
@@ -71,7 +73,9 @@ char APserver_path[APL_PATH_MAX + 1];
                  << " not found (this is OK when apl was started\n"
                     "    from the src directory): " << strerror(errno) << endl;
 
-        snprintf(APserver_path, APL_PATH_MAX, "%s/APs/APserver", bin_dir);
+        const int slen = snprintf(APserver_path, APL_PATH_MAX,
+                                  "%s/APs/APserver", bin_dir);
+        if (slen >= APL_PATH_MAX)   APserver_path[APL_PATH_MAX] = 0;
         if (access(APserver_path, X_OK) != 0)   // no APs/APserver either
            {
              get_CERR() << "Executable " << APserver_path << " not found.\n"
@@ -88,12 +92,14 @@ char APserver_path[APL_PATH_MAX + 1];
 
 char popen_args[APL_PATH_MAX + 1];
    {
+     int slen;
      if (server_sockname)
-        snprintf(popen_args, APL_PATH_MAX,
+        slen = snprintf(popen_args, APL_PATH_MAX,
                  "%s --path %s --auto", APserver_path, server_sockname);
      else
-        snprintf(popen_args, APL_PATH_MAX,
+        slen = snprintf(popen_args, APL_PATH_MAX,
                  "%s --port %u --auto", APserver_path, APserver_port);
+     if (slen >= APL_PATH_MAX)   popen_args[APL_PATH_MAX] = 0;
    }
 
    logit && get_CERR() << "Starting " << popen_args << "..." << endl;
@@ -108,7 +114,7 @@ FILE * fp = popen(popen_args, "r");
 
    for (int cc; (cc = getc(fp)) != EOF;)
        {
-         logit && get_CERR() << (char)cc;
+         logit && get_CERR() << char(cc);
        }
 
    logit && get_CERR() << endl;
@@ -148,14 +154,16 @@ char peer[100];
              return NO_TCP_SOCKET;
            }
 
-        snprintf(peer, sizeof(peer), "%s", server_sockname);
+        const unsigned int slen = snprintf(peer, sizeof(peer),
+                                           "%s", server_sockname);
+        if (slen >= sizeof(peer))   peer[sizeof(peer) - 1] = 0;
       }
 #else // use TCP
       {
         server_sockname = 0;
         logit && get_CERR() << "Using TCP socket towards APserver..."
                             << endl;
-        sock = (TCP_socket)(socket(AF_INET, SOCK_STREAM, 0));
+        sock = TCP_socket(socket(AF_INET, SOCK_STREAM, 0));
         if (sock == NO_TCP_SOCKET)
            {
              get_CERR() << "*** socket(AF_INET, SOCK_STREAM, 0) failed at "
@@ -171,19 +179,21 @@ char peer[100];
 
         // bind local port to 127.0.0.1
         //
-        sockaddr_in local;
-        memset(&local, 0, sizeof(sockaddr_in));
-        local.sin_family = AF_INET;
-        local.sin_addr.s_addr = htonl(0x7F000001);
+        SockAddr local;
+        memset(&local, 0, sizeof(SockAddr));
+        local.inet.sin_family = AF_INET;
+        local.inet.sin_addr.s_addr = htonl(0x7F000001);
 
-        if (::bind(sock, (const sockaddr *)&local, sizeof(sockaddr_in)))
+        if (::bind(sock, &local.addr, sizeof(sockaddr_in)))
            {
              get_CERR() << "bind(127.0.0.1) failed:" << strerror(errno) << endl;
              ::close(sock);
              return NO_TCP_SOCKET;
            }
 
-        snprintf(peer, sizeof(peer), "127.0.0.1 TCP port %d", APserver_port);
+        const unsigned int slen = snprintf(peer, sizeof(peer),
+                                  "127.0.0.1 TCP port %d", APserver_port);
+        if (slen >= sizeof(peer))   peer[sizeof(peer) - 1] = 0;
       }
 #endif
 
@@ -198,25 +208,25 @@ char peer[100];
 #if HAVE_SYS_UN_H
         if (server_sockname)
            {
-             sockaddr_un remote;
-             memset(&remote, 0, sizeof(sockaddr_un));
-             remote.sun_family = AF_UNIX;
-             strcpy(remote.sun_path + ABSTRACT_OFFSET, server_sockname);
+             SockAddr remote;
+             memset(&remote, 0, sizeof(SockAddr));
+             remote.uNix.sun_family = AF_UNIX;
+             strcpy(remote.uNix.sun_path + ABSTRACT_OFFSET, server_sockname);
 
-             if (::connect(sock, (sockaddr *)&remote,
-                           sizeof(remote)) == 0)   break;   // success
+             if (::connect(sock, &remote.addr, sizeof(sockaddr_un)) == 0)
+                break;   // success
            }
         else   // TCP
 #endif
            {
-             sockaddr_in remote;
+             SockAddr remote;
              memset(&remote, 0, sizeof(sockaddr_in));
-             remote.sin_family = AF_INET;
-             remote.sin_port = htons(APserver_port);
-             remote.sin_addr.s_addr = htonl(0x7F000001);
+             remote.inet.sin_family = AF_INET;
+             remote.inet.sin_port = htons(APserver_port);
+             remote.inet.sin_addr.s_addr = htonl(0x7F000001);
 
-             if (::connect(sock, (sockaddr *)&remote,
-                           sizeof(remote)) == 0)   break;   // success
+             if (::connect(sock, &remote.addr, sizeof(sockaddr_in)) == 0)
+                break;   // success
            }
 
          // ::connect() to APserver failed. If
@@ -269,7 +279,7 @@ char peer[100];
    usleep(50000);
    logit && get_CERR() << "connected to APserver, socket is " << sock << endl;
 
-   return (TCP_socket)sock;
+   return TCP_socket(sock);
 }
 //-----------------------------------------------------------------------------
 void
@@ -299,13 +309,15 @@ const int sock = Svar_DB::get_DB_tcp();
 
 char * del = 0;
 char buffer[2*MAX_SIGNAL_CLASS_SIZE + sizeof(Svar_record)];
-ostream * log = (LOG_startup || LOG_Svar_DB_signals) ? & cerr : 0;
+ostream * log = (LOG_startup != 0 || LOG_Svar_DB_signals != 0) ? & cerr : 0;
 Signal_base * response = Signal_base::recv_TCP(sock, buffer, sizeof(buffer),
                                                del, log);
    if (response)
       {
-        memcpy(&cache, response->get__SVAR_RECORD_IS__record().data(),
+        memcpy(static_cast<void *>(&cache),
+               response->get__SVAR_RECORD_IS__record().data(),
                sizeof(Svar_record));
+
         delete response;
       }
    else   get_CERR() << "Svar_record_P() failed at " << LOC << endl;
@@ -347,7 +359,7 @@ uint32_t vname1[MAX_SVAR_NAMELEN];
         else                break;
       }
 
-string vname((const char *)vname1, MAX_SVAR_NAMELEN*sizeof(uint32_t));
+string vname(charP(vname1), MAX_SVAR_NAMELEN*sizeof(uint32_t));
 
 MATCH_OR_MAKE_c request(tcp, vname,
                              to.proc,      to.parent,      to.grand,
@@ -382,12 +394,12 @@ GET_EVENTS_c request(tcp, id.proc, id.parent, id.grand);
 
 char * del = 0;
 char buffer[2*MAX_SIGNAL_CLASS_SIZE + 16];
-Signal_base * response = Signal_base::recv_TCP(tcp, buffer, sizeof(buffer),
-                                               del, 0);
+Signal_base * response =
+                    Signal_base::recv_TCP(tcp, buffer, sizeof(buffer), del, 0);
 
    if (response)
       {
-        events = (Svar_event)response->get__EVENTS_ARE__events();
+        events = Svar_event(response->get__EVENTS_ARE__events());
         const SV_key ret = response->get__EVENTS_ARE__key();
         delete response;
         return ret;
@@ -417,7 +429,7 @@ Signal_base * response = Signal_base::recv_TCP(tcp, buffer, sizeof(buffer),
 
    if (response)
       {
-        const Svar_event ret = (Svar_event)response->get__EVENTS_ARE__events();
+        const Svar_event ret = Svar_event(response->get__EVENTS_ARE__events());
         delete response;
         return ret;
       }
@@ -525,18 +537,18 @@ Signal_base * response = Signal_base::recv_TCP(tcp, buffer, sizeof(buffer),
 
    if (response)
       {
-         offering_id.proc   = (AP_num)(response->get__OFFERING_ID_IS__proc());
-         offering_id.parent = (AP_num)(response->get__OFFERING_ID_IS__parent());
-         offering_id.grand  = (AP_num)(response->get__OFFERING_ID_IS__grand());
+         offering_id.proc = AP_num(response->get__OFFERING_ID_IS__proc());
+         offering_id.parent = AP_num(response->get__OFFERING_ID_IS__parent());
+         offering_id.grand = AP_num(response->get__OFFERING_ID_IS__grand());
         delete response;
       }
-       
+
    return offering_id;
 }
 //-----------------------------------------------------------------------------
 void
 Svar_DB::get_offering_processors(AP_num to_proc,
-                                 Simple_string<AP_num, false> & processors)
+                                 std::vector<AP_num> & processors)
 {
 const TCP_socket tcp = get_Svar_DB_tcp(__FUNCTION__);
    if (tcp == NO_TCP_SOCKET)   return;
@@ -551,17 +563,17 @@ Signal_base * response = Signal_base::recv_TCP(tcp, buffer, sizeof(buffer),
    if (response)
       {
         const string & op = response->get__OFFERING_PROCS_ARE__offering_procs();
-        const AP_num * procs = (const AP_num *)(op.data());
+        const AP_num * procs = reinterpret_cast<const AP_num *>(op.data());
         const size_t count = op.size() / sizeof(AP_num);
 
-        loop(c, count)   processors.append(*procs++);
+        loop(c, count)   processors.push_back(*procs++);
         delete response; 
       }
 }
 //-----------------------------------------------------------------------------
 void
 Svar_DB::get_offered_variables(AP_num to_proc, AP_num from_proc,
-                               Simple_string<uint32_t, false> & varnames)
+                               std::vector<uint32_t> & varnames)
 {
 const TCP_socket tcp = get_Svar_DB_tcp(__FUNCTION__);
    if (tcp == NO_TCP_SOCKET)   return;
@@ -576,10 +588,10 @@ Signal_base * response = Signal_base::recv_TCP(tcp, buffer, sizeof(buffer),
    if (response)
       {
         const string & ov = response->get__OFFERED_VARS_ARE__offered_vars();
-        const uint32_t * names = (const uint32_t *)(ov.data());
+        const uint32_t * names = reinterpret_cast<const uint32_t *>(ov.data());
         const size_t count = ov.size() / sizeof(uint32_t);
 
-        loop(c, count)   varnames.append(*names++);
+        loop(c, count)   varnames.push_back(*names++);
         delete response; 
       }
 }
