@@ -36,20 +36,21 @@
 #include "Svar_signals.hh"
 #include "SystemVariable.hh"
 #include "UserFunction.hh"
-#include "Value.icc"
+#include "Value.hh"
+#include "ValueHistory.hh"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
-Symbol::Symbol(ID::Id id)
+Symbol::Symbol(Id id)
    : NamedObject(id),
      next(0),
-     name(ID::get_name(id)),
+     name(ID::get_name_UCS(id)),
      monitor_callback(0)
 {
    push();
 }
 //-----------------------------------------------------------------------------
-Symbol::Symbol(const UCS_string & ucs, ID::Id id)
+Symbol::Symbol(const UCS_string & ucs, Id id)
    : NamedObject(id),
      next(0),
      name(ucs),
@@ -68,7 +69,7 @@ ostream &
 Symbol::print_verbose(ostream & out) const
 {
    out << "Symbol ";
-   print(out) << " " << (const void *)this << endl;
+   print(out) << " " << voidP(this) << endl;
 
    loop(v, value_stack.size())
        {
@@ -91,8 +92,8 @@ Symbol::print_verbose(ostream & out) const
               case NC_VARIABLE:
                    {
                       Value_P val = item.apl_val;
-                      out << "Variable at " << (const void *)val.get() << endl;
-                      val->print_properties(out, 8);
+                      out << "Variable at " << voidP(val.get()) << endl;
+                      val->print_properties(out, 8, false);
                       out << endl;
                    }
                    break;
@@ -104,10 +105,8 @@ Symbol::print_verbose(ostream & out) const
                      Assert(fun);
 
                      fun->print_properties(out, 4);
-                     out << "    ⎕NC:            " << item.name_class
-                         << endl
-                         << "    addr:           " << (const void *)fun
-                         << endl;
+                     out << "    ⎕NC:            " << item.name_class << endl
+                         << "    addr:           " << voidP(fun) << endl;
 
                      out << endl;
                    }
@@ -128,12 +127,12 @@ Symbol::assign(Value_P new_value, bool clone, const char * loc)
    if (!new_value->is_complete())
       {
         CERR << "Incomplete value at " LOC << endl;
-        new_value->print_properties(CERR, 0);
-        print_history(CERR, new_value.get(), LOC);
+        new_value->print_properties(CERR, 0, false);
+        VH_entry::print_history(CERR, new_value.get(), LOC);
         Assert(0);
       }
 
-ValueStackItem & vs = value_stack.last();
+ValueStackItem & vs = value_stack.back();
 
    switch(vs.name_class)
       {
@@ -145,6 +144,8 @@ ValueStackItem & vs = value_stack.last();
              if (monitor_callback)   monitor_callback(*this, SEV_ASSIGNED);
              return;
 
+        case NC_LABEL:
+             MORE_ERROR() << "attempt to assign a value to label " << get_name();
         case NC_VARIABLE:
              if (vs.apl_val == new_value)   return;   // X←X
 
@@ -293,9 +294,9 @@ const ShapeItem ec_B = B->element_count();
 const Cell * cB = &B->get_ravel(0);
 const int incr_B = (ec_B == 1) ? 0 : 1;
 
-   while (!mult.done())
+   while (mult.more())
       {
-        const ShapeItem offset_A = mult.next();
+        const ShapeItem offset_A = mult++;
         if (offset_A < 0)                     INDEX_ERROR;
         if (offset_A >= A->element_count())   INDEX_ERROR;
         Cell & dest = A->get_ravel(offset_A);
@@ -307,20 +308,39 @@ const int incr_B = (ec_B == 1) ? 0 : 1;
    if (monitor_callback)   monitor_callback(*this, SEV_ASSIGNED);
 }
 //-----------------------------------------------------------------------------
-void
+bool
 Symbol::assign_named_lambda(Function * lambda, const char * loc)
 {
-ValueStackItem & vs = value_stack.last();
+ValueStackItem & vs = value_stack.back();
 UserFunction * ufun = lambda->get_ufun1();
    Assert(ufun);
+const Executable * uexec = ufun;
+   Assert(uexec);
 
    switch(vs.name_class)
       {
         case NC_FUNCTION:
         case NC_OPERATOR:
-             if (!vs.sym_val.function->is_lambda())   SYNTAX_ERROR;
-             Assert(vs.sym_val.function->get_ufun1());
-             vs.sym_val.function->get_ufun1()->decrement_refcount(LOC);
+             {
+               Function * old_fun = vs.sym_val.function;
+               Assert(old_fun);
+               if (!old_fun->is_lambda())   SYNTAX_ERROR;
+               UserFunction * old_ufun = old_fun->get_ufun1();
+               Assert(old_ufun);
+               for (StateIndicator * si = Workspace::SI_top();
+                    si; si = si->get_parent())
+                   {
+//                   if (uexec == si->get_executable())
+                     if (si->uses_function(old_ufun))
+                        {
+                          MORE_ERROR() << "function " << get_name()
+                                       << " is suspended or used";
+                          return true;
+                        }
+                   }
+
+               vs.sym_val.function->get_ufun1()->decrement_refcount(LOC);
+             }
 
              /* fall through */
 
@@ -331,10 +351,12 @@ UserFunction * ufun = lambda->get_ufun1();
              vs.sym_val.function = ufun;
              ufun->increment_refcount(LOC);
              if (monitor_callback)   monitor_callback(*this, SEV_ASSIGNED);
-             return;
+             return false;
 
         default: SYNTAX_ERROR;
       }
+
+   return false;
 }
 //-----------------------------------------------------------------------------
 void
@@ -343,10 +365,10 @@ Symbol::pop()
    if (value_stack.size() == 0)
       {
         CERR << "Symbol is: '" << get_name() << "' at " << LOC << endl;
-        Assert(0 && LOC);
+        FIXME;
       }
 
-const ValueStackItem & vs = value_stack.last();
+const ValueStackItem & vs = value_stack.back();
 
    if (vs.name_class == NC_VARIABLE)
       {
@@ -356,15 +378,10 @@ const ValueStackItem & vs = value_stack.last();
              CERR << "-pop-value " << name
                   << " flags " << ret->get_flags() << " ";
              if (value_stack.size() == 0)   CERR << " (last)";
-             CERR << " addr " << (const void *)ret.get() << endl;
+             CERR << " addr " << voidP(ret.get()) << endl;
            }
 
-        // unlike vector<ValueStackItem>, Simple_string<ValueStackItem>.pop()
-        // does not destruct vs.apl_val, so we do it here.
-        //
-        ptr_clear(value_stack.last().apl_val, LOC);
-
-        value_stack.pop();
+        value_stack.pop_back();
         if (monitor_callback)   monitor_callback(*this, SEV_POPED);
       }
    else
@@ -376,7 +393,7 @@ const ValueStackItem & vs = value_stack.last();
              if (value_stack.size() == 0)   CERR << " (last)";
              CERR << endl;
            }
-        value_stack.pop();
+        value_stack.pop_back();
         if (monitor_callback)   monitor_callback(*this, SEV_POPED);
       }
 }
@@ -391,7 +408,7 @@ Symbol::push()
         CERR << endl;
       }
 
-   value_stack.append(ValueStackItem());
+   value_stack.push_back(ValueStackItem());
    if (monitor_callback)   monitor_callback(*this, SEV_PUSHED);
 }
 //-----------------------------------------------------------------------------
@@ -405,7 +422,7 @@ Symbol::push_label(Function_Line label)
         CERR << endl;
       }
 
-   value_stack.append(ValueStackItem(label));
+   value_stack.push_back(ValueStackItem(label));
    if (monitor_callback)   monitor_callback(*this, SEV_PUSHED);
 }
 //-----------------------------------------------------------------------------
@@ -414,7 +431,7 @@ Symbol::push_function(Function * function)
 {
    Log(LOG_SYMBOL_push_pop)
       {
-        CERR << "+push_function " << name << " " << (const void *)function;
+        CERR << "+push_function " << name << " " << voidP(function);
         if (value_stack.size() == 0)   CERR << " (initial)";
         CERR << endl;
       }
@@ -423,7 +440,7 @@ ValueStackItem vs;
    if (function->is_operator())   vs.name_class = NC_OPERATOR;
    else                           vs.name_class = NC_FUNCTION;
    vs.sym_val.function = function;
-   value_stack.append(vs);
+   value_stack.push_back(vs);
    if (monitor_callback)   monitor_callback(*this, SEV_PUSHED);
 }
 //-----------------------------------------------------------------------------
@@ -431,7 +448,7 @@ void
 Symbol::push_value(Value_P value)
 {
 ValueStackItem vs;
-   value_stack.append(vs);
+   value_stack.push_back(vs);
    if (monitor_callback)   monitor_callback(*this, SEV_PUSHED);
    assign(value, true, LOC);
 
@@ -440,14 +457,14 @@ ValueStackItem vs;
         CERR << "+push-value " << name << " flags ";
         print_flags(CERR, get_value()->get_flags()) << " ";
         if (value_stack.size() == 0)   CERR << " (initial)";
-        CERR << " addr " << (const void *)get_value().get() << endl;
+        CERR << " addr " << voidP(get_value().get()) << endl;
       }
 }
 //-----------------------------------------------------------------------------
 int
 Symbol::get_ufun_depth(const UserFunction * ufun)
 {
-const Function * fun = (Function *)ufun;
+const Function * fun = ufun;
 const int sym_stack_size = value_stack_size();
 
    loop(s, sym_stack_size)
@@ -468,7 +485,7 @@ const int sym_stack_size = value_stack_size();
 Value_P
 Symbol::get_value()
 {
-const ValueStackItem & vs = value_stack.last();
+const ValueStackItem & vs = value_stack.back();
 
    if (vs.name_class == NC_VARIABLE)   return vs.apl_val;
 
@@ -481,9 +498,9 @@ Symbol::cant_be_defined() const
 // if (value_stack.size() > 1)         return "symbol was localized";
    if (Workspace::is_called(name))   return "function is called";
 
-   if (value_stack.last().name_class == NC_UNUSED_USER_NAME)   return 0;   // OK
-   if (value_stack.last().name_class == NC_FUNCTION)           return 0;   // OK
-   if (value_stack.last().name_class == NC_OPERATOR)           return 0;   // OK
+   if (value_stack.back().name_class == NC_UNUSED_USER_NAME)   return 0;   // OK
+   if (value_stack.back().name_class == NC_FUNCTION)           return 0;   // OK
+   if (value_stack.back().name_class == NC_OPERATOR)           return 0;   // OK
    return "bad name class";
 }
 //-----------------------------------------------------------------------------
@@ -491,24 +508,24 @@ Value_P
 Symbol::get_apl_value() const
 {
    Assert(value_stack.size() > 0);
-   if (value_stack.last().name_class != NC_VARIABLE)
-      throw_symbol_error(get_name(), LOC);
+   if (value_stack.back().name_class != NC_VARIABLE)
+      Error::throw_symbol_error(get_name(), LOC);
 
-   return value_stack.last().apl_val;
+   return value_stack.back().apl_val;
 }
 //-----------------------------------------------------------------------------
 const Cell *
 Symbol::get_first_cell() const
 {
    Assert(value_stack.size() > 0);
-   if (value_stack.last().name_class != NC_VARIABLE)   return 0;
-   return &value_stack.last().apl_val->get_ravel(0);
+   if (value_stack.back().name_class != NC_VARIABLE)   return 0;
+   return &value_stack.back().apl_val->get_ravel(0);
 }
 //-----------------------------------------------------------------------------
 bool
 Symbol::can_be_assigned() const
 {
-   switch (value_stack.last().name_class)
+   switch (value_stack.back().name_class)
       {
         case NC_UNUSED_USER_NAME:
         case NC_VARIABLE:
@@ -522,26 +539,26 @@ Symbol::get_SV_key() const
 {
    Assert(value_stack.size() > 0);
 
-   if (value_stack.last().name_class != NC_SHARED_VAR)   return SV_key(0);
+   if (value_stack.back().name_class != NC_SHARED_VAR)   return SV_key(0);
 
-   return value_stack.last().sym_val.sv_key;
+   return value_stack.back().sym_val.sv_key;
 }
 //-----------------------------------------------------------------------------
 void
 Symbol::set_SV_key(SV_key key)
 {
-   value_stack.last().name_class = NC_SHARED_VAR;
-   value_stack.last().sym_val.sv_key = key;
+   value_stack.back().name_class = NC_SHARED_VAR;
+   value_stack.back().sym_val.sv_key = key;
 }
 //-----------------------------------------------------------------------------
 const Function *
 Symbol::get_function() const
 {
    Assert(value_stack.size() > 0);
-   switch(value_stack.last().name_class)
+   switch(value_stack.back().name_class)
       {
         case NC_FUNCTION:
-        case NC_OPERATOR: return value_stack.last().sym_val.function;
+        case NC_OPERATOR: return value_stack.back().sym_val.function;
         default:          return 0;
       }
 }
@@ -549,7 +566,7 @@ Symbol::get_function() const
 Function *
 Symbol::get_function()
 {
-const ValueStackItem & vs = value_stack.last();
+const ValueStackItem & vs = value_stack.back();
 
    if (vs.name_class == NC_FUNCTION)   return vs.sym_val.function;
    if (vs.name_class == NC_OPERATOR)   return vs.sym_val.function;
@@ -560,7 +577,7 @@ const ValueStackItem & vs = value_stack.last();
 void
 Symbol::get_attributes(int mode, Cell * dest) const
 {
-const ValueStackItem & vs = value_stack.last();
+const ValueStackItem & vs = value_stack.back();
 bool has_result = false;
 int fun_valence = 0;
 int oper_valence = 0;
@@ -657,11 +674,11 @@ Symbol::resolve(Token & tok, bool left_sym)
 
    Assert1(value_stack.size());
 
-const ValueStackItem & vs = value_stack.last();
+const ValueStackItem & vs = value_stack.back();
    switch(vs.name_class)
       {
         case NC_UNUSED_USER_NAME:
-             if (!left_sym)   throw_symbol_error(get_name(), LOC);
+             if (!left_sym)   Error::throw_symbol_error(get_name(), LOC);
              return;   // leave symbol as is
 
         case NC_LABEL:
@@ -716,16 +733,16 @@ Symbol::resolve_lv(const char * loc)
    Assert(value_stack.size());
 
    // if this is not a variable, then re-use the error handling of resolve().
-   if (value_stack.last().name_class != NC_VARIABLE)
+   if (value_stack.back().name_class != NC_VARIABLE)
       {
         CERR << "Symbol '" << get_name()
              << "' has changed type from variable to name class "
-             << value_stack.last().name_class << endl
+             << value_stack.back().name_class << endl
              << " while executing an assignment" << endl;
         throw_apl_error(E_LEFT_SYNTAX_ERROR, loc);
       }
 
-Value_P val = value_stack.last().apl_val;
+Value_P val = value_stack.back().apl_val;
    return Token(TOK_APL_VALUE1, val->get_cellrefs(loc));
 }
 //-----------------------------------------------------------------------------
@@ -734,7 +751,7 @@ Symbol::resolve_class(bool left)
 {
    Assert1(value_stack.size());
 
-   switch(value_stack.last().name_class)
+   switch(value_stack.back().name_class)
       {
         case NC_LABEL:
         case NC_VARIABLE:
@@ -743,7 +760,7 @@ Symbol::resolve_class(bool left)
 
         case NC_FUNCTION:
              {
-               const int valence = value_stack.last().sym_val.function
+               const int valence = value_stack.back().sym_val.function
                                  ->get_fun_valence();
                if (valence == 2)   return TC_FUN2;
                if (valence == 1)   return TC_FUN1;
@@ -752,7 +769,7 @@ Symbol::resolve_class(bool left)
 
         case NC_OPERATOR:
              {
-               const int valence = value_stack.last().sym_val.function
+               const int valence = value_stack.back().sym_val.function
                                  ->get_oper_valence();
                return (valence == 2) ? TC_OPER2 : TC_OPER1;
              }
@@ -766,7 +783,7 @@ Symbol::expunge()
 {
    if (value_stack.size() == 0)   return 1;   // empty stack
 
-ValueStackItem & vs = value_stack.last();
+ValueStackItem & vs = value_stack.back();
 
    if (vs.name_class == NC_VARIABLE)
       {
@@ -814,7 +831,7 @@ ValueStackItem & vs = value_stack.last();
 void
 Symbol::set_nc(NameClass nc)
 {
-ValueStackItem & vs = value_stack.last();
+ValueStackItem & vs = value_stack.back();
 
    if (vs.name_class == NC_UNUSED_USER_NAME)
       {
@@ -828,7 +845,7 @@ ValueStackItem & vs = value_stack.last();
 void
 Symbol::share_var(SV_key key)
 {
-ValueStackItem & vs = value_stack.last();
+ValueStackItem & vs = value_stack.back();
 
    if (vs.name_class == NC_UNUSED_USER_NAME)   // new shared variable
       {
@@ -862,7 +879,7 @@ Symbol::unshare_var()
 {
    if (value_stack.size() == 0)   return NO_COUPLING;
 
-ValueStackItem & vs = value_stack.last();
+ValueStackItem & vs = value_stack.back();
    if (vs.name_class != NC_SHARED_VAR)   return NO_COUPLING;
 
 const SV_key key = get_SV_key();
@@ -879,13 +896,13 @@ const SV_Coupling old_coupling = Svar_DB::get_coupling(key);
 void
 Symbol::set_nc(NameClass nc, Function * fun)
 {
-ValueStackItem & vs = value_stack.last();
+ValueStackItem & vs = value_stack.back();
 
 const bool can_set = (vs.name_class == NC_FUNCTION) ||
                      (vs.name_class == NC_OPERATOR) ||
                      (vs.name_class == NC_UNUSED_USER_NAME);
-             
-   Assert(nc == NC_FUNCTION || NC_OPERATOR || NC_UNUSED_USER_NAME);
+
+   Assert(nc == NC_FUNCTION || nc == NC_OPERATOR || nc == NC_UNUSED_USER_NAME);
 
    if (!can_set)   DEFN_ERROR;
    vs.sym_val.function = fun;
@@ -903,7 +920,7 @@ Symbol::list(ostream & out)
 
    if (is_erased())   out << "   ERASED";
    Assert(value_stack.size());
-const NameClass nc = value_stack.last().name_class;
+const NameClass nc = value_stack.back().name_class;
    if      (nc == NC_INVALID)            out << "   INVALID NC";
    else if (nc == NC_UNUSED_USER_NAME)   out << "   Unused";
    else if (nc == NC_LABEL)              out << "   Label";
@@ -948,7 +965,7 @@ UCS_string data;
 
                for (char * cp = buffer + strlen(buffer);
                     cp < (buffer + 72); )   *cp++ = ' ';
-                sprintf(buffer + 72, "%8.8lld\r\n", (long long)(seq++));
+                sprintf(buffer + 72, "%8.8lld\r\n", long_long(seq++));
                 fwrite(buffer, 1, 82, out);
 
                // write function record(s)
@@ -975,7 +992,7 @@ UCS_string data;
              buffer[1 + uu] = cc;
            }
 
-        sprintf(buffer + 72, "%8.8lld\r\n", (long long)(seq++));
+        sprintf(buffer + 72, "%8.8lld\r\n", long_long(seq++));
         fwrite(buffer, 1, 82, out);
       }
 }
@@ -1040,7 +1057,7 @@ int count = 0;
                         {
                           char cc[100];
                           snprintf(cc, sizeof(cc), "    VS[%lld] ",
-                                   (long long)v);
+                                   long_long(v));
                           count += ufun->show_owners(cc, out, value);
                         }
                    }
@@ -1054,12 +1071,11 @@ int count = 0;
 }
 //-----------------------------------------------------------------------------
 void
-Symbol::vector_assignment(Simple_string<Symbol *, false> & symbols,
-                          Value_P values)
+Symbol::vector_assignment(std::vector<Symbol *> & symbols, Value_P values)
 {
    if (values->get_rank() > 1)   RANK_ERROR;
    if (!values->is_scalar() &&
-       values->element_count() != symbols.size())   LENGTH_ERROR;
+       size_t(values->element_count()) != symbols.size())   LENGTH_ERROR;
 
 const int incr = values->is_scalar() ? 0 : 1;
 const Cell * cV = &values->get_ravel(0);
@@ -1149,7 +1165,7 @@ const ValueStackItem & vs = value_stack[0];
              //
              loop(l, ufun->local_var_count())
                  {
-                   out << ";" << ufun->get_local_var(l).get_name();
+                   out << ";" << ufun->get_local_var(l)->get_name();
                  }
              out << UNI_ASCII_R_CURLY << endl;
            }
@@ -1158,8 +1174,23 @@ const ValueStackItem & vs = value_stack[0];
              out << "∇";
              loop(u, text.size())
                 {
-                   out << text[u];
-                   if (text[u] == '\n' && u < text.size() - 1)   out << " ";
+                  const Unicode cc = text[u];
+                  if (cc != UNI_ASCII_SPACE)   out << cc;
+                  else   // blank: display unless trailing
+                     {
+                       for (int uu = u + 1; uu < text.size(); ++uu)
+                           {
+                             const Unicode cuu = text[uu];
+                              if (cuu == UNI_ASCII_SPACE)   continue;
+                              if (cuu != UNI_ASCII_LF)   out << cc;
+                              break;
+                           }
+                     }
+
+                   // indent the text lines with 1 blank so that the column
+                   // below ∇ is blank
+                   //
+                   if (cc == UNI_ASCII_LF && u < text.size() - 1)   out << " ";
                 }
              if (ufun->get_exec_properties()[0])   out << "⍫";
              else                                  out << "∇";
@@ -1168,6 +1199,34 @@ const ValueStackItem & vs = value_stack[0];
       }
 }
 //-----------------------------------------------------------------------------
+int
+Symbol::get_SI_level(const Function * fun) const
+{
+   loop(v, value_stack.size())
+       {
+         const ShapeItem from_tos = value_stack.size() - v - 1;
+         const ValueStackItem & item = value_stack[from_tos];
+         if (item.sym_val.function == fun)
+            return Workspace::SI_top()->nth_push(this, from_tos);
+       }
+
+   FIXME;
+}
+//-----------------------------------------------------------------------------
+int
+Symbol::get_SI_level(const Value * val) const
+{
+   loop(v, value_stack.size())
+       {
+         const ShapeItem from_tos = value_stack.size() - v - 1;
+         const ValueStackItem & item = value_stack[from_tos];
+         if (item.apl_val.get() == val)
+            return Workspace::SI_top()->nth_push(this, from_tos);
+       }
+
+   FIXME;
+}
+//-----------------------------------------------------------------------------
 void
 Symbol::clear_vs()
 {
@@ -1218,7 +1277,7 @@ CDR_string cdr;
    CDR::to_CDR(cdr, *new_value);
    if (cdr.size() > MAX_SVAR_SIZE)   LIMIT_ERROR_SVAR;
 
-string data((const char *)cdr.get_items(), cdr.size());
+string data(charP(cdr.get_items()), cdr.size());
 
    // wait for shared variable to be ready
    //
@@ -1244,7 +1303,7 @@ const bool ws_to_ws = Svar_DB::is_ws_to_ws(get_SV_key());
                    for (const uint32_t * varname =
                                          Svar_DB::get_svar_name(get_SV_key());
                         varname && *varname; ++varname)
-                       CERR << (Unicode)(*varname++);
+                       CERR << Unicode(*varname++);
                    CERR << " is blocked on set. Waiting ...";
                  }
             }
@@ -1335,7 +1394,7 @@ const bool ws_to_ws = Svar_DB::is_ws_to_ws(get_SV_key());
                    for (const uint32_t * varname =
                                          Svar_DB::get_svar_name(get_SV_key());
                         varname && *varname; ++varname)
-                       CERR << (Unicode)(*varname++);
+                       CERR << Unicode(*varname++);
                    CERR << " is blocked on use. Waiting ...";
                  }
             }
@@ -1377,7 +1436,7 @@ const TCP_socket tcp = Svar_DB::get_DB_tcp();
            }
 
         CDR_string cdr;
-        loop(d, data.size())   cdr.append(data[d]);
+        loop(d, data.size())   cdr.push_back(data[d]);
         delete response;
         if (del)   delete del;
 
@@ -1420,7 +1479,7 @@ const ErrorCode err(ErrorCode(response->get__VALUE_IS__error()));
 
 const string & data = response->get__VALUE_IS__cdr_value();
 CDR_string cdr;
-   loop(d, data.size())   cdr.append(data[d]);
+   loop(d, data.size())   cdr.push_back(data[d]);
    delete response;
    if (del)   delete del;
 
