@@ -33,7 +33,7 @@
 #include "SystemVariable.hh"
 #include "Tokenizer.hh"
 #include "UserFunction.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
@@ -59,7 +59,7 @@ const uint32_t hash = compute_hash(sym_name);
              CERR << "Symbol " << sym_name << " has hash " << HEX(hash) << endl;
            }
 
-        Symbol * new_symbol = new Symbol(sym_name, ID::USER_SYMBOL);
+        Symbol * new_symbol = new Symbol(sym_name, ID_USER_SYMBOL);
         symbol_table[hash] =  new_symbol;
         return new_symbol;
       }
@@ -79,11 +79,26 @@ const uint32_t hash = compute_hash(sym_name);
 
    // no symbol with name sym_name exists. The second walk:
    //
-Symbol * sym = new Symbol(sym_name, ID::USER_SYMBOL);
+Symbol * sym = new Symbol(sym_name, ID_USER_SYMBOL);
    add_symbol(sym);
    return sym;
 }
 //-----------------------------------------------------------------------------
+UCS_string
+SymbolTable::find_lambda_name(const UserFunction * lambda)
+{
+   loop(s, SYMBOL_HASH_TABLE_SIZE)
+       {
+         for (Symbol * sym = symbol_table[s]; sym; sym = sym->next)
+             {
+               if (sym->is_erased())   continue;
+               if (sym->get_ufun_depth(lambda) != -1)   return sym->get_name();
+             }
+       }
+
+   return UCS_string();
+}
+//-----------------------------------------------------------------------------
 ostream &
 SymbolTable::list_symbol(ostream & out, const UCS_string & buf1) const
 {
@@ -120,15 +135,16 @@ UCS_string to;
      const bool bad_from_to = Command::parse_from_to(from, to, from_to);
      if (bad_from_to)
         {
-          CERR << "bad range argument " << from_to
-               << ", expecting from - to" << endl;
+          CERR << "bad range argument" << endl;
+          MORE_ERROR() << "bad range argument " << from_to
+               << ", expecting from-to";
           return;
         }
    }
 
    // put those symbols into 'list' that satisfy 'which'
    //
-Simple_string<Symbol *, false> list;
+std::vector<Symbol *> list;
 int symbol_count = 0;
    loop(s, SYMBOL_HASH_TABLE_SIZE)
        {
@@ -154,16 +170,13 @@ int symbol_count = 0;
 
                if (sym->value_stack.size() == 0)
                   {
-                    if (which == LIST_ALL)
-                       {
-                         list.append(sym);
-                       }
+                    if (which == LIST_ALL)   list.push_back(sym);
                     continue;
                   }
 
                if (sym->is_erased() && !(which & LIST_ERASED))   continue;
 
-               const NameClass nc = sym->value_stack.last().name_class;
+               const NameClass nc = sym->value_stack.back().name_class;
                if (((nc == NC_VARIABLE)         && (which & LIST_VARS))    ||
                    ((nc == NC_FUNCTION)         && (which & LIST_FUNS))    ||
                    ((nc == NC_OPERATOR)         && (which & LIST_OPERS))   ||
@@ -172,7 +185,7 @@ int symbol_count = 0;
                    ((nc == NC_INVALID)          && (which & LIST_INVALID)) ||
                    ((nc == NC_UNUSED_USER_NAME) && (which & LIST_UNUSED)))
                    {
-                     list.append(sym);
+                     list.push_back(sym);
                    }
              }
        }
@@ -197,9 +210,9 @@ UCS_string_vector names;
         if (which == LIST_NAMES)   // append .NC
            {
              name.append(UNI_ASCII_FULLSTOP);
-             name.append_number(list[l]->value_stack.last().name_class);
+             name.append_number(list[l]->value_stack.back().name_class);
            }
-        names.append(name);
+        names.push_back(name);
       }
 
    names.sort();
@@ -207,13 +220,14 @@ UCS_string_vector names;
    // figure column widths
    //
    enum { tabsize = 4 };
-Simple_string<int, false> col_width = names.compute_column_width(tabsize);
+std::vector<int> col_widths;
+   names.compute_column_width(tabsize, col_widths);
 
    loop(c, count)
       {
-        const int col = c % col_width.size();
+        const size_t col = c % col_widths.size();
         out << names[c];
-        if (col == (col_width.size() - 1) || c == (count - 1))
+        if (col == (col_widths.size() - 1) || c == (count - 1))
            {
              // last column or last item: print newline
              //
@@ -223,7 +237,7 @@ Simple_string<int, false> col_width = names.compute_column_width(tabsize);
            {
              // intermediate column: print spaces
              //
-             const int len = tabsize*col_width[col] - names[c].size();
+             const int len = tabsize*col_widths[col] - names[c].size();
              Assert(len > 0);
              loop(l, len)   out << " ";
            }
@@ -418,17 +432,17 @@ ValueStackItem & tos = symbol->value_stack[0];
    return true;
 }
 //-----------------------------------------------------------------------------
-Simple_string<const Symbol *, false>
+std::vector<const Symbol *>
 SymbolTable::get_all_symbols() const
 {
-Simple_string<const Symbol *, false> ret;
+std::vector<const Symbol *> ret;
    ret.reserve(1000);
 
    loop(hash, SYMBOL_HASH_TABLE_SIZE)
       {
         for (const Symbol * sym = symbol_table[hash]; sym; sym = sym->next)
             {
-              ret.append(sym);
+              ret.push_back(sym);
             }
       }
 
@@ -438,14 +452,14 @@ Simple_string<const Symbol *, false> ret;
 void
 SymbolTable::dump(ostream & out, int & fcount, int & vcount) const
 {
-Simple_string<const Symbol *, false> symbols;
+std::vector<const Symbol *> symbols;
    loop(hash, SYMBOL_HASH_TABLE_SIZE)
       {
         for (const Symbol * sym = symbol_table[hash]; sym; sym = sym->next)
             {
               if (sym->is_erased())              continue;
               if (sym->value_stack_size() < 1)   continue;
-              symbols.append(sym);
+              symbols.push_back(sym);
             }
       }
 
@@ -453,7 +467,7 @@ Simple_string<const Symbol *, false> symbols;
    //
    loop(d, symbols.size())
       {
-        for (ShapeItem j = d + 1; j < symbols.size(); ++j)
+        for (ShapeItem j = d + 1; j < ShapeItem(symbols.size()); ++j)
             {
               if (symbols[d]->get_name().compare(symbols[j]->get_name()) > 0)
                  {
@@ -512,7 +526,7 @@ SystemName * sym = symbol_table[hash];
 }
 //-----------------------------------------------------------------------------
 void
-SystemSymTab::add_fun_or_var(const UCS_string & name, ID::Id id,
+SystemSymTab::add_fun_or_var(const UCS_string & name, Id id,
                        QuadFunction * function, SystemVariable * variable)
 {
    // name should not yet exist
