@@ -46,7 +46,7 @@
 #include "SystemVariable.hh"
 #include "UserFunction.hh"
 #include "UserPreferences.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "Workspace.hh"
 
 UCS_string Quad_QUOTE::prompt;
@@ -126,7 +126,7 @@ SystemVariable::get_attributes(int mode, Cell * dest) const
 }
 //=============================================================================
 Quad_AV::Quad_AV()
-   : RO_SystemVariable(ID::Quad_AV)
+   : RO_SystemVariable(ID_Quad_AV)
 {
    Assert1(MAX_AV == 256);
 Value_P AV(MAX_AV, LOC);
@@ -151,7 +151,7 @@ Quad_AV::indexed_at(uint32_t pos)
 }
 //=============================================================================
 Quad_AI::Quad_AI()
-   : RO_SystemVariable(ID::Quad_AI),
+   : RO_SystemVariable(ID_Quad_AI),
      session_start(now()),
      user_wait(0)
 {
@@ -176,7 +176,7 @@ Value_P ret(5, LOC);
 }
 //=============================================================================
 Quad_ARG::Quad_ARG()
-   : RO_SystemVariable(ID::Quad_ARG)
+   : RO_SystemVariable(ID_Quad_ARG)
 {
 }
 //-----------------------------------------------------------------------------
@@ -192,7 +192,7 @@ Value_P Z(argc, LOC);
         UTF8_string utf(arg);
         UCS_string ucs(utf);
         Value_P sub(ucs, LOC);
-        new (Z->next_ravel())   PointerCell(sub, Z.getref());
+        new (Z->next_ravel())   PointerCell(sub.get(), Z.getref());
       }
 
    Z->check_value(LOC);
@@ -200,7 +200,7 @@ Value_P Z(argc, LOC);
 }
 //=============================================================================
 Quad_CT::Quad_CT()
-   : SystemVariable(ID::Quad_CT)
+   : SystemVariable(ID_Quad_CT)
 {
    Symbol::assign(FloatScalar(DEFAULT_Quad_CT, LOC), false, LOC);
 }
@@ -241,9 +241,9 @@ const Error * err = 0;
 
    for (StateIndicator * si = Workspace::SI_top(); si; si = si->get_parent())
        {
-         if (si->get_error().error_code != E_NO_ERROR)
+         if (StateIndicator::get_error(si).get_error_code() != E_NO_ERROR)
             {
-              err = &si->get_error();
+              err = &StateIndicator::get_error(si);
               break;
             }
          if (si->get_parse_mode() == PM_FUNCTION)   break;
@@ -253,7 +253,7 @@ const Error * err = 0;
       {
         // return 3 0⍴' '
         //
-        Shape sh((ShapeItem)3, (ShapeItem)0);
+        Shape sh(ShapeItem(3), ShapeItem(0));
         Value_P Z(sh, LOC);
         Z->set_proto_Spc();
         Z->check_value(LOC);
@@ -261,8 +261,8 @@ const Error * err = 0;
       }
 
 PrintBuffer pb;
-   pb.append_ucs(err->get_error_line_1());
-   pb.append_ucs(err->get_error_line_2());
+   pb.append_ucs(UCS_string(UTF8_string(err->get_error_line_1())));
+   pb.append_ucs(UCS_string(UTF8_string(err->get_error_line_2())));
    pb.append_ucs(err->get_error_line_3());
    pb.pad_to_spaces();   // replace pad char by spaces
 Value_P Z(pb, LOC);
@@ -273,30 +273,31 @@ Value_P Z(pb, LOC);
 Value_P
 Quad_ET::get_apl_value() const
 {
-StateIndicator * si = Workspace::SI_top();
+Value_P Z(2, LOC);
 ErrorCode ec = E_NO_ERROR;
 
-   for (; si; si = si->get_parent())
+   for (StateIndicator * si = Workspace::SI_top(); si; si = si->get_parent())
        {
-         ec = si->get_error().error_code;
+         ec = StateIndicator::get_error(si).get_error_code();
          if (ec != E_NO_ERROR)
             {
-              break;
+              new (Z->next_ravel()) IntCell(Error::error_major(ec));
+              new (Z->next_ravel()) IntCell(Error::error_minor(ec));
+              goto done;
             }
 
          if (si->get_parse_mode() == PM_FUNCTION)   break;
        }
 
-Value_P Z(2, LOC);
-
-   new (Z->next_ravel()) IntCell(Error::error_major(ec));
-   new (Z->next_ravel()) IntCell(Error::error_minor(ec));
+   new (Z->next_ravel()) IntCell(Error::error_major(E_NO_ERROR));
+   new (Z->next_ravel()) IntCell(Error::error_minor(E_NO_ERROR));
 
+done:
    Z->check_value(LOC);
    return Z;
 }
 //=============================================================================
-Quad_FC::Quad_FC() : SystemVariable(ID::Quad_FC)
+Quad_FC::Quad_FC() : SystemVariable(ID_Quad_FC)
 {
 Value_P QFC(6, LOC);
    new (QFC->next_ravel()) CharCell(UNI_ASCII_FULLSTOP);
@@ -408,7 +409,7 @@ Value_P new_val(ucs, LOC);
 }
 //=============================================================================
 Quad_IO::Quad_IO()
-   : SystemVariable(ID::Quad_IO)
+   : SystemVariable(ID_Quad_IO)
 {
    Symbol::assign(IntScalar(1, LOC), false, LOC);
 }
@@ -429,7 +430,7 @@ Quad_IO::assign(Value_P value, bool clone, const char * loc)
 }
 //=============================================================================
 Quad_L::Quad_L()
- : NL_SystemVariable(ID::Quad_L)
+ : NL_SystemVariable(ID_Quad_L)
 {
    Symbol::assign(IntScalar(0, LOC), false, LOC);
 }
@@ -442,7 +443,7 @@ StateIndicator * si = Workspace::SI_top_fun();
 
    // ignore assignments if error was SYNTAX ERROR or VALUE ERROR
    //
-   if (si->get_error().is_syntax_or_value_error())   return;
+   if (StateIndicator::get_error(si).is_syntax_or_value_error())   return;
 
    si->set_L(value);
 }
@@ -461,7 +462,7 @@ StateIndicator * si = Workspace::SI_top_error();
 }
 //=============================================================================
 Quad_LC::Quad_LC()
-   : RO_SystemVariable(ID::Quad_LC)
+   : RO_SystemVariable(ID_Quad_LC)
 {
    Symbol::assign(IntScalar(0, LOC), false, LOC);
 }
@@ -493,7 +494,7 @@ Value_P Z(len, LOC);
 }
 //=============================================================================
 Quad_LX::Quad_LX()
-   : NL_SystemVariable(ID::Quad_LX)
+   : NL_SystemVariable(ID_Quad_LX)
 {
    Symbol::assign(Str0(LOC), false, LOC);
 }
@@ -508,7 +509,7 @@ Quad_LX::assign(Value_P value, bool clone, const char * loc)
 }
 //=============================================================================
 Quad_PP::Quad_PP()
-   : SystemVariable(ID::Quad_PP)
+   : SystemVariable(ID_Quad_PP)
 {
 Value_P value(LOC);
 
@@ -529,7 +530,7 @@ APL_Integer pp = value->get_sole_integer();
 }
 //=============================================================================
 Quad_PR::Quad_PR()
-   : SystemVariable(ID::Quad_PR)
+   : SystemVariable(ID_Quad_PR)
 {
    Symbol::assign(CharScalar(UNI_ASCII_SPACE, LOC), false, LOC);
 }
@@ -545,27 +546,140 @@ UCS_string ucs = value->get_UCS_ravel();
 }
 //=============================================================================
 Quad_PS::Quad_PS()
-   : SystemVariable(ID::Quad_PS)
+   : SystemVariable(ID_Quad_PS),
+     print_quotients(false),
+     style(Command::boxing_format)
 {
-   Symbol::assign(IntScalar(0, LOC), false, LOC);
+Value_P val(2, LOC);
+   new (val->next_ravel())   IntCell(print_quotients);
+   new (val->next_ravel())   IntCell(style);
+   val->check_value(LOC);
+   Symbol::assign(val, false, LOC);
 }
 //-----------------------------------------------------------------------------
 void
-Quad_PS::assign(Value_P value, bool clone, const char * loc)
+Quad_PS::assign(Value_P B, bool clone, const char * loc)
 {
-const APL_Integer ps = value->get_sole_integer();
-   switch(ps)
+APL_Integer B_quot  = 0;
+APL_Integer B_style = 0;
+
+   if (B->get_rank() > 1)        RANK_ERROR;
+
+   if (B->element_count() < 1)   LENGTH_ERROR;
+   if (B->element_count() > 2)   LENGTH_ERROR;
+
+   if (!B->get_ravel(0).is_near_bool())
       {
-        case 0:
-        case 1:
-        case 2:
-        case 3:  Symbol::assign(IntScalar(ps, LOC), false, LOC);   return;
-        default: DOMAIN_ERROR;
+        MORE_ERROR() << "Bad quot in ⎕PS←quot style: quot is not near bool";
+        DOMAIN_ERROR;
       }
+
+   if (B->element_count() == 1)
+      {
+        // for compatibility with old workspaces
+        //
+        B_style = B->get_ravel(0).get_near_int();
+      }
+   else
+      {
+        B_quot  = B->get_ravel(0).get_near_bool();
+        B_style = B->get_ravel(1).get_near_int();
+      }
+
+   switch(B_style) // boxing format
+      {
+        case -29:
+        case -25: case -24: case -23:
+        case -22: case -21: case -20:
+        case -9: case  -8: case  -7:
+        case -4: case  -3: case  -2:
+        case  0:
+        case  2: case   3: case   4:
+        case  7: case   8: case   9:
+        case 20: case  21: case  22:
+        case 23: case  24: case  25:
+        case 29: break;   // OK
+
+        default: MORE_ERROR() <<
+                 "Invalid style in ⎕PS←quot style or ⎕PS←style:\n"
+                 "style is not ± 0, 2-4, 7-9, 20-25, or 29";
+                  DOMAIN_ERROR;
+      }
+
+   // values in B are valid
+   //
+   Command::boxing_format = B_style;
+   print_quotients = B_quot;
+   style = B_style;
+
+Value_P B2(2, LOC);
+   new (B2->next_ravel())   IntCell(B_quot);
+   new (B2->next_ravel())   IntCell(B_style);
+   B2->check_value(LOC);
+
+   Symbol::assign(B2, false, LOC);
+   return;
+}
+//-----------------------------------------------------------------------------
+void
+Quad_PS::assign_indexed(Value_P X, Value_P B)
+{
+   if (!(X->is_int_scalar() || X->is_int_vector()))   INDEX_ERROR;
+   if (!(B->is_int_scalar() || B->is_int_vector()))   DOMAIN_ERROR;
+   if (X->element_count() != B->element_count())      LENGTH_ERROR;
+
+const ShapeItem ec = X->element_count();
+const APL_Integer qio = Workspace::get_IO();
+
+APL_Integer Z_quot = print_quotients;
+APL_Integer Z_style = style;
+   loop(e, ec)
+      {
+        const APL_Integer x = X->get_ravel(e).get_near_int() - qio;
+        const APL_Integer b = B->get_ravel(e).get_near_int();
+
+        if (x == 0)   // display quotients
+           {
+             if (b < 0)   DOMAIN_ERROR;
+             if (b > 1)   DOMAIN_ERROR;
+             Z_quot = b;
+           }
+        else if (x == 1)   switch(b) // boxing format
+           {
+             case -29:
+             case -25: case -24: case -23:
+             case -22: case -21: case -20:
+             case -9: case  -8: case  -7:
+             case -4: case  -3: case  -2:
+             case  0:
+             case  2: case   3: case   4:
+             case  7: case   8: case   9:
+             case 20: case  21: case  22:
+             case 23: case  24: case  25:
+             case 29: Z_style = b;
+                  Command::boxing_format = b;
+                  break;
+ 
+             default:
+                  MORE_ERROR() <<
+                       "Invalid print style (try 0, 2-4, 7-9, 20-25, or 29)";
+                  DOMAIN_ERROR;
+           }
+        else
+           {
+             INDEX_ERROR;
+           }
+      }
+
+Value_P Z(2, LOC);
+   new (Z->next_ravel())   IntCell(Z_quot);
+   new (Z->next_ravel())   IntCell(Z_style);
+   Z->check_value(LOC);
+   Symbol::assign(Z, false, LOC);
 }
 //=============================================================================
 Quad_PW::Quad_PW()
-   : SystemVariable(ID::Quad_PW)
+   : SystemVariable(ID_Quad_PW)
 {
    Symbol::assign(IntScalar(uprefs.initial_pw, LOC), false, LOC);
 }
@@ -585,7 +699,7 @@ const APL_Integer pw = value->get_sole_integer();
 }
 //=============================================================================
 Quad_Quad::Quad_Quad()
- : SystemVariable(ID::Quad_Quad)
+ : SystemVariable(ID_Quad_Quad)
 {
 }
 //-----------------------------------------------------------------------------
@@ -619,7 +733,7 @@ UCS_string line;
 }
 //=============================================================================
 Quad_QUOTE::Quad_QUOTE()
- : SystemVariable(ID::QUOTE_Quad)
+ : SystemVariable(ID_QUOTE_Quad)
 {
    // we assign a dummy value so that ⍞ is not undefined.
    //
@@ -638,7 +752,7 @@ Quad_QUOTE::done(bool with_LF, const char * loc)
    if (prompt.size())
       {
         if (with_LF)   COUT << endl;
-        prompt.shrink(0);
+        prompt.clear();
       }
 }
 //-----------------------------------------------------------------------------
@@ -714,7 +828,7 @@ Value_P Z(line, LOC);
 }
 //=============================================================================
 Quad_R::Quad_R()
- : NL_SystemVariable(ID::Quad_R)
+ : NL_SystemVariable(ID_Quad_R)
 {
    Symbol::assign(IntScalar(0, LOC), false, LOC);
 }
@@ -727,7 +841,7 @@ StateIndicator * si = Workspace::SI_top_fun();
 
    // ignore assignments if error was SYNTAX ERROR or VALUE ERROR
    //
-   if (si->get_error().is_syntax_or_value_error())   return;
+   if (StateIndicator::get_error(si).is_syntax_or_value_error())   return;
 
    si->set_R(value);
 }
@@ -805,14 +919,14 @@ const APL_Integer qio = Workspace::get_IO();
            }
         else if (x == SYL_VALUE_COUNT_LIMIT)   // value count limit
            {
-             if (b && (b < (Value::value_count + 10)))    // too low
+             if (b && (b < APL_Integer(Value::value_count + 10)))    // too low
                 DOMAIN_ERROR;
              value_count_limit = b;
            }
         else if (x == SYL_RAVEL_BYTES_LIMIT)   // ravel bytes limit
            {
              const int cells = b / sizeof(Cell);
-             if (cells && (cells < (Value::total_ravel_count + 1000)))
+             if (cells && (cells < APL_Integer(Value::total_ravel_count + 1000)))
                 DOMAIN_ERROR;
 
              ravel_count_limit = cells;
@@ -820,7 +934,8 @@ const APL_Integer qio = Workspace::get_IO();
         else if (x == SYL_CURRENT_CORES)   // number of cores
            {
 #if PARALLEL_ENABLED
-             if (Parallel::set_core_count((CoreCount)b, false))   DOMAIN_ERROR;
+             if (Parallel::set_core_count(CoreCount(b), false))
+                DOMAIN_ERROR;
 #else
              DOMAIN_ERROR;
 #endif
@@ -830,6 +945,17 @@ const APL_Integer qio = Workspace::get_IO();
              if (b < 1000)   DOMAIN_ERROR;
              print_length_limit = b;
            }
+        else if (x == SYL_WA_MARGIN)   // ⎕WA safety margin
+           {
+             if (b < 1000000)   DOMAIN_ERROR;
+             Quad_WA::WA_margin = b;
+           }
+        else if (x == SYL_WA_SCALE)   // ⎕WA memory scale (45% - 200% %
+           {
+             if (b <  45)   DOMAIN_ERROR;
+             if (b > 200)   DOMAIN_ERROR;
+             Quad_WA::WA_scale = b;
+           }
         else
            {
              INDEX_ERROR;
@@ -846,7 +972,8 @@ Value_P Z(sh, LOC);
 #define syl2(n, e, v) syl1(n, e, v)
 #define syl3(n, e, v) syl1(n, e, v)
 #define syl1(n, _e, v) \
-  new (Z->next_ravel()) PointerCell(Value_P(UCS_string(UTF8_string(n)), LOC), Z.getref()); \
+  new (Z->next_ravel()) \
+      PointerCell(Value_P(UCS_string(UTF8_string(n)), LOC).get(), Z.getref()); \
   new (Z->next_ravel()) IntCell(v);
 #include "SystemLimits.def"
 
@@ -855,7 +982,7 @@ Value_P Z(sh, LOC);
 }
 //=============================================================================
 Quad_TC::Quad_TC()
-   : RO_SystemVariable(ID::Quad_TC)
+   : RO_SystemVariable(ID_Quad_TC)
 {
 Value_P QCT(3, LOC);
    new (QCT->next_ravel()) CharCell(UNI_ASCII_BS);
@@ -867,7 +994,7 @@ Value_P QCT(3, LOC);
 }
 //=============================================================================
 Quad_TS::Quad_TS()
-   : RO_SystemVariable(ID::Quad_TS)
+   : RO_SystemVariable(ID_Quad_TS)
 {
    Symbol::assign(IntScalar(0, LOC), false, LOC);
 }
@@ -892,7 +1019,7 @@ Value_P Z(7, LOC);
 }
 //=============================================================================
 Quad_TZ::Quad_TZ()
-   : SystemVariable(ID::Quad_TZ)
+   : SystemVariable(ID_Quad_TZ)
 {
    offset_seconds = compute_offset();
 
@@ -934,6 +1061,25 @@ int diff_minutes = local_minutes - gm_minutes;
    return 60*diff_minutes;
 }
 //-----------------------------------------------------------------------------
+ostream &
+Quad_TZ::print_timestamp(ostream & out, APL_time_us when) const
+{
+const APL_time_us offset = get_offset();
+const YMDhmsu time(when + 1000000*offset);
+const char * tz_sign = (offset < 0) ? "" : "+";
+
+ostringstream os;
+   os << setfill('0') << time.year  << "-"
+      << setw(2)      << time.month << "-"
+      << setw(2)      << time.day   << "  "
+      << setw(2)      << time.hour  << ":"
+      << setw(2)      << time.minute << ":"
+      << setw(2)      << time.second << " (GMT"
+      << tz_sign      << offset/3600 << ")";
+
+   return out << os.str();
+}
+//-----------------------------------------------------------------------------
 void
 Quad_TZ::assign(Value_P value, bool clone, const char * loc)
 {
@@ -954,10 +1100,10 @@ Cell & cell = value->get_ravel(0);
 
    if (cell.is_float_cell())
       {
-        const APL_Float fval = cell.get_real_value();
-        if (fval < -12.1)   return;
-        if (fval > 14.1)    return;
-        offset_seconds = int(0.5 + fval*3600);
+        const double hours = cell.get_real_value();
+        if (hours < -12.1)   return;
+        if (hours > 14.1)    return;
+        offset_seconds = int(0.5 + hours*3600);
         Symbol::assign(value, clone, LOC);
         return;
       }
@@ -966,7 +1112,7 @@ Cell & cell = value->get_ravel(0);
 }
 //=============================================================================
 Quad_UL::Quad_UL()
-   : RO_SystemVariable(ID::Quad_UL)
+   : RO_SystemVariable(ID_Quad_UL)
 {
    Symbol::assign(get_apl_value(), false, LOC);
 }
@@ -998,77 +1144,8 @@ Value_P Z(LOC);
    return Z;
 }
 //=============================================================================
-Quad_WA::Quad_WA()
-   : RO_SystemVariable(ID::Quad_WA)
-{
-   Symbol::assign(IntScalar(0, LOC), false, LOC);
-}
-//-----------------------------------------------------------------------------
-Value_P
-Quad_WA::get_apl_value() const
-{
-Value_P Z(LOC);
-
-uint64_t total = total_memory;   // max memory as reported by RLIM_INFINITY
-uint64_t proc_mem = 0;            // memory as reported proc/mem_info
-
-   {
-     FILE * pm = fopen("/proc/meminfo", "r");
-
-     if (pm)   // /proc/meminfo exists
-        {
-          for (;;)
-              {
-                char buffer[2000];
-                if (fgets(buffer, sizeof(buffer) - 1, pm) == 0)   break;
-                buffer[sizeof(buffer) - 1] = 0;
-                if (!strncmp(buffer, "MemFree:", 8))
-                   {
-                     proc_mem = atoi(buffer + 8);
-                     proc_mem *= 1024;
-                     break;
-                   }
-              }
-
-          fclose(pm);
-        }
-   }
-
-   if (proc_mem == 0)   // nothing from /proc/meminfo
-      {
-        if (total == RLIM_INFINITY)   // no rlimit
-           {
-             CERR << "Cannot properly determine ⎕WA (process has no memory"
-                     " limit and /proc/meminfo provided no info)" << endl;
-             total = 1000000;
-           }
-        else                          // process has rlimit
-           {
-             total -= Value::total_ravel_count * sizeof(Cell)
-                    + Value::value_count * sizeof(Value);
-           }
-      }
-   else
-      {
-        if (total == RLIM_INFINITY)   // no rlimit
-           {
-             total = proc_mem;   // use value from .proc/meminfo
-           }
-        else                          // process has rlimit
-           {
-             total -= Value::total_ravel_count * sizeof(Cell)
-                    + Value::value_count * sizeof(Value);
-             if (total > proc_mem)   total = proc_mem;   // use minimum
-           }
-      }
-
-   new (Z->next_ravel())   IntCell(total);
-   Z->check_value(LOC);
-   return Z;
-}
-//=============================================================================
 Quad_X::Quad_X()
- : NL_SystemVariable(ID::Quad_X)
+ : NL_SystemVariable(ID_Quad_X)
 {
    Symbol::assign(IntScalar(0, LOC), false, LOC);
 }
@@ -1081,7 +1158,7 @@ StateIndicator * si = Workspace::SI_top_fun();
 
    // ignore assignments if error was SYNTAX ERROR or VALUE ERROR
    //
-   if (si->get_error().is_syntax_or_value_error())   return;
+   if (StateIndicator::get_error(si).is_syntax_or_value_error())   return;
 
    si->set_X(value);
 }
