@@ -152,7 +152,7 @@ operator << (ostream & out, const Token & token)
              UCS_string ucs = pb.get_line(0).no_pad();
              if (ucs.size() > 20)
                 {
-                  ucs.shrink(20);
+                  ucs.resize(20);
                   more = true;
                 }
              out << ucs;
@@ -176,7 +176,7 @@ operator << (ostream & out, const Token & token)
         return out << token.get_Id() << token.get_int_val();
       }
 
-   if (token.get_Id() > ID::No_ID2)   return out << token.get_Id();
+   if (token.get_Id() > ID_No_ID2)   return out << token.get_Id();
 
    switch(token.get_Class())
       {
@@ -239,7 +239,7 @@ Token::ChangeTag(TokenTag new_tag)
 {
    Assert((tag & TV_MASK) == (new_tag & TV_MASK));
    // tag is ia const TokenTag, so we cheat a little here.
-   (TokenTag &)tag = new_tag;
+   const_cast<TokenTag &>(tag) = new_tag;
 }
 //-----------------------------------------------------------------------------
 int
@@ -303,13 +303,17 @@ Token::print_function(ostream & out) const
         case TOK_F2_EQUAL:      return out << UNI_ASCII_EQUAL;
         case TOK_F2_GREATER:    return out << UNI_ASCII_GREATER;
         case TOK_F2_AND:        return out << UNI_AND;
+        case TOK_F2_AND_B:      return out << UNI_AND << UNI_AND;
         case TOK_F2_OR:         return out << UNI_OR;
+        case TOK_F2_OR_B:       return out << UNI_OR << UNI_OR;
         case TOK_F2_INDEX:      return out << UNI_SQUISH_Quad;
         case TOK_F2_LEQ:        return out << UNI_LESS_OR_EQUAL;
         case TOK_F2_MEQ:        return out << UNI_MORE_OR_EQUAL;
         case TOK_F2_UNEQ:       return out << UNI_NOT_EQUAL;
         case TOK_F2_NOR:        return out << UNI_NOR;
+        case TOK_F2_NOR_B:      return out << UNI_NOR << UNI_NOR;
         case TOK_F2_NAND:       return out << UNI_NAND;
+        case TOK_F2_NAND_B:     return out << UNI_NAND << UNI_NAND;
 
         case TOK_F12_BINOM:     return out << UNI_ASCII_EXCLAM;
         case TOK_F12_CIRCLE:    return out << UNI_CIRCLE;
@@ -374,7 +378,7 @@ Token::print_function(ostream & out) const
 
    out << "{ unknown function Token " << tag;
 
-   if (get_Id() != ID::No_ID)   // tag has an Id
+   if (get_Id() != ID_No_ID)   // tag has an Id
       {
         out << ", Id = " << get_Id();
       }
@@ -449,7 +453,7 @@ Token::print_value(ostream & out) const
 
    out << "{ unknown value Token " << tag;
 
-   if (get_Id() != ID::No_ID)   // tag has an Id
+   if (get_Id() != ID_No_ID)   // tag has an Id
       {
         out << ", Id = " << get_Id();
       }
@@ -462,9 +466,9 @@ Token::show_trace(ostream & out, const UCS_string & fun_name,
                   Function_Line line) const
 {
 UCS_string fn = fun_name;
-   fn.append_utf8("[");
+   fn.append_UTF8("[");
    fn.append_number(line);
-   fn.append_utf8("] ");
+   fn.append_UTF8("] ");
 
    out << fn;
 
@@ -517,7 +521,7 @@ const Value & val = *get_apl_val();
       }
    else                  // matrix or higher
       {
-        pctx.set_style((PrintStyle)(pctx.get_style() | PST_NO_FRACT_0));
+        pctx.set_style(PrintStyle(pctx.get_style() | PST_NO_FRACT_0));
       }
 
 PrintBuffer pb(val, pctx, 0);
@@ -591,19 +595,19 @@ UCS_string ucs;
         case TC_L_PARENT:
         case TC_R_CURLY:
         case TC_L_CURLY:
-             return ID::get_name(get_Id());
+             return ID::get_name_UCS(get_Id());
 
         case TC_FUN0:
         case TC_FUN12:
         case TC_OPER1:
         case TC_OPER2:
-             if (get_Id() == ID::No_ID)   return get_function()->get_name();
-             return ID::get_name(get_Id());
+             if (get_Id() == ID_No_ID)   return get_function()->get_name();
+             return ID::get_name_UCS(get_Id());
 
         default:
              CERR << "Token: " << HEX4(tag) << " " << *this
                   << " at " << LOC << endl;
-             Q1((void *)get_Class())
+             Q1(get_Class())
              Backtrace::show(__FILE__, __LINE__);
              FIXME;
       }
@@ -687,19 +691,12 @@ Token::short_class_name(TokenClass cls)
    return "???";
 }
 //-----------------------------------------------------------------------------
-void
-Token::print_token_list(ostream & out, const Source<Token, true> & src)
-{
-const int len = src.rest();
-   loop(t, len)   out << "`" << src[t] << "  ";
-   out << endl;
-}
-//-----------------------------------------------------------------------------
 ostream &
 operator << (ostream & out, const Token_string & tos)
 {
    out << "[" << tos.size() << " token]: ";
-   Token::print_token_list(out, tos);
+   loop(t, tos.size())   CERR << "`" << tos[t] << "  ";
+   out << endl;
    out << endl;
    return out;
 }
@@ -749,7 +746,7 @@ Token::class_name(TokenClass tc)
    return "*** Obscure token class ***";
 }
 //-----------------------------------------------------------------------------
-inline void 
+inline void
 Token::copy_N(const Token & src)
 {
    tag = src.tag;
@@ -804,7 +801,7 @@ Token::move_1(Token & src, const char * loc)
         const Value * val = src.value.apl_val.get();
         if (val)
            {
-	     ADD_EVENT(val, VHE_TokMove1, src.value_use_count() - 1, loc);
+             ADD_EVENT(val, VHE_TokMove1, src.value_use_count() - 1, loc);
              src.clear(loc);
            }
       }
@@ -822,7 +819,7 @@ Token::move_2(const Token & src, const char * loc)
         if (val)
            {
              ADD_EVENT(val, VHE_TokMove2, src.value_use_count() - 1, loc);
-             ((Token &)src).clear(loc);
+             const_cast<Token &>(src).clear(loc);
            }
       }
 }
@@ -834,24 +831,24 @@ Token * t1 = &at(from);
 Token * t2 = &at(to);
    Assert(0 <= from);
    Assert(from <= to);
-   Assert(to <= size());
+   Assert(to <= ShapeItem(size()));
 
-   for (;t1 < t2; ++t1, --t2)
-       {
-         char tt[sizeof(Token)];
-         memcpy(tt, t1, sizeof(Token));
-         memcpy(t1, t2,   sizeof(Token));
-         memcpy(t2, tt, sizeof(Token));
-       }
+   while (t1 < t2)   t1++->Hswap(*t2--);
 }
 //-----------------------------------------------------------------------------
 void
-Token_string::print(ostream & out) const
+Token_string::print(ostream & out, bool details) const
 {
    loop(t, size())
        {
          const Token & tok = at(t);
-         out << "`" << tok << "  ";
+         out << "`" << tok;
+         if (details)   switch(tok.get_ValueType())
+            {
+              case TV_FLT: out << " " << tok.get_flt_val();   break;
+              default:                                        break;
+            }
+         out << "  ";
        }
 
    out << endl;
