@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2016  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2019  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,7 +20,9 @@
 
 #include <math.h>
 #include <string.h>
+#include <vector>
 
+#include "Backtrace.hh"
 #include "Common.hh"
 #include "FloatCell.hh"
 #include "Heapsort.hh"
@@ -29,7 +31,7 @@
 #include "PrintOperator.hh"
 #include "UCS_string.hh"
 #include "UTF8_string.hh"
-#include "Value.icc"
+#include "Value.hh"
 
 ShapeItem UCS_string::total_count = 0;
 ShapeItem UCS_string::total_id = 0;
@@ -41,44 +43,53 @@ UCS_string::UCS_string()
 }
 //-----------------------------------------------------------------------------
 UCS_string::UCS_string(Unicode uni)
-   : Simple_string<Unicode, false>(1, uni)
+   : basic_string<Unicode>(1, uni)
 {
   create(LOC);
 }
 //-----------------------------------------------------------------------------
 UCS_string::UCS_string(const Unicode * data, size_t len)
-   : Simple_string<Unicode, false>(data, len)
+   : basic_string<Unicode>(data, len)
 {
    create(LOC);
 }
 //-----------------------------------------------------------------------------
 UCS_string::UCS_string(size_t len, Unicode uni)
-   : Simple_string<Unicode, false>(len, uni)
+   : basic_string<Unicode>(len, uni)
 {
    create(LOC);
 }
 //-----------------------------------------------------------------------------
 UCS_string::UCS_string(const UCS_string & ucs)
-   : Simple_string<Unicode, false>(ucs)
+   : basic_string<Unicode>(ucs)
 {
    create(LOC);
 }
 //-----------------------------------------------------------------------------
 UCS_string::UCS_string(const UCS_string & ucs, size_t pos, size_t len)
-   : Simple_string<Unicode, false>(ucs, pos, len)
+   : basic_string<Unicode>(ucs, pos, len)
 {
    create(LOC);
 }
 //-----------------------------------------------------------------------------
 UCS_string::UCS_string(const char * cstring)
 {
+   // calling this constructor with and utf8-encoded C string is usually wrong.
+   //
+   // Instead of:   UCS_string(const char * utf8_string)
+   // one should:   UCS_string(UTF8_string(utf8_string))
+   //
+   // so that the constructor below takes care of the utf8-decoding.
+   //
+   // For ASCII C strings this constuctor is fine.
+   //
    create(LOC);
 
-const int len = strlen(cstring);
-   extend(len + 1);
-
-   while (items_valid < len)   items[items_valid++] = Unicode(*cstring++);
-   items[items_valid] = UNI_ASCII_NUL;
+   while (*cstring)
+      {
+        Assert((0x80 & *cstring) == 0);   // ASCII
+        *this += Unicode(*cstring++);
+      }
 }
 //-----------------------------------------------------------------------------
 UCS_string::UCS_string(const UTF8_string & utf)
@@ -88,7 +99,7 @@ UCS_string::UCS_string(const UTF8_string & utf)
    Log(LOG_char_conversion)
       CERR << "UCS_string::UCS_string(): utf = " << utf << endl;
 
-   for (int i = 0; i < utf.size();)
+   for (size_t i = 0; i < utf.size();)
       {
         const uint32_t b0 = utf[i++];
         uint32_t bx = b0;
@@ -159,38 +170,34 @@ int quad_pp = pctx.get_PP();
    if (quad_pp > MAX_Quad_PP)   quad_pp = MAX_Quad_PP;
    if (quad_pp < MIN_Quad_PP)   quad_pp = MIN_Quad_PP;
 
-const bool negative = (value < 0);
+const bool negative = (value < 0.0);
    if (negative)   value = -value;
 
 int expo = 0;
 
    if (value >= 10.0)   // large number, positive exponent
       {
-       while (value >= 1e16)
-             {
-               if (value > 1e307)
-                  {
-                    if (negative)   append_utf8("¯∞");
-                    else            append_utf8("∞");
-                    FloatCell::map_FC(*this);
-                    return;
-                  }
-               value *= 1e-16;   expo += 16;
-             }
+        if (value > 1e307)
+           {
+             if (negative)   append_UTF8("¯∞");
+             else            append_UTF8("∞");
+             FloatCell::map_FC(*this);
+             return;
+           }
+
+       while (value >= 1e16)   { value *= 1e-16;   expo += 16; }
        while (value >= 1e4)    { value *= 1e-4;    expo +=  4; }
        while (value >= 1e1)    { value *= 1e-1;    ++expo;     }
       }
    else if (value < 1.0)   // small number, negative exponent
       {
-       while (value < 1e-16)
-             {
-               if (value < 1e-305)   // very small number: make it 0
-                  {
-                    append(UNI_ASCII_0);
-                    return;
-                  }
-               value *= 1e16;   expo -= 16;
-             }
+       if (value < 1e-305)   // very small number: make it 0
+          {
+            append(UNI_ASCII_0);
+            return;
+          }
+
+       while (value < 1e-16)   { value *= 1e16;   expo -= 16; }
        while (value < 1e-4)    { value *= 1e4;    expo -=  4; }
        while (value < 1.0)     { value *= 10.0;   --expo;     }
       }
@@ -206,31 +213,36 @@ UCS_string digits;
       {
         if (value >= 10.0)
            {
-             digits.append((Unicode)(10 + '0'));
-             value = 0.0;
+             // 10.0 or more is a rounding error from 9,999...
+             digits.append(Unicode(10 + '0'));
+             while (digits.size() < (quad_pp + 2))   digits.append(UNI_ASCII_0);
+             break;
            }
         else if (value < 0.0)
            {
-             digits.append((Unicode)('0'));
+             // less than 0.0 is a rounding error from 0.000...
+             while (digits.size() < (quad_pp + 2))   digits.append(UNI_ASCII_0);
+             break;
+             digits.append(UNI_ASCII_0);
              value = 0.0;
            }
         else
            {
-             const int dig = (int)value;
+             const int dig = int(value);
              value -= dig;
-             value *= 10;
-             digits.append((Unicode)(dig + '0'));
+             value *= 10.0;
+             digits.append(Unicode(dig + '0'));
            }
       }
 
-   if (digits[0] != '0')   digits.pop();
+   if (digits[0] != '0')   digits.pop_back();
 
    // round last digit
    //
-const Unicode last = digits.last();
-   digits.pop();
+const Unicode last = digits.back();
+   digits.pop_back();
 
-   if (last >= '5')   digits.last() = (Unicode)(digits.last() + 1);
+   if (last >= '5')   digits.back() = Unicode(digits.back() + 1);
  
    // adjust carries of 2nd to last digit
    //
@@ -238,8 +250,8 @@ const Unicode last = digits.last();
        {
         if (digits[d] > '9')
            {
-             digits[d] =     (Unicode)(digits[d]     - 10);
-             digits[d - 1] = (Unicode)(digits[d - 1] +  1);
+             digits[d] =     Unicode(digits[d]     - 10);
+             digits[d - 1] = Unicode(digits[d - 1] +  1);
            }
        }
 
@@ -247,15 +259,15 @@ const Unicode last = digits.last();
    //
    if (digits[0] > '9')
       {
-        digits[0] = (Unicode)(digits[0] - 10);
-        digits.insert_before(0, UNI_ASCII_1);
+        digits[0] = Unicode(digits[0] - 10);
+        digits.insert(0, UNI_ASCII_1);
         ++expo;
-        digits.pop();
+        digits.pop_back();
       }
 
    // remove trailing zeros
    //
-   while (digits.size() && digits.last() == UNI_ASCII_0)   digits.pop();
+   while (digits.size() > 1 && digits.back() == UNI_ASCII_0)  digits.pop_back();
 
    // force scaled format if:
    //
@@ -331,7 +343,7 @@ UCS_string::UCS_string(const PrintBuffer & pb, Rank rank, int quad_PW)
 
 const int total_width = pb.get_width(0);
 
-Simple_string<int, false> breakpoints;
+std::vector<int> breakpoints;
    breakpoints.reserve(2*total_width/quad_PW);
 
    // print rows, breaking at breakpoints
@@ -348,14 +360,14 @@ Simple_string<int, false> breakpoints;
               if (row == 0)   // first row: set up breakpoints
                  {
                    chunk_len = pb.get_line(0).compute_chunk_length(quad_PW,col);
-                   breakpoints.append(chunk_len);
+                   breakpoints.push_back(chunk_len);
                  }
               else
                  {
                    chunk_len = breakpoints[b++];
                  }
 
-              if (col)   append_utf8("\n      ");
+              if (col)   append_UTF8("\n      ");
               UCS_string trow(pb.get_line(row), col, chunk_len);
               trow.remove_trailing_padchars();
               append(trow);
@@ -372,6 +384,32 @@ Simple_string<int, false> breakpoints;
        }
 }
 //-----------------------------------------------------------------------------
+/// constructor
+UCS_string::UCS_string(const Value & value)
+{
+   create(LOC);
+
+   if (value.get_rank() > 1) RANK_ERROR;
+
+const ShapeItem ec = value.element_count();
+   reserve(ec);
+
+   loop(e, ec)   append(value.get_ravel(e).get_char_value());
+}
+//-----------------------------------------------------------------------------
+UCS_string::UCS_string(istream & in)
+{
+   create(LOC);
+
+   for (;;)
+      {
+        const Unicode uni = UTF8_string::getc(in);
+        if (uni == Invalid_Unicode)   return;
+        if (uni == UNI_ASCII_LF)      return;
+        append(uni);
+      }
+}
+//-----------------------------------------------------------------------------
 int
 UCS_string::compute_chunk_length(int quad_PW, int col) const
 {
@@ -412,7 +450,7 @@ UCS_string::remove_trailing_padchars()
          if (at(u) == UNI_LINE_VERT2)   break;
          if (at(u) == UNI_iPAD_L0)
             {
-              shrink(0);
+              clear();
               return;
             }
        }
@@ -435,7 +473,7 @@ UCS_string::remove_trailing_padchars()
 void
 UCS_string::remove_trailing_whitespaces()
 {
-   while (size() && last() <= UNI_ASCII_SPACE)   pop();
+   while (size() && back() <= UNI_ASCII_SPACE)   pop_back();
 }
 //-----------------------------------------------------------------------------
 void
@@ -449,8 +487,8 @@ int count = 0;
       }
 
    if (count == 0)        return;      // no leading whitspaces
-   if (count == size())   shrink(0);   // only whitespaces
-   else                   *this = UCS_string(*this, count, size() - count);
+   if (count == size())   clear();     // only whitespaces
+   else                   basic_string::erase(0, count);
 }
 //-----------------------------------------------------------------------------
 void
@@ -465,46 +503,18 @@ UCS_string::split_ws(UCS_string & rest)
               ShapeItem arg = clen;
               while (arg < size() && Avec::is_white(at(arg)))   ++arg;
               while (arg < size())   rest.append(at(arg++));
-              shrink(clen);
+              resize(clen);
               return;
             }
        }
 }
 //-----------------------------------------------------------------------------
-/// constructor
-UCS_string::UCS_string(const Value & value)
-   : Simple_string<Unicode, false>(0, 0)
-{
-   create(LOC);
-
-   if (value.get_rank() > 1) RANK_ERROR;
-
-const ShapeItem ec = value.element_count();
-   reserve(ec);
-
-   loop(e, ec)   append(value.get_ravel(e).get_char_value());
-}
-//-----------------------------------------------------------------------------
-UCS_string::UCS_string(istream & in)
-   : Simple_string<Unicode, false>(0, 0)
-{
-   create(LOC);
-
-   for (;;)
-      {
-        const Unicode uni = UTF8_string::getc(in);
-        if (uni == Invalid_Unicode)   return;
-        if (uni == UNI_ASCII_LF)      return;
-        append(uni);
-      }
-}
-//-----------------------------------------------------------------------------
 void
 UCS_string::copy_black(UCS_string & dest, int & idx) const
 {
-   while (idx < size() && operator[](idx) <= ' ')   ++idx;
-   while (idx < size() && operator[](idx) >  ' ')   dest.append(operator[](idx++));
-   while (idx < size() && operator[](idx) <= ' ')   ++idx;
+   while (idx < size() && at(idx) <= ' ')   ++idx;
+   while (idx < size() && at(idx) >  ' ')   dest.append(at(idx++));
+   while (idx < size() && at(idx) <= ' ')   ++idx;
 }
 //-----------------------------------------------------------------------------
 ShapeItem
@@ -556,9 +566,20 @@ UCS_string::starts_with(const char * prefix) const
         if (uni != Unicode(pc))   return false;
       }
 
-   // strings agree, but prefix is longer
+   // strings match, but prefix is longer
    //
-   return false;   
+   return false;
+}
+//-----------------------------------------------------------------------------
+bool 
+UCS_string::ends_with(const char * suffix) const
+{
+const ShapeItem s_len = strlen(suffix);
+   if (size() < s_len)   return false;
+
+   suffix += s_len;    // goto end of suffix
+   loop(s, s_len)   if (at(size() - s - 1) != *--suffix)   return false;
+   return true;
 }
 //-----------------------------------------------------------------------------
 bool 
@@ -665,10 +686,119 @@ UCS_string::is_comment_or_label() const
    return false;
 }
 //-----------------------------------------------------------------------------
-void
-UCS_string::append_utf8(const UTF8 * str)
+ShapeItem
+UCS_string::double_quote_count(bool in_quote2) const
 {
-const size_t len = strlen((const char *)str);
+ShapeItem count = 0;
+bool in_quote1 = false;
+   loop(s, size())
+       {
+        const Unicode uni = at(s);
+        switch(uni)
+           {
+             case UNI_SINGLE_QUOTE:
+                  if (!in_quote2)   in_quote1 = ! in_quote1;
+                  break;
+
+             case UNI_ASCII_DOUBLE_QUOTE:
+                  if (!in_quote1)
+                     {
+                       ++count;
+                       in_quote2 = ! in_quote2;
+                     }
+                  break;
+
+             case UNI_ASCII_BACKSLASH:
+                  if (in_quote2)    ++s;   // ignore next char inside ""
+                  break;
+
+             case UNI_ASCII_NUMBER_SIGN:
+             case UNI_COMMENT:
+                  if (!(in_quote1 || in_quote2))   return count;
+
+             default:                            ;
+           }
+       }
+
+   return count;
+}
+//-----------------------------------------------------------------------------
+ShapeItem
+UCS_string::double_quote_first() const
+{
+bool in_quote1 = false;
+bool in_quote2 = true;
+   loop(s, size())
+       {
+        const Unicode uni = at(s);
+        switch(uni)
+           {
+             case UNI_SINGLE_QUOTE:
+                  if (!in_quote2)   in_quote1 = ! in_quote1;
+                  break;
+
+             case UNI_ASCII_DOUBLE_QUOTE:
+                  if (!in_quote1)   return s;
+                  break;
+
+             case UNI_ASCII_BACKSLASH:
+                  if (in_quote2)    ++s;   // ignore next char inside ""
+                  break;
+
+             case UNI_ASCII_NUMBER_SIGN:
+             case UNI_COMMENT:
+                  if (in_quote1 || in_quote2)   ; // ignore # and ⍝ in atrings
+                  else                          s = size();
+                  break;
+
+             default:                            ;
+           }
+       }
+
+
+   return -1;   // no un-commented and un-escaped " found
+}
+//-----------------------------------------------------------------------------
+ShapeItem
+UCS_string::double_quote_last() const
+{
+ShapeItem ret = -1;
+bool in_quote1 = false;
+bool in_quote2 = false;
+   loop(s, size())
+       {
+        const Unicode uni = at(s);
+        switch(uni)
+           {
+             case UNI_SINGLE_QUOTE:
+                  if (!in_quote2)   in_quote1 = ! in_quote1;
+                  break;
+
+             case UNI_ASCII_DOUBLE_QUOTE:
+                  if (!in_quote1)   ret = s;
+                  break;
+
+             case UNI_ASCII_BACKSLASH:
+                  if (in_quote2)    ++s;   // ignotr next char inside ""
+                  break;
+
+             case UNI_ASCII_NUMBER_SIGN:
+             case UNI_COMMENT:
+                  if (in_quote1 || in_quote2)   ; // ignore # and ⍝ in atrings
+                  else                          s = size();
+                  break;
+
+             default:                            ;
+           }
+       }
+
+   return ret;
+}
+//-----------------------------------------------------------------------------
+void
+UCS_string::append_UTF8(const UTF8 * str)
+{
+const size_t len = strlen(charP(str));
 const UTF8_string utf(str, len);
 const UCS_string ucs(utf);
 
@@ -689,16 +819,10 @@ UCS_string::append_quoted(const UCS_string & other)
 }
 //-----------------------------------------------------------------------------
 void
-UCS_string::append_ascii(const char * str)
-{
-   while (*str)   append(Unicode(*str++));
-}
-//-----------------------------------------------------------------------------
-void
 UCS_string::append_number(ShapeItem num)
 {
 char cc[40];
-   snprintf(cc, sizeof(cc) - 1, "%lld", (long long)num);
+   snprintf(cc, sizeof(cc) - 1, "%lld", long_long(num));
    loop(c, sizeof(cc))
       {
         if (cc[c])   append(Unicode(cc[c]));
@@ -711,7 +835,7 @@ UCS_string::append_hex(ShapeItem num, bool uppercase)
 {
 const char * format = uppercase ? "%llX" : "%llx";
 char cc[40];
-   snprintf(cc, sizeof(cc) - 1, format, (long long)num);
+   snprintf(cc, sizeof(cc) - 1, format, long_long(num));
    loop(c, sizeof(cc))
       {
         if (cc[c])   append(Unicode(cc[c]));
@@ -736,10 +860,10 @@ UCS_string::append_shape(const Shape & shape)
 }
 //-----------------------------------------------------------------------------
 void
-UCS_string::append_float(double num)
+UCS_string::append_float(APL_Float num)
 {
 char cc[60];
-   snprintf(cc, sizeof(cc) - 1, "%lf", num);
+   snprintf(cc, sizeof(cc) - 1, "%lf", double(num));
    loop(c, sizeof(cc))
       {
         if (cc[c])   append(Unicode(cc[c]));
@@ -747,35 +871,203 @@ char cc[60];
       }
 }
 //-----------------------------------------------------------------------------
+UCS_string
+UCS_string::un_escape(bool double_quoted, bool keep_LF) const
+{
+const char * hex = "0123456789abcdef";
+UCS_string ret;
+   ret.reserve(size());
+
+   if (double_quoted)
+      {
+        loop(s, size())
+            {
+             const Unicode uni = at(s);
+             if (uni != UNI_ASCII_BACKSLASH)   // normal char
+                {
+                  ret.append(uni);
+                  continue;
+                }
+
+             if (s >= (size() - 1))   // \ at end of string
+                {
+                  ret.append(UNI_ASCII_BACKSLASH);
+                  break;
+                }
+
+             const Unicode uni1 = at(++s);
+             switch(uni1)
+                 {
+                  case UNI_ASCII_a:            ret << UNI_ASCII_BEL;   continue;
+                  case UNI_ASCII_b:            ret << UNI_ASCII_BS;    continue;
+                  case UNI_ASCII_f:            ret << UNI_ASCII_FF;    continue;
+                  case UNI_ASCII_n:            if (keep_LF)   break;
+                                               ret << UNI_ASCII_LF;    continue;
+                  case UNI_ASCII_r:            ret << UNI_ASCII_CR;    continue;
+                  case UNI_ASCII_t:            ret << UNI_ASCII_BS;    continue;
+                  case UNI_ASCII_v:            ret << UNI_ASCII_VT;    continue;
+                  case UNI_ASCII_DOUBLE_QUOTE:
+                  case UNI_ASCII_BACKSLASH:
+                                               ret << uni1;            continue;
+                  default:                     break;
+                 }
+
+             int max_len = 0;
+             if (uni1 == UNI_ASCII_u)
+                {
+                  max_len = 4;
+                }
+             else if (uni1 == UNI_ASCII_x)
+                {
+                  max_len = 2;
+                }
+             else   // \n or \": keep them escaped
+                {
+                  ret.append(uni);
+                  ret.append(uni1);
+                  continue;
+                }
+
+               // \x or \u
+               //
+               int value = 0;
+               loop(m, max_len)
+                   {
+                     if (s >= (size() - 1))   break;
+                     const int dig = at(s+1);
+                     const char * pos = strchr(hex, dig);
+                     if (pos == 0)   break;   // non-hex character
+
+                     value = value << 4 | (pos - hex);
+                     ++s;
+                   }
+               ret.append(Unicode(value));
+            }
+      }
+   else
+      {
+        bool got_quote = false;
+        loop(s, size())
+           {
+             const Unicode uni = at(s);
+             if (uni == UNI_SINGLE_QUOTE)
+                {
+                  if (got_quote)
+                     {
+                        ret.append(UNI_SINGLE_QUOTE);
+                        got_quote = false;
+                     }
+                  else
+                     {
+                        ret.append(UNI_SINGLE_QUOTE);
+                        got_quote = true;
+                     }
+                }
+             else
+                {
+                  if (got_quote)   ret.append(UNI_SINGLE_QUOTE);   // mal-formed
+                  ret.append(uni);
+                  got_quote = false;
+                }
+           }
+      }
+
+   return ret;
+}
+//-----------------------------------------------------------------------------
+UCS_string
+UCS_string::do_escape(bool double_quoted) const
+{
+const char * hex = "0123456789abcdef";
+UCS_string ret;
+   ret.reserve(size());
+
+   if (double_quoted)
+      {
+        loop(s, size())
+           {
+             const Unicode uni = at(s);
+             switch(uni)
+                {
+                  case UNI_ASCII_BEL:            ret << "\\a";    continue;
+                  case UNI_ASCII_BS:             ret << "\\b";    continue;
+                  case UNI_ASCII_HT:             ret << "\\t";    continue;
+                  case UNI_ASCII_LF:             ret << "\\n";    continue;
+                  case UNI_ASCII_VT:             ret << "\\v";    continue;
+                  case UNI_ASCII_FF:             ret << "\\f";    continue;
+                  case UNI_ASCII_CR:             ret << "\\r";    continue;
+                  case UNI_ASCII_DOUBLE_QUOTE:   ret << "\\\"";   continue;
+                  case UNI_ASCII_BACKSLASH:      ret << "\\\\";   continue;
+                  default:                       break;
+                }
+
+             // none of the above
+             //
+             if (uni >= UNI_ASCII_SPACE && uni < UNI_ASCII_DELETE)
+                {
+                  ret.append(uni);
+                  continue;
+                }
+
+             if (uni <= 0x0F)   // small ASCII
+                {
+                  ret << "\\x0";
+                  ret << Unicode(hex[uni]);
+                }
+             else if (uni <= 0xFF)   // other ASCII
+                {
+                  ret << "\\x";
+                  ret << Unicode(hex[uni >> 4 & 0x0F]);
+                  ret << Unicode(hex[uni      & 0x0F]);
+                }
+             else
+                {
+                  ret.append(uni);
+                }
+           }
+      }
+   else   // single-quoted
+      {
+        loop(s, size())
+           {
+             const Unicode uni = at(s);
+             ret.append(uni);
+             if (uni == UNI_SINGLE_QUOTE)   ret.append(uni);   // another '
+           }
+      }
+
+   return ret;
+}
+//-----------------------------------------------------------------------------
 size_t
 UCS_string::to_vector(UCS_string_vector & result) const
 {
 size_t max_len = 0;
 
-   result.shrink(0);
+   result.clear();
    if (size() == 0)   return max_len;
 
-   result.append(UCS_string());
+   result.push_back(UCS_string());
    loop(s, size())
       {
         const Unicode uni = at(s);
         if (uni == UNI_ASCII_LF)    // line done
            {
-             const size_t len = result.last().size();
+             const size_t len = result.back().size();
              if (max_len < len)   max_len = len;
 
              if (s < size() - 1)   // more coming
-                result.append(UCS_string());
+                result.push_back(UCS_string());
            }
         else
            {
              if (uni != UNI_ASCII_CR)         // ignore \r.
-                result[result.size() - 1].append(uni);
+                result.back().append(uni);
            }
       }
 
    // if the last line lacked a \n we check max_len here again.
-const size_t len = result.last().size();
+const size_t len = result.back().size();
    if (max_len < len)   max_len = len;
 
    return max_len;
@@ -812,32 +1104,32 @@ bool negative = false;
 ostream &
 operator << (ostream & os, Unicode uni)
 {       
-   if (uni < 0x80)      return os << (char)uni;
+   if (uni < 0x80)      return os << char(uni);
         
-   if (uni < 0x800)     return os << (char)(0xC0 | (uni >> 6))
-                                  << (char)(0x80 | (uni & 0x3F));
+   if (uni < 0x800)     return os << char(0xC0 | (uni >> 6))
+                                  << char(0x80 | (uni & 0x3F));
         
-   if (uni < 0x10000)    return os << (char)(0xE0 | (uni >> 12))
-                                   << (char)(0x80 | (uni >>  6 & 0x3F))
-                                   << (char)(0x80 | (uni       & 0x3F));
+   if (uni < 0x10000)    return os << char(0xE0 | (uni >> 12))
+                                   << char(0x80 | (uni >>  6 & 0x3F))
+                                   << char(0x80 | (uni       & 0x3F));
 
-   if (uni < 0x200000)   return os << (char)(0xF0 | (uni >> 18))
-                                   << (char)(0x80 | (uni >> 12 & 0x3F))
-                                   << (char)(0x80 | (uni >>  6 & 0x3F))
-                                   << (char)(0x80 | (uni       & 0x3F));
+   if (uni < 0x200000)   return os << char(0xF0 | (uni >> 18))
+                                   << char(0x80 | (uni >> 12 & 0x3F))
+                                   << char(0x80 | (uni >>  6 & 0x3F))
+                                   << char(0x80 | (uni       & 0x3F));
 
-   if (uni < 0x4000000)  return os << (char)(0xF8 | (uni >> 24))
-                                   << (char)(0x80 | (uni >> 18 & 0x3F))
-                                   << (char)(0x80 | (uni >> 12 & 0x3F))
-                                   << (char)(0x80 | (uni >>  6 & 0x3F))
-                                   << (char)(0x80 | (uni       & 0x3F));
+   if (uni < 0x4000000)  return os << char(0xF8 | (uni >> 24))
+                                   << char(0x80 | (uni >> 18 & 0x3F))
+                                   << char(0x80 | (uni >> 12 & 0x3F))
+                                   << char(0x80 | (uni >>  6 & 0x3F))
+                                   << char(0x80 | (uni       & 0x3F));
 
-   return os << (char)(0xFC | (uni >> 30))
-             << (char)(0x80 | (uni >> 24 & 0x3F))
-             << (char)(0x80 | (uni >> 18 & 0x3F))
-             << (char)(0x80 | (uni >> 12 & 0x3F))
-             << (char)(0x80 | (uni >>  6 & 0x3F))
-             << (char)(0x80 | (uni       & 0x3F));
+   return os << char(0xFC | (uni >> 30))
+             << char(0x80 | (uni >> 24 & 0x3F))
+             << char(0x80 | (uni >> 18 & 0x3F))
+             << char(0x80 | (uni >> 12 & 0x3F))
+             << char(0x80 | (uni >>  6 & 0x3F))
+             << char(0x80 | (uni       & 0x3F));
 }
 //-----------------------------------------------------------------------------
 ostream &
@@ -880,7 +1172,7 @@ UCS_string::dump(ostream & out) const
    out << right << hex << uppercase << setfill('0');
    loop(s, size())
       {
-        out << " U+" << setw(4) << (int)at(s);
+        out << " U+" << setw(4) << int(at(s));
       }
 
    return out << left << dec << nouppercase << setfill(' ');
@@ -916,9 +1208,9 @@ UCS_string ret;
 }
 //-----------------------------------------------------------------------------
 UCS_string
-UCS_string::from_big(double & val)
+UCS_string::from_big(APL_Float & val)
 {
-   Assert(val >= 0);
+   Assert(val >= 0.0);
 
 long double value = val;
 int digits[320];   // DBL_MAX is 1.79769313486231470E+308
@@ -929,7 +1221,7 @@ long double fract;
    for (; value >= 1.0; ++d)
       {
          fract = modf(value / 10.0, &value);   // U.x -> .U
-         *d = (int)((fract + .02) * 10.0);
+         *d = int((fract + .02) * 10.0);
          fract -= 0.1 * *d;
       }
 
@@ -943,7 +1235,7 @@ UCS_string ret;
 }
 //-----------------------------------------------------------------------------
 UCS_string
-UCS_string::from_double_expo_prec(double v, int fract_digits)
+UCS_string::from_double_expo_prec(APL_Float v, int fract_digits)
 {
 UCS_string ret;
 
@@ -960,7 +1252,7 @@ UCS_string ret;
         return ret;
       }
 
-   if (v < 0)   { ret.append(UNI_OVERBAR);   v = - v; }
+   if (v < 0.0)   { ret.append(UNI_OVERBAR);   v = - v; }
 
 int expo = 0;
    while (v >= 1.0E1)
@@ -968,33 +1260,33 @@ int expo = 0;
         if (v >= 1.0E9)
            if (v >= 1.0E13)
               if (v >= 1.0E15)
-                 if     (v >= 1.0E16)      { v *= 1.0E-16;   expo += 16; }
-                 else /* v >= 1.0E15 */    { v *= 1.0E-15;   expo += 15; }
+                 if     (v >= 1.0E16)      { v = v * 1.0E-16;   expo += 16; }
+                 else /* v >= 1.0E15 */    { v = v * 1.0E-15;   expo += 15; }
               else
-                 if     (v >= 1.0E14)      { v *= 1.0E-14;   expo += 14; }
-                 else /* v >= 1.0E13 */    { v *= 1.0E-13;   expo += 13; }
+                 if     (v >= 1.0E14)      { v = v * 1.0E-14;   expo += 14; }
+                 else /* v >= 1.0E13 */    { v = v * 1.0E-13;   expo += 13; }
            else
               if (v >= 1.0E11)
-                 if     (v >= 1.0E12)      { v *= 1.0E-12;   expo += 12; }
-                 else /* v >= 1.0E11 */    { v *= 1.0E-11;   expo += 11; }
+                 if     (v >= 1.0E12)      { v = v * 1.0E-12;   expo += 12; }
+                 else /* v >= 1.0E11 */    { v = v * 1.0E-11;   expo += 11; }
               else
-                 if     (v >= 1.0E10)      { v *= 1.0E-10;   expo += 10; }
-                 else /* v >= 1.0E9 */     { v *= 1.0E-9;    expo += 9;  }
+                 if     (v >= 1.0E10)      { v = v * 1.0E-10;   expo += 10; }
+                 else /* v >= 1.0E9 */     { v = v * 1.0E-9;    expo += 9;  }
         else
            if (v >= 1.0E5)
               if (v >= 1.0E7)
-                 if     (v >= 1.0E8)       { v *= 1.0E-8;    expo += 8;  }
-                 else /* v >= 1.0E7 */     { v *= 1.0E-7;    expo += 7;  }
+                 if     (v >= 1.0E8)       { v = v * 1.0E-8;    expo += 8;  }
+                 else /* v >= 1.0E7 */     { v = v * 1.0E-7;    expo += 7;  }
               else
-                 if     (v >= 1.0E6)       { v *= 1.0E-6;    expo += 6;  }
-                 else /* v >= 1.0E5 */     { v *= 1.0E-5;    expo += 5;  }
+                 if     (v >= 1.0E6)       { v = v * 1.0E-6;    expo += 6;  }
+                 else /* v >= 1.0E5 */     { v = v * 1.0E-5;    expo += 5;  }
            else
               if (v >= 1.0E3)
-                 if     (v >= 1.0E4)       { v *= 1.0E-4;    expo += 4;  }
-                 else /* v >= 1.0E3 */     { v *= 1.0E-3;    expo += 3;  }
+                 if     (v >= 1.0E4)       { v = v * 1.0E-4;    expo += 4;  }
+                 else /* v >= 1.0E3 */     { v = v * 1.0E-3;    expo += 3;  }
               else
-                 if     (v >= 1.0E2)       { v *= 1.0E-2;    expo += 2;  }
-                 else /* v >= 1.0E1 */     { v *= 1.0E-1;    expo += 1;  }
+                 if     (v >= 1.0E2)       { v = v * 1.0E-2;    expo += 2;  }
+                 else /* v >= 1.0E1 */     { v = v * 1.0E-1;    expo += 1;  }
       }
 
    while (v < 1.0E0)
@@ -1002,33 +1294,33 @@ int expo = 0;
         if (v < 1.0E-8)
            if (v < 1.0E-12)
               if (v < 1.0E-14)
-                 if     (v < 1.0E-15)      { v *= 1.0E-16;   expo += 16; }
-                 else /* v < 1.0E-14 */    { v *= 1.0E-15;   expo += 15; }
+                 if     (v < 1.0E-15)      { v = v * 1.0E-16;   expo += 16; }
+                 else /* v < 1.0E-14 */    { v = v * 1.0E-15;   expo += 15; }
               else
-                 if     (v < 1.0E-13)      { v *= 1.0E-14;   expo += 14; }
-                 else /* v < 1.0E-12 */    { v *= 1.0E-13;   expo += 13; }
+                 if     (v < 1.0E-13)      { v = v * 1.0E-14;   expo += 14; }
+                 else /* v < 1.0E-12 */    { v = v * 1.0E-13;   expo += 13; }
            else
               if (v < 1.0E-10)
-                 if     (v < 1.0E-11)      { v *= 1.0E12;   expo += -12; }
-                 else /* v < 1.0E-10 */    { v *= 1.0E11;   expo += -11; }
+                 if     (v < 1.0E-11)      { v = v * 1.0E12;   expo += -12; }
+                 else /* v < 1.0E-10 */    { v = v * 1.0E11;   expo += -11; }
               else
-                 if     (v < 1.0E-9 )      { v *= 1.0E10;   expo += -10; }
-                 else /* v < 1.0E-8 */     { v *= 1.0E9;    expo += -9;  }
+                 if     (v < 1.0E-9 )      { v = v * 1.0E10;   expo += -10; }
+                 else /* v < 1.0E-8 */     { v = v * 1.0E9;    expo += -9;  }
         else
            if (v < 1.0E-4)
               if (v < 1.0E-6)
-                 if     (v < 1.0E-7)       { v *= 1.0E8;    expo += -8;  }
-                 else /* v < 1.0E-6 */     { v *= 1.0E7;    expo += -7;  }
+                 if     (v < 1.0E-7)       { v = v * 1.0E8;    expo += -8;  }
+                 else /* v < 1.0E-6 */     { v = v * 1.0E7;    expo += -7;  }
               else
-                 if     (v < 1.0E-5)       { v *= 1.0E6;    expo += -6;  }
-                 else /* v < 1.0E-4 */     { v *= 1.0E5;    expo += -5;  }
+                 if     (v < 1.0E-5)       { v = v * 1.0E6;    expo += -6;  }
+                 else /* v < 1.0E-4 */     { v = v * 1.0E5;    expo += -5;  }
            else
               if (v < 1.0E-2)
-                 if     (v < 1.0E-3)       { v *= 1.0E4;    expo += -4;  }
-                 else /* v < 1.0E-2 */     { v *= 1.0E3;    expo += -3;  }
+                 if     (v < 1.0E-3)       { v = v * 1.0E4;    expo += -4;  }
+                 else /* v < 1.0E-2 */     { v = v * 1.0E3;    expo += -3;  }
               else
-                 if     (v < 1.0E-1)       { v *= 1.0E2;    expo += -2;  }
-                 else /* v < 1.0E0  */     { v *= 1.0E1;    expo += -1;  }
+                 if     (v < 1.0E-1)       { v = v * 1.0E2;    expo += -2;  }
+                 else /* v < 1.0E0  */     { v = v * 1.0E1;    expo += -1;  }
       }
 
    Assert(v >= 1.0);
@@ -1055,11 +1347,11 @@ UCS_string mantissa = from_double_fixed_prec(v, fract_digits);
 }
 //-----------------------------------------------------------------------------
 UCS_string
-UCS_string::from_double_fixed_prec(double v, int fract_digits)
+UCS_string::from_double_fixed_prec(APL_Float v, int fract_digits)
 {
 UCS_string ret;
 
-   if (v < 0)   { ret.append(UNI_OVERBAR);   v = - v; }
+   if (v < 0.0)   { ret.append(UNI_OVERBAR);   v = - v; }
 
    // in the loop below, there could be rounding errors when casting float
    // to int. We therefore increase v slighly (by 0.3 of the rounded digit)
@@ -1073,7 +1365,7 @@ UCS_string ret;
 
    loop(f, fract_digits + 1)
       {
-        v *= 10;
+        v = v * 10.0;
         const int vv = v;   // subject to rounding errors!
         ret.append(Unicode(UNI_ASCII_0 + vv));
         v -= vv;
@@ -1109,8 +1401,8 @@ UCS_string::round_last_digit()
             }
       }
 
-   shrink(size() - 1);
-   if (items[size() - 1] == UNI_ASCII_FULLSTOP)   shrink(size() - 1);
+   pop_back();
+   if (back() == UNI_ASCII_FULLSTOP)   pop_back();
 }
 //----------------------------------------------------------------------------
 bool
@@ -1147,3 +1439,47 @@ UCS_string ret;
    return ret;
 }
 //----------------------------------------------------------------------------
+UCS_string
+UCS_string::to_HTML(int offset, bool preserve_ws) const
+{
+UCS_string ret;
+   for (;offset < size(); ++offset)
+      {
+        const Unicode uni = at(offset);
+        switch(uni)
+           {
+             case ' ':  if (preserve_ws)   ret.append_ASCII("&nbsp;");
+                        else               ret.append(uni);
+                        break;
+             case '#':  ret.append_ASCII("&#35;");   break;
+             case '%':  ret.append_ASCII("&#37;");   break;
+             case '&':  ret.append_ASCII("&#38;");   break;
+             case '<':  ret.append_ASCII("&lt;");    break;
+             case '>':  ret.append_ASCII("&gt;");    break;
+             default:   ret.append(uni);
+           }
+      }
+
+   return ret;
+}
+//----------------------------------------------------------------------------
+#if UCS_tracking
+UCS_string::~UCS_string()
+{
+   --total_count;
+   cerr << setfill('0') << endl << "@@ " << setw(5) << instance_id
+        << " DEL ##" << total_count
+        << " c= " << Backtrace::caller(3) << setfill(' ') << endl;
+}
+//----------------------------------------------------------------------------
+void UCS_string::create(const char * loc)
+{
+   ++total_count;
+   instance_id = ++total_id;
+   cerr << setfill('0') << endl << "@@ " << setw(5) << instance_id
+        << " NEW ##" << total_count << " " << loc
+        << " c= " << Backtrace::caller(3) << setfill(' ') << endl;
+}
+
+#endif
+//----------------------------------------------------------------------------
