@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2017  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -26,20 +26,24 @@
 #include <sys/mman.h>
 #include <errno.h>
 
+#include <vector>
+
+#include "Bif_F12_TAKE_DROP.hh"
 #include "Error.hh"
 #include "Output.hh"
 #include "Parser.hh"
 #include "StateIndicator.hh"
 #include "Symbol.hh"
 #include "UserFunction.hh"
-#include "Value.icc"
+#include "UserPreferences.hh"
+#include "Value.hh"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
 UserFunction::UserFunction(const UCS_string txt, const char * loc,
                            const UTF8_string & _creator, bool tolerant,
                            bool macro)
-  : Function(ID::USER_SYMBOL, TOK_FUN2),
+  : Function(ID_USER_SYMBOL, TOK_FUN2),
     Executable(txt, true, PM_FUNCTION, loc),
     header(txt, macro),
     creator(_creator),
@@ -59,7 +63,7 @@ UserFunction::UserFunction(const UCS_string txt, const char * loc,
    exec_properties[2] = 0;
    exec_properties[3] = 0;
 
-   line_starts.append(Function_PC_0);   // will be set later.
+   line_starts.push_back(Function_PC_0);   // will be set later.
 
    if (header.get_error() != E_NO_ERROR)   // bad header
       {
@@ -88,7 +92,7 @@ UserFunction::UserFunction(const UCS_string txt, const char * loc,
 //-----------------------------------------------------------------------------
 UserFunction::UserFunction(Fun_signature sig, int lambda_num,
                            const UCS_string & text, Token_string & lambda_body)
-  : Function(ID::USER_SYMBOL, TOK_FUN0),
+  : Function(ID_USER_SYMBOL, TOK_FUN0),
     Executable(sig, lambda_num, text, LOC),
     header(sig, lambda_num),
     creator(UNI_LAMBDA),
@@ -115,18 +119,22 @@ UserFunction::UserFunction(Fun_signature sig, int lambda_num,
    else                    tag = TOK_FUN0;
 
    while (lambda_body.size() > 2 &&
-          lambda_body.last().get_tag() == TOK_SYMBOL &&
+          lambda_body.back().get_tag() == TOK_SYMBOL &&
           lambda_body[lambda_body.size() - 2].get_tag() == TOK_SEMICOL)
       {
-        header.add_local_var(lambda_body.last().get_sym_ptr());
-        lambda_body.pop();   // varname
-        lambda_body.pop();   // semicolon
+        header.add_local_var(lambda_body.back().get_sym_ptr());
+        lambda_body.pop_back();   // varname
+        lambda_body.pop_back();   // semicolon
       }
 
+   // order of local vars is reversed. Fix that.
+   //
+   header.reverse_local_vars();
+
    parse_body_line(Function_Line_0, lambda_body, false, false, LOC);
    setup_lambdas();
-   line_starts.append(Function_PC(lambda_body.size() - 1));
-   line_starts.append(Function_PC_0);
+   line_starts.push_back(Function_PC(lambda_body.size() - 1));
+   line_starts.push_back(Function_PC_0);
    error_line = -1;   // no error
    error_info = 0;
 }
@@ -185,6 +193,7 @@ UserFunction::eval_XB(Value_P X, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
+   if (!header.X())    AXIS_ERROR;
    if (header.LO())    SYNTAX_ERROR;   // defined as operator
 
    Workspace::push_SI(this, LOC);
@@ -236,6 +245,7 @@ UserFunction::eval_AXB(Value_P A, Value_P X, Value_P B)
 
    if (header.LO())    SYNTAX_ERROR;    // defined as operator
    if (!header.A())    VALENCE_ERROR;   // monadic
+   if (!header.X())    AXIS_ERROR;
 
    Workspace::push_SI(this, LOC);
 
@@ -287,6 +297,7 @@ UserFunction::eval_LXB(Token & LO, Value_P X, Value_P B)
       }
 
    if (header.RO())    SYNTAX_ERROR;   // dyadic operator called monadically
+   if (!header.X())    AXIS_ERROR;
 
    Workspace::push_SI(this, LOC);
 
@@ -345,6 +356,7 @@ UserFunction::eval_ALXB(Value_P A, Token & LO, Value_P X, Value_P B)
 
    if (header.RO())    SYNTAX_ERROR;    // defined as dyadic operator
    if (!header.A())    VALENCE_ERROR;   // monadic
+   if (!header.X())    AXIS_ERROR;
 
    Workspace::push_SI(this, LOC);
 
@@ -402,7 +414,8 @@ UserFunction::eval_LRXB(Token & LO, Token & RO, Value_P X, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   if (!header.RO())    SYNTAX_ERROR;   // not defined as dyadic operator
+   if (!header.RO())   SYNTAX_ERROR;   // not defined as dyadic operator
+   if (!header.X())    AXIS_ERROR;
 
    Workspace::push_SI(this, LOC);
 
@@ -468,6 +481,7 @@ UserFunction::eval_ALRXB(Value_P A, Token & LO, Token & RO,
 
    if (!header.RO())   SYNTAX_ERROR;   // defined monadic op called dyadically
    if (!header.A())    VALENCE_ERROR;   // monadic
+   if (!header.X())    AXIS_ERROR;
 
    Workspace::push_SI(this, LOC);
 
@@ -502,24 +516,34 @@ Value_P Z = B->clone(LOC);
 void
 UserFunction::set_locked_error_info(Error & error) const
 {
-UCS_string & message_2 = error.error_message_2;
+UCS_string message_2(UTF8_string(error.get_error_line_2()));
 
+#define SHORT 0
    if (header.A())
       {
+#if SHORT
+        message_2.append(header.A()->get_name());
+        message_2.append(UNI_ASCII_SPACE);
+#else
         Value_P val_A = header.A()->get_value();
         if (!!val_A)
            {
-             PrintContext pctx(PR_APL_FUN);
+             PrintContext pctx(PR_BOXED_GRAPHIC);
              PrintBuffer pb(*val_A, pctx, 0);
              message_2.append(UCS_string(pb, 1, DEFAULT_Quad_PW));
              message_2.append(UNI_ASCII_SPACE);
            }
+#endif
       }
 
    message_2.append(header.get_name());
 
    if (header.B())
       {
+#if SHORT
+        message_2.append(UNI_ASCII_SPACE);
+        message_2.append(header.B()->get_name());
+#else
         Value_P val_B = header.B()->get_value();
         if (!!val_B)
            {
@@ -528,55 +552,308 @@ UCS_string & message_2 = error.error_message_2;
              PrintBuffer pb(*val_B, pctx, 0);
              message_2.append(UCS_string(pb, 1, DEFAULT_Quad_PW));
            }
+#endif
       }
 
-   error.right_caret = error.left_caret + message_2.size() - 7;
+   {
+     UTF8_string utf(message_2);
+     error.set_error_line_2(utf.c_str());
+   }
+
+   error.set_right_caret(error.get_left_caret() + message_2.size() - 7);
 }
 //-----------------------------------------------------------------------------
 void
-UserFunction::set_trace_stop(Simple_string<Function_Line, false> & lines,
-                             bool stop)
+UserFunction::set_trace_stop(std::vector<Function_Line> & lines, bool stop)
 {
    // Sort lines
    //
-Simple_string<bool, false> ts_lines;
-   ts_lines.reserve(line_starts.size());
+std::vector<bool> ts_lines;
 
-   loop(ts, line_starts.size())   ts_lines.append(false);
+   loop(ts, line_starts.size())   ts_lines.push_back(false);
    loop(ll, lines.size())
       {
         Function_Line line = lines[ll];
-        if (line >= 1 && line < (int)line_starts.size())  ts_lines[line] = true;
+        if (line >= 1 && line < int(line_starts.size()))
+           ts_lines[line] = true;
       }
 
    if (stop)
       {
-        stop_lines.shrink(0);
+        stop_lines.clear();
         loop(ts, line_starts.size())
            {
-             if (ts_lines[ts])   stop_lines.append((Function_Line)ts);
+             if (ts_lines[ts])
+                stop_lines.push_back(Function_Line(ts));
            }
       }
    else
       {
-        trace_lines.shrink(0);
+        trace_lines.clear();
         loop(ts, line_starts.size())
            {
-             if (ts_lines[ts])   trace_lines.append((Function_Line)ts);
+             if (ts_lines[ts])
+                trace_lines.push_back(Function_Line(ts));
            }
       }
 
    parse_body(LOC, false, false);
 }
 //-----------------------------------------------------------------------------
+ErrorCode
+UserFunction::transform_multi_line_strings()
+{
+  /* convert a set of lines like
+
+     [k+1] PREFIX "L1
+     [k+2] L2 ...
+     [k+N] Ln" SUFFIX
+
+    into:
+
+    [k+1] PREFIX "L1" "L2" ... "LN" SUFFIX
+    [...] (empty)
+    [k+N] (empty)
+   */
+
+enum Line_status
+   {
+     Function_header = 0,
+     APL_text        = 1,
+     Start_of_string = 2,
+     Inside_string   = 3,
+     End_of_string   = 4
+   };
+
+char status[get_text_size()];   status[0] = Function_header;
+Line_status current = APL_text;
+
+   // determine line status of each line...
+   //
+   for (int l = 1; l < get_text_size(); ++l)
+       {
+         const int count = get_text(l).double_quote_count(false);
+         if (count & 1)   // start or end of string
+            {
+              if (current == APL_text)   // start of multi-line string
+                 {
+                    status[l] = Start_of_string;
+                    current = Inside_string;
+                 }
+              else                       // end of multi-line string
+                 {
+                    status[l] = End_of_string;
+                    current = APL_text;
+                 }
+            }
+         else              // no status change
+            {
+              status[l] = current;
+            }
+       }
+
+   if (current == Start_of_string || current == Inside_string)
+      {
+         // multi-line string started but not ended.
+         //
+         return E_DEFN_ERROR;
+      }
+
+   // modify lines...
+   //
+   for (int l = 1; l < get_text_size();)
+       {
+         if (status[l] == APL_text)   { ++l;   continue; }
+         const int start = l;
+         Assert1(status[l] == Start_of_string);
+         UCS_string accu = get_text(l++);
+         accu << "\" ";
+         while (status[l] == Inside_string)
+               {
+                 accu << " \"" << get_text(l).do_escape(true) << "\"";
+                 clear_text(l++);
+               }
+         Assert(status[l] == End_of_string);
+         accu << "\"" << get_text(l);
+         set_text(start, accu);
+         clear_text(l++);
+       }
+
+   // remove trailing empty lines...
+   //
+   while (get_text_size() && get_text(get_text_size() - 1).size() == 0)
+         text.pop_back();
+
+// CERR << endl;
+// loop(l, get_text_size())   CERR << "[" << l << "]  " << get_text(l) << endl;
+
+   return E_NO_ERROR;
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+UserFunction::transform_multi_line_strings_3()
+{
+  /* convert a set of lines like
+
+     [k+1] PREFIX """
+     [k+2] L2 ...
+     [k+N] """
+
+    into:
+
+    [k+1] PREFIX "L2" ... "LN"
+    [...] (empty)
+    [k+N] (empty)
+
+     In contrast to transform_multi_line_strings(), line k+N may only
+     consist of spaces and the terminating """.
+   */
+
+enum Line_status
+   {
+     Function_header = 0,
+     APL_text        = 1,
+     Start_of_string = 2,
+     Inside_string   = 3,
+     End_of_string   = 4
+   };
+
+char status[get_text_size()];   status[0] = Function_header;
+Line_status current = APL_text;
+
+   // determine line status of each line...
+   //
+   for (ShapeItem l = 1; l < get_text_size(); ++l)
+       {
+         const UCS_string & line = get_text(l);
+         const ShapeItem len = line.size();
+         if (line.ends_with("\"\"\""))
+            {
+              if (current == APL_text)   // start of multi-line string
+                 {
+                    status[l] = Start_of_string;
+                    current = Inside_string;
+                 }
+              else                       // end of multi-line string
+                 {
+                   bool blanks_only = true;
+                   for (ShapeItem c = 0; c < (len - 3); ++c)
+                       {
+                          if (line[c] != UNI_ASCII_SPACE &&
+                              line[c] != UNI_ASCII_HT)   blanks_only = false;
+                       }
+
+                    if (blanks_only)
+                       {
+                         status[l] = End_of_string;
+                         current = APL_text;
+                       }
+                 }
+            }
+         else              // no status change
+            {
+              status[l] = current;
+            }
+       }
+
+   if (current == Start_of_string || current == Inside_string)
+      {
+         // multi-line string started but not ended.
+         //
+         return E_DEFN_ERROR;
+      }
+
+   // modify lines...
+   //
+   for (int l = 1; l < get_text_size(); )
+       {
+         if (status[l] == APL_text)   { ++l;   continue; }
+         const int start = l;
+         Assert1(status[l] == Start_of_string);
+         UCS_string accu = get_text(l++);
+         accu.resize(accu.size() - 3);   // remove trailing """
+         accu << " ";
+         while (status[l] == Inside_string)
+               {
+                 accu << " \"" << get_text(l).do_escape(true) << "\"";
+                 clear_text(l++);
+               }
+         Assert(status[l] == End_of_string);
+         set_text(start, accu);
+         clear_text(l++);
+       }
+
+   // remove trailing empty lines...
+   //
+   while (get_text_size() && get_text(get_text_size() - 1).size() == 0)
+         text.pop_back();
+
+// CERR << endl;
+// loop(l, get_text_size())   CERR << "[" << l << "]  " << get_text(l) << endl;
+
+   return E_NO_ERROR;
+}
+//-----------------------------------------------------------------------------
 void
 UserFunction::parse_body(const char * loc, bool tolerant, bool macro)
 {
-   line_starts.shrink(0);
-   line_starts.append(Function_PC_0);   // will be set later.
+   line_starts.clear();
+   line_starts.push_back(Function_PC_0);   // will be set later.
 
+UCS_string_vector original_text;
+   //
+   // The text is modified for parsing but restored afterwards so that e,g,
+   // ∇FUN[⎕]∇ shows the text enterd by the user.
+   //
+   // original_text is only set if text was modified.
+   //
    clear_body();
 
+   if (uprefs.multi_line_strings_3)   // new-style multi-line strings allowed
+      {
+        for (int l = 1; l < get_text_size(); ++l)
+            {
+              const UCS_string & line = get_text(l);
+              const ShapeItem len = line.size();
+              if (len >= 3 &&
+                  line[len - 1] == UNI_ASCII_DOUBLE_QUOTE &&
+                  line[len - 2] == UNI_ASCII_DOUBLE_QUOTE &&
+                  line[len - 3] == UNI_ASCII_DOUBLE_QUOTE)
+                 {
+                   original_text = text;
+                   const ErrorCode ec = transform_multi_line_strings_3();
+                   if (ec)
+                      {
+                        text = original_text;   // restore function text
+                        error_line = l;
+                        return;
+                      }
+
+                   break;
+                 }
+            }
+      }
+
+   if (uprefs.multi_line_strings)   // old-style multi-line strings allowed
+      {
+        for (int l = 1; l < get_text_size(); ++l)
+            {
+              if (get_text(l).double_quote_count(false) & 1)
+                 {
+                   original_text = text;
+                   const ErrorCode ec = transform_multi_line_strings();
+                   if (ec)
+                      {
+                        text = original_text;   // restore function text
+                        error_line = l;
+                        return;
+                      }
+
+                   break;
+                 }
+            }
+      }
+
    for (int l = 1; l < get_text_size(); ++l)
       {
         bool stop_line = false;
@@ -600,13 +877,13 @@ UserFunction::parse_body(const char * loc, bool tolerant, bool macro)
            }
 
         error_line = l;   // assume error
-        line_starts.append(Function_PC(body.size()));
+        line_starts.push_back(Function_PC(body.size()));
 
         if (stop_line)
            {
-             body.append(Token(TOK_STOP_LINE));
+             body.push_back(Token(TOK_STOP_LINE));
              const int64_t tr = 0;
-             body.append(Token(TOK_END, tr));
+             body.push_back(Token(TOK_END, tr));
            }
 
         const UCS_string & line = get_text(l);
@@ -624,7 +901,7 @@ UserFunction::parse_body(const char * loc, bool tolerant, bool macro)
                         << header.get_name() << endl;
                  }
             }
-        catch(Error err)
+        catch(const Error & err)
             {
               return;
             }
@@ -642,8 +919,11 @@ UserFunction::parse_body(const char * loc, bool tolerant, bool macro)
    // let [0] be the end of the function.
    line_starts[0] = Function_PC(body.size());
 
-   if (header.Z())   body.append(Token(TOK_RETURN_SYMBOL, header.Z()), LOC);
-   else              body.append(Token(TOK_RETURN_VOID), LOC);
+   if (header.Z())   body.push_back(Token(TOK_RETURN_SYMBOL, header.Z()));
+   else              body.push_back(Token(TOK_RETURN_VOID));
+
+   // restore the original text (before multi-line expansion)
+   if (original_text.size())   text = original_text;
 }
 //-----------------------------------------------------------------------------
 UserFunction *
@@ -655,11 +935,11 @@ UserFunction * fun = 0;
       {
         load(workspace, function, fun);
       }
-   catch (Error err)
+   catch (Error & err)
       {
         delete fun;
 
-        err.print(CERR);
+        err.print(CERR, LOC);
       }
    catch (...)
       {
@@ -682,7 +962,7 @@ char filename[FILENAME_MAX + 10];
    if (strlen(filename) > FILENAME_MAX)
       {
         CERR << "file name '" << filename << "' is too long" << endl;
-        throw_apl_error(E_SYSTEM_LIMIT_FILENAME, LOC);
+        throw_apl_error(E_SYS_LIMIT_FILENAME, LOC);
       }
 
 int in = open(filename, O_RDONLY);
@@ -705,7 +985,7 @@ struct stat st;
 off_t len = st.st_size;
 void * start = mmap(0, len, PROT_READ, MAP_SHARED, in, 0);
 
-   if (start == (const void *)-1)
+   if (start == reinterpret_cast<const void *>(-1))
       {
         CERR << "Can't mmap() workspace file '" 
              << filename << "': " << strerror(errno) << endl;
@@ -713,16 +993,17 @@ void * start = mmap(0, len, PROT_READ, MAP_SHARED, in, 0);
         throw_apl_error(E_WS_MMAP, LOC);
       }
 
-UTF8_string utf((const UTF8 *)start, len);
+UTF8_string utf(utf8P(start), len);
 
    // skip trailing \r and \n.
    //
-   while (utf.size() && (utf.last() == '\r' || utf.last() == '\n'))   utf.pop();
+   while (utf.size() &&
+          (utf.back() == '\r' || utf.back() == '\n'))   utf.pop_back();
 
-UCS_string ucs(utf);
-   munmap((char *)start, st.st_size);
+   munmap(start, st.st_size);
    close(in);
 
+UCS_string ucs(utf);
 int error_line = -1;
    fun = fix(ucs, error_line, false, LOC, filename, false);
 }
@@ -730,8 +1011,9 @@ int error_line = -1;
 Function_PC
 UserFunction::pc_for_line(Function_Line line) const
 {
-   if (line < 1)                          return Function_PC(body.size() - 1);
-   if (line >= (int)line_starts.size())   return Function_PC(body.size() - 1);
+   if (line < 1 || line >= int(line_starts.size()))
+      return Function_PC(body.size() - 1);
+
    return line_starts[line];
 }
 //-----------------------------------------------------------------------------
@@ -750,23 +1032,28 @@ UserFunction * ufun = new UserFunction(text, loc, creator, tolerant, false);
 const char * info = ufun->get_error_info();
    err_line = ufun->get_error_line();
 
-   if (info || err_line != -1)   // something went wrong
+const bool bad_function = info || err_line != -1;
+   if (bad_function)   // something went wrong
       {
         if (info)
            {
              Log(LOG_UserFunction__fix)   CERR << "Error: " << info << endl;
              MORE_ERROR() << info;
            }
+         else info = "Error";
 
          if (err_line == 0)
            {
-             MORE_ERROR() << " (function header)";
+             MORE_ERROR() << info << "in function header";
              Log(LOG_UserFunction__fix) CERR << "Bad header line" <<  endl;
            }
          else if (err_line > 0)
            {
-             MORE_ERROR() << " (function line [" << err_line << "] of:\n"
-                          << text;
+             UCS_string & more = MORE_ERROR();
+             more << info << " in function line [" << err_line << "] of:\n";
+             loop(l, ufun->get_text_size())
+                more << "[" << l << "] " << ufun->get_text(l) << "\n";
+
              Log(LOG_UserFunction__fix)
                 CERR << "Bad function line: " << err_line << endl;
            }
@@ -792,7 +1079,7 @@ Function * old_function = symbol->get_function();
         delete ufun;
         return 0;
       }
-     
+
    if (old_function)
       {
         const UserFunction * old_ufun = old_function->get_ufun1();
@@ -807,7 +1094,7 @@ Function * old_function = symbol->get_function();
 
    Log(LOG_UserFunction__fix)
       {
-        CERR << " addr " << (const void *)ufun << endl;
+        CERR << " addr " << voidP(ufun) << endl;
         ufun->print(CERR);
       }
 
@@ -845,7 +1132,7 @@ ShapeItem semi = -1;
 
               case UNI_ASCII_SEMICOLON: if (semi == -1)   semi = t - 1;
                                         continue;
-                                        
+
               case UNI_ASCII_LF:        break;   // header done
               default:                  continue;
             }
@@ -860,15 +1147,15 @@ ShapeItem semi = -1;
 UCS_string body_text;
    for (; t < text.size(); ++t)   body_text.append(text[t]);
 
-   while (body_text.last() == UNI_ASCII_LF)  body_text.pop();
+   while (body_text.back() == UNI_ASCII_LF)  body_text.pop_back();
 
 Token_string body;
    {
      Token ret_lambda(TOK_RETURN_SYMBOL, &Workspace::get_v_LAMBDA());
-     body.append(ret_lambda);
+     body.push_back(ret_lambda);
      const int64_t trace = 0;
      Token tok_endl(TOK_ENDL, trace);
-     body.append(tok_endl);
+     body.push_back(tok_endl);
    }
 
    if (semi != -1)
@@ -887,7 +1174,7 @@ const ErrorCode ec = parser.parse(body_text, body);
         return 0;
       }
 
-UserFunction * ufun = new UserFunction((Fun_signature)signature, 0,
+UserFunction * ufun = new UserFunction(Fun_signature(signature), 0,
                                        body_text, body);
    return ufun;
 }
@@ -901,6 +1188,73 @@ UserFunction::destroy()
    else               delete this;
 }
 //-----------------------------------------------------------------------------
+bool
+UserFunction::pushes_sym(const Symbol * sym) const
+{
+   if (sym == header.Z())   return true;
+   if (sym == header.A())   return true;
+   if (sym == header.LO())   return true;
+   if (sym == header.X())   return true;
+   if (sym == header.RO())   return true;
+   if (sym == header.B())   return true;
+
+   loop(l, local_var_count())
+       {
+         if (sym == get_local_var(l))   return true;
+       }
+
+   return false;
+}
+//-----------------------------------------------------------------------------
+void 
+UserFunction::help(ostream & out) const
+{
+   CERR << "    Header: " << get_text(0) << endl;
+
+   if (is_lambda())
+      {
+         UCS_string body(get_text(1), 2, get_text(1).size() - 2);
+         CERR << "Lambda: { " << body << " ";
+         loop(v, local_var_count())
+            {
+              const Symbol & sym = *get_local_var(v);
+              CERR << ";" << sym.get_name();
+            }
+         CERR << " }" << endl;
+         return;
+      }
+
+bool got_lamps = false;
+bool toronto = false;
+const UCS_string two_lamps(UTF8_string("⍝⍝"));
+   for (int l = 1; l < get_text_size(); ++l)
+       {
+         UCS_string line(get_text(l));
+         line.remove_leading_and_trailing_whitespaces();
+         if (line.size() < 2)          continue;   // too short
+
+         if (line[0] != UNI_COMMENT)   // not a comment
+            {
+              toronto = false;
+              continue;
+            }
+
+         const bool double_lamps = line[1] == UNI_COMMENT;   // ⍝⍝ line
+         if (line[1] == UNI_ASCII_FULLSTOP)                  // ⍝. line
+            {
+              toronto = true;
+            }
+
+         if (double_lamps || toronto)
+            {
+              got_lamps = true;
+              CERR << "    " << line << endl;
+            }
+       }
+
+   if (!got_lamps)   CERR << "    (no ⍝⍝ or ⍝. comment lines)" << endl;
+}
+//-----------------------------------------------------------------------------
 ostream &
 UserFunction::print(ostream & out) const
 {
@@ -933,6 +1287,11 @@ UCS_string
 UserFunction::get_name_and_line(Function_PC pc) const
 {
 UCS_string ret = header.get_name();
+   if (ret.size() && ret[0] == UNI_LAMBDA)
+      {
+        UCS_string name = Workspace::find_lambda_name(this);
+        if (name.size())   ret = name;
+      }
    ret.append(UNI_ASCII_L_BRACK);
 
    // pc may point to the next token already. If that is the case then
@@ -982,20 +1341,21 @@ UserFunction::adjust_line_starts()
    // this function is called from Executable::setup_lambdas() just before
    // Parser::remove_void_token(body) in order to adjust line_starts
    //
-Simple_string<ShapeItem, false> gaps;
+std::vector<ShapeItem> gaps;
    gaps.reserve(line_starts.size());   // count TOK_VOID in every line
    loop(ls, line_starts.size())
       {
-         gaps.append(0);
+         gaps.push_back(0);
          if (ls == 0)   continue;   // function header (has no TOK_VOID)
 
          const ShapeItem from = line_starts[ls];
          ShapeItem to = body.size();    // end of function (for last line)
-         if (ls < (line_starts.size() - 1))   to = line_starts[ls + 1];
+         if (ls < (ShapeItem(line_starts.size()) - 1))
+            to = line_starts[ls + 1];
 
         for (ShapeItem b = from; b < to; ++b)
             {
-             if (body[b].get_tag() == TOK_VOID)   ++gaps.last();
+             if (body[b].get_tag() == TOK_VOID)   ++gaps.back();
             }
       }
 
@@ -1010,13 +1370,13 @@ int total_gaps = 0;
 Function_PC
 UserFunction::line_start(Function_Line line) const
 {
-   if (line < 0 || line >= line_starts.size())
+   if (line < 0 || size_t(line) >= line_starts.size())
       {
         Q1(line)
         Q1(line_starts.size())
         Assert(0);
       }
-      
+
    return line_starts[line];
 }
 //-----------------------------------------------------------------------------
@@ -1032,18 +1392,12 @@ UserFunction::print_val_or_fun(ostream & out, Token & tok)
 }
 //-----------------------------------------------------------------------------
 UCS_string
-UserFunction::line_prefix(Function_Line l)
+UserFunction::line_prefix(Function_Line l) const
 {
-UCS_string ucs;
-   ucs.append(UNI_ASCII_L_BRACK);
-   if (l < Function_Line_10)   ucs.append(UNI_ASCII_SPACE);
-
 char cc[40];
-   snprintf(cc, sizeof(cc), "%d", l);
-   for (const char * s = cc; *s; ++s)   ucs.append(Unicode(*s));
-
-   ucs.append(UNI_ASCII_R_BRACK);
-   ucs.append(UNI_ASCII_SPACE);
-   return ucs;
+   if      (text.size() > 100)   snprintf(cc, sizeof(cc), "[%3d] ", l);
+   else if (text.size() > 10)    snprintf(cc, sizeof(cc), "[%2d] ", l);
+   else                          snprintf(cc, sizeof(cc), "[%d] ",  l);
+   return UCS_string(cc);
 }
 //-----------------------------------------------------------------------------
