@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2017  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#include <errno.h>
 #include <fcntl.h>
 #include <limits.h>
 #include <string.h>
@@ -42,7 +43,7 @@ extern const char * configure_args;
 #include "UserPreferences.hh"
 #include "Workspace.hh"
 
-#include "Value.icc"
+#include "Value.hh"
 
 UserPreferences uprefs;
 
@@ -84,8 +85,10 @@ char cc[4000];
 #endif
 
    snprintf(cc, sizeof(cc),
+"    -C new_root          do chroot(new_root) before starting APL (root only)\n"
 "    --cfg                show ./configure options used and exit\n"
 "    --noCIN              do not echo input (for scripting)\n"
+"    --CPU_limit_secs sec set CPU time limit to sec seconds\n"
 "    --echoCIN            echo (final) input to COUT\n"
 "    --rawCIN             do not emit escape sequences\n"
 "    --[no]Color          start with ]XTERM ON [OFF])\n"
@@ -95,6 +98,7 @@ char cc[4000];
 "    --gpl                show license (GPL) and exit\n"
 "    -L wsname            )LOAD wsname (and not SETUP or CONTINUE) on startup\n"
 "    --LX expr            execute APL expression expr first\n"
+"    --OFF                automatically )OFF after last input file\n"
 "    -p N                 use profile N in preferences files\n"
 "    --par proc           use processor parent ID proc (default: no parent)\n"
 "    --PW value           initial value of ⎕PW\n"
@@ -117,6 +121,7 @@ char cc[4000];
 "                         4:   exit after first error\n"
 "    --TR                 randomize order of testfiles\n"
 "    --TS                 append to (rather than override) summary.log\n"
+"    -u UID               run as user UID (root only)\n"
 "    -v, --version        show version information and exit\n"
 "    -w milli             wait milli milliseconds at startup\n"
 "    --                   end of options for %s\n", prog);
@@ -133,7 +138,7 @@ UserPreferences::show_GPL(ostream & out)
 "    This program is GNU APL, a free implementation of the\n"
 "    ISO/IEC Standard 13751, \"Programming Language APL, Extended\"\n"
 "\n"
-"    Copyright (C) 2008-2015  Dr. Jürgen Sauermann\n"
+"    Copyright (C) 2008-2017  Dr. Jürgen Sauermann\n"
 "\n"
 "    This program is free software: you can redistribute it and/or modify\n"
 "    it under the terms of the GNU General Public License as published by\n"
@@ -182,7 +187,7 @@ bool
 UserPreferences::parse_argv_1()
 {
 bool log_startup = false;
-   for (int a = 1; a < expanded_argv.size(); )
+   for (size_t a = 1; a < expanded_argv.size(); )
        {
          const char * opt = expanded_argv[a++];
          const char * val = (a < expanded_argv.size()) ? expanded_argv[a] : 0;
@@ -212,6 +217,49 @@ bool log_startup = false;
               user_profile = atoi(val);
               continue;
             }
+
+         if (!strcmp(opt, "-C"))
+            {
+              ++a;
+              if (!val)
+                 {
+                   CERR << "-C without directory" << endl;
+                   exit(a);
+                 }
+
+              if (chroot(val))
+                 {
+                   CERR << "chroot(" << val << ") failed: "
+                        << strerror(errno) << endl;
+                   exit(a);
+                 }
+
+              if (chdir("/"))
+                 {
+                   CERR << "chdir(\"/\") failed: "
+                        << strerror(errno) << endl;
+                   exit(a);
+                 }
+
+              continue;
+            }
+
+         if (!strcmp(opt, "-u"))
+            {
+              ++a;
+              if (!val)
+                 {
+                   CERR << "-u without user ID" << endl;
+                   exit(a);
+                 }
+              if (setuid(strtol(val, 0, 10)))
+                 {
+                   CERR << "setuid(" << val << ") failed: "
+                        << strerror(errno) << endl;
+                   exit(a);
+                 }
+              continue;
+            }
        }
 
    return log_startup;
@@ -239,9 +287,9 @@ UserPreferences::parse_argv_2(bool logit)
          }
    }
 
-   for (int a = 1; a < expanded_argv.size(); )
+   for (size_t a = 1; a < expanded_argv.size(); )
        {
-         if ((int)a == script_argc)   { ++a;   continue; }   // skip scriptname
+         if (a == script_argc)   { ++a;   continue; }   // skip scriptname
 
          const char * opt = expanded_argv[a++];
          const char * val = (a < expanded_argv.size()) ? expanded_argv[a] : 0;
@@ -273,6 +321,24 @@ UserPreferences::parse_argv_2(bool logit)
               exit(0);
             }
 
+         if (!strcmp(opt, "-C"))
+            {
+              ++a;
+              continue;   // -C already handled in parse_argv_1()
+            }
+
+         if (!strcmp(opt, "--CPU_limit_secs"))
+            {
+              ++a;
+              if (!val)
+                 {
+                   CERR << "--CPU_limit_secs without seconds" << endl;
+                   exit(a);
+                 }
+              CPU_limit_secs = atoi(val);
+              continue;
+            }
+
          if (!strcmp(opt, "--Color"))
             {
               do_Color = true;
@@ -316,8 +382,8 @@ UserPreferences::parse_argv_2(bool logit)
 
               const UTF8_string & filename(val);
               InputFile fam(filename, 0, false, !do_not_echo, true, no_LX);
-              InputFile::files_todo.append(fam);
-              InputFile::files_orig.append(fam);
+              InputFile::files_todo.push_back(fam);
+              InputFile::files_orig.push_back(fam);
               continue;
             }
 
@@ -425,6 +491,12 @@ UserPreferences::parse_argv_2(bool logit)
               continue;
             }
 
+         if (!strcmp(opt, "--OFF"))
+            {
+              auto_OFF = true;
+              continue;
+            }
+
          if (!strcmp(opt, "-p"))
             {
               ++a;
@@ -562,8 +634,8 @@ UserPreferences::parse_argv_2(bool logit)
                        }
                     const UTF8_string & filename = expanded_argv[a];
                     InputFile fam(filename, 0, true, true, false, no_LX);
-                    InputFile::files_todo.append(fam);
-                    InputFile::files_orig.append(fam);
+                    InputFile::files_todo.push_back(fam);
+                    InputFile::files_orig.push_back(fam);
                   }
 
               // 
@@ -607,6 +679,12 @@ UserPreferences::parse_argv_2(bool logit)
               show_version(cout);
               exit(0);
             }
+         if (!strcmp(opt, "-u"))
+            {
+              ++a;
+              continue;   // -u already handled in parse_argv_1()
+            }
+
          if (!strcmp(opt, "-w"))
             {
               ++a;
@@ -638,7 +716,7 @@ UserPreferences::parse_argv_2(bool logit)
       {
         const UTF8_string & filename = expanded_argv[script_argc];
         InputFile fam(filename, 0, false, !do_not_echo, true, no_LX);
-        InputFile::files_todo.insert_before(0, fam);
+        InputFile::files_todo.insert(InputFile::files_todo.begin(), fam);
       }
 
    // count number of testfiles
@@ -681,8 +759,8 @@ UserPreferences::expand_argv(int argc, const char ** argv)
 {
    loop(a, argc)
        {
-         original_argv.append(argv[a]);
-         expanded_argv.append(argv[a]);
+         original_argv.push_back(argv[a]);
+         expanded_argv.push_back(argv[a]);
        }
 
    if (argc <= 1)   // no args at all
@@ -712,7 +790,7 @@ const char * apl_args = argv[1];   // the args after e.g. /usr/bin/apl
         return;
       }
 
-   expanded_argv.erase(1);
+   expanded_argv.erase(expanded_argv.begin() + 1);
    --script_argc;
    for (int index = 1;;)
        {
@@ -724,7 +802,7 @@ const char * apl_args = argv[1];   // the args after e.g. /usr/bin/apl
 
          if (!strcmp(apl_args, "--"))        // "--" at end of apl_args
             {
-              expanded_argv.insert_before(index++, "--");
+              expanded_argv.insert(expanded_argv.begin() + index++, "--");
               ++script_argc;
               break;   // done
             }
@@ -732,7 +810,7 @@ const char * apl_args = argv[1];   // the args after e.g. /usr/bin/apl
          if (!strncmp(apl_args, "-- ", 2) &&
                apl_args[2] <= ' ')   // "--" somewhere in apl_args
             {
-              expanded_argv.insert_before(index++, "--");
+              expanded_argv.insert(expanded_argv.begin() + index++, "--");
               ++script_argc;
               apl_args += 2;   // skip "--"
               continue;
@@ -743,20 +821,18 @@ const char * apl_args = argv[1];   // the args after e.g. /usr/bin/apl
             {
               const int arg_len = arg_end - apl_args;
               const char * arg = strndup(apl_args, arg_len);
-              expanded_argv.insert_before(index++, arg);
+              expanded_argv.insert(expanded_argv.begin() + index++, arg);
               ++script_argc;
               apl_args += arg_len;
             }
          else           // last argument
             {
               const char * arg = strdup(apl_args);
-              expanded_argv.insert_before(index++, arg);
+              expanded_argv.insert(expanded_argv.begin() + index++, arg);
               ++script_argc;
               break;
             }
        }
-
-   return;
 }
 //-----------------------------------------------------------------------------
 void
@@ -778,7 +854,7 @@ UserPreferences::show_version(ostream & out)
          out << *bt << endl;
        }
 
-   out << "   Archive SVN:    " << ARCHIVE_SVN << endl;
+   out << "    Archive SVN:   " << ARCHIVE_SVN << endl;
 
    Output::set_color_mode(Output::COLM_OUTPUT);
 }
@@ -798,6 +874,10 @@ UserPreferences::show_configure_options()
         << is_default(ASSERT_LEVEL_WANTED == 1)
    << endl <<
 
+   "    SECURITY_LEVEL_WANTED=" << SECURITY_LEVEL_WANTED
+        << is_default(SECURITY_LEVEL_WANTED == 0)
+   << endl <<
+
    "    APSERVER_PATH=" << APSERVER_PATH
         << is_default(!strcmp(APSERVER_PATH, "/tmp/GNU-APL/APserver"))
    << endl <<
@@ -835,6 +915,13 @@ UserPreferences::show_configure_options()
         << is_default(MAX_RANK_WANTED == 8)
    << endl <<
 
+#ifdef RATIONAL_NUMBERS_WANTED
+   "    RATIONAL_NUMBERS_WANTED=yes"
+#else
+   "    RATIONAL_NUMBERS_WANTED=no (default)"
+#endif
+   << endl <<
+
    "    SHORT_VALUE_LENGTH_WANTED=" << SHORT_VALUE_LENGTH_WANTED
         << is_default(SHORT_VALUE_LENGTH_WANTED == 1)
         << ", therefore:" << endl <<
@@ -883,7 +970,7 @@ FILE *
 UserPreferences::open_user_file(const char * fname, char * filename,
                                 bool sys, bool log_startup)
 {
-   if (sys)   // eg. /etc/gnu-apl.d/preferences
+   if (sys)   // filename in /etc/gnu-apl.d/preferences
       {
         snprintf(filename, APL_PATH_MAX,
                  "%s/gnu-apl.d/%s", Makefile__sysconfdir, fname);
@@ -918,7 +1005,7 @@ FILE * f = fopen(filename, "r");
       {
          if (log_startup)
             CERR << "Not reading config file " << filename
-                 << " (not found/readable)" << endl;
+                 << " (" << strerror(errno) << ")" << endl;
          return 0;
       }
 
@@ -1007,8 +1094,33 @@ int file_profile = 0;   // the current profile in the preferences file
               file_profile = atoi(arg);
               continue;
             }
+
          if (file_profile && (file_profile != user_profile))   continue;
 
+         // set up pointers to arguments...
+         //
+         const char * sargs[100];
+         char * p = buffer;
+         int sargs_idx = 0;
+         for (;;)
+            {
+              while (*p && *p <= ' ')   ++p;    // discard leading whitespace
+              if (*p == 0)   break;
+
+              if (p[0] == '/' && p[1] == '/')   break;   // comment
+
+              const char * arg = p;
+              sargs[sargs_idx++] = arg;         // store start of argument
+              while (*p > ' ')   ++p;           // skip non-whitespaces
+              if (*p)   *p++ = 0;
+              else      break;
+            }
+
+         // some args could be non-numeric
+         //
+         Assert(sargs_idx >= count);
+         sargs[sargs_idx] = "NUL";              // terminating 0
+
          if (!strcasecmp(opt, "Color"))
             {
               if (yes_no)   // user said "Yes" or "No"
@@ -1054,90 +1166,85 @@ int file_profile = 0;   // the current profile in the preferences file
             }
          else if (!strcasecmp(opt, "Logging"))
             {
+              d[0] = strtoll(arg, 0, 10);   // decimal!
               Log_control(LogId(d[0]), true);
+              log_startup && CERR << "    logging facility " << d[0]
+                                  << " enabled in " << filename << endl;
             }
          else if (!strcasecmp(opt, "CIN-SEQUENCE"))
             {
-              loop(p, count - 1)   Output::color_CIN[p] = (char)(d[p] & 0xFF);
-              Output::color_CIN[count - 1] = 0;
+              loop(sa, sargs_idx)
+                  Output::color_CIN[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "COUT-SEQUENCE"))
             {
-              loop(p, count - 1)   Output::color_COUT[p] = (char)(d[p] & 0xFF);
-              Output::color_COUT[count - 1] = 0;
+              loop(sa, sargs_idx)
+                  Output::color_COUT[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "CERR-SEQUENCE"))
             {
-              loop(p, count - 1)   Output::color_CERR[p] = (char)(d[p] & 0xFF);
-              Output::color_CERR[count - 1] = 0;
+              loop(sa, sargs_idx)
+                  Output::color_CERR[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "UERR-SEQUENCE"))
             {
-              loop(p, count - 1)   Output::color_UERR[p] = (char)(d[p] & 0xFF);
-              Output::color_UERR[count - 1] = 0;
+              loop(sa, sargs_idx)
+                  Output::color_UERR[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "RESET-SEQUENCE"))
             {
-              loop(p, count - 1)   Output::color_RESET[p] = (char)(d[p] & 0xFF);
-              Output::color_RESET[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::color_RESET[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "CLEAR-EOL-SEQUENCE"))
             {
-              loop(p, count - 1)   Output::clear_EOL[p] = (char)(d[p] & 0xFF);
-              Output::clear_EOL[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::clear_EOL[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "CLEAR-EOS-SEQUENCE"))
             {
-              loop(p, count - 1)   Output::clear_EOS[p] = (char)(d[p] & 0xFF);
-              Output::clear_EOS[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::clear_EOS[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "KEY-CURSOR-UP"))
             {
-              loop(p, count - 1)
-                  Output::ESC_CursorUp[p] = (char)(d[p] & 0xFF);
-              Output::ESC_CursorUp[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::ESC_CursorUp[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "KEY-CURSOR-DOWN"))
             {
-              loop(p, count - 1)
-                  Output::ESC_CursorDown[p] = (char)(d[p] & 0xFF);
-              Output::ESC_CursorDown[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::ESC_CursorDown[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "KEY-CURSOR-RIGHT"))
             {
-              loop(p, count - 1)
-                  Output::ESC_CursorRight[p] = (char)(d[p] & 0xFF);
-              Output::ESC_CursorRight[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::ESC_CursorRight[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "KEY-CURSOR-LEFT"))
             {
-              loop(p, count - 1)
-                  Output::ESC_CursorLeft[p] = (char)(d[p] & 0xFF);
-              Output::ESC_CursorLeft[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::ESC_CursorLeft[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "KEY-CURSOR-END"))
             {
-              loop(p, count - 1)
-                  Output::ESC_CursorEnd[p] = (char)(d[p] & 0xFF);
-              Output::ESC_CursorEnd[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::ESC_CursorEnd[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "KEY-CURSOR-HOME"))
             {
-              loop(p, count - 1)
-                  Output::ESC_CursorHome[p] = (char)(d[p] & 0xFF);
-              Output::ESC_CursorHome[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::ESC_CursorHome[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "KEY-INSMODE"))
             {
-              loop(p, count - 1)
-                  Output::ESC_InsertMode[p] = (char)(d[p] & 0xFF);
-              Output::ESC_InsertMode[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::ESC_InsertMode[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "KEY-DELETE"))
             {
-              loop(p, count - 1)
-                  Output::ESC_Delete[p] = (char)(d[p] & 0xFF);
-              Output::ESC_Delete[count - 1] = 0;
+              loop(sa, sargs_idx)
+                 Output::ESC_Delete[sa] = decode_ASCII(sargs[sa + 1]);
             }
          else if (!strcasecmp(opt, "CIN-FOREGROUND"))
             {
@@ -1182,7 +1289,7 @@ int file_profile = 0;   // the current profile in the preferences file
                    continue;
                  }
 
-              LibPaths::set_lib_dir((LibRef)lib_ref, arg,
+              LibPaths::set_lib_dir(LibRef(lib_ref), arg,
                                     sys ? LibPaths::LibDir::CSRC_PREF_SYS
                                         : LibPaths::LibDir::CSRC_PREF_HOME);
             }
@@ -1223,10 +1330,33 @@ int file_profile = 0;   // the current profile in the preferences file
                  CERR << "bad value " << arg << " for INITIAL-⎕PW (ignored)"
                       << endl;
             }
+         else if (!strcasecmp(opt, "Multi-Line-Strings"))
+            {
+              multi_line_strings = yes;
+              multi_line_strings_3 = yes;
+            }
+         else if (!strcasecmp(opt, "New-Multi-Line-Strings"))
+            {
+              multi_line_strings_3 = yes;
+            }
+         else if (!strcasecmp(opt, "Old-Multi-Line-Strings"))
+            {
+              multi_line_strings = yes;
+            }
          else if (!strcasecmp(opt, "WINCH-SETS-⎕PW"))
             {
               WINCH_sets_pw = yes;
             }
+         else if (!strcasecmp(opt, "AUTO-OFF"))
+            {
+              auto_OFF = yes;
+            }
+
+         // security facilities...
+         //
+#define sec_def(X) else if (!strcasecmp(opt, #X)) X = yes;
+#include "Security.def"
+
          else
             {
               CERR << "Invalid option '" << opt << " in preferences file "
@@ -1237,6 +1367,48 @@ int file_profile = 0;   // the current profile in the preferences file
    fclose(f);
 }
 //-----------------------------------------------------------------------------
+int
+UserPreferences::decode_ASCII(const char * strg)
+{
+   if (!strcmp(strg, "NUL"))   return 0x00;
+   if (!strcmp(strg, "SOH"))   return 0x01;
+   if (!strcmp(strg, "STX"))   return 0x02;
+   if (!strcmp(strg, "ETX"))   return 0x03;
+   if (!strcmp(strg, "EOT"))   return 0x04;
+   if (!strcmp(strg, "ENQ"))   return 0x05;
+   if (!strcmp(strg, "ACK"))   return 0x06;
+   if (!strcmp(strg, "BEL"))   return 0x07;
+   if (!strcmp(strg, "BS"))   return 0x08;
+   if (!strcmp(strg, "HT"))    return 0x09;
+   if (!strcmp(strg, "LF"))    return 0x0A;
+   if (!strcmp(strg, "VT"))    return 0x0B;
+   if (!strcmp(strg, "FF"))    return 0x0C;
+   if (!strcmp(strg, "CR"))    return 0x0D;
+   if (!strcmp(strg, "SO"))    return 0x0E;
+   if (!strcmp(strg, "SI"))    return 0x0F;
+   if (!strcmp(strg, "DLE"))   return 0x10;
+   if (!strcmp(strg, "DC1"))   return 0x11;
+   if (!strcmp(strg, "DC2"))   return 0x12;
+   if (!strcmp(strg, "DC3"))   return 0x13;
+   if (!strcmp(strg, "DC4"))   return 0x14;
+   if (!strcmp(strg, "NAK"))   return 0x15;
+   if (!strcmp(strg, "SYN"))   return 0x16;
+   if (!strcmp(strg, "ETB"))   return 0x17;
+   if (!strcmp(strg, "CAN"))   return 0x18;
+   if (!strcmp(strg, "EM"))    return 0x19;
+   if (!strcmp(strg, "SUB"))   return 0x1A;
+   if (!strcmp(strg, "ESC"))   return 0x1B;
+   if (!strcmp(strg, "FS"))    return 0x1C;
+   if (!strcmp(strg, "GS"))    return 0x1D;
+   if (!strcmp(strg, "RS"))    return 0x1E;
+   if (!strcmp(strg, "US"))    return 0x1F;
+
+   if (strlen(strg) == 1)   return *strg;                 // single char
+   if (strlen(strg) == 2)   return strtol(strg, 0, 16);   // hex value, e.g. 4C
+   CERR << "invalid parameter " << strg << " in preferences file" << endl;
+   return -1;
+}
+//-----------------------------------------------------------------------------
 void
 UserPreferences::read_threshold_file(bool sys, bool log_startup)
 {
