@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2017  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -19,9 +19,11 @@
 */
 
 #include <sys/types.h>
+#include <vector>
 
 #include "CDR_string.hh"
 #include "CharCell.hh"
+#include "ComplexCell.hh"
 #include "Common.hh"
 #include "Error.hh"
 #include "FloatCell.hh"
@@ -38,31 +40,38 @@
 #include "SystemVariable.hh"
 #include "UCS_string.hh"
 #include "UserFunction.hh"
-#include "Value.icc"
+#include "Value.hh"
 #include "ValueHistory.hh"
 #include "Workspace.hh"
 
-ShapeItem Value::value_count = 0;
-ShapeItem Value::total_ravel_count = 0;
+extern uint64_t top_of_memory();
+
+uint64_t Value::value_count = 0;
+uint64_t Value::total_ravel_count = 0;
 
 // the static Value instances are defined in StaticObjects.cc
 
-void * Value::deleted_values = 0;
+_deleted_value * Value::deleted_values = 0;
 int Value::deleted_values_count = 0;
 uint64_t Value::fast_new = 0;
 uint64_t Value::slow_new = 0;
+uint64_t Value::alloc_size = 0;
+
 
 //-----------------------------------------------------------------------------
-inline void
+void
 Value::init_ravel()
 {
    owner_count = 0;
    pointer_cell_count = 0;
    nz_subcell_count = 0;
+   check_ptr = 0;
+   new (short_value)   IntCell(0);
+   ravel = short_value;
 
    ++value_count;
    if (Quad_SYL::value_count_limit &&
-       Quad_SYL::value_count_limit < value_count)
+       Quad_SYL::value_count_limit < APL_Integer(value_count))
       {
         MORE_ERROR() <<
 "the system limit on the APL value count (as set in ⎕SYL) was reached\n"
@@ -77,60 +86,162 @@ Value::init_ravel()
 
 const ShapeItem length = shape.get_volume();
 
-   if (length > SHORT_VALUE_LENGTH_WANTED)
+   // small values always succeed...
+   //
+   if (length <= SHORT_VALUE_LENGTH_WANTED)
       {
-        total_ravel_count += length;
-        if (Quad_SYL::ravel_count_limit &&
-            Quad_SYL::ravel_count_limit < total_ravel_count)
-           {
-             total_ravel_count -= length;
-
-             // make sure that the value is properly initialized
-             //
-             ravel = short_value;
-             new (&shape) Shape();
-             new (ravel)   IntCell(42);
-
-             MORE_ERROR() <<
-             "the system limit on the total ravel size (as set in ⎕SYL) "
-             "was reached\n(and to avoid lock-up, the limit in ⎕SYL "
-             "was automatically cleared).";
-
-             // reset the limit so that we don't get stuck here.
-             //
-             Quad_SYL::ravel_count_limit = 0;
-             set_attention_raised(LOC);
-             set_interrupt_raised(LOC);
-           }
-
-        try
-           {
-             Cell * long_ravel = (Cell *)(new char[length * sizeof(Cell)]);
-//           Cell * long_ravel =  new Cell[length];
-             ravel = long_ravel;
-           }
-        catch (...)
-           {
-             // for some unknown reason, this object gets cleared after
-             // throwing the E_WS_FULL below (which destroys the prev and
-             // next pointers. We therefore unlink() the object here (where
-             // prev and next pointers are still intact).
-             //
-             unlink();
-             throw_apl_error(E_WS_FULL, alloc_loc);
-           }
+        check_ptr = charP(this) + 7;
+        return;
       }
-   else
+
+   // large Value. If the compiler uses 4-byte pointers, then do not allow APL
+   // values that are too large. The reason is that new[] may return
+   // a non-0 pointer (thus pretending everything is OK), but a subsequent
+   // attempt to initialize the value might then throw a segfault.
+   //
+   enum { MAX_LEN = 2000000000 / sizeof(Cell) };
+   if (length > MAX_LEN && sizeof(void *) < 6)
+      throw_apl_error(E_WS_FULL, alloc_loc);
+
+   if (Quad_SYL::ravel_count_limit &&
+       Quad_SYL::ravel_count_limit < APL_Integer(total_ravel_count + length))
       {
+        // make sure that the value is properly initialized
+        //
+        new (&shape) Shape();
         ravel = short_value;
+        new (ravel)   IntCell(42);
+
+        MORE_ERROR() <<
+"the system limit on the total ravel size (as set in ⎕SYL) was reached\n"
+"(and to avoid lock-up, the limit in ⎕SYL was automatically cleared).";
+
+        // reset the limit so that we don't get stuck here.
+        //
+        Quad_SYL::ravel_count_limit = 0;
+        set_attention_raised(LOC);
+        set_interrupt_raised(LOC);
       }
 
+   alloc_size = length * sizeof(Cell);
+   ravel = reinterpret_cast<Cell *>(new char[alloc_size]);
+
+/*
+   ravel = 0;   // assume new() fails
+// try
+//    {
+        ravel = reinterpret_cast<Cell *>(new char[length * sizeof(Cell)]);
+        if (ravel == 0)   // new failed (without trowing an exception)
+           {
+              Log(LOG_Value_alloc)
+                 CERR << "new char[" << (length * sizeof(Cell))
+                      << "] (aka. long ravel allocation) returned 0 at " LOC
+                      << endl;
+
+              MORE_ERROR() << "The instatiation of a Value object succeeded, "
+                              "but allocation of its (large) ravel failed.";
+              new (&shape) Shape();
+              ravel = short_value;
+              throw_apl_error(E_WS_FULL, alloc_loc);
+           }
+      }
+   catch (...)
+      {
+        // for some unknown reason, this object gets cleared after
+        // throwing the E_WS_FULL below (which destroys the prev and
+        // next pointers. We therefore unlink() the object here (where
+        // prev and next pointers are still intact).
+        //
+        unlink();
+
+        Log(LOG_Value_alloc)
+           {
+             CERR << "new char[" << (length * sizeof(Cell))
+                  << "] (aka. long ravel allocation) threw an exception at " LOC
+                  << endl;
+           }
+
+        MORE_ERROR() << "The instatiation of a Value object succeeded, "
+                        "but allocation of its (large) ravel failed.";
+        new (&shape) Shape();
+        ravel = short_value;
+        throw_apl_error(E_WS_FULL, alloc_loc);
+      }
+*/
+
    // init the first ravel element to (prototype) 0 so that we can avoid
    // many empty checks all over the place
    //
    new (ravel)   IntCell(0);
+   check_ptr = charP(this) + 7;
+   total_ravel_count += length;
+}
+//-----------------------------------------------------------------------------
+bool
+Value::check_WS_FULL(const char * args, ShapeItem requested_cell_count,
+                     const char * loc)
+{
+const uint64_t used_memory
+               = (total_ravel_count + requested_cell_count) * sizeof(Cell)
+               + (value_count + 1) * sizeof(Value)
+               + Workspace::SI_entry_count() * sizeof(StateIndicator);
 
-   check_ptr = (const char *)this + 7;
+   if ((Quad_WA::total_memory*Quad_WA::WA_scale/100) >
+       (used_memory + Quad_WA::WA_margin))   return false;   // OK
+
+   Log(LOG_Value_alloc) CERR
+   << "    value_count:       " << value_count             << endl
+   << "    total_ravel_count: " << total_ravel_count       << " cells" << endl
+   << "    new cell_count:    " << requested_cell_count    << " cells" << endl
+   << "    total_memory:      " << Quad_WA::total_memory   << " bytes" << endl
+   << "    used_memory:       " << used_memory             << " bytes" << endl
+   << "    ⎕WA margin:        " << Quad_WA::WA_margin      << " bytes" << endl
+   << "    ⎕WA scale:         " << Quad_WA::WA_scale       << "%" << endl
+
+           << " at " << LOC << endl;
+   return true;
+}
+//-----------------------------------------------------------------------------
+void
+Value::catch_Error(const Error & error, const char * args, const char * loc)
+{
+   Log(LOG_Value_alloc)   CERR << "Ravel allocation failed" << endl;
+   MORE_ERROR() << "new Value(" << args
+                << ") failed (APL error in ravel allocation)";
+   throw error;   // rethrow
+}
+//-----------------------------------------------------------------------------
+void
+Value::catch_exception(const exception & ex, const char * args,
+                      const char * caller,  const char * loc)
+{
+const uint64_t used_memory
+               = total_ravel_count * sizeof(Cell)
+               + value_count * sizeof(Value)
+               + Workspace::SI_entry_count() * sizeof(StateIndicator);
+
+   CERR << "Value_P::Value_P(" << args << ") failed at " << loc
+        << " (caller: "        << caller << ")" << endl
+        << " what: "           << ex.what() << endl
+        << " initial sbrk(): 0x" << hex << Quad_WA::initial_sbrk << endl
+        << " current sbrk(): 0x" << top_of_memory() << endl
+        << " alloc_size:     0x" << alloc_size << dec << " ("
+                                 << alloc_size << ")" << endl
+        << "  used memory:   0x" << hex  << used_memory << dec
+                                 << " (" << used_memory << ")" << endl;
+
+   MORE_ERROR() << "new Value(" << args << ") failed (" << ex.what() << ")";
+   WS_FULL;
+}
+//-----------------------------------------------------------------------------
+void
+Value::catch_ANY(const char * args, const char * caller, const char * loc)
+{
+   Log(LOG_Value_alloc)
+      CERR << "Value_P::Value_P(Shape " << args << " failed at " << loc
+           << " (caller: " << caller << ")" << endl;
+   MORE_ERROR() << "new Value(" << args << ") failed (ANY)";
+   WS_FULL;
 }
 //-----------------------------------------------------------------------------
 Value::Value(const char * loc)
@@ -198,7 +309,8 @@ Value::Value(const UTF8_string & utf, const char * loc)
    init_ravel();
 
    new (&get_ravel(0)) CharCell(UNI_ASCII_SPACE);   // prototype
-   loop(l, utf.size())   new (next_ravel()) CharCell((Unicode)(utf[l] & 0xFF));
+   loop(l, utf.size())
+       new (next_ravel()) CharCell(Unicode(utf[l] & 0xFF));
    set_complete();
 }
 //-----------------------------------------------------------------------------
@@ -238,7 +350,7 @@ const ShapeItem width = pb.get_width(0);
 //-----------------------------------------------------------------------------
 Value::Value(const char * loc, const Shape * sh)
    : DynamicObject(loc, &all_values),
-     shape((ShapeItem)sh->get_rank()),
+     shape(ShapeItem(sh->get_rank())),
      flags(VF_NONE),
      valid_ravel_items(0)
 {
@@ -260,7 +372,10 @@ Value::~Value()
 
 const ShapeItem length = nz_element_count();
 
+#if !APL_Float_is_class
+   // APL_Float is NOT a class, Therefore release only PointerCells
    if (get_pointer_cell_count() > 0)
+#endif
       {
         Cell * cZ = &get_ravel(0);
         if (is_complete())   // OK to release
@@ -277,13 +392,15 @@ const ShapeItem length = nz_element_count();
 
    --value_count;
 
-   if (ravel != short_value)
+   if (ravel == 0)   return;   // new() failed
+
+   if (ravel != short_value)   // long value
       {
         total_ravel_count -= length;
         delete [] ravel;
       }
 
-   Assert(check_ptr == (const char *)this + 7);
+   Assert(check_ptr == charP(this) + 7);
    check_ptr = 0;
 }
 //-----------------------------------------------------------------------------
@@ -331,7 +448,7 @@ const int src_incr = new_value->is_scalar() ? 0 : 1;
         Cell * dest = C->get_lval_value();   // can be 0!
         if (dest)   dest->release(LOC);   // free sub-values etc (if any)
 
-        new (dest)   PointerCell(new_value, *cellowner);
+        new (dest)   PointerCell(new_value.get(), *cellowner);
         return;
       }
 
@@ -363,16 +480,14 @@ const ShapeItem ec = nz_element_count();
 
    // find the first lval cell with a non-0 owner
    //
-const Cell * C = &get_ravel(0);
    loop(e, ec)
       {
-        if (C->is_lval_cell())
-           return ((const LvalCell *)C)->get_cell_owner();
+        const Cell & cell = get_ravel(e);
+        if (cell.is_lval_cell())
+           return cell.cLvalCell().get_cell_owner();
 
-        if (C->is_pointer_cell())
-           return  C->get_pointer_value()->get_lval_cellowner();
-
-        ++C;
+        if (cell.is_pointer_cell())
+           return  cell.get_pointer_value()->get_lval_cellowner();
       }
 
    return 0;
@@ -407,9 +522,9 @@ const char * sc = set ? " SET " : " CLEAR ";
 const int new_flags = set ? flags | flag : flags & ~flag;
 const char * chg = flags == new_flags ? " (no change)" : " (changed)";
 
-   CERR << "Value " << (const void *)this
-        << sc << flag_name << " (" << (const void *)flag << ")"
-        << " at " << loc << " now = " << (const void *)(intptr_t)new_flags
+   CERR << "Value " << voidP(this)
+        << sc << flag_name << " (" << HEX(flag) << ")"
+        << " at " << loc << " now = " << HEX(new_flags)
         << chg << endl;
 }
 //-----------------------------------------------------------------------------
@@ -425,11 +540,10 @@ Value::init()
 void
 Value::mark_all_dynamic_values()
 {
-   for (const DynamicObject * dob = DynamicObject::all_values.get_prev();
+   for (DynamicObject * dob = DynamicObject::all_values.get_prev();
         dob != &all_values; dob = dob->get_prev())
        {
-         Value * val = (Value *)dob;
-         val->set_marked();
+         dob->pValue()->set_marked();
        }
 }
 //-----------------------------------------------------------------------------
@@ -457,16 +571,16 @@ Value::rollback(ShapeItem items, const char * loc)
    //
    while (items < nz_element_count())   new (&get_ravel(items++)) IntCell(0);
 
-   ((Value *)this)->alloc_loc = loc;
+   const_cast<Value *>(this)->alloc_loc = loc;
 }
 //-----------------------------------------------------------------------------
 void
 Value::erase_all(ostream & out)
 {
-   for (DynamicObject * vb = DynamicObject::all_values.get_next();
-        vb != &DynamicObject::all_values; vb = vb->get_next())
+   for (const DynamicObject * dob = DynamicObject::all_values.get_next();
+        dob != &DynamicObject::all_values; dob = dob->get_next())
        {
-         Value * v = (Value *)vb;
+         const Value * v = dob->pValue();
          out << "erase_all sees Value:" << endl
              << "  Allocated by " << v->where_allocated() << endl
              << "  ";
@@ -482,31 +596,32 @@ int count = 0;
    Log(LOG_Value__erase_stale)
       CERR << endl << endl << "erase_stale() called from " << loc << endl;
 
-   for (DynamicObject * obj = all_values.get_next();
-        obj != &all_values; obj = obj->get_next())
+   for (DynamicObject * dob = all_values.get_next();
+        dob != &all_values; dob = dob->get_next())
        {
-         if (obj == obj->get_next())   // a loop
+         Value * v = dob->pValue();
+         if (dob == dob->get_next())   // a loop
             {
               CERR << "A loop in DynamicObject::all_values (detected in "
                       "function erase_stale() at object "
-                   << (const void *)obj << "): " << endl;
+                   << voidP(dob) << "): " << endl;
               all_values.print_chain(CERR);
               CERR << endl;
 
-              CERR << " DynamicObject: " << obj << endl;
-              CERR << " Value:         " << (Value *)obj << endl;
-              CERR << *(Value *)obj << endl;
+              CERR << " DynamicObject: " << dob << endl;
+              CERR << " Value:         " << v   << endl;
+              CERR << *v                        << endl;
             }
 
-         Assert(obj != obj->get_next());
-         Value * v = (Value *)obj;
+         Assert(dob != dob->get_next());
          if (v->owner_count)   continue;
 
          ADD_EVENT(v, VHE_Stale, v->owner_count, loc);
 
          Log(LOG_Value__erase_stale)
             {
-              CERR << "Erasing stale Value " << (const void *)obj << ":" << endl
+              CERR << "Erasing stale Value "
+                   << voidP(dob) << ":" << endl
                    << "  Allocated by " << v->where_allocated() << endl
                    << "  ";
               v->list_one(CERR, false);
@@ -516,12 +631,12 @@ int count = 0;
          //
          ++count;
 
-         obj->unlink();
+         dob->unlink();
 
          // v->erase(loc) could mess up the chain, so we start over
          // rather than continuing
          //
-         obj = &all_values;
+         dob = &all_values;
        }
 
    return count;
@@ -548,24 +663,24 @@ int count = 0;
              << "finish_incomplete() called from " << loc << endl;
       }
 
-   for (DynamicObject * obj = all_values.get_next();
-        obj != &all_values; obj = obj->get_next())
+   for (DynamicObject * dob = all_values.get_next();
+        dob != &all_values; dob = dob->get_next())
        {
-         if (obj == obj->get_next())   // a loop
+         Value * v = dob->pValue();
+         if (dob == dob->get_next())   // a loop
             {
               CERR << "A loop in DynamicObject::all_values (detected in "
                       "function Value::finish_incomplete() at object "
-                   << (const void *)obj << "): " << endl;
+                   << voidP(dob) << "): " << endl;
               all_values.print_chain(CERR);
               CERR << endl;
 
-              CERR << " DynamicObject: " << obj << endl;
-              CERR << " Value:         " << (Value *)obj << endl;
-              CERR << *(Value *)obj << endl;
+              CERR << " DynamicObject: " << dob << endl;
+              CERR << " Value:         " << v   << endl;
+              CERR << *v                        << endl;
             }
 
-         Assert(obj != obj->get_next());
-         Value * v = (Value *)obj;
+         Assert(dob != dob->get_next());
          if (v->flags & VF_complete)   continue;
 
          ADD_EVENT(v, VHE_Completed, v->owner_count, LOC);
@@ -578,7 +693,7 @@ int count = 0;
          Log(LOG_Value__erase_stale)
             {
               CERR << "Fixed incomplete Value "
-                   << (const void *)obj << ":" << endl
+                   << voidP(dob) << ":" << endl
                    << "  Allocated by " << v->where_allocated() << endl
                    << "  ";
               v->list_one(CERR, false);
@@ -591,7 +706,7 @@ int count = 0;
 }
 //-----------------------------------------------------------------------------
 ostream &
-Value::list_one(ostream & out, bool show_owners)
+Value::list_one(ostream & out, bool show_owners) const
 {
    if (flags)
       {
@@ -613,7 +728,7 @@ Value::list_one(ostream & out, bool show_owners)
 
    // print owners...
    //
-   out << "Owners of " << (const void *)(intptr_t)this << ":" << endl;
+   out << "Owners of " << voidP(this) << ":" << endl;
 
    Workspace::show_owners(out, *this);
 
@@ -625,11 +740,11 @@ ostream &
 Value::list_all(ostream & out, bool show_owners)
 {
 int num = 0;
-   for (DynamicObject * vb = all_values.get_prev();
-        vb != &all_values; vb = vb->get_prev())
+   for (const DynamicObject * dob = all_values.get_prev();
+        dob != &all_values; dob = dob->get_prev())
        {
          out << "Value #" << num++ << ":";
-         ((Value *)vb)->list_one(out, show_owners);
+         dob->pValue()->list_one(out, show_owners);
        }
 
    return out << endl;
@@ -693,7 +808,8 @@ ShapeItem ec = element_count();
             }
          else if (left)
             {
-              new (dest++) LvalCell((Cell *)&cell, &dest_owner);
+              new (dest++)
+                  LvalCell(const_cast<Cell *>(&cell), &dest_owner);
             }
          else
             {
@@ -1013,11 +1129,11 @@ const ShapeItem ec_z = Z->element_count();
    // construct iterators.
    // We go from lower indices to higher indices in IX, which
    // means from higher indices to lower indices in this and Z
-   
+   //
    loop(z, ec_z)
-       Z->next_ravel()->init(get_ravel(mult.next()), Z.getref(), LOC);
+       Z->next_ravel()->init(get_ravel(mult++), Z.getref(), LOC);
 
-   Assert(mult.done());
+   Assert(!mult.more());
    Z->check_value(LOC);
    return Z;
 }
@@ -1056,8 +1172,10 @@ const Cell * cI = &X->get_ravel(0);
          const ShapeItem idx = cI++->get_near_int() - qio;
          if (idx < 0 || idx >= max_idx)
             {
-              MORE_ERROR() << "⎕IO=" << qio << " offending index=" << idx
-                           << " max index=" << max_idx;
+              MORE_ERROR() << "min index=⎕IO (=" << qio
+                           <<  "), offending index=" << (idx + qio)
+                           << ", max index=⎕IO+" << (max_idx - 1)
+                           << " (=" << (max_idx + qio - 1) << ")";
               Z->rollback(Z->valid_ravel_items, LOC);
               INDEX_ERROR;
             }
@@ -1178,7 +1296,7 @@ Value_P Z(len_A + 1, LOC);
       }
    else
       {
-        new (Z->next_ravel()) PointerCell(B, Z.getref());
+        new (Z->next_ravel()) PointerCell(B.get(), Z.getref());
       }
 
    Z->check_value(LOC);
@@ -1208,7 +1326,7 @@ Value_P Z(len_B + 1, LOC);
       }
    else
       {
-        new (Z->next_ravel()) PointerCell(A, Z.getref());
+        new (Z->next_ravel()) PointerCell(A.get(), Z.getref());
       }
 
    loop(b, len_B)   Z->next_ravel()->init(B->get_ravel(b), Z.getref(), LOC);
@@ -1236,7 +1354,7 @@ Value_P Z(2, LOC);
       }
    else
       {
-        new (Z->next_ravel()) PointerCell(A, Z.getref());
+        new (Z->next_ravel()) PointerCell(A.get(), Z.getref());
       }
 
    if (B->is_simple_scalar())
@@ -1245,7 +1363,7 @@ Value_P Z(2, LOC);
       }
    else
       {
-        new (Z->next_ravel()) PointerCell(B, Z.getref());
+        new (Z->next_ravel()) PointerCell(B.get(), Z.getref());
       }
 
    Z->check_value(LOC);
@@ -1469,12 +1587,12 @@ PrintContext pctx = Workspace::get_PrintContext(PR_APL);
            {
              return out << endl;
            }
-            
+
         pctx.set_style(PR_APL_MIN);
       }
    else                  // matrix or higher
       {
-        pctx.set_style((PrintStyle)(pctx.get_style() | PST_NO_FRACT_0));
+        pctx.set_style(PrintStyle(pctx.get_style() | PST_NO_FRACT_0));
       }
 
 PrintBuffer pb(*this, pctx, &out);   // constructor prints it
@@ -1494,27 +1612,47 @@ int style = pctx.get_style();
         style |= PST_NO_FRACT_0;
       }
 
-   pctx.set_style((PrintStyle)style);
+   pctx.set_style(PrintStyle(style));
 
 PrintBuffer pb(*this, pctx, &out);
    return out;
 }
 //-----------------------------------------------------------------------------
 ostream &
-Value::print_properties(ostream & out, int indent) const
+Value::print_properties(ostream & out, int indent, bool help) const
 {
 UCS_string ind(indent, UNI_ASCII_SPACE);
-   out << ind << "Addr:    " << (const void *)this << endl
-       << ind << "Rank:    " << get_rank()  << endl
-       << ind << "Shape:   " << get_shape() << endl
-       << ind << "Flags:   " << get_flags();
-   if (is_complete())   out << " VF_complete";
-   if (is_marked())     out << " VF_marked";
-   out << endl
-       << ind << "First:   " << get_ravel(0)  << endl
-       << ind << "Dynamic: ";
+   if (help)
+      {
+        out << ind << "Rank:  " << get_rank()  << endl
+            << ind << "Shape:";
+        loop(r, get_rank())   out << " " << get_shape_item(r);
+        out << endl
+            << ind << "Depth: " << compute_depth()   << endl
+            << ind << "Type:  ";
 
-   DynamicObject::print(out);
+       const CellType types = deep_cell_types();
+       if (!(types & CT_CHAR))           // no chars in this value
+          out << "numeric";
+       else if (!(types & CT_NUMERIC))   // no numbers in this value
+          out << "character";
+       else                              // chars and numbers in this value
+          out << "mixed";
+      }
+   else
+      {
+        out << ind << "Addr:    " << voidP(this) << endl
+            << ind << "Rank:    " << get_rank()  << endl
+            << ind << "Shape:   " << get_shape() << endl
+            << ind << "Flags:   " << get_flags();
+        if (is_complete())   out << " VF_complete";
+        if (is_marked())     out << " VF_marked";
+        out << endl
+             << ind << "First:   " << get_ravel(0)  << endl
+             << ind << "Dynamic: ";
+
+        DynamicObject::print(out);
+      }
    return out;
 }
 //-----------------------------------------------------------------------------
@@ -1569,7 +1707,7 @@ Value::print_structure(ostream & out, int indent, ShapeItem idx) const
 {
    loop(i, indent)   out << "    ";
    if (indent)   out << "[" << idx << "] ";
-   out << "addr=" << (const void *)this
+   out << "addr=" << voidP(this)
        << " ≡" << compute_depth()
        << " ⍴" << get_shape()
        << " flags: " << HEX4(get_flags()) << "   "
@@ -1588,6 +1726,31 @@ const Cell * c = &get_ravel(0);
 }
 //-----------------------------------------------------------------------------
 Value_P
+Value::clone(const char * loc) const
+{
+#ifdef PERFORMANCE_COUNTERS_WANTED
+const uint64_t start_1 = cycle_counter();
+#endif
+
+Value_P Z(get_shape(), loc);
+
+const Cell * src = &get_ravel(0);
+Cell * dst = &Z->get_ravel(0);
+const ShapeItem count = nz_element_count();
+
+   loop(c, count)   src++->init_other(dst++, Z.getref(), LOC);
+
+   Z->check_value(LOC);
+
+#ifdef PERFORMANCE_COUNTERS_WANTED
+const uint64_t end_1 = cycle_counter();
+   Performance::fs_clone_B.add_sample(end_1 - start_1, count);
+#endif
+
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
 Value::prototype(const char * loc) const
 {
    // the type of an array is an array with the same structure, but all numbers
@@ -1596,61 +1759,21 @@ Value::prototype(const char * loc) const
    // the prototype of an array is the type of the first element of the array.
 
 const Cell & first = get_ravel(0);
-
+   if (first.is_integer_cell())     return IntScalar(0, LOC);
+   if (first.is_character_cell())   return CharScalar(UNI_ASCII_SPACE, LOC);
    if (first.is_pointer_cell())
       {
         Value_P B0 = first.get_pointer_value();
         Value_P Z(B0->get_shape(), loc);
         const ShapeItem ec_Z =  Z->element_count();
 
-        if (ec_Z == 0)
-           {
-             Z->get_ravel(0).init_type(B0->get_ravel(0), Z.getref(), LOC);
-           }
-
         loop(z, ec_Z)
             Z->next_ravel()->init_type(B0->get_ravel(z), Z.getref(), LOC);
         Z->set_complete();
         return Z;
       }
-   else
-      {
-        Value_P Z(loc);
 
-        Z->next_ravel()->init_type(first, Z.getref(), LOC);
-        Z->set_complete();
-        return Z;
-      }
-
-}
-//-----------------------------------------------------------------------------
-Value_P
-Value::clone(const char * loc) const
-{
-#ifdef PERFORMANCE_COUNTERS_WANTED
-#ifdef HAVE_RDTSC
-const uint64_t start_1 = cycle_counter();
-#endif
-#endif
-
-Value_P ret(get_shape(), loc);
-
-const Cell * src = &get_ravel(0);
-Cell * dst = &ret->get_ravel(0);
-const ShapeItem count = nz_element_count();
-
-   loop(c, count)   dst++->init(*src++, ret.getref(), LOC);
-
-   ret->check_value(LOC);
-
-#ifdef PERFORMANCE_COUNTERS_WANTED
-#ifdef HAVE_RDTSC
-const uint64_t end_1 = cycle_counter();
-   Performance::fs_clone_B.add_sample(end_1 - start_1, count);
-#endif
-#endif
-
-   return ret;
+   DOMAIN_ERROR;
 }
 //-----------------------------------------------------------------------------
 /// lrp p.138: S←⍴⍴A + NOTCHAR (per column)
@@ -1706,19 +1829,19 @@ const ShapeItem rows = ec/cols;
 int
 Value::print_incomplete(ostream & out)
 {
-Simple_string<Value *, false> incomplete;
+std::vector<const Value *> incomplete;
 bool goon = true;
 
    for (const DynamicObject * dob = all_values.get_prev();
         goon && (dob != &all_values); dob = dob->get_prev())
        {
-         Value * val = (Value *)dob;
+         const Value * val = dob->pValue();
          goon = (dob != dob->get_prev());
 
          if (val->is_complete())   continue;
 
-         out << "incomplete value at " << (const void *)val << endl;
-         incomplete.append(val);
+         out << "incomplete value at " << voidP(val) << endl;
+         incomplete.push_back(val);
 
          if (!goon)
             {
@@ -1731,10 +1854,8 @@ bool goon = true;
    //
    loop(s, incomplete.size())
       {
-        incomplete[s]->print_stale_info(out,
-               (const DynamicObject *)
-               incomplete[s]);
-       }
+        incomplete[s]->print_stale_info(out, incomplete[s]);
+      }
 
    return incomplete.size();
 }
@@ -1742,8 +1863,8 @@ bool goon = true;
 int
 Value::print_stale(ostream & out)
 {
-Simple_string<Value *, false> stale_vals;
-Simple_string<const DynamicObject *, false> stale_dobs;
+std::vector<const Value *> stale_vals;
+std::vector<const DynamicObject *> stale_dobs;
 bool goon = true;
 int count = 0;
 
@@ -1752,14 +1873,14 @@ int count = 0;
    for (const DynamicObject * dob = all_values.get_prev();
         goon && (dob != &all_values); dob = dob->get_prev())
        {
-         Value * val = (Value *)dob;
+         const Value * val = dob->pValue();
          goon = (dob == dob->get_prev());
 
          if (val->owner_count)   continue;
 
-         out << "stale value at " << (const void *)val << endl;
-         stale_vals.append(val);
-         stale_dobs.append(dob);
+         out << "stale value at " << voidP(val) << endl;
+         stale_vals.push_back(val);
+         stale_dobs.push_back(dob);
 
          if (!goon)
             {
@@ -1773,7 +1894,7 @@ int count = 0;
    loop(s, stale_vals.size())
       {
         const DynamicObject * dob = stale_dobs[s];
-        Value * val = stale_vals[s];
+        const Value * val = stale_vals[s];
         val->print_stale_info(out, dob);
         ++count;
        }
@@ -1789,7 +1910,7 @@ int count = 0;
    for (const DynamicObject * dob = all_values.get_prev();
         dob != &all_values; dob = dob->get_prev())
        {
-         Value * val = (Value *)dob;
+         const Value * val = dob->pValue();
 
          // don't print values found in the previous round.
          //
@@ -1816,12 +1937,12 @@ int count = 0;
 }
 //-----------------------------------------------------------------------------
 void
-Value::print_stale_info(ostream & out, const DynamicObject * dob)
+Value::print_stale_info(ostream & out, const DynamicObject * dob) const
 {
    out << "print_stale_info():   alloc(" << dob->where_allocated()
        << ") flags(" << get_flags() << ")" << endl;
 
-   print_history(out, (const Value *)dob, LOC);
+   VH_entry::print_history(out, dob->pValue(), LOC);
 
    try 
       {
@@ -1843,3 +1964,81 @@ operator<<(ostream & out, const Value & v)
    return out;
 }
 //-----------------------------------------------------------------------------
+Value_P
+IntScalar(APL_Integer val, const char * loc)
+{
+Value_P Z(loc);
+   new (Z->next_ravel())   IntCell(val);
+   Z->check_value(LOC);
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+FloatScalar(APL_Float val, const char * loc)
+{
+Value_P Z(loc);
+   new (Z->next_ravel())   FloatCell(val);
+   Z->check_value(LOC);
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+ComplexScalar(APL_Complex val, const char * loc)
+{
+Value_P Z(loc);
+   new (Z->next_ravel()) ComplexCell(val);
+   Z->check_value(LOC);
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+CharScalar(Unicode uni, const char * loc)
+{
+Value_P Z(loc);
+   new (Z->next_ravel()) CharCell(uni);
+   Z->check_value(LOC);
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+Idx0(const char * loc)
+{
+Value_P Z(ShapeItem(0), loc);
+   Z->check_value(LOC);
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+Str0(const char * loc)
+{
+Value_P Z(ShapeItem(0), loc);
+   Z->set_proto_Spc();
+   Z->check_value(LOC);
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+Str0_0(const char * loc)
+{
+Shape sh(ShapeItem(0), ShapeItem(0));
+Value_P Z(sh, loc);
+   Z->set_proto_Spc();
+   Z->check_value(LOC);
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+Idx0_0(const char * loc)
+{
+Shape sh(ShapeItem(0), ShapeItem(0));
+Value_P Z(sh, loc);
+   Z->check_value(LOC);
+   return Z;
+}
+//-----------------------------------------------------------------------------
+ostream & operator << (ostream & out, const AP_num3 & ap3)
+{
+   return out << ap3.proc << "." << ap3.parent << "." << ap3.grand;
+}
+//-----------------------------------------------------------------------------
+
