@@ -32,7 +32,10 @@ using namespace std;
 #include "IO_Files.hh"
 #include "LibPaths.hh"
 #include "Output.hh"
+#include "Quad_FFT.hh"
 #include "Quad_FX.hh"
+#include "Quad_GTK.hh"
+#include "Quad_PLOT.hh"
 #include "Quad_SQL.hh"
 #include "Quad_TF.hh"
 #include "SystemVariable.hh"
@@ -48,14 +51,14 @@ Workspace::Workspace()
      top_SI(0)
 {
 #define ro_sv_def(x, str, _txt)                                   \
-   if (*str) { UCS_string q(UNI_Quad_Quad);   q.append_utf8(str); \
-   distinguished_names.add_variable(q, ID::x, &v_ ## x); }
+   if (*str) { UCS_string q(UNI_Quad_Quad);   q.append_UTF8(str); \
+   distinguished_names.add_variable(q, ID_ ## x, &v_ ## x); }
 
 #define rw_sv_def ro_sv_def
 
 #define sf_def(x, str, _txt)                                      \
-   if (*str) { UCS_string q(UNI_Quad_Quad);   q.append_utf8(str); \
-   distinguished_names.add_function(q, ID::x, x::fun); }
+   if (*str) { UCS_string q(UNI_Quad_Quad);   q.append_UTF8(str); \
+   distinguished_names.add_function(q, ID_ ## x, x::fun); }
 
 #include "SystemVariable.def"
 }
@@ -65,7 +68,6 @@ Workspace::push_SI(Executable * fun, const char * loc)
 {
    Assert1(fun);
 
-
    if (Quad_SYL::si_depth_limit && SI_top() &&
        Quad_SYL::si_depth_limit <= SI_top()->get_level())
       {
@@ -78,7 +80,45 @@ Workspace::push_SI(Executable * fun, const char * loc)
         set_interrupt_raised(LOC);
       }
 
-   the_workspace.top_SI = new StateIndicator(fun, SI_top());
+   if (Value::check_WS_FULL(__FUNCTION__, 1000, LOC))
+      {
+        if (SI_top() && SI_top()->get_executable() &&
+             the_workspace.SI_entry_count() > 100)
+            MORE_ERROR() <<
+          "Value::check_WS_FULL() complained when calling a defined function.\n"
+          "This was most likely caused by some infinite recursion involving "
+             << SI_top()->get_executable()->get_name();
+        else
+           MORE_ERROR() <<
+           "Value::check_WS_FULL() complained when calling a defined function";
+        WS_FULL;
+      }
+
+   try
+      {
+         StateIndicator * old_top = SI_top();
+         the_workspace.top_SI = new StateIndicator(fun, old_top);
+         if (the_workspace.top_SI == 0)
+            {
+              MORE_ERROR() <<
+              "malloc() → 0 when calling a defined function";
+              the_workspace.top_SI = old_top;
+              WS_FULL;
+            }
+      }
+   catch (const std::bad_alloc & e)
+      {
+         if (SI_top() && SI_top()->get_executable() &&
+             the_workspace.SI_entry_count() > 100)
+            MORE_ERROR() <<
+            "std::bad_alloc exception when calling a defined function\n"
+            "This was most likely caused by some infinite recursion involving "
+            << SI_top()->get_executable()->get_name();
+         else
+            MORE_ERROR() <<
+            "std::bad_alloc exception when calling a defined function";
+         WS_FULL;
+      }
 
    Log(LOG_StateIndicator__push_pop)
       {
@@ -87,7 +127,7 @@ Workspace::push_SI(Executable * fun, const char * loc)
              << " exec=" << fun << " "
              << fun->get_name();
 
-        CERR << " new SI is " << (const void *)SI_top()
+        CERR << " new SI is " << voidP(SI_top())
              << " at " << loc << endl;
       }
 }
@@ -107,7 +147,8 @@ const Executable * exec = SI_top()->get_executable();
 
         if (exec->get_ufun())   CERR << exec->get_ufun()->get_name();
         else                    CERR << SI_top()->get_parse_mode_name();
-        CERR << " " << (const void *)SI_top() << " at " << loc << endl;
+        CERR << " " << voidP(SI_top())
+             << " at " << loc << endl;
       }
 
    // remove the top SI
@@ -141,7 +182,7 @@ Workspace::clear_error(const char * loc)
    //
    for (StateIndicator * si = the_workspace.SI_top(); si; si = si->get_parent())
        {
-         si->get_error().init(E_NO_ERROR, LOC);
+         new (&StateIndicator::get_error(si)) Error(E_NO_ERROR, LOC);
          if (si->get_parse_mode() == PM_FUNCTION)   break;
        }
 }
@@ -162,7 +203,8 @@ Workspace::SI_top_error()
 {
    for (StateIndicator * si = SI_top(); si; si = si->get_parent())
        {
-         if (si->get_error().error_code != E_NO_ERROR)   return si;
+         if (StateIndicator::get_error(si).get_error_code() != E_NO_ERROR)
+            return si;
        }
 
    return 0;   // no context with an error
@@ -177,22 +219,34 @@ Workspace::immediate_execution(bool exit_on_error)
            {
               Command::process_line();
            }
-         catch (Error err)
+         catch (Error & err)
             {
               if (!err.get_print_loc())
                  {
-                   if (err.error_code != E_DEFN_ERROR)
+                   if (err.get_error_code() != E_DEFN_ERROR)
                       {
                         err.print_em(UERR, LOC);
                         CERR << __FUNCTION__ << "() caught APL error "
-                             << HEX(err.error_code) << " ("
-                             << err.error_name(err.error_code) << ")" << endl;
+                             << HEX(err.get_error_code()) << " ("
+                             << err.error_name(err.get_error_code()) << ")"
+                             << endl;
 
                         IO_Files::apl_error(LOC);
                       }
                  }
               if (exit_on_error)   return Token(TOK_OFF);
             }
+         catch (const std::bad_alloc & e)
+            {
+              CERR << "*** " << __FUNCTION__
+                   << "() caught bad_alloc: " << e.what() << " ***" << endl;
+              if (the_workspace.SI_entry_count() > 100)
+                 CERR << ")SI depth is " << the_workspace.SI_entry_count()
+                      <<  "; )SIC is strongly recommended !!!"
+                      << endl;
+              try { WS_FULL } catch (...) {}
+              if (exit_on_error)   return Token(TOK_OFF);
+            }
          catch (...)
             {
               CERR << "*** " << __FUNCTION__
@@ -313,7 +367,7 @@ Workspace::is_called(const UCS_string & funname)
 Symbol * current_referent = lookup_existing_symbol(funname);
    if (current_referent == 0)   return false;   // no such symbol
 
-   Assert(current_referent->get_Id() == ID::USER_SYMBOL);
+   Assert(current_referent->get_Id() == ID_USER_SYMBOL);
 
 const NameClass nc = current_referent->get_nc();
    if (nc != NC_FUNCTION && nc != NC_OPERATOR)   return false;
@@ -363,7 +417,7 @@ Workspace::write_OUT(FILE * out, uint64_t & seq, const UCS_string_vector
                    continue;
                  }
 
-              if (obj->get_Id() == ID::USER_SYMBOL)   // user defined name
+              if (obj->get_Id() == ID_USER_SYMBOL)   // user defined name
                  {
                    const Symbol * sym = lookup_existing_symbol(objects[o]);
                    Assert(sym);
@@ -434,7 +488,8 @@ int count = 0;
    loop(f, the_workspace.expunged_functions.size())
        {
          char cc[100];
-         snprintf(cc, sizeof(cc), "    ⎕EX[%lld] ", (long long)f);
+         const long long lf(f);
+         snprintf(cc, sizeof(cc), "    ⎕EX[%lld] ", lf);
          count += the_workspace.expunged_functions[f]
                                ->show_owners(cc, out, value);
        }
@@ -457,8 +512,8 @@ const int ret = the_workspace.expunged_functions.size();
 
    while(the_workspace.expunged_functions.size())
        {
-         const UserFunction * ufun = the_workspace.expunged_functions.last();
-         the_workspace.expunged_functions.pop();
+         const UserFunction * ufun = the_workspace.expunged_functions.back();
+         the_workspace.expunged_functions.pop_back();
          out << "finally deleting " << ufun->get_name() << "...";
          delete ufun;
          out << " OK" << endl;
@@ -473,7 +528,7 @@ Workspace::clear_WS(ostream & out, bool silent)
 {
    // remove user-defined commands
    //
-   get_user_commands().shrink(0);
+   get_user_commands().clear();
 
    // clear the SI (pops all localized symbols)
    //
@@ -485,7 +540,7 @@ Workspace::clear_WS(ostream & out, bool silent)
 
    // clear the )MORE error info
    //
-   more_error().shrink(0);
+   more_error().clear();
 
    // ⎕PW and ⎕TZ shall survive )CLEAR (lrm p. 260);
    //
@@ -512,6 +567,9 @@ const int tz = the_workspace.v_Quad_TZ.get_offset();
    Workspace::set_PW(pw, LOC);
    the_workspace.v_Quad_TZ.set_offset(tz);
 
+   // close open windows in ⎕GTK
+   Quad_GTK::fun->clear();
+
    // close open files in ⎕FIO
    Quad_FIO::fun->clear();
 
@@ -536,62 +594,32 @@ Workspace::list_SI(ostream & out, SI_mode mode)
    for (const StateIndicator * si = SI_top(); si; si = si->get_parent())
        si->list(out, mode);
 
-   if (mode && SIM_debug)   out << endl;
+   if (mode & SIM_debug)   out << endl;
 }
 //-----------------------------------------------------------------------------
 void
-Workspace::save_WS(ostream & out, UCS_string_vector & lib_ws)
+Workspace::save_WS(ostream & out, LibRef libref, const UCS_string & wsname,
+                   bool name_from_WSID)
 {
-   // )SAVE
-   // )SAVE wsname
-   // )SAVE libnum wsname
-
-bool name_from_WSID = false;
-   if (lib_ws.size() == 0)   // no argument: use )WSID value
-      {
-        name_from_WSID = true;
-        lib_ws.append(the_workspace.WS_name);
-      }
-   else if (lib_ws.size() > 2)   // too many arguments
-      {
-        out << "BAD COMMAND+" << endl;
-        MORE_ERROR() << "too many parameters in command )SAVE";
-        return;
-      }
-
-   // at this point, lib_ws.size() is 1 or 2.
-
-LibRef libref = LIB_NONE;
-UCS_string wname = lib_ws.last();
-   if (lib_ws.size() == 2)   libref = (LibRef)(lib_ws[0].atoi());
-UTF8_string filename = LibPaths::get_lib_filename(libref, wname, false,
+UTF8_string filename = LibPaths::get_lib_filename(libref, wsname, false,
                                                   ".xml", 0);
 
-   if (wname.compare(UCS_string("CLEAR WS")) == 0)   // don't save CLEAR WS
-      {
-        COUT << "NOT SAVED: THIS WS IS CLEAR WS" << endl;
-        MORE_ERROR() <<
-        "the workspace was not saved because 'CLEAR WS' is a special \n"
-        "workspace name that cannot be saved. Use )WSID <name> first.";
-        return;
-      }
-
-   // dont save if workspace names differ and file exists
+   // dont )SAVE if wsname differs from wsid and the file exists
    //
 const bool file_exists = access(filename.c_str(), W_OK) == 0;
    if (file_exists)
       {
-        if (wname.compare(the_workspace.WS_name) != 0)   // names differ
+        if (wsname.compare(the_workspace.WS_name) != 0)   // names differ
            {
-             COUT << "NOT SAVED: THIS WS IS "
-                  << the_workspace.WS_name << endl;
+             out << "NOT SAVED: THIS WS IS "
+                 << the_workspace.WS_name << endl;
 
              MORE_ERROR() << "the workspace was not saved because:\n"
                   << "   the workspace name '" << the_workspace.WS_name
-                  << "' of )WSID\n   does not match the name '" << wname
+                  << "' of )WSID\n   does not match the name '" << wsname
                   << "' used in the )SAVE command\n"
                   << "   and the workspace file\n   " << filename
-                  << "\n   already exists. Use )WSID " << wname << " first."; 
+                  << "\n   already exists. Use )WSID " << wsname << " first."; 
              return;
            }
       }
@@ -608,34 +636,20 @@ const bool file_exists = access(filename.c_str(), W_OK) == 0;
 ofstream outf(filename.c_str(), ofstream::out);
    if (!outf.is_open())   // open failed
       {
-        CERR << "Unable to )SAVE workspace '" << wname
+        CERR << "Unable to )SAVE workspace '" << wsname
              << "'. " << strerror(errno) << endl;
         return;
       }
 
-   the_workspace.WS_name = wname;
+   the_workspace.WS_name = wsname;
 
 XML_Saving_Archive ar(outf);
    ar.save();
 
    // print time and date to COUT
-   {
-     const APL_time_us offset = get_v_Quad_TZ().get_offset();
-     const YMDhmsu time(now() + 1000000*offset);
-     const char * tz_sign = (offset < 0) ? "" : "+";
-
-     ostringstream os;
-     os << setfill('0') << time.year  << "-"
-        << setw(2)      << time.month << "-"
-        << setw(2)      << time.day   << "  " 
-        << setw(2)      << time.hour  << ":"
-        << setw(2)      << time.minute << ":"
-        << setw(2)      << time.second << " (GMT"
-        << tz_sign      << offset/3600 << ")";
-
-     if (name_from_WSID)   os << " " << the_workspace.WS_name;
-     out << os.str() << endl;
-   }
+   get_v_Quad_TZ().print_timestamp(out, now());
+   if (name_from_WSID)   out << " " << the_workspace.WS_name;
+   out << endl;
 }
 //-----------------------------------------------------------------------------
 bool
@@ -646,7 +660,7 @@ Workspace::backup_existing_file(const char * filename)
    if (access(filename, F_OK) != 0)   return false;   // OK
 
 UTF8_string backup_filename = filename;
-   backup_filename.append_str(".bak");
+   backup_filename.append_ASCII(".bak");
 
    // 2. if backup file exists then remove it...
    //
@@ -699,19 +713,8 @@ Workspace::load_DUMP(ostream & out, const UTF8_string & filename, int fd,
    {
      struct stat st;
      fstat(fd, &st);
-     const int offset = Quad_TZ::compute_offset();  // call BEFORE localtime() !
-     tm * t = localtime(&st.st_mtime);
-     const char * tz_sign = offset < 0 ? "" : "+";
-
-      out << "DUMPED "
-          << setfill('0') << (1900 + t->tm_year) << "-"
-          << setw(2)      << (t->tm_mon + 1)     << "-"
-          << setw(2)      << t->tm_mday          << " "
-          << setw(2)      << t->tm_hour          << ":"
-          << setw(2)      << t->tm_min           << ":"
-          << setw(2)      << t->tm_sec           << " (GMT"
-          << tz_sign      << offset/3600 << ")"
-          << setfill(' ') <<  endl;
+     const APL_time_us when = 1000000ULL * st.st_mtime;
+     Workspace::get_v_Quad_TZ().print_timestamp(out << "DUMPED ", when) << endl;
    }
 
 FILE * file = fdopen(fd, "r");
@@ -737,7 +740,7 @@ InputFile fam(filename, file, false, false, true, with_LX);
         fam.set_COPY();
         ++Bif_F1_EXECUTE::copy_pending;
       }
-   InputFile::files_todo.insert_before(0, fam);
+   InputFile::files_todo.insert(InputFile::files_todo.begin(), fam);
 }
 //-----------------------------------------------------------------------------
 /// a streambuf that escapes certain HTML characters
@@ -754,12 +757,12 @@ public:
       {
         switch(c & 0xFF)
            {
-              case '#':  out_file << "&#35;";    break;
+              case '#':  out_file << "&#35;";   break;
               case '%':  out_file << "&#37;";   break;
               case '&':  out_file << "&#38;";   break;
               case '<':  out_file << "&lt;";    break;
               case '>':  out_file << "&gt;";    break;
-              default:   out_file << (char)(c & 0xFF);
+              default:   out_file << char(c & 0xFF);
            }
         return 0;
       }
@@ -780,35 +783,19 @@ public:
 };
 //-----------------------------------------------------------------------------
 void
-Workspace::dump_WS(ostream & out, UCS_string_vector & lib_ws, bool html,
-                   bool silent)
+Workspace::dump_WS(ostream & out, LibRef libref, const UCS_string & wsname,
+                   bool html, bool silent)
 {
    // )DUMP
    // )DUMP wsname
    // )DUMP libnum wsname
 
-   if (lib_ws.size() == 0)   // no argument: use )WSID value
-      {
-         lib_ws.append(the_workspace.WS_name);
-      }
-   else if (lib_ws.size() > 2)   // too many arguments
-      {
-        out << "BAD COMMAND+" << endl;
-        MORE_ERROR() << "too many parameters in command )DUMP";
-        return;
-      }
-
-   // at this point, lib_ws.size() is 1 or 2.
-
-LibRef libref = LIB_NONE;
-const UCS_string wname = lib_ws.last();
-   if (lib_ws.size() == 2)   libref = (LibRef)(lib_ws[0].atoi());
 const char * extension = html ? ".html" : ".apl";
-UTF8_string filename = LibPaths::get_lib_filename(libref, wname, false,
+UTF8_string filename = LibPaths::get_lib_filename(libref, wsname, false,
                                                   extension, 0);
-   if (wname.compare(UCS_string("CLEAR WS")) == 0)   // don't save CLEAR WS
+   if (wsname.compare(UCS_string("CLEAR WS")) == 0)   // don't save CLEAR WS
       {
-        COUT << "NOT DUMPED: THIS WS IS " << wname << endl;
+        COUT << "NOT DUMPED: THIS WS IS " << wsname << endl;
         MORE_ERROR() <<
         "the workspace was not dumped because 'CLEAR WS' is a special\n"
         "workspace name that cannot be dumped. Use )WSID <name> first.";
@@ -826,8 +813,9 @@ ofstream outf(filename.c_str(), ofstream::out);
 
    if (!outf.is_open())   // open failed
       {
-        CERR << "Unable to )DUMP workspace '" << wname
-             << "': " << strerror(errno) << endl;
+        CERR << "Unable to )DUMP workspace '" << wsname
+             << "': " << strerror(errno) << "." << endl
+             << "    NOTE: filename: " << filename << endl;
         return;
       }
 
@@ -838,10 +826,9 @@ ostream * sout = &outf;
    // print header line, workspace name, time, and date to outf
    //
 const APL_time_us offset = get_v_Quad_TZ().get_offset();
-const YMDhmsu time(now() + 1000000*offset);
-const char * tz_sign = (offset < 0) ? "" : "+";
+const APL_time_us gmt = now();
+const YMDhmsu time(gmt + 1000000*offset);
    {
-
      if (html)
         {
           outf <<
@@ -849,7 +836,7 @@ const char * tz_sign = (offset < 0) ? "" : "+";
 "                      \"http://www.w3.org/TR/html4/strict.dtd\">"  << endl <<
 "<html>"                                                            << endl <<
 "  <head>"                                                          << endl <<
-"    <title>" << wname << ".apl</title> "                           << endl <<
+"    <title>" << wsname << ".apl</title> "                          << endl <<
 "    <meta http-equiv=\"content-type\" "                            << endl <<
 "          content=\"text/html; charset=UTF-8\">"                   << endl <<
 "    <meta name=\"author\" content=\"??????\">"                     << endl <<
@@ -891,16 +878,8 @@ const char * tz_sign = (offset < 0) ? "" : "+";
 
      *sout << " ⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝" << endl
           << "⍝" << endl
-          << "⍝ " << wname << " "
-          << setfill('0') << time.year  << "-"
-          << setw(2)      << time.month << "-"
-          << setw(2)      << time.day   << " " 
-          << setw(2)      << time.hour  << ":"
-          << setw(2)      << time.minute << ":"
-          << setw(2)      << time.second << " (GMT"
-          << tz_sign      << offset/3600 << ")"
-          << setfill(' ') << endl
-          << "⍝" << endl
+          << "⍝ " << wsname << " ";
+     get_v_Quad_TZ().print_timestamp(*sout, gmt) << endl
           << " ⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝" << endl
           << endl;
    }
@@ -913,7 +892,7 @@ int variable_count = 0;
    // system variables
    //
 #define ro_sv_def(x, _str, _txt)
-#define rw_sv_def(x, _str, _txt) if (ID:: x != ID::Quad_SYL) \
+#define rw_sv_def(x, _str, _txt) if (ID_ ## x != ID_Quad_SYL) \
    { get_v_ ## x().dump(*sout);   ++variable_count; }
 #include "SystemVariable.def"
 
@@ -921,18 +900,11 @@ int variable_count = 0;
 
    if (silent)
       {
-        out << setfill('0') << time.year        << "-"
-            << setw(2)      << time.month       << "-"
-            << setw(2)      << time.day         << " "
-            << setw(2)      << time.hour        << ":"
-            << setw(2)      << time.minute      << ":"
-            << setw(2)      << time.second      << " (GMT"
-            << tz_sign      << offset/3600 << ")"
-            << setfill(' ') <<  endl;
+        get_v_Quad_TZ().print_timestamp(out, gmt) << endl;
       }
    else
       {
-        out << "DUMPED WORKSPACE '" << wname << "'" << endl
+        out << "DUMPED WORKSPACE '" << wsname << "'" << endl
             << " TO FILE '" << filename << "'" << endl
             << " (" << function_count << " FUNCTIONS, " << variable_count
             << " VARIABLES)" << endl;
@@ -942,7 +914,7 @@ int variable_count = 0;
 void
 Workspace::dump_commands(ostream & out)
 {
-Simple_string<Command::user_command, false> & cmds = get_user_commands();
+vector<Command::user_command> & cmds = get_user_commands();
 
    loop(c, cmds.size())
       out << "]USERCMD " << cmds[c].prefix
@@ -954,30 +926,31 @@ Simple_string<Command::user_command, false> & cmds = get_user_commands();
 //-----------------------------------------------------------------------------
 // )LOAD WS, set ⎕LX of loaded WS on success
 void
-Workspace::load_WS(ostream & out, const UCS_string_vector & lib_ws,
+Workspace::load_WS(ostream & out, LibRef libref, const UCS_string & wsname,
                    UCS_string & quad_lx, bool silent)
 {
-   if (lib_ws.size() < 1 || lib_ws.size() > 2)   // no or too many argument(s)
+   // )LOAD wsname                              wsname /absolute or relative
+   // )LOAD libnum wsname                       wsname relative
+   //
+   if (UTF8_string(wsname).ends_with(".bak"))
       {
         out << "BAD COMMAND+" << endl;
-        MORE_ERROR() << "too many parameters in command )LOAD";
+        MORE_ERROR() << wsname <<
+        " is a backup file! You need to rename it before it can be )LOADed";
         return;
       }
 
-LibRef libref = LIB_NONE;
-   if (lib_ws.size() == 2)   libref = (LibRef)(lib_ws[0].atoi());
-UCS_string wname = lib_ws.last();
-UTF8_string filename = LibPaths::get_lib_filename(libref, wname, true,
+UTF8_string filename = LibPaths::get_lib_filename(libref, wsname, true,
                                                   ".xml", ".apl");
 
 int dump_fd = -1;
 XML_Loading_Archive in(filename.c_str(), dump_fd);
 
-   if (dump_fd != -1)
+   if (dump_fd != -1)   // wsname.apl
       {
         the_workspace.clear_WS(out, true);
 
-        Log(LOG_command_IN)   out << "LOADING " << wname << " from file '"
+        Log(LOG_command_IN)   out << "LOADING " << wsname << " from file '"
                                   << filename << "' ..." << endl;
 
         load_DUMP(out, filename, dump_fd, do_LX, silent, 0);   // closes dump_fd
@@ -993,24 +966,23 @@ XML_Loading_Archive in(filename.c_str(), dump_fd);
            wsid_end[-3] == 'a' &&
            wsid_end[-2] == 'p' &&
            wsid_end[-1] == 'l')   wsid_end -= 4;   // skip .apl extension
-        const UTF8_string wsid_utf8((const UTF8 *)wsid_start,
-                                    wsid_end - wsid_start);
+        const UTF8_string wsid_utf8(utf8P(wsid_start), wsid_end - wsid_start);
         const UCS_string wsid_ucs(wsid_utf8);
-        wsid(out, wsid_ucs, true);
+        wsid(out, wsid_ucs, libref, true);
 
         // we cant set ⎕LX because it was not executed yet.
         return;
       }
-   else
+   else   // wsname.xml
       {
         if (!in.is_open())   // open failed
            {
-             out << ")LOAD " << wname << " (file " << filename
+             out << ")LOAD " << wsname << " (file " << filename
                  << ") failed: " << strerror(errno) << endl;
              return;
            }
 
-        Log(LOG_command_IN)   out << "LOADING " << wname << " from file '"
+        Log(LOG_command_IN)   out << "LOADING " << wsname << " from file '"
                                   << filename << "' ..." << endl;
 
         // got open file. We assume that from here on everything will be fine.
@@ -1024,39 +996,15 @@ XML_Loading_Archive in(filename.c_str(), dump_fd);
 }
 //-----------------------------------------------------------------------------
 void
-Workspace::copy_WS(ostream & out, UCS_string_vector & lib_ws_objects,
-                   bool protection)
+Workspace::copy_WS(ostream & out, LibRef libref, const UCS_string & wsname,
+                   UCS_string_vector & lib_ws_objects, bool protection)
 {
    // )COPY wsname                              wsname /absolute or relative
    // )COPY libnum wsname                       wsname relative
    // )COPY wsname objects...                   wsname /absolute or relative
    // )COPY libnum wsname objects...            wsname relative
 
-   if (lib_ws_objects.size() < 1)   // at least workspace name is required
-      {
-        out << "BAD COMMAND+" << endl;
-        MORE_ERROR() << "missing parameter(s) in command )COPY or )PCOPY";
-        return;
-      }
-
-LibRef libref = LIB_NONE;
-   if (Avec::is_digit(lib_ws_objects[0][0]))
-      {
-        libref = (LibRef)(lib_ws_objects[0].atoi());
-        lib_ws_objects.erase(0);
-      }
-
-   if (lib_ws_objects.size() < 1)   // at least workspace name is required
-      {
-        out << "BAD COMMAND+" << endl;
-        MORE_ERROR() << "missing parameter(s) in command )COPY or )PCOPY";
-        return;
-      }
-
-UCS_string wname = lib_ws_objects[0];
-   lib_ws_objects.erase(0);
-
-UTF8_string filename = LibPaths::get_lib_filename(libref, wname, true,
+UTF8_string filename = LibPaths::get_lib_filename(libref, wsname, true,
                                                   ".xml", ".apl");
 
 int dump_fd = -1;
@@ -1070,12 +1018,12 @@ XML_Loading_Archive in(filename.c_str(), dump_fd);
 
    if (!in.is_open())   // open failed: try filename.xml unless already .xml
       {
-        CERR << ")COPY " << wname << " (file " << filename
+        CERR << ")COPY " << wsname << " (file " << filename
              << ") failed: " << strerror(errno) << endl;
         return;
       }
 
-   Log(LOG_command_IN)   CERR << "LOADING " << wname << " from file '"
+   Log(LOG_command_IN)   CERR << "LOADING " << wsname << " from file '"
                               << filename << "' ..." << endl;
 
    in.set_protection(protection, lib_ws_objects);
@@ -1084,10 +1032,9 @@ XML_Loading_Archive in(filename.c_str(), dump_fd);
 }
 //-----------------------------------------------------------------------------
 void
-Workspace::wsid(ostream & out, UCS_string arg, bool silent)
+Workspace::wsid(ostream & out, UCS_string arg, LibRef lib, bool silent)
 {
-   while (arg.size() && arg[0] <= ' ')       arg.erase(0);
-   while (arg.size() && arg.back() <= ' ')   arg.pop();
+   arg.remove_leading_and_trailing_whitespaces();
 
    if (arg.size() == 0)   // inquire workspace name
       {
@@ -1095,6 +1042,17 @@ Workspace::wsid(ostream & out, UCS_string arg, bool silent)
         return;
       }
 
+   if (lib == LIB_WSNAME)     // i.e. arg may start with a library number
+      {
+        if (arg[0] >= '0' && arg[0] <= '9')   // it does
+           {
+             lib = LibRef(arg[0] - '0');
+             arg.erase(0);                       // skip the library number
+             arg.remove_leading_whitespaces();   // and blanks after it
+           }
+        else lib = LIB0;
+      }
+
    loop(a, arg.size())
       {
         if (arg[a] <= ' ')
@@ -1105,7 +1063,22 @@ Workspace::wsid(ostream & out, UCS_string arg, bool silent)
       }
 
    if (!silent)   out << "WAS " << the_workspace.WS_name << endl;
+
+   // prepend the lib number unless it is 0.
+   //
+   if (lib != LIB0)
+      {
+        arg.prepend(UNI_ASCII_SPACE);
+        arg.prepend(Unicode('0' + lib));
+      }
    the_workspace.WS_name = arg;
 }
 //-----------------------------------------------------------------------------
+UCS_string &
+MORE_ERROR()
+{
+   Workspace::more_error().clear();
+   return Workspace::more_error();
+}
+//-----------------------------------------------------------------------------
 
