@@ -28,9 +28,9 @@ void SystemVariableCommand::run_command( NetworkConnection &conn, const std::vec
 {
     stringstream out;
 
-#define ro_sv_def(VAR, _str, txt) out << ID::get_name( ID::VAR ) << "\n";
-#define rw_sv_def(VAR, _str, txt) out << ID::get_name( ID::VAR ) << "\n";
-#define sf_def(FUN, _str, txt)    out << ID::get_name( ID::FUN ) << "\n";
+#define ro_sv_def(VAR, _str, txt) out << ID::get_name( ID_VAR ) << "\n";
+#define rw_sv_def(VAR, _str, txt) out << ID::get_name( ID_VAR ) << "\n";
+#define sf_def(FUN, _str, txt)    out << ID::get_name( ID_FUN ) << "\n";
 #include "../SystemVariable.def"
 
     out << END_TAG << "\n";
