@@ -53,17 +53,17 @@ void TraceData::remove_listener( NetworkConnection *connection )
 void TraceData::display_value_for_trace( ostream &out, const Value_P &value, int cr_level )
 {
     if( cr_level < 0 ) {
-        PrintContext context( PST_NONE, Workspace::get_PrintContext().get_PP(), 100000 );
+        PrintContext context( PST_NONE, Workspace::get_PP(), 100000 );
         value->print1( out, context );
     }
     else {
         if( cr_level < 1 || cr_level > 9 ) {
             throw ConnectionError( "Illegal CR level" );
         }
-        PrintContext context( PST_NONE, Workspace::get_PrintContext().get_PP(), 100000 );
-        Value_P cr_formatted = Quad_CR::do_CR( cr_level, *value, context );
+        PrintContext context( PST_NONE, Workspace::get_PP(), 100000 );
+        Value_P cr_formatted = Quad_CR::do_CR( cr_level, value.get(), context);
 
-        PrintContext context2( PST_NONE, Workspace::get_PrintContext().get_PP(), 100000 );
+        PrintContext context2( PST_NONE, Workspace::get_PP(), 100000 );
         cr_formatted->print1( out, context2 );
     }
 }
