@@ -22,16 +22,17 @@
 #include <sstream>
 #include <ostream>
 
-#include <Workspace.hh>
 #include <Command.hh>
 #include <ComplexCell.hh>
 #include <DiffOut.hh>
 #include <Error.hh>
-#include <LineInput.hh>
 #include <FloatCell.hh>
+#include <LineInput.hh>
+#include <Macro.hh>
 #include <PointerCell.hh>
 #include <Tokenizer.hh>
 #include <UserPreferences.hh>
+#include <Workspace.hh>
 
 #include "libapl.h"
 
@@ -48,27 +49,27 @@ int_scalar(int64_t val, const char * loc)
 {
 Value_P Z(loc);
    new (Z->next_ravel()) IntCell(val);
-   Value_P::increment_owner_count(Z.get(), loc);   // keep value
+   Z.get()->increment_owner_count(loc);   // keep value
    return Z.get();
 }
 //-----------------------------------------------------------------------------
 /// A new floating point scalar.
 APL_value
-double_scalar(double val, const char * loc)
+double_scalar(APL_Float val, const char * loc)
 {
 Value_P Z(loc);
    new (Z->next_ravel()) FloatCell(val);
-   Value_P::increment_owner_count(Z.get(), loc);   // keep value
+   Z.get()->increment_owner_count(loc);   // keep value
    return Z.get();
 }
 //-----------------------------------------------------------------------------
 /// A new complex scalar.
 APL_value
-complex_scalar(double real, double imag, const char * loc)
+complex_scalar(APL_Float real, APL_Float imag, const char * loc)
 {
 Value_P Z(loc);
    new (Z->next_ravel()) ComplexCell(real, imag);
-   Value_P::increment_owner_count(Z.get(), loc);   // keep value
+   Z.get()->increment_owner_count(loc);   // keep value
    return Z.get();
 }
 //-----------------------------------------------------------------------------
@@ -77,8 +78,8 @@ APL_value
 char_scalar(int uni, const char * loc)
 {
 Value_P Z(loc);
-   new (Z->next_ravel()) CharCell((Unicode)uni);
-   Value_P::increment_owner_count(Z.get(), loc);   // keep value
+   new (Z->next_ravel()) CharCell(Unicode(uni));
+   Z.get()->increment_owner_count(loc);   // keep value
    return Z.get();
 }
 //-----------------------------------------------------------------------------
@@ -94,7 +95,7 @@ Value_P Z(sh, loc);
    while (Cell * cell = Z->next_ravel())   new (cell)   IntCell(0);
 
    Z->check_value(LOC);
-   Value_P::increment_owner_count(Z.get(), loc);   // keep value
+   Z.get()->increment_owner_count(loc);   // keep value
    return Z.get();
 }
 //-----------------------------------------------------------------------------
@@ -106,7 +107,7 @@ UTF8_string utf8(str);
 UCS_string ucs(utf8);
 
 Value_P Z(ucs, loc);
-   Value_P::increment_owner_count(Z.get(), loc);   // keep value
+   Z.get()->increment_owner_count(loc);   // keep value
    return Z.get();
 }
 
@@ -118,8 +119,8 @@ Value_P Z(ucs, loc);
 void
 release_value(const APL_value val, const char * loc)
 {
-Value * v = (Value *) val;
-   if (val)   Value_P::decrement_owner_count(v, loc);
+Value * v = const_cast<Value *>(val);
+   if (val)   v->decrement_owner_count(loc);
 }
 
 
@@ -140,8 +141,7 @@ get_rank(const APL_value val)
 int64_t
 get_axis(const APL_value val, unsigned int axis)
 {
-   return axis < (unsigned int)(val->get_rank())
-          ? val->get_shape_item(axis) : -1;
+   return uRank(axis) < val->get_rank() ? val->get_shape_item(axis) : -1;
 }
 //-----------------------------------------------------------------------------
 
@@ -156,7 +156,7 @@ get_element_count(const APL_value val)
 int
 get_type(const APL_value val, uint64_t idx)
 {
-   if (idx >= (uint64_t)(val->nz_element_count()))   return 0;
+   if (idx >= uint64_t(val->nz_element_count()))   return 0;
    return val->get_ravel(idx).get_cell_type();
 }
 //-----------------------------------------------------------------------------
@@ -185,7 +185,7 @@ get_int(const APL_value val, uint64_t idx)
 //-----------------------------------------------------------------------------
 
 /// return the real part of val[idx] (after having checked is_numeric())
-double
+APL_Float
 get_real(const APL_value val, uint64_t idx)
 {
    return val->get_ravel(idx).get_real_value();
@@ -193,7 +193,7 @@ get_real(const APL_value val, uint64_t idx)
 //-----------------------------------------------------------------------------
 
 /// return the imag part of val[idx] (after having checked is_numeric())
-double
+APL_Float
 get_imag(const APL_value val, uint64_t idx)
 {
    return val->get_ravel(idx).get_imag_value();
@@ -207,7 +207,7 @@ APL_value
 get_value(const APL_value val, uint64_t idx)
 {
 Value_P sub = val->get_ravel(idx).get_pointer_value();
-   Value_P::increment_owner_count(sub.get(), LOC);
+   sub.get()->increment_owner_count(LOC);   // keep value
    return sub.get();
 }
 /******************************************************************************
@@ -232,17 +232,18 @@ Value_P Z(sh, LOC);
         // and the caller is responsiblr fopr decrementing it when the
         // value is no longer needed.
         //
-        Value_P::increment_owner_count(Z.get(), LOC);
+        Z.get()->increment_owner_count(LOC);   // keep value
         return Z.get();
       }
 
-   if (!Avec::is_first_symbol_char((Unicode)*var_name_ucs))   return 0;
+   if (!Avec::is_first_symbol_char(Unicode(*var_name_ucs)))
+      return 0;
 
 UCS_string var_name;
    var_name.reserve(40);
    while (*var_name_ucs)
       {
-        const Unicode uni = (Unicode)*var_name_ucs++;
+        const Unicode uni = Unicode(*var_name_ucs++);
         if (!Avec::is_symbol_char(uni))   return 0;
         var_name.append(uni);
       }
@@ -268,10 +269,10 @@ Cell * cell = &val->get_ravel(idx);
    if (cell->is_pointer_cell())
       {
         Value * v = cell->get_pointer_value().get();
-        Value_P::decrement_owner_count(v, LOC);
+        v->decrement_owner_count(LOC);
       }
 
-   new (cell)   CharCell((Unicode)new_char);
+   new (cell)   CharCell(Unicode(new_char));
 }
 //-----------------------------------------------------------------------------
 
@@ -283,7 +284,7 @@ Cell * cell = &val->get_ravel(idx);
    if (cell->is_pointer_cell())
       {
         Value * v = cell->get_pointer_value().get();
-        Value_P::decrement_owner_count(v, LOC);
+        v->decrement_owner_count(LOC);
       }
 
    new (cell)   IntCell(new_int);
@@ -292,13 +293,13 @@ Cell * cell = &val->get_ravel(idx);
 
 /// val[idx]←new_double
 void
-set_double(double new_double, APL_value val, uint64_t idx)
+set_double(APL_Float new_double, APL_value val, uint64_t idx)
 {
 Cell * cell = &val->get_ravel(idx);
    if (cell->is_pointer_cell())
       {
         Value * v = cell->get_pointer_value().get();
-        Value_P::decrement_owner_count(v, LOC);
+        v->decrement_owner_count(LOC);
       }
 
    new (cell)   FloatCell(new_double);
@@ -307,13 +308,13 @@ Cell * cell = &val->get_ravel(idx);
 
 /// val[idx]←new_complex
 void
-set_complex(double new_real, double new_imag, APL_value val, uint64_t idx)
+set_complex(APL_Float new_real, APL_Float new_imag, APL_value val, uint64_t idx)
 {
 Cell * cell = &val->get_ravel(idx);
    if (cell->is_pointer_cell())
       {
         Value * v = cell->get_pointer_value().get();
-        Value_P::decrement_owner_count(v, LOC);
+        v->decrement_owner_count(LOC);
       }
 
    new (cell)   ComplexCell(new_real, new_imag);
@@ -326,7 +327,7 @@ Cell * cell = &val->get_ravel(idx);
    if (cell->is_pointer_cell())
       {
         Value * v = cell->get_pointer_value().get();
-        Value_P::decrement_owner_count(v, LOC);
+        v->decrement_owner_count(LOC);
       }
 
    if (new_value->is_simple_scalar())   // e.g. ⊂5 is 5
@@ -337,11 +338,13 @@ Cell * cell = &val->get_ravel(idx);
       {
         const Cell & src = new_value->get_ravel(0);
         if (!src.is_pointer_cell())   DOMAIN_ERROR;
-        new (cell)   PointerCell(src.get_pointer_value()->clone(LOC), *val);
+        Value_P sub = src.get_pointer_value()->clone(LOC);
+        new (cell)   PointerCell(sub.get(), *val);
       }
    else
       {
-        new (cell)   PointerCell(new_value->clone(LOC), *val);
+        Value_P sub = new_value->clone(LOC);
+        new (cell)   PointerCell(sub.get(), *val);
       }
 }
 
@@ -359,7 +362,7 @@ const StateIndicator * si = Workspace::SI_top();
    if (si == Workspace::SI_top())   return E_NO_ERROR;
 
    si = Workspace::SI_top_error();
-   if (si)   return si->get_error().error_code;
+   if (si)   return StateIndicator::get_error(si).get_error_code();
    return E_UNKNOWN_ERROR;
 } 
 //-----------------------------------------------------------------------------
@@ -368,16 +371,16 @@ apl_exec_ucs(const unsigned int * line_ucs)
 { 
 UCS_string line;
    line.reserve(200);
-   while (*line_ucs)   line.append((Unicode)*line_ucs++);
+   while (*line_ucs)   line.append(Unicode(*line_ucs++));
 
 const StateIndicator * si = Workspace::SI_top();
   Command::process_line(line);
    if (si == Workspace::SI_top())   return E_NO_ERROR;
 
    si = Workspace::SI_top_error();
-   if (si)   return si->get_error().error_code;
+   if (si)   return StateIndicator::get_error(si).get_error_code();
    return E_UNKNOWN_ERROR;
-} 
+}
 //-----------------------------------------------------------------------------
 const char *
 apl_command(const char * command)
@@ -396,7 +399,7 @@ apl_command_ucs(const unsigned int * command)
 {
 UCS_string command_ucs;
    command_ucs.reserve(200);
-   while (*command)   command_ucs.append((Unicode)*command++);
+   while (*command)   command_ucs.append(Unicode(*command++));
 
 ostringstream out;
   Command::do_APL_command(out, command_ucs);
@@ -406,8 +409,8 @@ UTF8_string result_utf8(out.str().c_str());
 
 UCS_string result_ucs(result_utf8);
 
-unsigned int * ret = (unsigned int *)
-                     malloc((result_ucs.size() + 1) * sizeof(int));
+unsigned int * ret = reinterpret_cast<unsigned int *>
+                     (malloc((result_ucs.size() + 1) * sizeof(int)));
    loop(l, result_ucs.size())   ret[l] = result_ucs[l];
    ret[result_ucs.size()] = 0;
    return ret;
@@ -424,7 +427,7 @@ get_function_ucs(const unsigned int * name, APL_function * L, APL_function * R)
 {
 UCS_string function_ucs;
    function_ucs.reserve(40);
-   while (*name)   function_ucs.append((Unicode)*name++);
+   while (*name)   function_ucs.append(Unicode(*name++));
 
    if (function_ucs.size() == 0)   return 0;   // empty name
 
@@ -434,7 +437,7 @@ Token_string tos;
 
    // resolve user defined names to user defined functions
    //
-   for (int j = 0; j < tos.size(); ++j)
+   for (int j = 0; j < int(tos.size()); ++j)
        {
         if (tos[j].get_ValueType() == TV_SYM)   // user defined function
            {
@@ -509,7 +512,7 @@ print_ucs(FILE * out, const unsigned int * string_ucs)
 {
 UCS_string ucs;
    ucs.reserve(200);
-   while (*string_ucs)   ucs.append((Unicode)*string_ucs++);
+   while (*string_ucs)   ucs.append(Unicode(*string_ucs++));
 
 UTF8_string utf8(ucs);
    fprintf(out, "%s", utf8.c_str());
@@ -527,7 +530,7 @@ Symbol * symbol = Workspace::lookup_existing_symbol(var_name_ucs);
 Value_P Z = symbol->get_value();
    if (!Z)                              return 0;
 
-   Value_P::increment_owner_count(Z.get(), loc);
+   Z.get()->increment_owner_count(loc);   // keep value
    return Z.get();
 }
 //-----------------------------------------------------------------------------
@@ -592,7 +595,7 @@ int
 UTF8_to_Unicode(const char * utf, int * length)
 {
 int len = 0;
-const Unicode uni = UTF8_string::toUni((const UTF8 *)utf, len, false);
+const Unicode uni = UTF8_string::toUni(utf8P(utf), len, false);
    if (length)   *length = len;
    return uni;
 }
@@ -600,11 +603,11 @@ const Unicode uni = UTF8_string::toUni((const UTF8 *)utf, len, false);
 void
 Unicode_to_UTF8(int uni, char * dest, int * length)
 {
-UCS_string ucs((Unicode)uni);
+UCS_string ucs(1, Unicode(uni));
 UTF8_string utf8(ucs);
-   memcpy(dest, &utf8[0], utf8.size());
-   if (length)   *length = utf8.size();
+   for (int d = 0; d < int(utf8.size()); ++d)   dest[d] = utf8[d];
    dest[utf8.size()] = 0;
+   if (length)   *length = utf8.size();
 }
 //-----------------------------------------------------------------------------
 extern void init_1(const char * argv0, bool log_startup);
@@ -619,6 +622,10 @@ init_libapl(const char * progname, int log_startup)
    uprefs.requested_id = 2000;
 
    init_1(progname, log_startup);
+
+   uprefs.read_config_file(true,  log_startup);   // in /etc/gnu-apl.d/
+   uprefs.read_config_file(false, log_startup);   // in $HOME/.config/gnu_apl/
+
    init_2(log_startup);
 }
 //-----------------------------------------------------------------------------
