@@ -81,7 +81,7 @@ signal_SEGV_handler(int)
            "SEGMENTATION FAULT" << endl;
 
 #if PARALLEL_ENABLED
-   CERR << "thread: " << (const void *)pthread_self() << endl;
+   CERR << "thread: " << reinterpret_cast<const void *>(pthread_self()) << endl;
    Thread_context::print_all(CERR);
 #endif // PARALLEL_ENABLED
 
@@ -194,6 +194,13 @@ show_argv(int argc, const char ** argv)
       CERR << "stdin is: CLOSED" << endl;
    else
       CERR << "stdin is: OPEN" << endl;
+
+   // tell if fd 3 is open or closed
+   //
+   if (fcntl(3, F_GETFD))
+      CERR << "fd 3 is:  CLOSED" << endl;
+   else
+      CERR << "fd 3 is:  OPEN" << endl;
 }
 //-----------------------------------------------------------------------------
 /// print a welcome message (copyright notice)
@@ -202,7 +209,7 @@ show_welcome(ostream & out, const char * argv0)
 {
 char c1[200];
 char c2[200];
-   snprintf(c1, sizeof(c1), "Welcome to GNU APL version %s",build_tag[1]);
+   snprintf(c1, sizeof(c1), "Welcome to GNU APL version %s", build_tag[1]);
    snprintf(c2, sizeof(c2), "for details run: %s --gpl.", argv0);
 
 const char * lines[] =
@@ -216,7 +223,7 @@ const char * lines[] =
   ""                                                                      ,
   c1                                                                      ,
   ""                                                                      ,
-  "Copyright (C) 2008-2016  Dr. Jürgen Sauermann"                         ,
+  "Copyright (C) 2008-2019  Dr. Jürgen Sauermann"                         ,
   "Banner by FIGlet: www.figlet.org"                                      ,
   ""                                                                      ,
   "This program comes with ABSOLUTELY NO WARRANTY;"                       ,
@@ -270,8 +277,8 @@ const bool log_startup = uprefs.parse_argv_1();
 
    init_1(argv[0], log_startup);
 
-   uprefs.read_config_file(true,  log_startup);   // /etc/gnu-apl.d/preferences
-   uprefs.read_config_file(false, log_startup);   // $HOME/.gnu_apl/preferences
+   uprefs.read_config_file(true,  log_startup);   // in /etc/gnu-apl.d/
+   uprefs.read_config_file(false, log_startup);   // in $HOME/.config/gnu_apl/
    uprefs.read_threshold_file(true,  log_startup);  // dito parallel_thresholds
    uprefs.read_threshold_file(false, log_startup);  // dito parallel_thresholds
 
@@ -318,7 +325,13 @@ const bool log_startup = uprefs.parse_argv_1();
 
    uprefs.parse_argv_2(log_startup);
 
-   Macro::init_macros();
+   if (uprefs.CPU_limit_secs)
+      {
+        rlimit rl;
+        getrlimit(RLIMIT_CPU, &rl);
+        rl.rlim_cur = uprefs.CPU_limit_secs;
+        setrlimit(RLIMIT_CPU, &rl);
+      }
 
    if (uprefs.emacs_mode)
       {
@@ -463,6 +476,7 @@ const bool log_startup = uprefs.parse_argv_1();
          Command::process_line(load_cmd);
       }
 
+   Quad_TZ::compute_offset();
    return 0;
 }
 //-----------------------------------------------------------------------------
