@@ -21,7 +21,7 @@
 #include "PostgresArgListBuilder.hh"
 
 #include <string.h>
-#include "../Simple_string.hh"
+#include <vector>
 
 template<class T>
 PostgresBindArg<T>::~PostgresBindArg()
@@ -157,12 +157,13 @@ static void update_double_cell( Cell *cell, char *content )
 
 Value_P PostgresArgListBuilder::run_query( bool ignore_result )
 {
-    int n = args.size();
-    int array_len = n == 0 ? 1 : n;
-    Simple_string<Oid, false>          types  (array_len, 0);
-    Simple_string<const char *, false> values (array_len, 0);
-    Simple_string<int, false>          lengths(array_len, 0);
-    Simple_string<int, false>          formats(array_len, 0);
+    const int n = args.size();
+    const int array_len = n == 0 ? 1 : n;
+    const char * cp_null = 0;
+    vector<Oid>          types  (array_len, 0);
+    vector<const char *> values (array_len, cp_null);
+    vector<int>          lengths(array_len, 0);
+    vector<int>          formats(array_len, 0);
 
     for( int i = 0 ; i < n ; i++ ) {
         PostgresArg *arg = args[i];
@@ -191,7 +192,7 @@ Value_P PostgresArgListBuilder::run_query( bool ignore_result )
                 for( int col = 0 ; col < cols ; col++ ) {
                     if( PQgetisnull( result.get_result(), row, col ) ) {
                         new (db_result_value->next_ravel())
-                            PointerCell( Idx0( LOC ), db_result_value.getref());
+                            PointerCell( Idx0(LOC).get(), db_result_value.getref());
                     }
                     else {
                         Oid col_type = PQftype( result.get_result(), col );
@@ -212,12 +213,12 @@ Value_P PostgresArgListBuilder::run_query( bool ignore_result )
                         else {
                             if( *value == 0 ) {
                                 new (db_result_value->next_ravel())
-                                    PointerCell( Str0( LOC ),
+                                    PointerCell( Str0( LOC ).get(),
                                                  db_result_value.getref() );
                             }
                             else {
                                 new (db_result_value->next_ravel())
-                                    PointerCell( make_string_cell( value, LOC ),
+                                    PointerCell( make_string_cell( value, LOC ).get(),
                                                  db_result_value.getref() );
                             }
                         }
